/**
*      CPP file for CellMap Class
*      CellMap provides basic functions for cell array
*      It has all cells information in a field;
*      
*      Created by:    Nan Chen
*      Created date:  8/7/05
*      Last Modified: 3/6/07
*/

#include "CellMap.h"
//#include <iostream>
#include <gdecs/gdecs.h>


CellMap::CellMap()
{
    cellnumber = 0;
}

CellMap::~CellMap()
{
    delete []cellarray;
    delete []isused;
                   
}


int CellMap::addnewcell(const Cell& ce)
{
    cellarray[cellnumber] = ce;
    isused[cellnumber] = true;
    cellnumber++;
}



void CellMap::initfield(SingleMap *sm)
{
    fprintf (syslog, "Field number %d\n", sm->sitenumber);
    fflush(syslog);
    for (int xx = 0; xx < sm->dimensionsize[0]; xx++)
        for (int yy = 0; yy < sm->dimensionsize[1]; yy++)    
            for (int zz = 0; zz < sm->dimensionsize[2]; zz++)    
            {
                Site te = sm->getsite(xx, yy, zz);	      
                checksite(sm->value[xx * sm->dimensionsize[2] * sm->dimensionsize[1] + yy * sm->dimensionsize[2] + zz],	  
                       xx + sm->location[0], yy + sm->location[1], zz + sm->location[2]);
            }
            
    for (int i = 0; i < cellnumber + 1; i++)
    {
        fprintf(syslog, "Cell %d  Vol %d\n", i, cellarray[i].volume);
    }
    
}

void CellMap::cleancellmap()
{
    cellnumber = 0;
    for (int i = 0; i < MAXCELLNUMBER; i++)
        isused[i] = false;
     
}

void CellMap::checksite(Site &t, int x, int y, int z)
{
    Cell ce;
    ce.cellID = t.cellID;
    ce.celltype = t.celltype;
    
    if (!isused[ce.cellID])
    {
        ce.volume = 1;
        ce.comass[0] = x;
        ce.comass[1] = y;
        ce.comass[2] = z;
	ce.fixed = false;
	cellarray[ce.cellID] = ce;
	isused[ce.cellID] = true;
	if (ce.cellID > cellnumber)
    	    cellnumber = ce.cellID;
	return;
    }
    int t1 = ce.cellID;
    cellarray[t1].volume++;
    cellarray[t1].comass[0] += x;
    cellarray[t1].comass[1] += y;    
    cellarray[t1].comass[2] += z;    
}
