/**
 *      CPP file of ControlReader Class
 *      ControlReader class is used to input control parameters from control file
 *    
 *      Created by:    Nan Chen
 *      Created date:  8/12/05
 *      Last Modified: 9/6/05
 */

#include "ControlReader.h"
// #include <stdio.h>
// #include <stdlib.h>
// #include <string.h>
// #include <iostream>

bool ControlReader::inputenergypara(EnergyPara * pa)
{
    char index[20];
    int t = 0;
    strcpy (index, "Adhesionenergy");
    for (int i = 0; i < termnumber; i++)
    {
        if (!strcmp(index, termname[i]))
        {
            t++;
            int i1,i2;
            double *p = termvalue[i];
            i1 = (int) (p[0]+0.1);
            i2 = (int) (p[1]+0.1);
            double va = p[2];
            pa->contactenergy[i1][i2] = va;
            pa->contactenergy[i2][i1] = va;
        }
    }
    if (t == 0)
    {
        std::cout<<"Can not find Adhesionenergy term in input file"<<std::endl;
        fprintf(syslog, "Can not find Adhesionenergy term in input file\n");
    }
    t = 0;
    strcpy (index, "Volume");
    for (int i = 0; i < termnumber; i++)
    {
        if (!strcmp(index, termname[i]))
        {
            t++;
            int i1,i2;
            double *p = termvalue[i];
            i1 = (int) (p[0]+0.1);
            double va = p[1];
            pa->volume[i1] = va;
        }
    }
    if (t == 0)
    {
        std::cout<<"Can not find Volume term in input file"<<std::endl;
        fprintf(syslog, "Can not find Volume term in input file\n");
    }
    t = 0;
    strcpy (index, "Volumeconstant");
    for (int i = 0; i < termnumber; i++)
    {
        if (!strcmp(index, termname[i]))
        {
            int i1,i2;
            t++;
            double *p = termvalue[i];
            i1 = (int) (p[0]+0.1);
            double va = p[1];
            pa->volumeconstant[i1] = va;
        }
    }
    if (t == 0)
    {
        std::cout<<"Can not find Volumeconstant term in input file"<<std::endl;
        fprintf(syslog, "Can not find Volumeconstant term in input file\n");
    }
    t = 0;
    strcpy (index, "Fluidconstant");
    for (int i = 0; i < termnumber; i++)
    {
        if (!strcmp(index, termname[i]))
        {
            int i1,i2;
            t++;
            double *p = termvalue[i];
            i1 = (int) (p[0]+0.1);
            double va = p[1];
            pa->fluidconstant[i1] = va;
        }
    }
    if (t == 0)
    {
        std::cout<<"Can not find Volumeconstant term in input file"<<std::endl;
        fprintf(syslog, "Can not find Volumeconstant term in input file\n");
    }

    return (true);
}

bool ControlReader::readfile(char *filename)
{
    termnumber = 0;
    fin = fopen(filename, "rt");
    if (!fin)
    {
        std::cout<<"can not open control file: "<<filename<<std::endl;
        return (false);
    }
    do 
    {
        bool testb = getoneterm();
		/*if (!getoneterm())
        {
        if (termnumber == 0)
        {
        std::cout<<"No term has been read from control file"<<std::endl;;
        fclose(fin);
        return (false);
    }
        std::cout<<termnumber<<" terms have been read from Control file"<<std::endl;
        fclose(fin);
        return (true);
    }*/

    }while (!feof(fin));
    if (termnumber == 0)
    {
        std::cout<<"No term has been read from control file"<<std::endl;;
        fclose(fin);
        return (false);
    }
    std::cout<<termnumber<<" terms have been read from Control file"<<std::endl;
    fprintf(syslog, "%d terms have been read from control file\n", termnumber);
    fclose(fin);
    return (true);
}

bool ControlReader::getoneterm()
{
    double value[20];
    if (feof(fin))
    {
        return (false);
    }
    char termn[100];
    int count = fscanf(fin, "%s", termn);
    if (count != 1)
        return (false);
    if (termn[0] == '\/') //It is a comment line
    {
        char ch1;
        do 
        {
            fscanf(fin, "%c", &ch1);
        }while (ch1!='\n');
        return (true);
    }

    strcpy(termname[termnumber], termn);
    double tedouble;
    count = fscanf(fin, "%lf", &tedouble);
    if (count != 1) //Not default input
    {
        char techar;
        fscanf(fin, "%c", &techar);
        if (techar == 'D' || techar == 'd') //User Deined data
        {
            return(getselfdefine());
        }
        std::cout<<"Undefined in term"<<termnumber << std::endl;
        return (false);
    }
    fprintf(syslog, "%s   ", termname[termnumber]);
    for (int ii = 0; ii < (30 - strlen(termname[termnumber])); ii++)
        fprintf(syslog, " ");
    value[0] = tedouble;
    fprintf(syslog, "%lf  ", tedouble);
        
    int subnumber = 1;
    do
    {
        count = fscanf(fin, "%lf", &tedouble);
        if (count != 1) //This term is finished
        {
            memcpy(termvalue[termnumber], value, sizeof(double) * subnumber);
            termvaluenumber[termnumber] = subnumber;
            termdefine[termnumber] = 0;
            termnumber++;
            fprintf(syslog, "\n");
            return (true);
        }
        fprintf(syslog, "%lf  ", tedouble);
        value[subnumber] = tedouble;
        subnumber++;
    } while (1);
}

bool ControlReader::getselfdefine()
{
    double value[400];

    int subnumber = 0;
    do
    {
        double tedouble;
        int count = fscanf(fin, "%lf", &tedouble);
        if (count != 1) //This term is finished
        {
            definevalue[termnumber] = (void*) new double[subnumber];
            memcpy(definevalue[termnumber], value, sizeof(double) * subnumber);
            termvaluenumber[termnumber] = subnumber;
            termdefine[termnumber] = 1;
            termnumber++;
            return (true);
        }
        value[subnumber] = tedouble;
        subnumber++;
    } while (1);
}

void * ControlReader::getvalue(char * index)
{
    int t = 0;
    for (int i = 0; i < termnumber; i++)
    {
        if (!strcmp(index, termname[i]))
        {
            t++;
            choice = i;
            if (termdefine[i] == 1)
                return (definevalue[i]);
            if (termdefine[i] == 0)
                return ((void *)termvalue[i]);

        }
    }
    if (t == 0)
    {
        std::cout<<"Can not find "<<index<<" term in input file"<<std::endl;
        fprintf(syslog, "Can not find %s term in input file\n", index);
    }
    return(NULL);
}
