/**
*      CPP file for Node Class
*      Node class is the simulation control class
*      It managed all simuation running on one node
*
*      Created by:    Nan Chen
*      Created date:  8/17/05
*      Last Modified: 3/6/05
*/


#include "Node.h"
#include "LogicNode.h"
#include "ControlReader.h"
// #include <stdlib.h>
// #include <ios>
// #include <fstream>
#include "Random.h"


using namespace std;

/**
*
*        Get system information, 
*
*/


void Node::getsysinf()
{
//    int long long initS = 0;
//    init_genrand(initS);   
    

}

void  Node::getV(double *v, int x, int y, int z, COMPONENT lcomp)
{
    int   dim = wv->rect_grid->dim, ix, iy, celltype, ic[MAXD], i; 
    RECT_GRID *gr = wv->rect_grid; 
    float *L = gr->L, *U = gr->U; 
    static Locstate  state = NULL;
    float crds[3];
    
    if(state == NULL)
        alloc_state(fr->interf,&state,fr->sizest);
    crds[0] = (double)x/500*(U[0]-L[0])+L[0];
    crds[1] = (double)y/250*(U[1]-L[1])+L[1];
    crds[2] = 0;
    hyp_solution(crds,lcomp,NULL,UNKNOWN_SIDE,
	    fr,wv,state,NULL);
    
    for(i = 0; i < dim; i++)
	v[i] = Vel(state)[i];


}

/*
*       Output Data
*/

void Node::writefile()
{
    char filename[200];     
    sprintf(filename, "%dN%d.dat", 0, currenttime);
    nodes[0].filewriter->nodewrite(filename, nodes[0].field);

    return;  
}

bool Node::initbuffer()
{
    return (true);    
}


bool Node::updatecells()
{
    return (true);

}

/**
*
*        Update LogicBuffer
*
*/

bool Node::updatebuffer()
{
    return (true);
}

/**
*       Read Control File
*
*       Input:   Control file name
*       Return:  True if it is successful
*       True:    Correct control file has been read
*       False:   Wrong in reading control file
*
*/ 

bool Node::initconreader(char * filename)
{
    fprintf(syslog, "------------ Reading Control Input File ---------  \n");      

    fprintf(syslog, "\n");     
    conreader = new (ControlReader);
    if (!conreader)
    {
        std::cout<<"Memory is not enough for control file reader"<<std::endl;
        fprintf(syslog, "Memory is not enough for control file reader\n");
        return (false);
    }
    conreader->syslog = syslog;

    fprintf(syslog, "Memory Allocation succeed in ControlReader Unit \n");     
    fprintf(syslog, "Begin to read Control file %s\n\n", filename);        
    bool tebool = conreader->readfile(filename);
    if (!tebool)
    {
        std::cout<<"Control input error"<<std::endl;
        fprintf(syslog, "Error in reading control file %s\n", filename);
        return (false);
    }

    std::cout<<"Control file reading succeed!"<<std::endl;
    fprintf(syslog, "Control file reading succeed!\n\n\n", filename);

    //Get system field size
    double x = *(double*)conreader->getvalue("Xsyssize");
    double y = *(double*)conreader->getvalue("Ysyssize");
    double z = *(double*)conreader->getvalue("Zsyssize");

    int ix = (int) (x+0.001);
    int iy = (int) (y+0.001);	
    int iz = (int) (z+0.001);	

    CPMPoint point1(0, 0, 0);
    syspt1 = point1;
    CPMPoint point2(ix, iy, iz);
    syspt2 = point2;

    // Get the topology running mode
    runningmode = 1;

    return (true);
}

/**
*   Potts model initilization
*
*/

bool Node::initpotts()
{
    fprintf(syslog, "\n\n-------------Potts Model Unit Initialization--------  \n");      

    pottsmodel = new (Potts);
    if (!pottsmodel)
    {
        std::cout<<"Memory is not enough for pottsmodel"<<std::endl;
        fprintf(syslog, "Memory is not enough for Potts Model\n");
        return (false);
    }
    pottsmodel->syslog = syslog;
    fprintf(syslog, "Memory Allocation succeed  \n");     
    if (!pottsmodel->initmember())
    {
        std::cout<<"Some error happens during Potts initialization"<<std::endl;
        fprintf(syslog, "Some error happens during Potts initialization\n");
        return (false);	
    }
    pottsmodel->node = this;
    fprintf(syslog, "Pottsl Model Unit initialization succeed  \n");  
    fflush(syslog);

    return (true);
}

/**
*   Field initilization
*
*/

void Node::initConfigure(SingleMap *field)
{

    int x = field->dimensionsize[0];
    int y = field->dimensionsize[1];
    int z = field->dimensionsize[2];
    CPMPoint center[600];
    int r[600];
    

    
// Multi    
  
    int cellnumber = 230;
    for (int i = 0; i < cellnumber; i++)
    {
L:
	int xx = 200 + (int)(rand01() *150);
	int yy = 5 + (int)(rand01() *50);
	CPMPoint te(xx,yy,0);
	if (i == 0)
	{
	    center[i] = te;
	    r[i] = 4;
	    continue;
	}
	for (int j =0; j< i; j++)
	{
	    int d = (xx - center[j].cor[0]) * (xx - center[j].cor[0])
		+ (yy - center[j].cor[1]) * (yy - center[j].cor[1]); 
	    if (d < 10)
		goto L;
	}
	center[i] = te;
	r[i] = 4;
    }
    for (int i=0; i< x; i++)
	for (int j = 0; j <y; j++)
	{
	    CPMPoint pt(i, j, 0);
	    Site t;
	    t.cellID = 1;
	    t.celltype = 1;
	    int x1 = i;
	    int y1 = j;
	    field->setsite(pt, t);
	    for (int k = 0; k < cellnumber; k++)
	    {
		int d = (x1 - center[k].cor[0]) * (x1 - center[k].cor[0])
		    + (y1 - center[k].cor[1]) * (y1 - center[k].cor[1]); 
		if (d < (r[k]*r[k]))
		{
		    Site t;
		    t.cellID = 3 + k;

		    t.celltype = 3;   
		    field->setsite(pt, t);
		}
	    }
	}


    for (int i=200; i< 350; i++)
	for (int j = 0; j <4; j++)
	{
	    CPMPoint pt(i, j, 0);
	    Site t;
	    t.cellID = 2;
	    t.celltype = 2;
	    field->setsite(pt, t);
	}	    	
	
}

//Add some other components for initial configure
void Node::addComponents(SingleMap *field)
{
//    return;
    int x = field->dimensionsize[0];
    int y = field->dimensionsize[1];
    int z = field->dimensionsize[2];
    CPMPoint center[600];
    int r[600];

    int cellnumber = 150;
    for (int i = 0; i < cellnumber; i++)
    {
L:
	int xx = 2 + (int)(rand01() *200);
	int yy = 2 + (int)(rand01() *200);
	CPMPoint te(xx,yy,0);
	if (i == 0)
	{
	    center[i] = te;
	    r[i] = 4;
	    continue;
	}
	for (int j =0; j< i; j++)
	{
	    int d = (xx - center[j].cor[0]) * (xx - center[j].cor[0])
		+ (yy - center[j].cor[1]) * (yy - center[j].cor[1]); 
	    if (d < 10)
		goto L;
	}
	center[i] = te;
	r[i] = 4;
    }
    for (int i=0; i< x; i++)
	for (int j = 0; j <y; j++)
	{
	    CPMPoint pt(i, j, 0);
	    Site t;
	    int x1 = i;
	    int y1 = j;
	    for (int k = 0; k < cellnumber; k++)
	    {
		int d = (x1 - center[k].cor[0]) * (x1 - center[k].cor[0])
		    + (y1 - center[k].cor[1]) * (y1 - center[k].cor[1]); 
		if (d < (r[k]*r[k]))
		{
		    Site t;
		    t.cellID = 321 + k;

		    t.celltype = 4;   
		    field->setsite(pt, t);
		}
	    }
	}
	
}
    	    	    

/**
 *   Field and logicnodes initilization
*
*/

bool Node::initfield()
{	
    if (runningmode == 1) //Single pc
    {
        //        field = new SingleMap;
        logicnodesnumber = 1;
        nodes = new LogicNode[logicnodesnumber];
        if (!nodes)
        {
            std::cout<<"Memory is not enough for field and logic node"<<std::endl;
            fprintf(syslog, "Memory is not enough for field and logic nodes\n");
            return (false);
        }
        double x = *(double*)conreader->getvalue("Xsyssize");
        double y = *(double*)conreader->getvalue("Ysyssize");
        double z = *(double*)conreader->getvalue("Zsyssize");

        int ix = (int) (x+0.001);
        int iy = (int) (y+0.001);	
        int iz = (int) (z+0.001);	
        long int allsize = ix * iy *iz;

        for (int i = 0; i < 1; i++)
        {
            nodes[i].field = new SingleMap;
            nodes[i].syslog = syslog;
            if (!nodes[i].field)
            {
                fprintf(syslog, "Memeory is not enough for Logic nodes mem\n ");
                return (false);
            }
            nodes[i].field->value = new Site[allsize];
            if (!nodes[i].field->value)
            {
                fprintf(syslog, "Memeory is not enough for Logic nodes mem\n ");
                return (false);
            }
            nodes[i].field->sitenumber = allsize;
            int dimm = (int) (*(double*)conreader->getvalue("Dimension")+0.1);
            nodes[i].field->dimension = dimm;
            nodes[i].field->dimensionsize[0] = ix;
            nodes[i].field->dimensionsize[1] = iy;	  
            nodes[i].field->dimensionsize[2] = iz;	  

            bool tb =nodes[i].init();
            if (!tb)
                return (tb);
	    
	    int readFile =(int)(*(double*)conreader->getvalue("FieldFromFile") + 0.2); 	
	    if (readFile > 0){    
                nodes[i].filereader->filemode = 1;
                nodes[i].filereader->syslog = syslog;
                bool t1 = nodes[i].getfield("in/field.dat");
                if (!t1)
                    return (t1);
		addComponents(nodes[i].field);
	    }
	    else
	        initConfigure(nodes[0].field);

        }
        field = nodes[0].field;
        pottsmodel->field = field;
        pottsmodel->neibor->setup(field);
        pottsmodel->neibor->initdis();	
        //	memcpy(z, conreader->getvalue("Z"), sizeof(double) * 4);	
        //return (true);

        cells = new CellMap[1];
        if (!cells)
        {
            fprintf(syslog, "Error in Cells\n");
            return (false);
        }


        // Allocate CellMap memory
        int cellnum = (int) (*(double*)conreader->getvalue("Cellnumber")+0.1);
        for (int i = 0; i< 1; i++)
        {    
            cells[i].syslog = syslog;
            nodenumber = 1;
            cells[i].init(cellnum);
        }
    }
    if (runningmode == 2)//MPI grid
    {
    }
    
    cells[0].initfield(nodes[0].field);
    //cells[i].localfield(nodes[i].field);
    	
    cells[0].cellarray[2].fixed = true;

    field = nodes[0].field;
    pottsmodel->field = field;
    pottsmodel->neibor->setup(field);
    pottsmodel->neibor->initdis();	
    fprintf(syslog, "Neighbour Caluclator is initialized\n\n");
    fflush(syslog);
    //	memcpy(z, conreader->getvalue("Z"), sizeof(double) * 4);	
    return (true);
}


/**
*  Topology and logicbuffer initilization
*
*/

bool Node::inittopo()
{
    return (true);
}

/**
*   Parameters initilization
*
*/

bool Node::initpara()
{

    // Energy parameters initialization
    fprintf(syslog, "----------- Parameters initialization --------------\n");  
    //pottsmodel->enpara->celltypenum = (int) (*(double*)conreader->getvalue("Celltypenum")+0.1);
    pottsmodel->enpara->celltypenum = 50;
    bool te = conreader->inputenergypara(pottsmodel->enpara);
    if (!te)
    {
        std::cout<<"Error in Energy parameters initialization"<<std::endl;
        fprintf(syslog, "Error in Energy parameters initialization\n");
        return (false);
    }
    // Parameters setting
    writefrequency = (int) (*(double*)conreader->getvalue("Writefrequency")+0.1);
    allsteps = (int) (*(double*)conreader->getvalue("Allsteps")+0.1);
    pottsmodel->acceptptr->constantk = (*(double*)conreader->getvalue("Kconstant")+0.1);
    pottsmodel->acceptptr->temperature = (*(double*)conreader->getvalue("Temperature")+0.1);
    pottsmodel->writefrequency = writefrequency;
    pottsmodel->allsteps = allsteps;
    pottsmodel->bufferupdatefrequency = (int) (*(double*)conreader->getvalue("Bufferupdate")+0.1);
    bufferupdatefrequency = pottsmodel->bufferupdatefrequency;

    fprintf(syslog, "\n\n");
    fflush(syslog);
 
    return (true);

}


/**
*
*   Node Initilization 
*
*/

bool Node::init()
{
    
    getsysinf();
    outputtitle();



    if (!initconreader("in/input.dat"))
        return (false);

    if (!inittopo())
        return (false);

    if (!initpotts())
        return (false);

    if (!initfield())
        return (false);


    if (!initbuffer())
        return (false);

    if (!initpara())
        return (false);
    return(true);
}


void Node::preRun()
{
    if (runningmode == 1) //Single Processor
    {
        fprintf(syslog, "\n\n-----------------Start running------------------\n");
        fprintf(syslog, "------------------------------------------------\n");

    
        currenttime = 0;
        pottsmodel->runnumber = writefrequency;
        char filename[20];
        nodes[0].filewriter->filemode = 1;
        writefile();
        currentlogic = 0;
        pottsmodel->field = nodes[0].field;
        pottsmodel->cells = &(cells[0]);

        // File Name list output
        filelist = fopen("filename.dat","wt");

        fprintf(filelist, "%d\n", nodenumber);
        int j = 0;
        sprintf(filename, "%dN%d.dat", j, currenttime);
        fprintf(filelist, "%s\n", filename);
        fflush(filelist);

    }
    
}


void Node::oneTimeStep()
{
    pottsmodel->getSurfaceVel();
    std::cout<<"Step:  "<<currenttime;
    fprintf(syslog, "Steps: %d\n", currenttime);
    fflush(syslog);

    //            for (int runnum = 0; runnum < bufferupdatefrequency; runnum++)
    for (int runnum = 0; runnum < writefrequency; runnum++)
	pottsmodel->onestep();
    currenttime += writefrequency;

    pottsmodel->updateType();
    // Write to file
    writefile();

    // Update filename list            
    for (int j = 0; j < 1; j++)
    {
	char filename[300];
        sprintf(filename, "%dN%d.dat", j, currenttime);
    
	fprintf(filelist, "%s\n", filename);
	fflush(filelist);
    }

}

/**
*   Node main loop
*
*/

bool Node::run()
{
    if (runningmode == 1) //Single Processor
    {
	preRun();

        //Main loop
        bool t;
        do{
	    oneTimeStep();

        } while (currenttime <= allsteps);

        // Check if the runing is successful
        t = true;
        if (!t)
        {
            std::cout<<"Potts model running error"<<std::endl;
            fprintf(syslog, "Potts model running error\n");
            return (false);
        }
        fprintf(syslog, "Simulation is successfully finished\n");
   
        return(t);
    }

    fprintf(syslog, "Not valid mode\n");
    return (false);

}


/** 
*
*   Node class memory release
*
*/

void Node::finish()
{
    if (nodeID == 0)
    {
        if (!filelist)
        {
	    fflush(filelist);
            fclose(filelist);
            filelist = NULL;
        }
    }

    return ;
    if (conreader)
    {
        delete conreader;
        conreader = NULL;
    }
    if (pottsmodel)
    {
        delete pottsmodel;
        pottsmodel = NULL;
    }
    if (nodes)
    {
        delete[] nodes;
        nodes = NULL;
    }

    if (cells)
    {
        delete[] cells;
        cells = NULL;
    }


}
void Node::outputtitle()
{
    fprintf(syslog,"\n");

    fprintf(syslog, "/*****************************************************************"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "******************************************************************"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                                                 *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                       Cell Clot    Version 0.2                  *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                                                 *"); 
    fprintf(syslog,"\n");                               
    fprintf(syslog, "*"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*       A CPM modelling tool for Clot formation simulation        *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                                                 *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                                                 *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                     By Nan chen (nchen1@nd.edu)                 *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                                                 *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*            Bio-comlexity Center of University of Notre Dame     *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                       March. 2007               *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "*                                                                 *"); 
    fprintf(syslog,"\n");
    fprintf(syslog, "****************************************************************** "); 
    fprintf(syslog,"\n");
    fprintf(syslog, "******************************************************************/ "); 
    fprintf(syslog,"\n\n\n");
}

