/**
*      Head file for NodeClass
*      Node class is the simulation control class
*      It managed all simuation running on one node
*
*      Created by:    Nan Chen
*      Created date:  8/17/05
*      Last Modified: 3/6/07
*/

#ifndef NODE_H
#define NODE_H

#include <gdecs/gdecs.h>

// #include <stdlib.h>
// #include <stdio.h>
#include "SingleMap.h"
#include "CellMap.h"
#include "Potts.h"
#include <cpm/cpmprotos.h>

class LogicNode;
class ControlReader;

class Potts;
class LogicBuffer;
class Grid2D;


class Node
{
private:
  SingleMap *field;
  Potts *pottsmodel;
  ControlReader *conreader;
  Grid2D *topo;

public:
  LogicNode *nodes;
  Wave *wv;
  Front *fr;
  LogicBuffer *buffer;
  CellMap *cells;
  FILE *syslog;
  FILE *filelist;
  CPMPoint pt1, pt2;
  CPMPoint syspt1, syspt2;
  int logicnodesnumber;
  int currentlogic;
  int bufferupdatefrequency;
  int nodenumber;
  int nodeID; 
  int currenttime;
  int allsteps, writefrequency;
  int runningmode; //1 Single Proc. 2 MPI
  
  void singlestep();
  void nodeswitch();
  void preRun();
  void oneTimeStep();
  void writefile();
  void savedata();
  void initConfigure(SingleMap *field);
  void addComponents(SingleMap *field);
  void constructfield();
  void getsysinf();
  bool initconreader(char *filename);
  bool init();
  bool run();
  void finish();
  bool initpara();
  bool inittopo();
  bool initbuffer();
  bool initfield();
  bool initpotts();
  
  bool updatebuffer();
  bool updatecells();
  void outputtitle();
  int getcelltype(int x, int y, int z)
  {	
    int k = pottsmodel->getcelltype(x, y, z); 
    return k;
  };
  void getV(double *, int , int , int , COMPONENT);
  Node(){nodes = NULL; buffer = NULL; topo = NULL; pottsmodel = NULL; field = NULL; conreader = NULL; cells = NULL;};
  ~Node(){  finish();};
};
#endif
