/**
 *      Cpp file of Potts Class
 *      Potts class is responsible for main Potts model loop
 *      
 *      Created by:    Nan Chen
 *      Created date:  8/17/05
 *      Last Modified: 3/6/07
 */

#include "Potts.h"
#include "Random.h"
#include "Node.h"
#define Ltype 1

Potts::Potts()
{
    field = NULL;
    acceptptr = NULL;
    enpara = NULL;
    automationptr = NULL;
    neibor = NULL;
    cells = NULL;
    chem = NULL;
}

Potts::~Potts()
{
    if (acceptptr)
    {
	delete acceptptr;
	acceptptr = NULL;
    }
    if (enpara)
    {
	delete enpara;
	enpara = NULL;
    }
    if (automationptr)
    {
	delete automationptr;
	automationptr = NULL;
    }
    if (neibor)
    {
	delete neibor;
	neibor = NULL;
    }

    if (chem)
    {
	delete chem;
	chem = NULL;
    }

}


bool Potts::initmember()
{
    acceptptr = new Accept;
    if (!acceptptr)
    {
	std::cout<<"Memory is not enough for pottsmodel"<<std::endl;
	fprintf(syslog, "Memory is not enough for Potts Model\n");
	return (false);	
    }

    fprintf(syslog, "Accept Unit initialization succeed  \n");     
    enpara = new EnergyPara;
    if (!enpara)
    {
	std::cout<<"Memory is not enough for pottsmodel"<<std::endl;
	fprintf(syslog, "Memory is not enough for Potts Model\n");
	return (false);
    }
    fprintf(syslog, "Energy Parameters Unit initialization succeed  \n");  
    automationptr = new Automation;
    if (!automationptr)
    {
	std::cout<<"Memory is not enough for pottsmodel"<<std::endl;
	fprintf(syslog, "Memory is not enough for Potts Model\n");
	return (false);
    }
    fprintf(syslog, "Automation Unit initialization succeed  \n");  
    neibor = new NeighborCal;
    if (!neibor)
    {
	std::cout<<"Memory is not enough for pottsmodel"<<std::endl;
	fprintf(syslog, "Memory is not enough for Potts Model\n");
	return (false);
    }
    fprintf(syslog, "Neighbor Calculator Unit initialization succeed  \n");  

    chem = new ChemField;
    if (!chem)
    {
	std::cout<<"Memory is not enough for pottsmodel"<<std::endl;
	fprintf(syslog, "Memory is not enough for Potts Model\n");
	return (false);
    }
    fprintf(syslog, "Chemical Field Unit initialization succeed  \n");  
    return (true);
}

bool Potts::init()
{
    currentenergy = 0;
    currentenergy += getvolume();
    currentenergy += getcontact();
    currentenergy += getchemical();
    //currentenergy += getFluid();
    return (true);
}
bool Potts::run()
{
    /*
       for (int i =0; i < runnumber; i++)
       {
       if ((i % bufferupdatefrequency) == 0)
       node->updatebuffer();
       if (i == 0)
       {
       init();

       CellMap &tece = *cells;
       fprintf(syslog, "  %5d  Cells\n", tece.cellnumber);
       fprintf(syslog, "  No.        ID           CellType           Volume        ObjVol.\n");
       for (int i = 1; i <= tece.cellnumber; i++)
       {
       Cell &tec = tece.cellarray[i];
       fprintf(syslog, "%4d         %4d          %4d              %6d    %lf\n",
       i, tec.cellID, tec.celltype, tec.volume, tec.objvolume);

       }            
       }
       onestep();
       }*/
    return (true);
}

void Potts::onestep()
{
    int onemcs = field->sitenumber;
    fprintf(syslog, "running\n");
    fflush(syslog);
    for (int i = 0; i<onemcs; i++)
    {

	oneflip();
	//std::cout<<i<<"   "<<currentenergy<<std::endl;
    }
}

void Potts::oneflip()
{
    Flip fltest = getrandomflip();
    if (!sitesuccess) return;
    double endiff = 0;
    endiff += getcontactchange (fltest);
    endiff += getvolumechange (fltest);
    endiff += getchemicalchange(fltest);
    endiff += getFluid(fltest);

    if (acceptptr->energyfunction(endiff))
    {
	acceptflip(fltest);  
	currentenergy += endiff;
    }        
}

Flip Potts::getrandomflip()
{
    CPMPoint po = randomposition();
    Site tesite;
    sitesuccess = testsite (tesite, po);
    Flip teflip;
    if (sitesuccess)
    {
	teflip.location = po;
	teflip.newsitevalue = tesite;

    }
    return (teflip);
}

bool Potts::testsite (Site &test, CPMPoint &po)
{
    //  neibor->getlist (po);
    Site tesite = field -> getsite (po);
    double choice = rand01() * neibor->disnum;
    int chid = (int) choice;
    if (chid == neibor ->disnum)
	chid --;
    CPMPoint tepo;
    tepo = po + neibor->displace[chid];

    bool outside = false;
    Site tout;

    tout = getsite(tepo);

    //neibor->correctrange(tepo);
    //Site tesite1 = field->getsite (tepo);
    if (tout.cellID == tesite.cellID)
	return (false);
    if (cells->cellarray[tout.cellID].fixed)
	return false;
    if (cells->cellarray[tesite.cellID].fixed)
	return false;
    test = tout;
    return (true);

}

/*
 *  Randomly choose one position from field
 *
 */

CPMPoint Potts::randomposition()
{
    CPMPoint tepoint;

    if (field->dimension == 2)
    {
	for (int i = 0; i < 2 ; i ++)
	{
	    int tsize = field -> getsize (i);
	    double loca = rand01() * tsize;
	    tepoint.cor[i] = (int) loca;
	    if (tepoint.cor[i] == tsize)
		tepoint.cor[i]--;
	}
	tepoint.cor[2] = 0;
	return (tepoint); 
    }
    CPMPoint t(0, 0, 0);
    return (t);
}

/*
 * Get the site value from location 'tepo'
 *
 * This function will determine whether the tepo is inside the field or outside
 * then get the site value from field or buffer
 *
 */

Site Potts::getsite(CPMPoint &tepo)
{
    bool outside = false;
    Site tout;
    for (int i = 0; i < field->dimension; i++)
    {
	int sized = field->dimensionsize[i];
	if (tepo.cor[i] < 0)
	{
	    tepo.cor[i]+=sized;                       
	}
	if (tepo.cor[i] >= sized)
	{
	    tepo.cor[i]-=sized;                                   
	}
    }

    Site tesite;
    tesite = field->getsite(tepo);
    return (tesite);

}


void Potts::getV(double *v, int x, int y, int z)
{
    node->getV(v, x, y, z, (COMPONENT)2);
}

void Potts::updateType()
{

    for (int i = 0; i < field ->dimensionsize [0]; i++)
	for (int j = 0; j < field->dimensionsize[1]; j++)
	{
	    CPMPoint tepo;
	    tepo.cor[0] = i;
	    tepo.cor[1] = j;
	    tepo.cor[2] = 0;
	    Site tesite;
	    tesite = field -> getsite (tepo);
	    tesite.celltype = cells->cellarray[tesite.cellID].celltype;
	    field->setsite(tepo, tesite);
	}

}

// Obtain surface and liquid surface number
double Potts::getSurfaceVel()
{
    if (field -> dimension == 2)
    {
	for (int i = 1; i< cells->cellnumber + 1; i++)
	{
	    cells->cellarray[i].surface = 0;
	    cells->cellarray[i].liquid = 0;
	    cells->cellarray[i].liquidV[0] = 0;
	    cells->cellarray[i].liquidV[1] = 0;
	    cells->cellarray[i].liquidV[2] = 0;    
	}    

	for (int i = 0; i < field ->dimensionsize [0]; i++)
	    for (int j = 0; j < field->dimensionsize[1]; j++)
	    {
		CPMPoint tepo;
		tepo.cor[0] = i;
		tepo.cor[1] = j;
		tepo.cor[2] = 0;		
		int type1, type2;
		Site tesite;
		tesite = field -> getsite (tepo);
		type1 = tesite.celltype;
		if (type1 == Ltype)
		    continue;
		neibor->getlist(tepo);
		int ID1 = tesite.cellID;
		for (int k=0; k < neibor-> resultnum; k++)
		{
		    CPMPoint tepo = neibor->result[k];
		    Site tesite1 = getsite(tepo);
		    type2 = tesite1.celltype;;
		    int ID2 = tesite1.cellID;
		    if (ID1 == ID2)
			continue;
		    cells->cellarray[ID1].surface++;
		    if (type2 == Ltype)	
		    {
			cells->cellarray[ID1].liquid++;
			double vv[3];
			getV(vv,i,j,0);
			for (int ind = 0; ind < 3; ind++)
			    cells->cellarray[ID1].liquidV[ind] +=vv[ind];
		    }
		}	     		
	    }

	for (int i = 1; i< cells->cellnumber + 1; i++)
	{
//	    if (cells->cellarray[i].liquid == 0)
//		continue;
	    for (int ind = 0; ind < 3; ind++)
		if (cells->cellarray[i].liquid != 0)
    		    cells->cellarray[i].liquidV[ind] /= cells->cellarray[i].liquid;
		
    	    fprintf(syslog, "Cell %d liquid: %d, surface %d celltype %d  v %lf\n ", i,cells->cellarray[i].liquid, 
	                cells->cellarray[i].surface,
			cells->cellarray[i].celltype,
			cells->cellarray[i].liquidV[0]);
	    
	}        	    
	return 0;    
    }  
}


/*
 *    Calculate system contact energy
 * 
 *
 */

double Potts::getcontact()
{
    if (field -> dimension == 2)
    {
	double energy = 0;
	for (int i = 0; i < field ->dimensionsize [0]; i++)
	    for (int j = 0; j < field->dimensionsize[1]; j++)
	    {
		CPMPoint tepo;
		tepo.cor[0] = i;
		tepo.cor[1] = j;
		tepo.cor[2] = 0;
		neibor->getlist(tepo);
		int type1, type2;
		Site tesite;
		tesite = field -> getsite (tepo);
		type1 = cells->cellarray[tesite.cellID].celltype;;
		for (int k=0; k < neibor-> resultnum; k++)
		{
		    CPMPoint tepo = neibor->result[k];
		    tesite = getsite(tepo);
		    type2 = cells->cellarray[tesite.cellID].celltype;;
		    if (type1 == type2)
			continue;
		    energy += enpara-> getcontact (type1, type2);
		}	     		
	    }
	energy = energy / 2;
	return (energy);    
    }  
}

double Potts::getvolume()
{
    if (field -> dimension == 2)
    {
	double energy = 0;
	for (int i = 0; i <= cells->cellnumber; i++)
	{
	    int cellty = cells->cellarray[i].celltype;
	    if (cellty < 0)
		continue;
	    double volume = (double) cells->cellarray[i].volume;
	    energy += enpara->getvolume(cellty, volume);
	}
	return (energy);    
    }  
}

double Potts::getchemical()
{
    return 0;
}

double Potts::getcontactchange(Flip &fl)
{
    if (field ->dimension == 2)
    {
	double newenergy=0;  
	double oldenergy=0;
	neibor->getlist(fl.location);
	int type1, type2;
	Site tesite;
	tesite = field -> getsite (fl.location);
	type1 = cells->cellarray[tesite.cellID].celltype;
	type2 = cells->cellarray[fl.newsitevalue.cellID].celltype;
	int ID1 = tesite.cellID;	
	for (int k=0; k < neibor-> resultnum; k++)
	{
	    CPMPoint tepo = neibor->result[k];
	    Site tout;
	    tesite = getsite(tepo);       
	    int type3 = cells->cellarray[tesite.cellID].celltype;           
	    int ID3 = tesite.cellID;
	    int ID2 = fl.newsitevalue.cellID;
	    if (ID1 != ID3)	    
		oldenergy += enpara->getcontact(type1, type3);
	    if (ID2 != ID3)
		newenergy += enpara->getcontact(type2, type3);	  
	}
	return (newenergy - oldenergy);	
    }
}


double Potts::getvolumechange(Flip &fl)
{
    double newenergy = 0;
    Site oldsite = field->getsite(fl.location);
    double oldvolume = cells->cellarray[oldsite.cellID].volume;
    int cellty = cells->cellarray[oldsite.cellID].celltype;

    newenergy += enpara->getvolume(cellty, oldvolume);
    newenergy -= enpara->getvolume(cellty,(oldvolume - 1));

    oldvolume = cells->cellarray[fl.newsitevalue.cellID].volume;
    cellty = cells->cellarray[fl.newsitevalue.cellID].celltype;  
    newenergy += enpara->getvolume(cellty, oldvolume);  
    newenergy -= enpara->getvolume(cellty, (oldvolume+1));

    newenergy= -newenergy;
    return (newenergy);
}



double Potts::getchemicalchange(Flip &fl)
{
    return 0;    
}

int Potts::getcelltype(int x, int y, int z)
{

 	Site  tesite = field -> getsite (x, y, z);
	return cells->cellarray[tesite.cellID].celltype;
	  
}
double Potts::getFluid(Flip &fl)
{


          //  return 0;

    if (field ->dimension == 2)
    {
	double newenergy=0;  
	double oldenergy=0;
	int typeOld, typeNew;
	int IDold, IDnew;
	Site tesite;

	IDold = tesite.cellID;
	IDnew = fl.newsitevalue.cellID;
	typeOld = cells->cellarray[tesite.cellID].celltype;
	typeNew = cells->cellarray[fl.newsitevalue.cellID].celltype;
	if (cells->cellarray[IDold].liquid ==0 && cells->cellarray[IDnew].liquid==0) //Not contact with liquid
	    return 0;


	double oldmove[3], newmove[3];
	for (int i = 0; i < 2; i++)
	{
	    oldmove[i] = (double)(cells->cellarray[IDold].comass[i]-fl.location.cor[i])/(cells->cellarray[IDold].volume-1)
		-(double)cells->cellarray[IDold].comass[i]/cells->cellarray[IDold].volume;
	    newmove[i] = (double)(cells->cellarray[IDnew].comass[i]+fl.location.cor[i])/(cells->cellarray[IDnew].volume+1)
		-(double)cells->cellarray[IDnew].comass[i]/cells->cellarray[IDnew].volume;
	}
	double oldDiff, newDiff;    	    
	if (cells->cellarray[IDold].liquid == 0)
	    oldDiff = 0;
	else{	  
	    oldDiff = cells->cellarray[IDold].liquidV[0] * oldmove[0] + cells->cellarray[IDold].liquidV[1] * oldmove[1];
	    oldDiff *= enpara->getfluid(typeOld) * cells->cellarray[IDold].liquid/cells->cellarray[IDold].surface;
	}
	if (cells->cellarray[IDnew].liquid == 0)
	    newDiff = 0;
	else{	  	
	    newDiff = cells->cellarray[IDnew].liquidV[0] * newmove[0] + cells->cellarray[IDnew].liquidV[1] * newmove[1];
	    newDiff *= enpara->getfluid(typeNew) * cells->cellarray[IDnew].liquid/cells->cellarray[IDnew].surface;
	}
	//	fprintf (syslog, "%lf\n", (-oldDiff - newDiff));


	return -(oldDiff + newDiff);	
    }

}
/*
 *  If flip is accepted
 *
 */

void Potts::acceptflip(Flip &f1)
{

    // Change volume
    Site oldsite = field->getsite(f1.location);

    cells->cellarray[oldsite.cellID].removesite(f1.location.cor[0],
	    f1.location.cor[1],f1.location.cor[2]);

    cells->cellarray[f1.newsitevalue.cellID].addsite(f1.location.cor[0],
	    f1.location.cor[1],f1.location.cor[2]);

    // Change site value
    field->setsite(f1.location, f1.newsitevalue);    
    // Check Diff
    //    if (cells->cellarray[oldsite.cellID].celltype == 4 && cells->cellarray[f1.newsitevalue.cellID].celltype ==3)
    //	cells->cellarray[oldsite.cellID].celltype = 3;
    if (cells->cellarray[oldsite.cellID].celltype == 3 && cells->cellarray[f1.newsitevalue.cellID].celltype ==4)
	cells->cellarray[f1.newsitevalue.cellID].celltype = 3;


}

void Potts::getcells()
{
    cells->cellnumber = 0;
    cells->cleancellmap();
}
