#include <ghyp/ghyp.h>
#include <gdecs/vecdecs.h>

#include <ghyp/ghyploc.h>

#if defined (PETSC)
#include "petscksp.h"
#endif /* if defined (PETSC) */


struct _RD_Field_Stencil{
        // viscosity and density before advection
        float       *c,   *cg_c;
        /* On the centers of the cell edges */
        VEL_ST      *u_states[MAXD], *cg_u_states[MAXD];

        struct {
            bool    c;
            bool    u_states;
            bool    soln;
        } alloc;

        RECT_GRID   *rect_gr;
        int         npts_sten;
        int         nnx, nny;

        Front       *fr, *newfr;
        Wave        *wave, *newwave;
        Chem_Stencil  *chem_sten;
};
typedef struct _RD_Field_Stencil RD_Field_Stencil;

#define Rect_chem(ic,field)                                            \
         ((field)->cg_c[(ic)[1]*(field)->nnx + (ic)[0]] )

/* Local global used for irregular point */
/* For efficiency */
LOCAL   int     gmax[MAXD], vsten_rad, is_vec_method;
LOCAL   int     dim;
LOCAL   int     lbuf[MAXD], ubuf[MAXD];
LOCAL   float   h[MAXD];
LOCAL   float   L[MAXD], VL[MAXD], VU[MAXD];

LOCAL   COMPONENT       ext_comp;
LOCAL   float           hdir, hndir, dir[MAXD];
LOCAL   GRID_DIRECTION  prev_side, next_side;
LOCAL   INTERFACE       *intfc = NULL;
LOCAL   int             endpt;
LOCAL   int             deb_sten = NO;
#if defined(PETSC)
LOCAL   float     K[2] = {0.5,0.5};
LOCAL   int       comm_size = 1;
#endif /* if defined(PETSC) */
/* End of Local global used for irregular point */

LOCAL RD_Field_Stencil *alloc_RD_field_space(RD_Field_Stencil*,Wave*);
LOCAL void             grid_chemical(RD_Field_Stencil*,Wave*,Front*,int);
LOCAL int              cell_edge_vel(RD_Field_Stencil*);
LOCAL void             advec_diff_chem(float,RD_Field_Stencil*);
LOCAL void             Chem_Crank_Nicolson_sten(float,int*,RD_Field_Stencil*,P_CELL2D*,int);
LOCAL void             petsc_cell_center_parab_solver(Wave*,P_CELL2D**,int,int,int,int,
                          int**,int**,float**);
LOCAL void             Runge_Kutta_step(RD_Field_Stencil*,Wave*,Front*,float);
/* End of local function declaration */




LOCAL RD_Field_Stencil *alloc_RD_field_space(
        RD_Field_Stencil      *field,
        Wave                  *wv)
{
        int       dim, nrad, i;
        RECT_GRID *gr = wv->rect_grid;
        RD_Field_Stencil     *rd_field;
        int        edgexmax, edgeymax, edgex_n_segs, edgey_n_segs;
        int        *gmax, *lbuf, *ubuf, ix, iy;

        dim = gr->dim;

        if(field == NULL)
        {
            scalar(&rd_field, sizeof(RD_Field_Stencil));
            zero_scalar(rd_field,sizeof(RD_Field_Stencil));

            /* Temp. set to use 2nd order scheme */
            rd_field->npts_sten = 5;

            nrad = (rd_field->npts_sten - rd_field->npts_sten%2)/2;

            switch(dim)
            {
            case 2:
                edgexmax = gr->gmax[0]+gr->lbuf[0]+gr->ubuf[0];
                edgeymax = gr->gmax[1]+gr->lbuf[1]+gr->ubuf[1];
                edgex_n_segs = edgexmax*(edgeymax+1);
                edgey_n_segs = edgeymax*(edgexmax+1);

                rd_field->nnx = edgexmax;
                rd_field->nny = edgeymax;

                // storage for chemicals
                vector(&(rd_field->c),edgexmax*edgeymax,sizeof(float));
                rd_field->cg_c = rd_field->c + edgexmax*gr->lbuf[1] + gr->lbuf[0];
                rd_field->alloc.c = YES;

                // storage for vel on cell edge
                vector(&(rd_field->u_states[0]),edgex_n_segs,sizeof(VEL_ST));
                vector(&(rd_field->u_states[1]),edgey_n_segs,sizeof(VEL_ST));

                rd_field->cg_u_states[0] = rd_field->u_states[0] +
                    edgexmax*gr->lbuf[1] + gr->lbuf[0];
                rd_field->cg_u_states[1] = rd_field->u_states[1] +
                    (edgexmax+1)*gr->lbuf[1] + gr->lbuf[0];
                rd_field->alloc.u_states = YES;
            break;
            default:
                printf("ERROR: in g_FD_NS_solver, implement alloc sten for dim %d\n",dim);
                clean_up(ERROR);
            }
            return rd_field;
        }
        return field;
}

LOCAL void grid_chemical(
        RD_Field_Stencil  *rd_field,
        Wave              *wv,
        Front             *fr,
        int               indx)
{
        int             smin[MAXD], smax[MAXD];
        TRI_GRID        *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        // Front           *nfr = ns_field->newfr;
        int             ix, iy, ic[MAXD], i, j, k, i1;
        COMPONENT       comp;
        int             deb_flag = NO;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }


        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                ic[0] = ix; ic[1] = iy;

                Rect_chem(ic,rd_field) = 0.0;

                comp = Rect_comp(ic,wv);
                if(comp != ext_comp && (! is_excluded_comp(comp,fr->interf)))
                {
                    Rect_chem(ic,rd_field) = Chem(Rect_state(ic,wv))[indx];
                }
            }
        }
}
 
/* This is a version of cell_edge_vector_u_ver4(), the velocity is from new wave*/
LOCAL  int cell_edge_vel(
        RD_Field_Stencil *rd_field)
{
        int             smin[MAXD], smax[MAXD];
        Wave            *newwv = rd_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(newwv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = rd_field->newfr, *fr = rd_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vnew[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh = newwv->rect_grid->h;
        int             vset;

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        // Vertical edge
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0]+1;  ix < smax[0];  ++ix)
            {
                ic[0] = ix-1; ic[1] = iy;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                   comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    for(i = 0; i < dim; i++)
                    {
                        v_line_vel(ic1,rd_field).v[i] = vnew[i] =
                          0.5*(Vel(Rect_state(ic,newwv))[i] + Vel(Rect_state(ic1,newwv))[i]);
                    }
                }
                else
                {
                    // Dirichlet boundary type, might need to re-consider
                    for(i = 0; i < dim; i++)
                        v_line_vel(ic1,rd_field).v[i] = 0.0;
                }
            }
        }        

        // Left vertical edge
        ic[0] = ix = smin[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                 continue;

            if(Rect_blk_edge(ic,newwv).edge[0][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic,rd_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,0,end,newwv,YES,sten);

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                if(sten->nc[-1] > 1)
                {
                    for(i = 0; i < sten->nc[-1]; i++)
                    {
                        if(wave_type(sten->crx[-1][i]->hs) == NEUMANN_BOUNDARY ||
                           wave_type(sten->crx[-1][i]->hs) == DIRICHLET_BOUNDARY)
                        {
                            newcrx = sten->crx[-1][i];
                            if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                comp = negative_component(newcrx->hs);
                            else
                                comp = positive_component(newcrx->hs);
                            break;
                        }
                    }
                }
                else
                    newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vel(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet
            for(i = 0; i < dim; i++)
                v_line_vel(ic,rd_field).v[i] = vnew[i];
        }

        // Right vertical edge
        ic[0] = ix = smax[0]-1;
        ic1[0] = smax[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            ic1[1] = ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(ic,newwv).edge[0][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic1,rd_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,0,end,newwv,YES,sten);

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                if(sten->nc[1] > 1)
                {
                    // printf("WARNING: cell_edge_vector_u_ver4, sten->nc[1] = %d\n",
                    //       sten->nc[1]);
                    // printf("wave type %d\n", wave_type(sten->crx[1][0]->hs));
                    for(i = 0; i < sten->nc[1]; i++)
                    {
                        if(wave_type(sten->crx[1][i]->hs) == NEUMANN_BOUNDARY ||
                           wave_type(sten->crx[1][i]->hs) == DIRICHLET_BOUNDARY)
                        {
                            newcrx = sten->crx[1][i];
                            if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                comp = negative_component(newcrx->hs);
                            else
                                comp = positive_component(newcrx->hs);

                            // printf("comp = %d\n", comp);
                            // print_general_vector("point crds",Coords(sten->p[1]),dim,"\n");
                            // print_curve(Curve_of_hs(newcrx->hs));
                            // print_crxings(newcrx, NO);
                            // clean_up(ERROR);
                            break;
                        }
                    }
                }
                else
                    newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver4(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet
            for(i = 0; i < dim; i++)
                v_line_vel(ic1,rd_field).v[i] = vnew[i];
        }

        // Horizontal edge
        for (iy = smin[1]+1;  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                ic[0] = ix; ic[1] = iy-1;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                   comp1 != POTS_COMP && comp != POTS_COMP)
                {
                    for(i = 0; i < dim; i++)
                    {
                        h_line_vel(ic1,rd_field).v[i] = vnew[i] =
                             0.5*(Vel(Rect_state(ic,newwv))[i] + Vel(Rect_state(ic1,newwv))[i]);
                    }
                }
                else
                {
                    for(i = 0; i < dim; i++)
                        h_line_vel(ic1,rd_field).v[i] = 0.0;
                }
            }
        }

        // bottom horizontal edge

        ic[1] = iy = smin[1];
        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(ic,newwv).edge[1][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic,rd_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,1,end,newwv,YES,sten);

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                if(sten->nc[-1] > 1)
                {
                    for(i = 0; i < sten->nc[-1]; i++)
                    {
                        if(wave_type(sten->crx[-1][i]->hs) == NEUMANN_BOUNDARY ||
                           wave_type(sten->crx[-1][i]->hs) == DIRICHLET_BOUNDARY)
                        {
                            newcrx = sten->crx[-1][i];
                            if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                comp = negative_component(newcrx->hs);
                            else
                                comp = positive_component(newcrx->hs);
                            break;
                        }
                    }
                }
                else
                    newcrx = sten->crx[-1][0];
                // newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vel(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet
            for(i = 0; i < dim; i++)
                h_line_vel(ic,rd_field).v[i] = vnew[i];
        }

        // top horizontal edge
        ic[1] = iy = smax[1]-1;
        ic1[1] =  smax[1];

        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            ic1[0] = ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(ic,newwv).edge[1][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic1,rd_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,1,end,newwv,YES,sten);

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                if(sten->nc[1] > 1)
                {
                    for(i = 0; i < sten->nc[1]; i++)
                    {
                        if(wave_type(sten->crx[1][i]->hs) == NEUMANN_BOUNDARY ||
                           wave_type(sten->crx[1][i]->hs) == DIRICHLET_BOUNDARY)
                        {
                            newcrx = sten->crx[1][i];
                            if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                comp = negative_component(newcrx->hs);
                            else
                                comp = positive_component(newcrx->hs);
                            break;
                        }
                    }
                }
                else
                    newcrx = sten->crx[1][0];
                // newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), top boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet
            for(i = 0; i < dim; i++)
                h_line_vel(ic1,rd_field).v[i] = vnew[i];

        }
}

LOCAL void advec_diff_chem(
        float      dt,
        RD_Field_Stencil *rd_field)
{
        P_CELL2D        ***p_cell, **tran_cell; // used for laplacian
        int             Nx = rd_field->rect_gr->gmax[0];
        int             Ny = rd_field->rect_gr->gmax[1];
        Wave            *newwv = rd_field->newwave;
        Wave            *wv = rd_field->wave;
        Front           *newfr = rd_field->newfr;
        Front           *fr = rd_field->fr;
        TRI_GRID        *ntg = wave_tri_soln(newwv)->tri_grid;
        TRI_GRID        *tg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        int             ix, iy, ic[MAXD], i, j, k;
        float           *dh = wv->rect_grid->h;

        int             *n_dist = NULL, myid, ICOORDS[2], n_chem;
        int             NLblocks,ilower, iupper, I;
        int             **I_to_ijk;   /* local  i, j, k, local I */
        int             **IJK_to_I;  /* global I, J, K, global I */
        int             idxm[1],  sten_index[3], indices;
        float           norm,min,max,x,y,*coords;
        float           val[1], values[3][3];

        float           **soln;

#if defined (PETSC)
#if defined(__MPI__)
        MPI_Comm_size(PETSC_COMM_WORLD,&comm_size);
#endif /* defined(__MPI__) */

        n_chem = rd_field->chem_sten->n_chem;
        tri_array(&p_cell,Nx,Ny,n_chem,sizeof(P_CELL2D));
        matrix(&tran_cell,Nx,Ny,sizeof(P_CELL2D));
        vector(&n_dist,comm_size,sizeof(int));
        matrix(&I_to_ijk,Nx*Ny,2,sizeof( int));
        matrix(&IJK_to_I,(Nx+4),(Ny+4),sizeof( int));
        matrix(&soln,(Nx+4),(Ny+4),sizeof(float));

        myid = pp_mynode();
        find_Cartesian_coordinates(myid,newfr->pp_grid,ICOORDS);

        NLblocks = 0;
        for(j = 0; j < Ny; j++)
        {
            ic[1] = j;
            for(i = 0; i < Nx; i++)
            {
                ic[0] = i;
                if(Rect_comp(ic,newwv) != ext_comp &&
                   !is_excluded_comp(Rect_comp(ic,newwv),newfr->interf))
                {
                    I_to_ijk[NLblocks][0] = i;
                    I_to_ijk[NLblocks][1] = j;
                    NLblocks++;
                }
            }
        }

#if defined(__MPI__)
        pp_all_gather((POINTER)&NLblocks,sizeof(int),(POINTER)n_dist,
              sizeof(int)*comm_size);
#else /* defined(__MPI__) */
        assign((POINTER)n_dist,(POINTER)&NLblocks,sizeof(int));
#endif /* defined(__MPI__) */

        ilower = 0;
        iupper = n_dist[0]-1;

        for (i = 1; i <= myid; i++)
        {
            ilower += n_dist[i-1];
            iupper += n_dist[i];
        }

        create_global_element_list(newfr,ilower,iupper,ICOORDS,I_to_ijk,IJK_to_I);

        for(iy = 0; iy < Ny; iy++)
        {
            for(ix = 0; ix < Nx; ix++)
            {
                ic[0] = ix; ic[1] = iy;
                for(i = 0; i < 3; ++i)
                    for(j = 0; j < 3; ++j)
                {
                    {
                       for(k = 0; k < n_chem; k++)
                       {
                           p_cell[ix][iy][k].set[i][j] = NO;
                           p_cell[ix][iy][k].stencil[i][j] = 0.0;
                       }
                    }
                }
                for(k = 0; k < n_chem; k++)
                    p_cell[ix][iy][k].r_side = 0.0;
                if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                   continue;

                Chem_Crank_Nicolson_sten(dt, ic, rd_field, p_cell[ix][iy],n_chem);
            }
        }

        for(k = 0; k < n_chem; k++)
        {
            for(j = 0; j < Ny; j++)
            {
                for(i = 0; i < Nx; i++)
                {
                    assign(&tran_cell[i][j], &p_cell[i][j][k], sizeof(P_CELL2D));
                }
            }

            petsc_cell_center_parab_solver(newwv, tran_cell, 
                    iupper, ilower, Nx, Ny, I_to_ijk, IJK_to_I, soln);

            for(j = 0; j < Ny; j++)
            {
                ic[1] = j; 
                for(i = 0; i < Nx; i++)
                {
                    ic[0] = i; 
                    Chem(Rect_state(ic,newwv))[k] = soln[i][j];
                }
            }
        }
   
        free_these(5,p_cell, n_dist, I_to_ijk, IJK_to_I, tran_cell);
        free(soln);

#endif /*defined (PETSC)*/
}

/* Here, no flux boundary condition is used for the chemicals on the Neumann_boundary.
 * On the Derichlet boundary, the constant flow boundary condition is used if the
 * Chemicals are not supplied from outside region. 
*/
LOCAL void Chem_Crank_Nicolson_sten(
        float     dt,
        int       *ic,
        RD_Field_Stencil *rd_field,
        P_CELL2D  *p_cell,
        int       n_chem)
{
        Wave      *newwv = rd_field->newwave;
        Wave      *wv = rd_field->wave;
        TRI_GRID  *tg = wave_tri_soln(wv)->tri_grid;
        TRI_GRID  *ntg = wave_tri_soln(newwv)->tri_grid;
        Front     *fr = rd_field->fr, *newfr = rd_field->newfr;
        COMPONENT comp, ncomp, nncomp;
        int       dir, ndir, i, j, k, l, end = 1; // 2nd order central diff
        static    Stencil *sten = NULL, *nsten;
        int       **icoords, **nicoords;
        int       iperm[2] = {0, 1}; // No need to permute
        int       label_storex[3], *labelx, label_storey[3], *labely, irregx, irregy;
        float     *dh = wv->rect_grid->h;
        BLK_EDGE  *blk_edge = ntg->blk_edge;
        float     c_xx, c_yy; // they are actually c_xx + c_yy.
        int       debug_flag = NO;
        float     values[3][3], *mu_chem;

        mu_chem = rd_field->chem_sten->mu_chem;
        labelx = label_storex + 1;
        labely = label_storey + 1;
        comp = Rect_comp(ic,wv);

        dir = 0; ndir = 1;

        if (sten == NULL)
        {
            sten = alloc_stencil(3,fr);
            nsten = alloc_stencil(3,fr);
        }

        icoords = sten->icoords;
        sten->st[0] = Rect_state(ic,wv);

        sten->fr  = fr;
        sten->wave = wv;
        sten->newfr = newfr; sten->newwave = newwv;

        nsten->fr  = fr;
        nsten->wave = wv;
        nsten->newfr = newfr; nsten->newwave = newwv;
        nicoords = nsten->icoords;
        nsten->st[0] = Rect_state(ic,wv);

        for(j = 0; j < dim; j++)
            nicoords[0][j] = icoords[0][j] = ic[j];

        for (j = -end; j <= end; ++j)
        {
            icoords[j][dir] = ic[dir] + j;
            icoords[j][(dir+1)%dim] = ic[(dir+1)%dim];
        }
        for (j = -end; j <= end; ++j)
        {
            nicoords[j][dir] = ic[dir];
            nicoords[j][(dir+1)%dim] = ic[(dir+1)%dim] + j;
        }

        for (j = 0; j < dim; ++j)
        {
            if (j == dir)
            {
                Coords(nsten->p[0])[dir] = Coords(sten->p[0])[dir] =
                         cell_center(ic[dir],dir,wv->rect_grid);
                for (l = 1; l <= end; ++l)
                {
                    Coords(sten->p[-l])[dir] =
                             Coords(sten->p[0])[dir]-l*dh[j];
                    Coords(sten->p[l])[dir]  =
                             Coords(sten->p[0])[dir]+l*dh[j];
                }
                for (l = 1; l <= end; ++l)
                {
                    Coords(nsten->p[-l])[dir] =
                             Coords(sten->p[0])[dir];
                    Coords(nsten->p[l])[dir]  =
                             Coords(sten->p[0])[dir];
                }
            }
            else
            {
                for (l = -end; l <= end; ++l)
                    Coords(sten->p[l])[j] = Rect_coords(ic,wv)[j];
                for (l = -end; l <= end; ++l)
                    Coords(nsten->p[l])[j] = Rect_coords(ic,wv)[j] + l*dh[j];
            }
        }

        prev_side = WEST;
        next_side = EAST;
        ncomp = set_NS_stencil_comps_and_crosses(dir,dir,iperm,sten,end);

        prev_side = SOUTH;
        next_side = NORTH;
        ncomp = set_NS_stencil_comps_and_crosses(ndir,ndir,iperm,nsten,end);

        for (l = -end; l <= end; ++l)
            labelx[l] = YES;
        irregx = Crank_N_sten_reg(dir,sten,labelx, NULL);

        for (l = -end; l <= end; ++l)
        {
            if(labelx[l] == YES)
            {
                for(i = 0; i < n_chem; i++)
                    p_cell[i].set[l+end][1] = YES;
            }
        }

        for (l = -end; l <= end; ++l)
            labely[l] = YES;
        irregy = Crank_N_sten_reg(ndir,nsten,labely, NULL);
        for (l = -end; l <= end; ++l)
        {
            if(labely[l] == YES)
            {
                for(i = 0; i < n_chem; i++)
                p_cell[i].set[1][l+end] = YES;
            }
        }

        for(i = 0; i < n_chem; i++)
        {
            c_xx = (Chem(sten->st[1])[i] - 2.0*Chem(sten->st[0])[i] +
                     Chem(sten->st[-1])[i])/sqr(dh[dir]);
            c_yy = (Chem(nsten->st[1])[i] - 2.0*Chem(nsten->st[0])[i] +
                     Chem(nsten->st[-1])[i])/sqr(dh[ndir]);

            p_cell[i].r_side = Chem(sten->st[0])[i] + 
                 0.5*dt*mu_chem[i]*(c_xx + c_yy); 
            p_cell[i].r_side -= dt*Rect_chem_der(ic,rd_field->chem_sten)[i];

            values[0][1] = values[2][1] = -0.5*dt/sqr(dh[dir])*mu_chem[i];
            values[1][2] = values[1][0] = -0.5*dt/sqr(dh[ndir])*mu_chem[i];
            values[1][1] = dt/sqr(dh[dir])*mu_chem[i]+
                       dt/sqr(dh[ndir])*mu_chem[i] + 1.0;

            // Account for boundary conditions
            if(irregx == YES)
            {
                for(l = 0; l < 3; l+=2)
                {
                    if(label_storex[l] == N_SURF)
                    {
                        values[1][1] -= 0.5*dt/sqr(dh[dir])*mu_chem[i]; 
                    }
                    else if(label_storex[l] == D_SURF)
                    {
                        values[1][1] -= 0.5*dt/sqr(dh[dir])*mu_chem[i]; 
                    }
                }
            }
            if(irregy == YES)
            {
                for(l = 0; l < 3; l+=2)
                {
                    if(label_storey[l] == N_SURF)
                    {
                        values[1][1] -= 0.5*dt/sqr(dh[ndir])*mu_chem[i]; 
                    }
                    else if(label_storey[l] == D_SURF)
                    {
                        values[1][1] -= 0.5*dt/sqr(dh[ndir])*mu_chem[i]; 
                    }
                }
            }
            //////
            for (l = 0; l < 3; ++l)
            {
                p_cell[i].stencil[1][l] = values[1][l];
                p_cell[i].stencil[l][1] = values[l][1];
            }
        }
}


LIB_LOCAL int g_React_Diff_solver(
        int        *iperm,
        float      *dh,
        float      dt,
        Wave       *wv,
        Wave       *nwv,
        Front      *fr,
        Front      *nfr,
        COMPONENT  max_comp,
        Chem_Stencil  *chem_sten)
{
        RECT_GRID  *gr = fr->rect_grid;
        int        k, i;
        RD_Field_Stencil *rd_field = NULL;
        // int        edgexmax, edgeymax, edgex_n_segs, edgey_n_segs;
        int        ix, iy;
        int        imin[3], imax[3], idirs[3];
        int        icoords[MAXD], i0min, i0max, vsize, i1;
        float      dir[MAXD];
        int        nrad;
        Vec_NS_Gas  *vgas = NULL;
        Vec_NS_Gas  *Tvgas[2] = {NULL, NULL};    /* Unsplit, for cross direvative.
                                                  * Tangential upper and lower state.
                                                  */
        DEBUG_ENTER(g_React_Diff_solver)

        /* Set local globals */
        dim = gr->dim;
        for (i = 0; i < dim; ++i)
        {
            gmax[i] = gr->gmax[i];
            h[i] = gr->h[i];
            L[i] = gr->L[i];
            VL[i] = gr->VL[i];
            VU[i] = gr->VU[i];
            lbuf[i] = gr->lbuf[i];
            ubuf[i] = gr->ubuf[i];
        }
        intfc = nfr->interf;
        ext_comp = exterior_component(intfc);
        /* END set local globals */

        rd_field = alloc_RD_field_space(rd_field,wv);
        rd_field->chem_sten = chem_sten;

        // Temp. set to use 2nd order scheme */
        nrad = (rd_field->npts_sten - rd_field->npts_sten%2)/2;

        rd_field->rect_gr = wv->rect_grid;
        rd_field->fr = fr;
        rd_field->newfr = nfr;
        rd_field->wave = wv;
        rd_field->newwave = nwv;

        // cell_edge_vel(rd_field);

        // grid_chemical(rd_field,rd_field->wave,rd_field->fr,0);

        advec_diff_chem(dt,rd_field);

        // RK step
        Runge_Kutta_step(rd_field,nwv,nfr,dt);

        free_these(2,rd_field->u_states[0],rd_field->u_states[1]);
        free(rd_field->c);
        free(rd_field);

        DEBUG_LEAVE(g_React_Diff_solver)
}

// RK step, 3-order TVD Runge-Kutta method
LOCAL void Runge_Kutta_step(
        RD_Field_Stencil *rd_field,
        Wave             *wv,
        Front            *fr,
        float            dt)
{
        int             smin[MAXD], smax[MAXD];
        int             ix, iy, ic[MAXD], i, j, k, indx;
        COMPONENT       comp;
        int             deb_flag = NO;
        float           coeff = 0.001;
        float           u0, u1, u2, u3;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        for(indx = 0; indx < rd_field->chem_sten->n_chem; indx++)
        {
            for (iy = smin[1];  iy < smax[1];  ++iy)
            {
                for (ix = smin[0];  ix < smax[0];  ++ix)
                {
                    ic[0] = ix; ic[1] = iy;

                    comp = Rect_comp(ic,wv);
                    if(comp != ext_comp && (! is_excluded_comp(comp,fr->interf)))
                    {
                        u0 = Chem(Rect_state(ic,wv))[indx];
                        u1 = u0 + dt*coeff*u0;
                        u2 = 3.0/4.0*u0 + 1.0/4.0*u1 + 1.0/4.0*dt*coeff*u1;
                        u3 = 1.0/3.0*u0 + 2.0/3.0*u2 + 2.0/3.0*dt*coeff*u2;  
                        Chem(Rect_state(ic,wv))[indx] = u3;
                    }
                }
            }
        }
}


LOCAL void petsc_cell_center_parab_solver(
        Wave              *wv_blk,
        P_CELL2D          **p_cell,
        int               iupper,
        int               ilower,
        int               Nx,
        int               Ny,
        int               **I_to_ijk,
        int               **IJK_to_I,
        float             **soln)
{
        int             ix, iy, ic[MAXD], i, j;
        int             I;
        int             idxm[1], sten_index[3], indices;
        float           norm,min,max,x,y,*coords;
        float           val[1], values[3][3];

#if defined (PETSC)
            /*PETSC Variables*/
        KSP                  ksp;
        Mat                  A;
        Vec                  phi, B, C;
        PetscScalar          *vphi;
        MatNullSpace         nullsp;
        PC                   pc;
        int                  its;
        PetscReal            norm_B, norm_C;
#endif /*defined (PETSC)*/

#if defined(__MPI__)
        MPI_Comm_size(PETSC_COMM_WORLD,&comm_size);
#endif /* defined(__MPI__) */

#if defined(__MPI__)
#else
        MatCreate(PETSC_COMM_WORLD, &A);
        MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,iupper-ilower+1,iupper-ilower+1);
        MatSetFromOptions(A);
#endif // if defined(__MPI__) //

        VecCreate(PETSC_COMM_WORLD,&B);
        VecSetSizes(B,(iupper-ilower + 1),PETSC_DECIDE);
        VecSetFromOptions(B);
        VecDuplicate(B,&phi);
        VecDuplicate(B,&C);


        for (I = ilower; I <= iupper; I++)
        {
            for(i = 0; i < dim; i++)
                ic[i] = I_to_ijk[I-ilower][i];
            idxm[0] = I;
            ix = ic[0]; iy = ic[1];
            if(Rect_blk_edge(ic,wv_blk).ctype == EXTERNAL)
                continue;

            for(i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                    values[i][j] = p_cell[ic[0]][ic[1]].stencil[i][j];
            }
            for( i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                {
                    sten_index[i] = IJK_to_I[ix+i+1][iy+j+1];
                    if(p_cell[ic[0]][ic[1]].set[i][j] == YES)
                    {
                        MatSetValues(A,1,idxm,1,&sten_index[i],
                            &values[i][j],INSERT_VALUES);
                    // TMP
                    // printf("cell ic[%d %d] set %d\n", ic[0], ic[1], sten_index[i]);
                    }
                }
            }
            val[0] = p_cell[ic[0]][ic[1]].r_side;
            VecSetValues(B,1,idxm,val,INSERT_VALUES);
        }

        MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
        MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);

        VecAssemblyBegin(B);
        VecAssemblyEnd(B);

        // MatNullSpaceCreate(PETSC_COMM_WORLD,PETSC_TRUE,0,PETSC_NULL,&nullsp);
        KSPCreate(PETSC_COMM_WORLD,&ksp);
        KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN);
        KSPGetPC(ksp,&pc);
        PCSetType(pc,PCJACOBI);

        // KSPSetNullSpace(ksp,nullsp);
        // KSPSetType(ksp,KSPBCGS);
        KSPSetTolerances(ksp,1.e-12,PETSC_DEFAULT,PETSC_DEFAULT,700);
        KSPSetComputeSingularValues(ksp, PETSC_TRUE);

        KSPSetFromOptions(ksp);
        // KSPSetUp(ksp);

        KSPSolve(ksp,B,phi);
        KSPGetIterationNumber(ksp,&its);
        KSPGetResidualNorm(ksp,&norm);

        KSPComputeExtremeSingularValues(ksp, &max, &min);
        // printf("\n X-component of velocity:\n"
        //     " Number of iteration  is %d,"
        //     "residual norm = %6.5e,extreme test SV is [%7.6e ; %7.6e]\n",its,norm,min,max);

        // MatView(A,PETSC_VIEWER_STDOUT_WORLD);
        // VecView(B,PETSC_VIEWER_STDOUT_WORLD);
        for (iy = 0; iy < Ny; iy++)
        {
            for (ix = 0; ix < Nx; ix++)
            {
                indices = IJK_to_I[ix+2][iy+2];
                ic[0] = ix; ic[1] = iy;

                if (indices >= 0)
                {
                   VecGetArray(phi,&vphi);
                   soln[ix][iy] = vphi[indices-ilower];

                   // printf("ic[%d %d] x-comp vel = %g\n",
                   //    ix, iy, vphi[indices-ilower]);
                   VecRestoreArray(phi,&vphi);
                }
            }
        }

        if(debugging("NS_vel"))
        {
            MatMult(A,phi,C);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            // printf("-------A*phi, print RHS----\n");
            // VecView(B,PETSC_VIEWER_STDOUT_WORLD);
            printf("After CN-X_component, norm B = %g, norm (A*phi) = %g\n",
                         norm_B, norm_C);
        }
        // END TEST

        VecDestroy(B);
        VecDestroy(phi);
        VecDestroy(C);
        MatDestroy(A);
        KSPDestroy(ksp);
        // MatNullSpaceDestroy(nullsp);
}


