/*
*                               ghyploc.h:
*
*       Copyright 1999 by The University at Stony Brook, All rights reserved.
*
*/

#if !defined(_GHYPLOC_H)
#define _GHYP_H

#include <gdecs/gdecs.h>

enum __EDGE_TYPE{
      P_BDRY = 0, // physical boundary
      REG_E       // interior edge
};
typedef enum __EDGE_TYPE _EDGE_TYPE;

struct _Gradient_Var{
      float dvar[MAXD];
};
typedef struct _Gradient_Var Gradient_Var;

struct _Vec_NS_Gas{
        float       *rho;       /* Mass Density */
        float       *p;         /* pressure */
        float       **m;        /* momentum density */
                                /* Split: 0 = normal; Unsplit: No permutation */
        float       **v;        /* velocity */
        Locstate    *state;
        float       **coords;
        int         nprms;
        int         *prms_jmp;
        float       **chem;        /* chemical compoents, chem[0], chem[1] for comp0, comp1 etc*/ 
        int         n_chem;
        float       *mu_chem;      /* diffusion coefficient of chemical compoents */
        struct {
            bool rho;
            bool p;
            bool m;
            bool v;
            bool state;
            bool coords;
            bool vst;
            bool prms_jmp;
            bool chem;
            bool mu_chem;
        } alloc;
        struct {
            bool rho;
            bool p;
            bool m;
            bool v;
            bool state;
            bool coords;
            bool vst;
            bool prms_jmp;
            bool chem;
            bool mu_chem;
        } set;
};
typedef struct _Vec_NS_Gas Vec_NS_Gas;

#define Vec_NS_Gas_field_allocated(vst,ptr)        (vst)->alloc.ptr
#define Vec_NS_Gas_field_set(vst,ptr)              (vst)->set.ptr

struct _Vec_NS_Src{
        float       *rho;       /* Mass Density */
        float       *p;         /* pressure */
                                /* Split: 0 = normal; Unsplit: No permutation */
        float       **v;        /* velocity */
        float    *radii, rmin;

        struct{
            bool rho;
            bool p;
            bool v;
            bool radii;
            bool src;
        } alloc;
};
typedef struct _Vec_NS_Src Vec_NS_Src;

struct _NS_Field_Stencil{
        // edge type flags are stored here
        // _EDGE_TYPE    *edge_type[MAXD], *cg_edge_type[MAXD];

        // viscosity and density before advection
        float       *rho,   *cg_rho;
        float       *mu,    *cg_mu;

        // Reconstructed convective derivatives (U\cdot \grad U)^{n+1/2}
        VEL_ST      *con_der, *cg_con_der;

        /* States for saving velocity on staggered grids - no more */
        /* On the centers of the cell edges */
        VEL_ST      *u_states[MAXD], *cg_u_states[MAXD];

        /* On the centers of the cell */
        // To save Crank-Nicolson step velocity
        VEL_ST      *u_star, *cg_u_star;

        /* Bell Colella central grid */
        /* On the centers of left side of cell edges */
        // indexed by edge
        // uL_states[0] : horizontal line L side states,
        // uL_states[1] : vertical line L side states
        VEL_ST      *uL_states[MAXD], *cg_uL_states[MAXD];

        /* On the centers of right side of cell edges */
        // indexed by edge
        // uR_states[0] : horizontal line R side states,
        // uR_states[1] : vertical line R side states
        VEL_ST      *uR_states[MAXD], *cg_uR_states[MAXD];

        /* Normal component of velocity on cell edges */
        float       *n_vel_edge[MAXD], *cg_n_vel_edge[MAXD];

        float       *hr_st, *cg_hr_st; // rho on horizontal cell edge after upwind
        float       *vr_st, *cg_vr_st; // rho on vertical cell edge after upwind

        Gradient_Var    *Gp, *cg_Gp;     // Gradient P
        float           *vlap[MAXD], *cg_vlap[MAXD];  // viscosity term
        float           *phi, *cg_phi;   // Projection scalar
        float           *rho_recip[2], *cg_rho_recip[2];   // On cell edge
                                         // 0 --- horizonal line
                                         // 1 --- vertical line
        struct {
            // bool    edge_type;
            bool    rho;
            bool    mu;
            bool    con_der;
            bool    n_vel_edge;
            bool    u_states;
            bool    uL_states;
            bool    uR_states;
            bool    Gp;
            bool    phi;
            bool    vlap;
            bool    I_to_ijk;
            bool    IJK_to_I;
            bool    u_star;
            bool    rho_recip;
            bool    vr_st;
            bool    hr_st;
        } alloc;

        /*  For regular grid points */
        Vec_NS_Gas  *vgas;
        Vec_NS_Gas  *Tvgas[2];        /* Unsplit, for G(p). Laplacian (viscosity) */
        Vec_Src     *src;             // For source terms

        RECT_GRID   *rect_gr;
        int         nnx, nny, nnz;   /* number of blocks in direction */
        int         npts_sten;

        /* irregular point stencil. */
        Stencil     *pt_sten;
        Front       *fr, *newfr;
        Wave        *wave, *newwave;

        int         **I_to_ijk;   /* local  i, j, k, local I */
        int         **IJK_to_I;   /* global I, J, K, global I */
        int         ilower, iupper;

        /* Density and viscosity used*/
        int         n_eos;
        float       **den_vis;
        COMPONENT   *comps;
        // Chemical component
        int         n_chem;
        struct _Chem_Stencil *chem_sten;
};
typedef struct _NS_Field_Stencil NS_Field_Stencil;

struct _Chem_Stencil { 
        float       **hel_st, **cg_hel_st; // chemical state on horizontal edge left side
        byte        *hedge_l_storage;
        float       **her_st, **cg_her_st; // chemical state on horizontal edge right side
        byte        *hedge_r_storage;

        float       **vel_st, **cg_vel_st; // chemical state on vertical edge left side
        byte        *vedge_l_storage;
        float       **ver_st, **cg_ver_st; // chemical state on vertical edge right side
        byte        *vedge_r_storage;

        float       **vem_st, **cg_vem_st; // chemical state on vertical edge
        byte        *vedge_m_storage;
        float       **hem_st, **cg_hem_st; // chemical state on horizontal edge
        byte        *hedge_m_storage;

        // Reconstructed convective derivatives (div (cU))^{n+1/2}
        float       **con_der, **cg_con_der;  
        byte        *con_der_storage; 

        int         nnx, nny, nnz;   /* number of blocks in direction */
        int         n_chem;
        float       *mu_chem;  
};
typedef struct _Chem_Stencil Chem_Stencil;

#define h_line_chem_L(icoord,field)                                \
        ((field)->cg_hel_st[nn_vx_indx((icoord),(field),(0))])
#define h_line_chem_R(icoord,field)                                \
        ((field)->cg_her_st[nn_vx_indx((icoord),(field),(0))])
#define v_line_chem_L(icoord,field)                                \
        ((field)->cg_vel_st[nn_vy_indx((icoord),(field),(0))])
#define v_line_chem_R(icoord,field)                                \
        ((field)->cg_ver_st[nn_vy_indx((icoord),(field),(0))])
#define h_line_chem(icoord,field)                                  \
        ((field)->cg_hem_st[nn_vx_indx((icoord),(field),(0))])
#define v_line_chem(icoord,field)                                  \
        ((field)->cg_vem_st[nn_vy_indx((icoord),(field),(0))])
#define Rect_chem_der(ic,field)                                    \
        ((field)->cg_con_der[(ic)[1]*(field)->nnx + (ic)[0]])

//                                  Y ^
// axis define for vertical line  --L | R ----> X
//             for horizontal line    | R
//                                    |--------> X
//                                    | L
// L refers to the lower side, and R refers
// to upper side in the axis direction

// For velocity field on cell edges

#define h_line_vel_L(icoord,field)                                    \
        ((field)->cg_uL_states[0][nn_vx_indx((icoord),(field),(0))])
#define h_line_vel_R(icoord,field)                                    \
        ((field)->cg_uR_states[0][nn_vx_indx((icoord),(field),(0))])
#define v_line_vel_L(icoord,field)                                    \
        ((field)->cg_uL_states[1][nn_vy_indx((icoord),(field),(0))])
#define v_line_vel_R(icoord,field)                                    \
        ((field)->cg_uR_states[1][nn_vy_indx((icoord),(field),(0))])

// Normal velocity on cell edges
#define  n_vel_h_edge(icoord,field)                                  \
         ((field)->cg_n_vel_edge[0][nn_vx_indx((icoord),(field),(0))])
#define  n_vel_v_edge(icoord,field)                                  \
         ((field)->cg_n_vel_edge[1][nn_vy_indx((icoord),(field),(0))])

//
#define field_velx_s(icoords,field)                                      \
        Reg_grid_xvel((icoords),(field),0)
#define field_velx_n(icoords,field)                                       \
        Reg_grid_xvel((icoords),(field),1)

#define field_vely_w(icoords,field)                                       \
        Reg_grid_yvel((icoords),(field),0)
#define field_vely_e(icoords,field)                                       \
        Reg_grid_yvel((icoords),(field),1)

// For velocity field on cell edges (not used)
#define Reg_grid_xvel(ic,field,i)      ((field)->cg_u_states[0][nn_vx_indx((ic),(field),(i))])
#define Reg_grid_yvel(ic,field,i)      ((field)->cg_u_states[1][nn_vy_indx((ic),(field),(i))])

// velocity on cell edges, computed after upwinding and limiting
#define h_line_vel(ic,field)      ((field)->cg_u_states[0][nn_vx_indx((ic),(field),(0))])
#define v_line_vel(ic,field)      ((field)->cg_u_states[1][nn_vy_indx((ic),(field),(0))])

// cell edge flags
// #define h_edge_flag(ic,field)      ((field)->cg_edge_type[0][nn_vx_indx((ic),(field),(0))])
// #define v_edge_flag(ic,field)      ((field)->cg_edge_type[1][nn_vy_indx((ic),(field),(0))])

//#define nn_vx_indx(ic,field,i) ( ((ic)[1]+(i))*((field)->rect_gr->gmax[0]) + (ic)[0] )
//#define nn_vy_indx(ic,field,i) ( ((ic)[1])*((field)->rect_gr->gmax[0]+1) + (ic)[0] + (i) )
#define nn_vx_indx(ic,field,i) ( ((ic)[1]+(i))*((field)->nnx) + (ic)[0] )
#define nn_vy_indx(ic,field,i) ( ((ic)[1])*((field)->nnx+1) + (ic)[0] + (i) )

/* P Field Gradient on cell centers */
// #define Rect_gradP(ic,field)                                          \
//         ( ((field)->cg_Gp + (ic)[1]*((field)->nnx) + (ic)[0] )->dvar )
#define Rect_gradP(ic,field)                                          \
         ((field)->cg_Gp[(ic)[1]*(field)->nnx + (ic)[0]].dvar )

// diffusion of the flow field on cell centers
#define Rect_diff(ic,i,field)                                         \
          ((field)->cg_vlap[i][(ic)[1]*(field)->nnx + (ic)[0]])

/* phi scalar field on cell centers */
#define Rect_phi(ic,field)                                            \
         ((field)->cg_phi[(ic)[1]*(field)->nnx + (ic)[0]] )
/* convective derivative field on cell centers */
#define Rect_con_der(ic,field)                                        \
         ((field)->cg_con_der[(ic)[1]*(field)->nnx + (ic)[0]] )
/* Average density and viscosity on cell centers */
#define Rect_den(ic,field)                                            \
         ((field)->cg_rho[(ic)[1]*(field)->nnx + (ic)[0]] )
#define Rect_visc(ic,field)                                           \
         ((field)->cg_mu[(ic)[1]*(field)->nnx + (ic)[0]] )

/* u^star velocity field on cell centers */
#define Rect_u_star(ic,field)                                        \
         ((field)->cg_u_star[(ic)[1]*(field)->nnx + (ic)[0]] )

// 1/rho on cell edges, computed by averaging
#define h_line_rhor(ic,field)      ((field)->cg_rho_recip[0][nn_vx_indx((ic),(field),(0))])
#define v_line_rhor(ic,field)      ((field)->cg_rho_recip[1][nn_vy_indx((ic),(field),(0))])

struct _Upwind_NS_Index {       /* Indices of states stored in u             */
        int  v[3];
        int  density;
        int  pressure;
        int  chem[20];          // static, should check 
};
typedef struct _Upwind_NS_Index Upwind_NS_Index;

struct _Upwind_NS{
        Vec_NS_Gas  *vst;
        Vec_NS_Gas  *Tvst[2];    /* Unsplit, for G(p). Laplacian (viscosity) */

        Upwind_NS_Index  index;
        int              vsize;
        int              dim;
        int              worksp_len;
        int              nfloats;
        int              offset;
        int              sten_rad;
        float            ***worksp;
        bool             is_src; /* YES if source terms are present */

        float   **CellEdge;      /* Coordinates of cell edges    */
        float   **coords;        /* Coordinates of cell centers  */

        float   **u;             /* Initial state data, u[], density, pressure from vst */

        Vec_Src *src;            /* Source terms */
        float   **source;        /* Array form of src */

                /* State reconstruction data */
        float   **central;       /* Central differences */
        float   **forward;       /* Forward differences */
        float   **backward;      /* Backward differences */

        float   **source_sum;    /* Source term contributions from 1/2 step */

        float   **sgn;           /* Signs of central differences */
        float   **du;            /* limited state slope */

        Vec_NS_Gas VLst;         /* Vec Gas form of left half step state    */
        float   **uL;            /* Half step evolved left state              */

        Vec_NS_Gas VRst;         /* Vec Gas form of right half step state    */
        float   **uR;            /* Half step evolved right state              */

        Vec_NS_Gas VMst;         /* Vec Gas form from mid state    */
        float   **uM;            /* Mid state              */
        float   *mu_chem;        /* Chemical diffusion coefficient */

        float   **gravity;       /* array of cell centered gravities. */
                                 /* For unsplit multiD. solver */
        struct {
            bool mu_chem;
            bool source;
            bool gravity;
        } alloc;
        struct {
            bool mu_chem;
            bool source;
            bool gravity;
        } set;
};
typedef struct _Upwind_NS   Upwind_NS;

struct _NS_PT_Stencil{
};
typedef struct _NS_PT_Stencil NS_PT_Stencil;


IMPORT  void     sten_pt_loc_in_dir(int*,int,int,Wave*,int,Stencil*);
IMPORT  int      velocity_from_boundary(Stencil*,CRXING*,
                       POINT*,COMPONENT,Locstate,float*);
IMPORT  COMPONENT  set_NS_stencil_comps_and_crosses(int,int,int*,Stencil*,int);
IMPORT  int      g_React_Diff_solver(int*,float*,float,Wave*,Wave*,Front*,Front*,
                       COMPONENT,Chem_Stencil*);
IMPORT  bool     Crank_N_sten_reg(int,Stencil*,int*,NS_Field_Stencil*);
#if defined(PETSC)
IMPORT  void      create_global_element_list(Front*,int,int,int*,int**,int**);
#endif /* if defined(PETSC) */

// #include <ghyp/ghypprotos.h>

#endif /* !defined(_GHYPLOC_H) */

