#if defined(TWOD) || defined(THREED)
#define DEBUG_STRING    "init_jet"
#include <ginit/ginit.h>

#if defined(TWOD) /* if defined(TWOD) */
LOCAL   float      x_range, hheight;
LOCAL   float      sine_len(float,POINTER);
LOCAL   void       insert_sine_curve_section(BOND*,CURVE*,float,float,float*,float,float,Front*);
// LOCAL   void       init_blood_bub(INIT_DATA*,INIT_PHYSICS*);

LOCAL   void       init_obstacle_bub(INIT_DATA*,INIT_PHYSICS*,CURVE*,CURVE*);

#if defined(__cplusplus)
extern "C" {
#endif /* defined(__cplusplus) */
    LOCAL   LSODE_FUNC  isine_len;
#if defined(__cplusplus)
}
#endif /* defined(__cplusplus) */


EXPORT void init_blood_vessel(
        INIT_DATA       *init,
	INIT_PHYSICS    *ip)
{
        Front      *front = ip->root->front;
        INTERFACE  *intfc = front->interf;
        RECT_GRID  *gr = front->rect_grid;
        float      *L = gr->GL;
        float      *U = gr->GU;
        float      *h = gr->h;
        float      coords[MAXD], scoords[MAXD];
        float      lv_s, lv_e, lv_h; 
        float      uv_s, uv_e, uv_h; 
        char       s[1024];
        NODE       *lns, *lne;
        CURVE      *lwall = NULL;
        NODE       *uns, *une;
        CURVE      *uwall = NULL;
	Gas_param  *paramsa;
        int        nprms, i;
        Gas_param  **prms_list; /* = gas_params_list(intfc);*/
        int        init_bub = NO;
        float      bub_rad, bub_cen[MAXD];
        NODE       *bns, *bne;
        CURVE      *cv;
	Gas_param  *paramsb;
        int        set_up_side = NO, set_low_side = NO;
 
        debug_print("jet","Entered init_blood_vessel()\n");

	printf("**************************\n");
	printf("Entered init_blood_vessel\n");

        set_obstacle_comp_type(comp_type(COMPOBST),front);

        if(debugging("pipe_only"))
        {
	    (void) prompt_for_eos_params(init,ip,YES,"");
	    paramsa = prompt_for_eos_params(init,ip,YES,"\n\tfor the vessel blood");
            prompt_for_ambient_state(comp_type(COMPA),paramsa,
	                " for the vessel blood",front,init);
            paramsa->dens = Dens(Ambient(comp_type(COMPA)));
            paramsa->comp = COMPA;
            nprms = return_params_list(&prms_list);
            for (i=0; i<nprms; i++)
                prms_list[i]->eos->_NS_eos = YES;
	    // rect_boundary_type(intfc,0,1) = DIRICHLET_BOUNDARY;
	    // rect_boundary_type(intfc,1,0) = NEUMANN_BOUNDARY;
	    // rect_boundary_type(intfc,1,1) = NEUMANN_BOUNDARY;
	    // rect_boundary_type(intfc,1,0) = MIXED_TYPE_BOUNDARY;
	    // rect_boundary_type(intfc,1,1) = MIXED_TYPE_BOUNDARY;
            return;
        }

        screen("\nBlood vessel geometry enters from below.\n");
        //Read upper boundary geometry 
        screen("The lower vessel wall is assumed to be parabola (y(default), n): ");
        (void) Gets(s);
        if (strncmp(s,"y",1) == 0 || strncmp(s,"Y",1) == 0)
        {
            set_low_side = YES;
            screen("The lower vessel is started from the left of the domain (percent): ");
            (void) Scanf("%f\n",&lv_s);
            if ((lv_s <= 0.0) || (lv_s >= 1.0))
            {
                screen("ERROR in  init_blood_vessel()"
                  "invalid starting point %g, (outside domain range)\n",lv_s);
                screen("should be between 0 and 1.0\n");
                clean_up(ERROR);
            }

            screen("The lower vessel is ended at (percent): ");
            (void) Scanf("%f\n",&lv_e);
            if ((lv_e >= 1.0) || (lv_e <= lv_s))
            {
                screen("ERROR in  init_blood_vessel()"
                  "invalid ending point %g, (outside domain range)\n",lv_e);
                screen("should be between 0 and 1.0 and greater than %g\n", lv_s);
                clean_up(ERROR);
            }

            screen("The maximum height of the lower vessel is: ");
            (void) Scanf("%f\n",&lv_h);
            if ((lv_h < 0) || (lv_h > (U[1]-L[1])))
            {
                screen("ERROR in  init_blood_vessel()"
                  "invalid maximum height %g, (outside domain range)\n",lv_h);
                screen("should be between %g and %g\n", L[1], U[1]);
                clean_up(ERROR);
            }

            // Draw the lower boundary
            scoords[0] = L[0] + lv_s*(U[0]-L[0]);
            scoords[1] = L[1];
            lns = make_node(Point(scoords));
            set_is_bdry(lns);
            node_type(lns) = FIXED_NODE;
        
            coords[0] = L[0] + lv_e*(U[0]-L[0]);
            coords[1] = L[1];
            lne = make_node(Point(coords));
            set_is_bdry(lne);
            node_type(lne) = FIXED_NODE;

            lwall = make_curve(COMPA,COMPOBST,lns,lne);
            wave_type(lwall) = NEUMANN_BOUNDARY;
            start_status(lwall) = end_status(lwall) = FIXED;

            x_range = coords[0]-scoords[0];
            hheight = lv_h;
            insert_sine_curve_section(lwall->last, lwall, 0.0, x_range, scoords, lv_h,
                 x_range,front);
            // set_is_bdry(lwall);

            prompt_for_adherence_coeff(Hyper_surf(lwall));
        }

        //Read upper boundary geometry 
        screen("The upper vessel wall is assumed to be parabola (y(default), n): ");
        (void) Gets(s);
        if (strncmp(s,"y",1) == 0 || strncmp(s,"Y",1) == 0)
        {
            set_up_side = YES;
            screen("The upper vessel is started from the left of the domain (percent): ");
            (void) Scanf("%f\n",&uv_s);
            if ((uv_s <= 0.0) || (uv_s >= 1.0))
            {
                screen("ERROR in  init_blood_vessel()"
                  "invalid upper starting point %g, (outside domain range)\n",uv_s);
                screen("should be between 0 and 1.0\n");
                clean_up(ERROR);
            }

            screen("The upper vessel is ended at (percent): ");
            (void) Scanf("%f\n",&uv_e);
            if ((uv_e >= 1.0) || (uv_e <= uv_s))
            {
                screen("ERROR in  init_blood_vessel()"
                  "invalid ending point %g, (outside domain range)\n",uv_e);
                screen("should be between 0 and 1.0 and greater than %g\n", uv_s);
                clean_up(ERROR);
            }

            screen("The minimum height of the upper vessel is (measured from upper boundary): ");
            (void) Scanf("%f\n",&uv_h);
            if ((uv_h > 0) || (uv_h+U[1] < L[1]))
            {
                screen("ERROR in  init_blood_vessel()"
                  "invalid minimum height %g, (outside domain range)\n",uv_h+U[1]);
                screen("should be between %g and %g\n", L[1], U[1]);
                clean_up(ERROR);
            }
        
            // Draw upper boundary geometry
            scoords[0] = L[0] + uv_s*(U[0]-L[0]);
            scoords[1] = U[1];
            uns = make_node(Point(scoords));
            set_is_bdry(uns);
            node_type(uns) = FIXED_NODE;
        
            coords[0] = L[0] + uv_e*(U[0]-L[0]);
            coords[1] = U[1];
            une = make_node(Point(coords));
            set_is_bdry(une);
            node_type(une) = FIXED_NODE;

            uwall = make_curve(COMPOBST,COMPA,uns,une);
            wave_type(uwall) = NEUMANN_BOUNDARY;
            start_status(uwall) = end_status(uwall) = FIXED;
            // set_is_bdry(uwall);

            x_range = coords[0]-scoords[0];
            hheight = uv_h;
            insert_sine_curve_section(uwall->last, uwall, 0.0, x_range, scoords, uv_h,
                 x_range,front);

            prompt_for_adherence_coeff(Hyper_surf(uwall));
        }
  
        // TMP, insert obstacle in the domain
        if(debugging("init_obstacle"))
        {
            init_obstacle_bub(init,ip,uwall,lwall);
        }
        // END TMP, insert obstacle in the domain
 
        screen("Type 'y' to initialize bubbles(y, n(dflt)): ");
        (void) Gets(s);
        if (strncmp(s,"y",1) == 0 || strncmp(s,"Y",1) == 0)        
        {
            screen("Type 'y' to initialize random bubbles(y, n(dflt)): ");
            (void) Gets(s);
            if (strncmp(s,"y",1) == 0 || strncmp(s,"Y",1) == 0)        
                init_blood_bub(init,ip,uwall,lwall);
            else 
            {
                float   coords[MAXD];
                float   pi = 3.14159265358979323846;
                int     n_theta, k;
                float   theta, d_theta;
                int     num_bubbles = 1, i;

                screen("Enter number of bubbles (default 1): ");
                (void) Scanf("%d\n",&num_bubbles);
                init_bub = YES;

                for(i = 0; i < num_bubbles; i++)
                {
                    screen("Enter bubble radius: ");
                    (void) Scanf("%f\n",&bub_rad);
                    screen("Enter bubble center: ");
                    Scanf("%f %f\n",&bub_cen[0],&bub_cen[1]);
                    coords[0] = bub_cen[0] + bub_rad;
                    coords[1] = bub_cen[1];
                    bns = make_node(Point(coords));
                    bne = bns;
                    node_type(bns) = CLOSED_NODE;
               // printf("enteed bubble radius %g, bub_cen %g %g\n",
               //             bub_rad, bub_cen[0], bub_cen[1]);

                    cv = make_curve(COMPA,COMPB,bns,bne);
                    n_theta = 4*(irint( pi*bub_rad/1.5/min(h[0], h[1])) + 1);
                    d_theta = 2*pi/n_theta;

                    theta = -d_theta;
                    for (k = 1; k < n_theta; k++)
                    {
                        coords[0] = bub_cen[0] + bub_rad*cos(theta);
                        coords[1] = bub_cen[1] + bub_rad*sin(theta);
                        Check_return(insert_point_in_bond(Point(coords),cv->last,cv),
                             init_blood_vessel);
                        if (debugging("vel_bubbles"))
                        {
                            printf("start = (%g, %g), point = (%g, %g),\n",
                               Coords(cv->last->start)[0],
                               Coords(cv->last->start)[1],
                               coords[0], coords[1]);
                        }
                        theta -= d_theta;
                    }
                    wave_type(cv) = CONTACT;
                    start_status(cv) = end_status(cv) = INCIDENT;
                }
            }
        }

	(void) prompt_for_eos_params(init,ip,YES,"");
	paramsa = prompt_for_eos_params(init,ip,YES,"\n\tfor the vessel blood");
        prompt_for_ambient_state(comp_type(COMPA),paramsa,
	                " for the vessel blood",front,init);
        paramsa->dens = Dens(Ambient(comp_type(COMPA)));
        paramsa->comp = COMPA;

        // if(init_bub == YES)
        {
            paramsb = prompt_for_eos_params(init,ip,YES,"\n\tfor the gas bubble");     
            prompt_for_ambient_state(comp_type(COMPB),paramsb,
                                 " for the gas bubble",front,init);    
            paramsb->dens = Dens(Ambient(comp_type(COMPB)));
            paramsb->comp = COMPB;
        }

        // printf("EOS blood den %g, comp %d, gas den %g, comp %d\n",
        //      paramsa->dens, paramsa->comp, paramsb->dens, paramsb->comp);

        nprms = return_params_list(&prms_list);
        for (i=0; i<nprms; i++)
            prms_list[i]->eos->_NS_eos = YES;

        if(set_low_side == YES)
	    rect_boundary_type(intfc,1,0) = MIXED_TYPE_BOUNDARY;

        if(set_up_side == YES)
	    rect_boundary_type(intfc,1,1) = MIXED_TYPE_BOUNDARY;

        //chemicals  
        screen("Type 'y' to initialize chemical diffusion coefficient(y, n(dflt)): ");
        (void) Gets(s);
        if (strncmp(s,"y",1) == 0 || strncmp(s,"Y",1) == 0)
            g_init_chem_coeff();

        debug_print("jet","Left init_blood_vessel()\n");
        
        printf("geomview plot init intfc\n");
        geomview_intfc_plot2d("gview_plot",10000,intfc,gr);
}

/* arc length integrand for sine(X/x_range*PI) */
LOCAL float sine_len(
        float    theta,
        POINTER  prms)
{
        return sqrt(1+sqr(hheight*PI/x_range*cos(theta*PI/x_range)));
}

LOCAL void insert_sine_curve_section(
        BOND  *b,
        CURVE *c,
        float ThetaS, /* scaled */
        float ThetaE, /* scaled */
        float *shift,
        float scaleh,
        float span,
        Front *fr)
{
        RECT_GRID   *gr = fr->rect_grid;
        float       *h = gr->h;
        float       coords[MAXD];

        float       space, Len, ds;
        float       epsabs, epsrel;
        float       abserr;
        int         neval;
        int         istate;
        int         i, N, dim = gr->dim;
        float       *theta;
        QUADRATURE_STATUS  ier;
        float       tol = MIN_SC_SEP(fr->interf);

        space = Front_spacing(fr,
                (wave_type(c) >= FIRST_VECTOR_PHYSICS_WAVE_TYPE) ?
                 VECTOR_WAVE : GENERAL_WAVE)*min(h[0],h[1]);

        /* Compute the arc length in the scaled metric of the curve
         * section to be inserted.
         */
        epsrel = 1.0e-6;
        epsabs = 0.01*min(h[0],h[1]);

        Len = dqng(sine_len,NULL,0.0,span,
                      epsabs,epsrel,&abserr,&neval,&ier);

        printf("sine arc length from 0 to PI = %g, ier = %d\n", Len, ier);

        switch (ier)
        {
        case INVALID_EPSILON:
            screen("ERROR in insert_sine_curve_section(), "
                   "invalid epsilons\n"
                   "epsabs = %"FFMT", epsrel = %"FFMT"\n",epsabs,epsrel);
            clean_up(ERROR);
            break;
        case INACCURATE_INTEGRAL:
            if ((fabs(abserr) > 20.0*epsabs) && (fabs(abserr) > 20.0*epsrel*Len))
            {
                /*
                 * Don't print a warning is we are close to satifying the
                 * error requirements
                 */
                (void) printf("WARNING in insert_sine_curve_section(), "
                              "inaccurate result\n \tneval = %d, "
                              "abserr = %"FFMT" result = %"FFMT"\n",
                              neval,abserr,Len);
            }
            break;
        case ACCURATE_INTEGRAL:
        default:
            break;
        }

        N = irint(fabs(Len)/space);

        // TMP
        // printf("Number of points on sine arc length = %d, space = %g\n",N, space);

        if (N < 1)      /*Nothing to do, the sine section would be */
            return;     /* too short */
 
        vector(&theta,N+1,FLOAT);
        ds = Len/N;

        /*
        *  Integrate with respect to arc length to find angles with
        *  eqi-distance separation in the scaled metric along the sine 
        */

        theta[0] = ThetaS;
        istate = 1;
        if ((!ode_solver(isine_len,theta,0.0,ds,N,epsrel,
                         epsrel*fabs(ThetaE-ThetaS),
                         abserr,&istate,NULL)) || (istate != 2))
        {
            screen("ERROR in insert_sine_curve_section() can't solve ODE\n"
                   "error flag = %d\n",istate);
            clean_up(ERROR);
        }
        theta[N] = ThetaE;

        // TMP
        /*
        for(i = 0; i <= N; i++)
        {
            if(i == 0)
                printf("theta[%d] = %g\n", i, theta[i]);
            else 
                printf("theta[%d] = %g  dx = %g\n", i, theta[i], theta[i]-theta[i-1]);
        }
        */

        for(i = 1; i < N; i++)
        {
            coords[0] = shift[0] + theta[i];
            coords[1] = shift[1] + scaleh*sin(theta[i]*PI/span);
            insert_point_in_bond(Point(coords),b,c);
            b = b->next;
        }
        free(theta);

        // TMP
        // print_curve(c);
}

LOCAL void init_obstacle_bub(
        INIT_DATA       *init,
        INIT_PHYSICS    *ip,
        CURVE           *uwall,
        CURVE           *lwall)
{
        Front      *front = ip->root->front;
        Wave       *wave = ip->root->wave;
        INTERFACE  *intfc = front->interf;
        RECT_GRID  *gr = front->rect_grid;
        float      *L = gr->L;
        float      *U = gr->U;
        float      *h = gr->h;
        float      coords[MAXD], scoords[MAXD], crds[MAXD];
        Gas_param  *paramsa;
        int        nprms, i, j, icrds[MAXD], dim = gr->dim, lnum_pt,unum_pt;
        Gas_param  **prms_list; /* = gas_params_list(intfc);*/
        NODE       *bns, *bne;
        CURVE      *cv;
        float      pi = 3.14159265358979323846;
        int        n_theta, k, alloc_site;
        float      theta, d_theta, spacing;
        int          bmin[MAXD], bmax[MAXD], Rbmin[MAXD], Rbmax[MAXD], imin[MAXD], imax[MAXD];
        float      **lpt = NULL, **upt = NULL, cl[MAXD], cu[MAXD];
        BOND       *b;
        float      bub_rad, bub_cen[5][MAXD];

        // return_bub_diameter_spacing(&diam,&spacing);
        bub_rad = 0.21;

        for(i = 0; i < dim; i++)
        {
            imin[i] = 0;
            imax[i] = gr->gmax[i];
        }

        if(lwall != NULL)
            lpt = polygon_of_curve(lwall,&lnum_pt);
        if(uwall != NULL)
            upt = polygon_of_curve(uwall,&unum_pt);

        bub_cen[0][0] = 2.0; bub_cen[0][1] = -0.657;  
        bub_cen[1][0] = 1.4; bub_cen[1][1] = -0.657;  
        bub_cen[2][0] = 2.6; bub_cen[2][1] = -0.657;  
        bub_cen[3][0] = 2.25; bub_cen[3][1] = -0.167;  

        for(alloc_site = 0; alloc_site < 4; alloc_site++)
        {
            coords[0] = bub_cen[alloc_site][0] + bub_rad;
            coords[1] = bub_cen[alloc_site][1];
            bns = make_node(Point(coords));
            bne = bns;
            node_type(bns) = CLOSED_NODE;
 
            cv = make_curve(COMPA,COMPOBST,bns,bne);
            n_theta = 4*(irint( pi*bub_rad/1.5/min(h[0], h[1])) + 1);
            d_theta = 2*pi/n_theta;

            theta = -d_theta;
            for (k = 1; k < n_theta; k++)
            {
                 coords[0] = bub_cen[alloc_site][0] + bub_rad*cos(theta);
                 coords[1] = bub_cen[alloc_site][1] + bub_rad*sin(theta);
                 Check_return(insert_point_in_bond(Point(coords),cv->last,cv),
                   init_obstacle_bub);
                 theta -= d_theta;
            }
            wave_type(cv) = NEUMANN_BOUNDARY;
            start_status(cv) = end_status(cv) = FIXED;
            // prompt_for_adherence_coeff(Hyper_surf(lwall));
            no_slip(Hyper_surf(cv)) = YES;
            adherence_coeff(Hyper_surf(cv)) = 1.0;
        }

}


#if defined(__cplusplus)
extern "C" {
#endif /* defined(__cplusplus) */

/*ARGSUSED*/
LOCAL   void isine_len(
        int             *neq,
        float           *x,
        float           *y,
        float           *yp)
{
        *yp = 1.0/sine_len(*y,NULL);
}               /*end iel_len*/

#if defined(__cplusplus)
}
#endif /* defined(__cplusplus) */


#endif /* defined(TWOD) */
#endif /* defined(TWOD) || defined(THREED) */
