#include <hyp/hlocaldecs.h> 
	
#include <time.h> 
#include <sys/time.h> 
#include <sys/times.h> 
 
// #define XMAX 200 
// #define YMAX 100 
// #define NUM_PLATELET 2000 
// #define SPEED_MAX 50 
// #define SPEED_MAX 50 
 
//functions declaration 
LOCAL int velocity_y(int, int, Pots_Loc*); 
LOCAL int velocity_x(int, int, Pots_Loc*); 
// LOCAL int pt_in_bubble(Front*,COMPONENT,COMPONENT,float*);

//variable declaration 
 
LOCAL   int  n_inactive = 0, num_in = 0; 

LOCAL   int i,j,k; 
LOCAL	int temp1,temp2,temp3,temp4; 
LOCAL	int totalstep,sum,step; 
LOCAL	int x,y,z; 
LOCAL   int vx,vy;
LOCAL	int energy; 
LOCAL	int velocity_abs,prob; 
LOCAL   int XMAX = 200;  //default grid size
LOCAL   int YMAX = 100;  //default grid size 
LOCAL   int  NUM_PLATELET = 2000; // default number of platelets 
LOCAL   int  SPEED_MAX = 50;  // default place where the flow speed is a maximum
 
// the grid is XMAX x YMAX, and the status[][] represent the status in each grid
// thrombim[][] represent the concentration of thrombim
		
LOCAL	int **status;   //=0 nothing   =1 unactive  =2 active  =3 in clot 
LOCAL	int **thrombim; 
LOCAL	int *curve; 
 
LOCAL	char ch[100]; 
LOCAL	char file_name[100]; 

FILE *fp; 
FILE *fout; 
	 
EXPORT void init_for_restart(
        Pots_Loc     *ploc)
{
        RECT_GRID    *pgr;

        pgr = &(ploc->comp_grid);

        XMAX = pgr->gmax[0];
        YMAX = pgr->gmax[1];
        // initial platelet
        n_inactive = sum=0;
        NUM_PLATELET=XMAX*YMAX/10;
        SPEED_MAX = YMAX*1/2; // 3/12; //1/3, 2/5

        status = ploc->status;
        thrombim = ploc->thrombim;  
        curve = ploc->curve;

        ploc->_velocity_y = velocity_y;
        ploc->_velocity_x = velocity_x;

        temp1=XMAX*6/10;
        temp2=XMAX*7/10;
}
 
//velocity function 
LOCAL	int velocity_x( 
        int i,  
        int j, 
        Pots_Loc *ploc)  
{  
    //return the x_velocity at position (i,j) 
    // return 1; 
    int  ic[MAXD]; 
    ic[0] = i; 
    ic[1] = j; 
    return Pots_vel(ic,ploc)[0]; 
} 
 
 
 
LOCAL int velocity_y( 
      int i,  
      int j, 
      Pots_Loc *ploc)  
{  
    //return the x_velocity at position (i,j) 
    // return 0; 
    int  ic[MAXD]; 
    ic[0] = i; 
    ic[1] = j; 
    return Pots_vel(ic,ploc)[1]; 
} 
 
EXPORT Pots_Loc *Clot_init(
    Wave        *wv,
    Front       *fr,
    int         nx,
    int         ny,
    COMPONENT   comp,
    COMPONENT   bubcomp) 
{ 
    int          ic[MAXD];
    int          i, ix, iy;
    int          n_reg_nodes;
    Pots_Loc     *ploc;
    RECT_GRID    *gr = wv->rect_grid;
    RECT_GRID    *pgr;
    float        *crds;

    scalar(&ploc,sizeof(Pots_Loc));
    pgr = &(ploc->comp_grid);

    pgr->dim = gr->dim;
    for(i = 0; i < gr->dim; i++)
    {
        pgr->L[i] = gr->L[i];
        pgr->U[i] = gr->U[i];
        pgr->GL[i] = gr->GL[i];
        pgr->GU[i] = gr->GU[i];
        pgr->lbuf[i] = 0;
        pgr->ubuf[i] = 0;
    }

    pgr->gmax[0] = nx;
    pgr->gmax[1] = ny;

    for(i = 0; i < gr->dim; i++)
    {
        pgr->h[i] = (pgr->U[i]-pgr->L[i])/pgr->gmax[i];
    }

    n_reg_nodes = pgr->gmax[0];
    for( i = 1; i < gr->dim; i++)
        n_reg_nodes *= pgr->gmax[i];
    ploc->nnx = pgr->gmax[0];
    ploc->nny = pgr->gmax[1];

    vector(&(ploc->node_points),n_reg_nodes, sizeof(TG_PT));
    ploc->cg_npts = ploc->node_points;
    vector(&(ploc->new_node_points),n_reg_nodes, sizeof(TG_PT));
    ploc->cg_new_npts = ploc->new_node_points;
    vector(&(ploc->components),n_reg_nodes, sizeof(COMPONENT));
    ploc->cg_comps = ploc->components;
    vector(&(ploc->pots_vel),n_reg_nodes, sizeof(POTS_V));
    ploc->cg_pots_vel = ploc->pots_vel;

    i = 0;
    for(iy = 0; iy < pgr->gmax[1]; iy++)
    {
        for(ix = 0; ix < pgr->gmax[0]; ix++)
        {
            Coords(ploc->new_node_points+i)[0] = Coords(ploc->node_points+i)[0]
                                           = pgr->L[0] + pgr->h[0]*(0.5+ix);
            Coords(ploc->new_node_points+i)[1] = Coords(ploc->node_points+i)[1]
                                           = pgr->L[1] + pgr->h[1]*(0.5+iy);
            i++;
        }
    }

    // initial platelet
    XMAX = nx;
    YMAX = ny;
    n_inactive = sum=0;
    NUM_PLATELET=XMAX*YMAX/10;
    SPEED_MAX = ny*1/3; //1/3, 2/5

    // Allocate storage
    vector(&curve,XMAX,sizeof(int));
    matrix(&status,   XMAX, YMAX, sizeof(int));
    matrix(&thrombim, XMAX, YMAX, sizeof(int));

    srand((unsigned)time( NULL));

    for(i=0;i<XMAX;i++)
        for(j=0;j<YMAX;j++)
        {
            status[i][j]=0;
            thrombim[i][j]=0;
        }

    temp1=XMAX*6/10;
    temp2=XMAX*7/10;

    for(i=temp1;i<=temp2;i++)
    {
        thrombim[i][0]=50;
    }

    while(sum<NUM_PLATELET)
    {
        x=rand()%(XMAX-1);
        y=rand()%YMAX;

        ic[0] = x;
        ic[1] = y;
        crds = Pots_crds(ic,ploc);
        if(pt_in_bubble(fr, comp, bubcomp, crds) == YES)
            continue;
 
        // NEW, do not init in the obstacle
        if( YES == is_excluded_comp(long_component(crds,fr->interf), 
                  fr->interf) )
            continue;

        if(status[x][y]==0)
        {
            status[x][y]=1;
            sum++;
        }
    }// end while

    i = 0;
    for(iy = 0; iy < pgr->gmax[1]; iy++)
    {
        for(ix = 0; ix < pgr->gmax[0]; ix++)
        {
            *(ploc->cg_comps+i) = status[ix][iy];
            i++;
        }
    }

    for(i = 0; i < pgr->gmax[0]; i++) 
        curve[i] = 0;  

    totalstep=0; 

    //initial chemical 
 
    ploc->status = status;
    ploc->curve = curve;
    ploc->thrombim = thrombim;
    ploc->xmax = XMAX;
    ploc->ymax = YMAX;

    ploc->_velocity_y = velocity_y;
    ploc->_velocity_x = velocity_x;
    return ploc;
}//end of Clot_init() 
 
EXPORT void Clot_driver( 
        Pots_Loc *ploc)  
{ 
 
            int          ic[MAXD], ii, num_left; 
            RECT_GRID    *pgr = &(ploc->comp_grid); 
            int          dim = pgr->dim; 
            int          ic_xy[MAXD], ix, iy;
            float        crds_xy[MAXD];
 
            // for(i = 0; i < pgr->gmax[0]; i++) 
            //    ploc->curve[i] = 0;  
 
	    // STEP 1: 
	    // chemical diffusion by activated cells 
		 
	    for(j=(YMAX-2);j>0;j--) 
		for(i=1;i<(XMAX-1);i++) 
		{ 
		    thrombim[i][j]+= 
			0.2*(thrombim[i-1][j]+thrombim[i+1][j]+thrombim[i][j-1]+
                                  thrombim[i-1][j-1]+thrombim[i+1][j-1]); 
		    if (thrombim[i][j]>10000) thrombim[i][j]=10000; 
		} 

		// STEP 2:  
		// move the platelet 
            // TMP
	    // for (i=(XMAX-2);i>=0;i--) 
	    for (i=(XMAX-2);i>=0;i--) 
	    { 
	        for (j=0;j<YMAX;j++) 
	        { 
		    // unactive or active, can be move 

                    if(status[i][j] == 0 ||
                       status[i][j] == 3)
                        continue;

                    vx=velocity_x(i,j,ploc);
                    vy=velocity_y(i,j,ploc);
                    x=i+vx;
                    y=j+vy;

                    ic_xy[0] = x; ic_xy[1] = y; 
                    for(ii = 0; ii < dim; ii++)
                       crds_xy[ii] = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                    
                    if(y == 0 && vy == -1)
                        vy = 0; 

                    if (status[x][y]==0) //if can move
                    {
                        status[x][y]=status[i][j];
                        status[i][j]=0;
                        for(ii = 0; ii < dim; ii++)
                            Pots_new_crds(ic_xy,ploc)[ii] = 
                            Pots_crds(ic_xy,ploc)[ii] = crds_xy[ii];    
                    }
                    else if((vx)==0 && status[x][y]!=0)
                    {
                        x=i;
                        y=j;
                    }
                    else if(status[x][y]==3 && y>0)
                    {  
                        if(status[x][y-1]==0)
                        //move right down if is occupide by clot
                        {
                            y--;
                            status[x][y]=status[i][j];
                            status[i][j]=0;
                            ic_xy[0] = x; ic_xy[1] = y; 
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] = 
                                Pots_crds(ic_xy,ploc)[ii] 
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                        }
                    }
                    else if(status[x][y]==3 && y<(YMAX-1) )
                    {       
                        if ( status[i][j+1]==0 && status[x][y+1]==0)
                        //move right up if is occupide by clot
                        {
                            y++;
                            status[x][y]=status[i][j];
                            status[i][j]=0;
                            ic_xy[0] = x; ic_xy[1] = y;
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                        }
                    }
                    else if(status[x][y]==3 && y<(YMAX-2) )
                    {       
                        if ( status[i][j+1]==0 && status[x][y+2]==0)
                        //move right up*2 if occupied by clot
                        {
                            y+=2;
                            status[x][y]=status[i][j];
                            status[i][j]=0;
                            ic_xy[0] = x; ic_xy[1] = y;
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                        }
                    }
                    else if(status[x][y]!=0 && vy==1 && status[x][y-1]==0)
                    {// can't move right up, so move right
                        y--;
                        status[x][y]=status[i][j];
                        status[i][j]=0;
                        ic_xy[0] = x; ic_xy[1] = y;
                        for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);

                    }
                    else if(status[x][y]!=0 && vy==-1 && status[x][y+1]==0)
                    {
                    // can't move right down, so move right
                        y++;
                        status[x][y]=status[i][j];
                        status[i][j]=0;
                        ic_xy[0] = x; ic_xy[1] = y;
                        for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);

                    }
                    else if(status[x][y]!=0 && vy==0 && y>0 && y<(YMAX-1) )
                    //move to right and block
                    {
                        if(status[x][y-1]==0 && status[x][y+1]!=0)
                        {// move right down
                            y--;
                            status[x][y]=status[i][j];
                            status[i][j]=0;
                            ic_xy[0] = x; ic_xy[1] = y;
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                        }
                        else if (status[x][y-1]!=0 && status[x][y+1]==0)
                        {// move right up
                            y++;
                            status[x][y]=status[i][j];
                            status[i][j]=0;
                            ic_xy[0] = x; ic_xy[1] = y;
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                        }
                        else if (status[x][y-1]==0 && status[x][y+1]==0)
                        {// move down/up randonly
                            switch(rand()%2)
                            {
                               case 0:
                               {
                                   y++;
                                   status[x][y]=status[i][j];
                                   status[i][j]=0;
                                   break;
                               }
                               case 1:
                               {
                                   y--;
                                   status[x][y]=status[i][j];
                                   status[i][j]=0;
                                   break;
                               }
                            }//end of switch
                            ic_xy[0] = x; ic_xy[1] = y;
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);
                        }
                    }

                    else if(status[x][y]!=0)
                    {
                        x=i;
                        y=j;
                        if(y<curve[x])
                           status[i][j]=3; //being stuck, add to clot
                    }

                    // if active can be attach to clot
                    if(x<(XMAX-1))
                    {
                        if(status[x][y]==2 && y>0 && status[x+1][y-1]==0 && status[x+1][y]==3)
                        {
                            status[x+1][y-1]=2;
                            status[x][y]=0;
                            ic_xy[0] = x+1; ic_xy[1] = y-1;
                            for(ii = 0; ii < dim; ii++)
                                Pots_new_crds(ic_xy,ploc)[ii] =
                                Pots_crds(ic_xy,ploc)[ii]
                                = pgr->L[ii] + pgr->h[ii]*(0.5+ic_xy[ii]);

                        }
                        else if(status[x][y]==2)
                        {
                            energy=0;
                            if(y>0)
                            {
                                if(status[x][y-1]==3) energy+=100;
                                if(status[x-1][y]==3) energy+=100;
                                if(status[x+1][y]==3) energy+=100;
                                if(status[x-1][y-1]==3) energy+=50;
                                if(status[x+1][y-1]==3) energy+=50;
                                velocity_abs=y;
                                prob=(velocity_abs-SPEED_MAX)*(velocity_abs-SPEED_MAX)/
                                            (SPEED_MAX*SPEED_MAX/100);
                                if(energy>=100 && rand()%100<prob )
                                {
                                    status[x][y]=3;
                                }
                            }
                            else if(y==0)
                            {
                                temp1=XMAX*13/20;
                                temp2=XMAX*7/10;
                                if(x>temp1 && x<temp2) status[x][y]=3;
                            }
                        }

                        //if non-activated can be active
                        // TMP, do not activate the platelets
                        if(! debugging("NO_ACT"))
                        {
                            if (status[x][y]==1)
                            {
                                if (thrombim[x][y]>20 && rand()%100<80)
                                    status[x][y]=2;
                            }
                        }
                    }//end if
                }//for j
            }// for i

 
	    // STEP 2.5 
	    // adjust the clot by fibrin energy 
	    // adjust(); 
 
	    // STEP 3 
	    // make new platelet from left 

	    for(j=0;j<YMAX;j++) status[XMAX-1][j]=0; 

	    // calculate how many platelets lie in the [0,XMAX/10]x[0,YMAX] region

	    temp1=0; 
	    for(i=0;i<(XMAX/10);i++) 
	        for(j=0;j<YMAX;j++) 
		    if (status[i][j]==1 || status[i][j]==2) temp1++; 
	    temp2=NUM_PLATELET/10-temp1; 
 
	    // generate #temp2 new platelets from left
	    sum=0; 
	    while(sum<temp2) 
	    { 
		y=rand()%(YMAX-1);   //0->99 

		if (status[0][y]==0) 
		{ 
		    status[0][y]=1; 
		    sum++; 
                    // mapped to Rect_grid, whose index starts with 0 
                    ic[0] = 0; ic[1] = y;
                    for(ii = 0; ii < dim; ii++)
                        Pots_new_crds(ic,ploc)[ii] = Pots_crds(ic,ploc)[ii] =
                             pgr->L[ii] + pgr->h[ii]*(0.5+ic[ii]);
		} 
	    } 
 
            // STEP 4 
	    // generate the interpolation curve 
	    for(i=0;i<XMAX;i++) 
	    {     
	        curve[i]=0; 
	        j=YMAX-3; 
	        while(status[i][j]!=3 && j>0) 
	        { 
	            j--; 
                } 
	        curve[i]=j; 
	    }     
	 
	//********************************************* 
	// passing the value of curve[200] to the fluid 
	// for(i=1;i<=200;i++) 
        // { 
        //     for(j=1;j<=100;j++) fprintf(fout,"%d",p[i][j]); 
	//     fprintf(fout,"\n"); 
	//} 
 
}// end of main() 
 
// Set up particale grid, a particale correspond 
// to a pots. 
// Is p = status ????
EXPORT Pots_Loc *Partical_Init_loc( 
        RECT_GRID    *gr, 
        int          nx, 
        int          ny) 
{ 
        int          i, ix, iy; 
        int          n_reg_nodes; 
        Pots_Loc     *ploc; 
        RECT_GRID    *pgr; 
 
        scalar(&ploc,sizeof(Pots_Loc)); 
        pgr = &(ploc->comp_grid); 
 
        pgr->dim = gr->dim; 
        for(i = 0; i < gr->dim; i++) 
        { 
            pgr->L[i] = gr->L[i]; 
            pgr->U[i] = gr->U[i]; 
            pgr->GL[i] = gr->GL[i]; 
            pgr->GU[i] = gr->GU[i]; 
            pgr->lbuf[i] = 0; 
            pgr->ubuf[i] = 0; 
        } 
 
        pgr->gmax[0] = nx; 
        pgr->gmax[1] = ny; 
 
        for(i = 0; i < gr->dim; i++) 
        { 
            pgr->h[i] = (pgr->U[i]-pgr->L[i])/pgr->gmax[i]; 
        } 
 
        n_reg_nodes = pgr->gmax[0]; 
        for( i = 1; i < gr->dim; i++) 
            n_reg_nodes *= pgr->gmax[i]; 
        ploc->nnx = pgr->gmax[0]; 
        ploc->nny = pgr->gmax[1]; 
 
        vector(&(ploc->node_points),n_reg_nodes, sizeof(TG_PT)); 
        ploc->cg_npts = ploc->node_points; 
        vector(&(ploc->new_node_points),n_reg_nodes, sizeof(TG_PT)); 
        ploc->cg_new_npts = ploc->new_node_points; 
        vector(&(ploc->components),n_reg_nodes, sizeof(COMPONENT)); 
        ploc->cg_comps = ploc->components; 
        vector(&(ploc->pots_vel),n_reg_nodes, sizeof(POTS_V)); 
        ploc->cg_pots_vel = ploc->pots_vel; 
 
        i = 0; 
        for(iy = 0; iy < pgr->gmax[1]; iy++) 
        { 
            for(ix = 0; ix < pgr->gmax[0]; ix++) 
            { 
                Coords(ploc->new_node_points+i)[0] = Coords(ploc->node_points+i)[0] 
                                               = pgr->L[0] + pgr->h[0]*(0.5+ix);                 
                Coords(ploc->new_node_points+i)[1] = Coords(ploc->node_points+i)[1] 
                                               = pgr->L[1] + pgr->h[1]*(0.5+iy);                 
                i++; 
            } 
        } 
 
        i = 0; 
        for(iy = 0; iy < pgr->gmax[1]; iy++) 
        { 
            for(ix = 0; ix < pgr->gmax[0]; ix++) 
            { 
                *(ploc->cg_comps+i) = status[ix][iy]; 
                i++; 
            } 
        } 
 
        // ploc->a = a; 
        ploc->status = status; 
        // ploc->ev = ev; 
        // ploc->Index = Index; 
        // ploc->tag = tag; 
        ploc->curve = curve; 
        ploc->thrombim = thrombim;
        ploc->xmax = XMAX;
        ploc->ymax = YMAX;
 
        ploc->_velocity_y = velocity_y; 
        ploc->_velocity_x = velocity_x; 
        return ploc; 
} 
 
EXPORT void free_pots_loc( 
        Pots_Loc        *ploc) 
{ 
        int  i; 
 
        free(ploc->node_points); 
        free(ploc->new_node_points); 
        free(ploc->components); 
        free(ploc->pots_vel); 
        free(ploc); 
 
        free(curve); 
        free(status); 
        free(thrombim); 
} 
 
EXPORT void gv_out_pots( 
        const char    *dname, 
        int           step, 
        Pots_Loc      *ploc) 
{ 
        FILE          *fp; 
        static char   *fname = NULL, *ppfname = NULL; 
        static size_t fname_len = 0, ppfname_len = 0; 
        char          outname[256],outdir[256]; 
        const char    *nstep; 
 
        RECT_GRID   *pgr = &(ploc->comp_grid); 
        int         i, ic[MAXD], ix, iy, dim = pgr->dim; 
        float       crds[3]; 
        float       *L = pgr->L; 
        float       *U = pgr->U; 
        float       amb1[3] = {1.0, 0.3, 0.3}; 
        float       dif1[3] = {1.0, 0.3, 0.3}; 
        float       amb2[3] = {0.3, 0.6, 0.6}; 
        float       dif2[3] = {0.3, 0.6, 0.6}; 
        int         num_c = 0; 
        BOND        *oldb; 
        float       color[4] = {0.898, 0.400, 0.000, 1.000}; 
        CURVE       **c, *oldc; 
        static float **clrmap = NULL; 
        INTERFACE    *intfc = ploc->intfc; 
 
        if(clrmap == NULL) 
        { 
            vector(&clrmap,6,sizeof(float*)); 
            for(i = 0; i < 6; i++) 
                vector(&clrmap[i],4,sizeof(float)); 
            set_default_clr_map(clrmap, 6); 
        } 
 
        sprintf(outdir,"%s/%s",dname,"geomintfc"); 
        ppfname = set_ppfname(ppfname,"intfc",&ppfname_len); 
        nstep = right_flush(step,7); 
        sprintf(outname,"%s.ts%s",ppfname,nstep); 
 
        fname = get_list_file_name(fname,outdir,outname,&fname_len); 
        if ((fp = fopen(fname,"w")) == NULL) 
        { 
            (void) printf("WARNING in geomview_intfc_plot2d(), " 
                           "can't open %s\n",fname); 
            return; 
        } 
 
        fprintf(fp,"{ # World\n"); 
        fprintf(fp,"INST   geom { LIST # World list\n"); 
        fprintf(fp,"{\n"); 
        fprintf(fp,"  geom {appearance {\n"); 
        fprintf(fp,"  material {\n"); 
        fprintf(fp,"     ambient %g %g %g\n", amb1[0], amb1[1], amb1[2]); 
        fprintf(fp,"     diffuse %g %g %g}}\n", dif1[0], dif1[1], dif1[2]); 
        fprintf(fp,"geom LIST\n"); 
        /* Begin LIST OBJ */ 
 
        for(iy = 0; iy < pgr->gmax[1]; iy++) 
        { 
            for(ix = 0; ix < pgr->gmax[0]; ix++) 
            { 
                ic[0] = ix; ic[1] = iy; 
 
                if(ploc->status[ix][iy] == 3) 
                { 
                    for(i = 0; i < dim; i++) 
                       crds[i] = pgr->L[i] + pgr->h[i]*(0.5+ic[i]); 
                    crds[2] = 1.0; 
                    fprintf(fp,"{ SPHERE\n"); 
                    fprintf(fp,"%f ", pgr->h[0]/2.0); 
                    fprintf(fp,"%f %f %f}\n",crds[0],crds[1],crds[2]); 
                } 
            } 
        } 
        /* end of LIST OBJ */ 
        fprintf(fp,"} } \n"); 
  
        // Inactivated  
        for(iy = 0; iy < pgr->gmax[1]; iy++) 
        { 
            for(ix = 0; ix < pgr->gmax[0]; ix++) 
            { 
                ic[0] = ix; ic[1] = iy; 
 
                if(ploc->status[ix][iy] == 1 ||
                   ploc->status[ix][iy] == 2) 
                { 
                    for(i = 0; i < dim; i++) 
                       crds[i] = pgr->L[i] + pgr->h[i]*(0.5+ic[i]); 
                    crds[2] = 1.0; 
                    fprintf(fp,"{ SPHERE\n"); 
                    fprintf(fp,"%f ", pgr->h[0]/2.0); 
                    fprintf(fp,"%f %f %f}\n",crds[0],crds[1],crds[2]); 
                } 
            } 
        } 
 
        /* beginning of writting Vect to file */ 
        for(i = 0, c = intfc->curves; c && *c;  c++, i++) 
        { 
            oldc = *c; 
            oldb = oldc->first; 
            fprintf(fp,"{ \n"); 
            fprintf(fp,"VECT\n"); 
            fprintf(fp,"%1d %1d %1d\n", 1, oldc->num_points, 1); 
            fprintf(fp,"%1d\n%1d\n", oldc->num_points,1); 
 
            while (oldb) 
            { 
                fprintf(fp,"%f %f %f \t",  Coords(oldb->start)[0], 
                                Coords(oldb->start)[1] , 1.0); 
                if (oldb == oldc->last) 
                    break; 
                oldb = oldb->next; 
            } 
            fprintf(fp,"%f %f %f \t\n",  Coords(oldb->end)[0], 
                  Coords(oldb->end)[1] , 1.0); 
            // fprintf(fp,"%f %f %f %f \t\n", 1.0, 0.2, 0.2, 0.8 ); 
            fprintf(fp,"%f %f %f %f \t\n", clrmap[i%6][0], clrmap[i%6][1], 
                       clrmap[i%6][2], clrmap[i%6][3]); 
            fprintf(fp,"}\n"); 
        } 
        /* end of writting Vect to file     */ 
 
         
 
        /*  computational grid SKEL*/ 
        /* WHITE YELLOW */ 
        fprintf(fp,"{ VECT\n"); 
        fprintf(fp,"%1d %1d %1d\n", 1, 5, 1); 
        fprintf(fp,"%1d\n%1d\n", 5,1); 
        (void) fprintf(fp,"%f %f %f \t",L[0],L[1],1.0); 
        (void) fprintf(fp,"%f %f %f \t",U[0],L[1],1.0); 
        (void) fprintf(fp,"%f %f %f \t",U[0],U[1],1.0); 
        (void) fprintf(fp,"%f %f %f \t\n",L[0],U[1],1.0); 
        (void) fprintf(fp,"%f %f %f \t",L[0],L[1],1.0); 
        fprintf(fp,"%f %f %f %f \t\n", 1.0, 1.0, 0.0, 1.0 ); 
        fprintf(fp,"}\n"); 
 
        fprintf(fp,"} \n }\n"); 
        fclose(fp); 
 
} 
 
 
 
 
 
