/*
*                               hEB.c:
*
*       Copyright 1999 by The University at Stony Brook, All rights reserved.
*
*/

#define DEBUG_STRING "hEB"

#include <hyp/hlocaldecs.h>

LOCAL int           dim, imin[MAXD], imax[MAXD]; // buffered comp grid size
LOCAL COMPONENT     ext_comp;
LOCAL INTERFACE     *intfc = NULL;

LOCAL int       cell_on_phy_boundary(int*,RECT_GRID*,BDRY_SIDE*);
LOCAL int       set_bdry_cell_edge(BLK_EDGE*,int*,float*,COMPONENT,BDRY_SIDE*,Wave*);

/* init embedded boundary structure */

EXPORT int init_EB(
       Wave   *wave,
       Front  *front)
{
        COMPONENT       comp, corn_comp[2][2];
        COMPONENT       cn, cs, cw, ce;
        int             comp_label[4], comp_count;
        float           *crds, *crds_s, *crds_e, area;
        int             icrds[MAXD], icrds_end[MAXD], ic[MAXD];
        int             icn[MAXD], ics[MAXD], icw[MAXD], ice[MAXD];
        int             ic_corn[MAXD];
        int             status = YES;
        int             i, j, k, l, ii;
        int             max[MAXD], min[MAXD];
        static CRXING   **crx[4] = {NULL,NULL,NULL,NULL};
        int             nc[4]; /* number of CRXING */
        TRI_GRID        *ntg = wave_tri_soln(wave)->tri_grid;
        RECT_GRID       *gr = wave->rect_grid;
        size_t          sizest = wave->sizest;
        BDRY_SIDE       side[4];
        int             on_bd;
        BLK_EDGE        *blk_edge_store = ntg->blk_edge_store;
        float           *a_frac;
        int             deb_flag = NO;

        DEBUG_ENTER(init_EB)

        // set local globals
        dim = wave->rect_grid->dim;
        for(i = 0; i < dim; i++)
        {
            imin[i] = -wave->rect_grid->lbuf[i];
            imax[i] = wave->rect_grid->gmax[i] + wave->rect_grid->ubuf[i];
        }
        ext_comp = exterior_component(front->interf);
        intfc = front->interf;
        // End set local globals

        area = 1.0;
        for(i = 0; i < dim; i++)
            area *= gr->h[i];

        if(NULL == crx[0])
        {
            for(i = 0; i < 4; i++)
                 vector(&crx[i],4,sizeof(CRXING *));
        }

        for (i = 0; i < 2; i++)
        {
            min[i] = 0;
            max[i] = ntg->comp_grid.gmax[i];
            min[i] -= ntg->comp_grid.lbuf[i];
            max[i] += ntg->comp_grid.ubuf[i];
        }

        // TMP
        if(blk_edge_store == NULL)
        {
            printf("ERROR: in init_EB, blk_edge = NULL front = %d, wave = %d\n", front, wave);
            clean_up(ERROR);
        }
  
        for (i=min[1]; i<max[1]; i++)
        {
            for (j=min[0]; j<max[0]; j++)
            {
                icrds[1] = i;
                icrds[0] = j;
                icrds_end[0] = icrds[0] + 1;
                icrds_end[1] = icrds[1] + 1;

                for(k = 0; k < dim; k++)
                {
                    icn[k] = ics[k] = icw[k] = ice[k] =
                          icrds[k];
                }
                icn[1]++; ics[1]--; icw[0]--; ice[0]++;

                Rect_blk_edge(icrds,wave).nc[1][0] = 
                    nc[0] = comp_crossings_in_direction(crx[0],icrds,EAST,ntg);
                Rect_blk_edge(icrds,wave).nc[0][0] = 
                    nc[2] = comp_crossings_in_direction(crx[2],icrds,NORTH,ntg);
                Rect_blk_edge(icrds,wave).nc[1][1] = 
                    nc[1] = comp_crossings_in_direction(crx[1],icrds_end,WEST,ntg);
                Rect_blk_edge(icrds,wave).nc[0][1] = 
                    nc[3] = comp_crossings_in_direction(crx[3],icrds_end,SOUTH,ntg);

                for(k = 0; k < nc[0]; k++)
                    Rect_blk_edge(icrds,wave).crx[1][0][k] = crx[0][k];
                for(k = 0; k < nc[2]; k++)
                    Rect_blk_edge(icrds,wave).crx[0][0][k] = crx[2][k];
                for(k = 0; k < nc[1]; k++)
                    Rect_blk_edge(icrds,wave).crx[1][1][k] = crx[1][k];
                for(k = 0; k < nc[3]; k++)
                    Rect_blk_edge(icrds,wave).crx[0][1][k] = crx[3][k];

                crds = Rect_coords(icrds,wave);
                comp = Rect_comp(icrds,wave);

                crds_s = Rect_compute_coords(icrds,wave);
                crds_e = Rect_compute_coords(icrds_end,wave);

                cn = Find_rect_comp_ver2(icn,wave);
                cs = Find_rect_comp_ver2(ics,wave);
                cw = Find_rect_comp_ver2(icw,wave);
                ce = Find_rect_comp_ver2(ice,wave);

                assign(ic_corn,icrds,dim*sizeof(int));
                Rect_blk_edge(icrds,wave).corn_comp[0][0] = 
                    corn_comp[0][0] = Rect_compute_comp(ic_corn,wave);
                ic_corn[0]++;
                Rect_blk_edge(icrds,wave).corn_comp[1][0] = 
                    corn_comp[1][0] = Rect_compute_comp(ic_corn,wave);
                ic_corn[0]--; ic_corn[1]++;
                Rect_blk_edge(icrds,wave).corn_comp[0][1] = 
                    corn_comp[0][1] = Rect_compute_comp(ic_corn,wave);
                ic_corn[0]++;
                Rect_blk_edge(icrds,wave).corn_comp[1][1] = 
                    corn_comp[1][1] = Rect_compute_comp(ic_corn,wave);

                if(comp == ext_comp ||
                   is_excluded_comp(comp,front->interf) ||
                   comp == POTS_COMP)
                {
                    Rect_blk_edge(icrds,wave).ctype = EXTERNAL;
                    continue;
                }

                comp_count = cell_nbhd_config(cn, cs, cw, ce, comp_label);

                // TMP
                // Note: The grid-based geometry is not strictly symmetric.
                // Use sym_pipe with 10 by 10 mesh debug
                /**
                if(icrds[0] == -4 && (icrds[1] == 0))
                {
                    printf("In init_EB\n");
                    printf("cell[%d %d] comp %d, corner comp[%d %d %d %d]\n",
                       icrds[0], icrds[1], comp, corn_comp[0][0],  
                        corn_comp[1][0], corn_comp[0][1], corn_comp[1][1]);
                    printf("crxing count[%d %d %d %d]\n",nc[0], nc[1], nc[2], nc[3]);
                    deb_flag = YES;
                }
                else 
                   deb_flag = NO;
                **/

                // TMP
                /**
                if((icrds[0] == 0 && icrds[1] == 3) ||
                   (icrds[0] == 0 && icrds[1] == 4))
                {
                    printf("In init_EB ic[%d %d]\n",  icrds[0], icrds[1]);
                    if(nc[0] != 0)
                        printf("EAST crxing (%9.8g %9.8g)\n", 
                             Coords(crx[0][0]->pt)[0], Coords(crx[0][0]->pt)[1]);
                    if(nc[2] != 0)
                        printf("North crxing (%9.8g %9.8g)\n", 
                             Coords(crx[2][0]->pt)[0], Coords(crx[2][0]->pt)[1]);
                    if(nc[1] != 0)
                        printf("WEST crxing (%9.8g %9.8g)\n", 
                             Coords(crx[1][0]->pt)[0], Coords(crx[1][0]->pt)[1]);
                    if(nc[3] != 0)
                        printf("SOUTH crxing (%9.8g %9.8g)\n", 
                             Coords(crx[3][0]->pt)[0], Coords(crx[3][0]->pt)[1]);
                    deb_flag = YES;
                }
                else
                    deb_flag = NO;
                **/ 

                /* This is for irregular domain*/
                if(comp_count == 0)
                    Rect_blk_edge(icrds,wave).bdry_ctype = INTERNAL;
                else
                {
                    Rect_blk_edge(icrds,wave).bdry_ctype = EXTERNAL;
                    Rect_blk_edge(icrds,wave).comp_label[1][1] = comp_label[0];
                    Rect_blk_edge(icrds,wave).comp_label[1][0] = comp_label[1];
                    Rect_blk_edge(icrds,wave).comp_label[0][0] = comp_label[2];
                    Rect_blk_edge(icrds,wave).comp_label[0][1] = comp_label[3];
                }

                a_frac = Rect_blk_edge(icrds,wave).a_frac;

                if(nc[1] == 0 && nc[0] == 0 &&
                   nc[2] == 0 && nc[3] == 0)
                {
                    // printf("ic[%d %d], not crossed\n", j, i);
                    if((on_bd = cell_on_phy_boundary(icrds,&(ntg->comp_grid),side))
                               == YES)
                    {
                        // printf("init_EB: ic[%d %d], onboundary, partial %d\n", j, i, PARTIAL);
                        // printf("side[%d %d %d %d]\n", side[0], side[1], side[2], side[3]);

                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                        set_bdry_cell_edge(&(Rect_blk_edge(icrds,wave)),
                                icrds,crds,comp,side,wave);
                    }
                    else if(comp_count != 0)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                        if(comp_label[0] == YES)
                            Rect_blk_edge(icrds,wave).edge[1][1].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[1][1].etype = FULL_SURF;
                        if(comp_label[1] == YES)
                            Rect_blk_edge(icrds,wave).edge[1][0].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[1][0].etype = FULL_SURF;
                        if(comp_label[2] == YES)
                            Rect_blk_edge(icrds,wave).edge[0][0].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[0][0].etype = FULL_SURF;
                        if(comp_label[3] == YES)
                            Rect_blk_edge(icrds,wave).edge[0][1].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[0][1].etype = FULL_SURF;
                        for(k = 0; k < dim; k++)
                        { 
                            Rect_blk_edge(icrds,wave).edge[k][0].length = gr->h[(k+1)%2];
                            Rect_blk_edge(icrds,wave).edge[k][1].length = gr->h[(k+1)%2];
                        }
                        if(comp_label[0] == YES && comp_label[1] == YES &&
                           comp_label[2] == YES && comp_label[3] == YES)
                            Rect_blk_edge(icrds,wave).ctype = EXTERNAL;
                    }
                    else
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTERNAL;

                        for(k = 0; k < dim; k++)
                        { 
                            Rect_blk_edge(icrds,wave).edge[k][0].etype = FULL_SURF;
                            Rect_blk_edge(icrds,wave).edge[k][1].etype = FULL_SURF;
                            Rect_blk_edge(icrds,wave).edge[k][0].length = gr->h[(k+1)%2];
                            Rect_blk_edge(icrds,wave).edge[k][1].length = gr->h[(k+1)%2];
                        }
                    }
                    Rect_blk_edge(icrds,wave).a_frac[0] = 1.0;
                    Rect_blk_edge(icrds,wave).config = 0;
                }
                else if(nc[0] == 1 && nc[3] == 1 &&
                        nc[1] == 0 && nc[2] == 0)
                {
                    if(wave_type(crx[0][0]->hs) < FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[3][0]->hs) < FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                    }
                    else if(wave_type(crx[0][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                            wave_type(crx[3][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTER_PART;
                    }
                    else
                    {
                        printf("ERROR: init_EB()\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        printf("Inconsistent wave types\n");
                        clean_up(ERROR); 
                    }
                    Rect_blk_edge(icrds,wave).edge[1][0].etype = PARTIAL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][1].etype = PARTIAL_SURF;

                    Rect_blk_edge(icrds,wave).edge[0][1].cen[0] = crds_e[0];
                    Rect_blk_edge(icrds,wave).edge[0][1].cen[1] = 
                               0.5*(crds_e[1]+Coords(crx[3][0]->pt)[1]);
                    Rect_blk_edge(icrds,wave).edge[0][1].length = 
                            sqrt(sqr(crds_e[0]-Coords(crx[3][0]->pt)[0]) + 
                                 sqr(crds_e[1]-Coords(crx[3][0]->pt)[1]));
                    Rect_blk_edge(icrds,wave).edge[1][0].cen[1] = crds_s[1];
                    Rect_blk_edge(icrds,wave).edge[1][0].cen[0] = 
                               0.5*(crds_s[0]+Coords(crx[0][0]->pt)[0]);
                    Rect_blk_edge(icrds,wave).edge[1][0].length = 
                            sqrt(sqr(crds_s[0]-Coords(crx[0][0]->pt)[0]) + 
                                 sqr(crds_s[1]-Coords(crx[0][0]->pt)[1]));

                    Rect_blk_edge(icrds,wave).edge[0][0].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][0].length = gr->h[1];
                    Rect_blk_edge(icrds,wave).edge[1][1].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][1].length = gr->h[0];

                    a_frac[1] = 0.5*(crds_e[0]-Coords(crx[0][0]->pt)[0])*
                                    (Coords(crx[3][0]->pt)[1]-crds_s[1])/area;
                    a_frac[0] = 1.0-a_frac[1];
                    Rect_blk_edge(icrds,wave).config = 1;
                    // TMP
                    if(deb_flag)
                        printf("ic[%d %d] crds_e[%9.8g %9.8g] crds_s[%9.8g %9.8g], crx[0] [%9.8g %9.8g],"
                              "crx[3] [%9.8g %9.8g]\n",
                               icrds[0], icrds[1], crds_e[0], crds_e[1],
                               crds_s[0], crds_s[1], 
                               Coords(crx[0][0]->pt)[0], Coords(crx[0][0]->pt)[1],
                               Coords(crx[3][0]->pt)[0], Coords(crx[3][0]->pt)[1]);
                }
                else if(nc[0] == 0 && nc[1] == 0 && nc[2] == 1 && nc[3] == 1)
                {
                    if(wave_type(crx[2][0]->hs) < FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[3][0]->hs) < FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                    }
                    else if(wave_type(crx[2][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                            wave_type(crx[3][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTER_PART;
                    }
                    else
                    {
                        printf("ERROR: init_EB()\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        printf("Inconsistent wave types\n");
                        clean_up(ERROR);
                    }

                    Rect_blk_edge(icrds,wave).edge[0][0].etype = PARTIAL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][1].etype = PARTIAL_SURF;

                    if(corn_comp[0][0] == comp)
                    {
                        Rect_blk_edge(icrds,wave).edge[0][0].cen[0] = crds_s[0];
                        Rect_blk_edge(icrds,wave).edge[0][0].cen[1] =
                               0.5*(crds_s[1]+Coords(crx[2][0]->pt)[1]);
                        Rect_blk_edge(icrds,wave).edge[0][0].length =
                            sqrt(sqr(crds_s[0]-Coords(crx[2][0]->pt)[0]) +
                                 sqr(crds_s[1]-Coords(crx[2][0]->pt)[1]));

                        Rect_blk_edge(icrds,wave).edge[1][0].etype = FULL_SURF;
                        Rect_blk_edge(icrds,wave).edge[1][0].length = gr->h[0];
                        // This is wired code
                        Rect_blk_edge(icrds,wave).edge[1][1].etype = N_SURF;
                        Rect_blk_edge(icrds,wave).edge[1][1].length = gr->h[0];
                    }
                    else
                    {
                        Rect_blk_edge(icrds,wave).edge[0][0].cen[0] = crds_s[0];
                        Rect_blk_edge(icrds,wave).edge[0][0].cen[1] =
                               0.5*(crds_s[1]+gr->h[1] + Coords(crx[2][0]->pt)[1]);
                        Rect_blk_edge(icrds,wave).edge[0][0].length =
                            sqrt(sqr(crds_s[0]-Coords(crx[2][0]->pt)[0]) +
                                 sqr(crds_s[1]+gr->h[1] -Coords(crx[2][0]->pt)[1]));

                        Rect_blk_edge(icrds,wave).edge[1][0].etype = N_SURF;
                        Rect_blk_edge(icrds,wave).edge[1][0].length = gr->h[0];
                        // This is wired code
                        Rect_blk_edge(icrds,wave).edge[1][1].etype = FULL_SURF;
                        Rect_blk_edge(icrds,wave).edge[1][1].length = gr->h[0];
                    }
                    if(corn_comp[1][0] == comp)
                    {
                        Rect_blk_edge(icrds,wave).edge[0][1].cen[0] = crds_e[0];
                        Rect_blk_edge(icrds,wave).edge[0][1].cen[1] =
                               0.5*(crds_e[1]-gr->h[1] + Coords(crx[3][0]->pt)[1]);
                        Rect_blk_edge(icrds,wave).edge[0][1].length =
                            sqrt(sqr(crds_e[0]-Coords(crx[3][0]->pt)[0]) +
                                 sqr(crds_e[1]-gr->h[1]-Coords(crx[3][0]->pt)[1]));
                    }
                    else
                    {
                        Rect_blk_edge(icrds,wave).edge[0][1].cen[0] = crds_e[0];
                        Rect_blk_edge(icrds,wave).edge[0][1].cen[1] =
                               0.5*(crds_e[1] + Coords(crx[3][0]->pt)[1]);
                        Rect_blk_edge(icrds,wave).edge[0][1].length =
                            sqrt(sqr(crds_e[0]-Coords(crx[3][0]->pt)[0]) +
                                 sqr(crds_e[1]-Coords(crx[3][0]->pt)[1]));
                    }
                    a_frac[0] = 0.5*(Coords(crx[3][0]->pt)[1]-crds_s[1] + 
                                     Coords(crx[2][0]->pt)[1]-crds_s[1])*gr->h[0]/area;
                    a_frac[1] = 1.0-a_frac[0];
                    Rect_blk_edge(icrds,wave).config = 2;
                    // TMP
                    if(deb_flag)
                    {
                        printf("Cell fraction %10.9g, %10.9g\n",  a_frac[0], a_frac[1]);
                    }
                }
                else if(nc[0] == 1 && nc[1] == 0 && nc[2] == 1 && nc[3] == 0)
                {
                    if(wave_type(crx[0][0]->hs) < FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[2][0]->hs) < FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                    }
                    else if(wave_type(crx[0][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                            wave_type(crx[2][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTER_PART;
                    }
                    else
                    {
                        printf("ERROR: init_EB()\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        printf("Inconsistent wave types\n");
                        clean_up(ERROR);
                    }
                    Rect_blk_edge(icrds,wave).edge[0][0].etype = PARTIAL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][0].etype = PARTIAL_SURF;

                    Rect_blk_edge(icrds,wave).edge[0][0].cen[0] = crds_s[0];
                    Rect_blk_edge(icrds,wave).edge[0][0].cen[1] =
                               0.5*(crds_s[1]+gr->h[1] + Coords(crx[2][0]->pt)[1]);
                    Rect_blk_edge(icrds,wave).edge[0][0].length =
                            sqrt(sqr(crds_e[0]-Coords(crx[2][0]->pt)[0]) +
                                 sqr(crds_e[1]+gr->h[1] - Coords(crx[2][0]->pt)[1]));
                    Rect_blk_edge(icrds,wave).edge[1][0].cen[1] = crds_s[1];
                    Rect_blk_edge(icrds,wave).edge[1][0].cen[0] =
                               0.5*(crds_e[0]+Coords(crx[0][0]->pt)[0]);
                    Rect_blk_edge(icrds,wave).edge[1][0].length =
                            sqrt(sqr(crds_e[0]-Coords(crx[0][0]->pt)[0]) +
                                 sqr(crds_s[1]-Coords(crx[0][0]->pt)[1]));

                    Rect_blk_edge(icrds,wave).edge[0][1].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][1].length = gr->h[1];
                    Rect_blk_edge(icrds,wave).edge[1][1].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][1].length = gr->h[0];

                    a_frac[0] = 0.5*(Coords(crx[0][0]->pt)[0]-crds_s[0])*
                                    (Coords(crx[2][0]->pt)[1]-crds_s[1])/area;
                    a_frac[1] = 1.0-a_frac[0];
                    Rect_blk_edge(icrds,wave).config = 3;
                }
                else if(nc[0] == 0 && nc[1] == 1 && nc[2] == 0 && nc[3] == 1)
                {
                    if(wave_type(crx[1][0]->hs) < FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[3][0]->hs) < FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                    }
                    else if(wave_type(crx[1][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                            wave_type(crx[3][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTER_PART;
                    }
                    else
                    {
                        printf("ERROR: init_EB()\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        printf("Inconsistent wave types\n");
                        clean_up(ERROR);
                    }
                    Rect_blk_edge(icrds,wave).edge[0][1].etype = PARTIAL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][1].etype = PARTIAL_SURF;

                    Rect_blk_edge(icrds,wave).edge[0][1].cen[0] = crds_e[0];
                    Rect_blk_edge(icrds,wave).edge[0][1].cen[1] =
                               0.5*(crds_s[1] + Coords(crx[3][0]->pt)[1]);
                    Rect_blk_edge(icrds,wave).edge[0][1].length =
                            sqrt(sqr(crds_e[0]-Coords(crx[3][0]->pt)[0]) +
                                 sqr(crds_s[1] - Coords(crx[3][0]->pt)[1]));
                    Rect_blk_edge(icrds,wave).edge[1][1].cen[1] = crds_e[1];
                    Rect_blk_edge(icrds,wave).edge[1][1].cen[0] =
                               0.5*(crds_s[0]+Coords(crx[1][0]->pt)[0]);
                    Rect_blk_edge(icrds,wave).edge[1][1].length =
                            sqrt(sqr(crds_s[0]-Coords(crx[1][0]->pt)[0]) +
                                 sqr(crds_e[1]-Coords(crx[1][0]->pt)[1]));

                    Rect_blk_edge(icrds,wave).edge[0][0].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][0].length = gr->h[1];
                    Rect_blk_edge(icrds,wave).edge[1][0].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][0].length = gr->h[0];

                    a_frac[1] = 0.5*(crds_e[0]-Coords(crx[1][0]->pt)[0])*
                                    (crds_e[1]-Coords(crx[3][0]->pt)[1])/area;
                    a_frac[0] = 1.0-a_frac[1];
                    Rect_blk_edge(icrds,wave).config = 4;
                    // TMP
                    if(deb_flag)
                        printf("ic[%d %d] crds_e[%9.8g %9.8g] crds_s[%9.8g %9.8g], crx[1] [%9.8g %9.8g],"
                              "crx[3] [%9.8g %9.8g]\n",
                               icrds[0], icrds[1], crds_e[0], crds_e[1],
                               crds_s[0], crds_s[1],
                               Coords(crx[1][0]->pt)[0], Coords(crx[1][0]->pt)[1],
                               Coords(crx[3][0]->pt)[0], Coords(crx[3][0]->pt)[1]);
                }
                else if(nc[0] == 0 && nc[1] == 1 && nc[2] == 1 && nc[3] == 0)
                {
                    if(wave_type(crx[1][0]->hs) < FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[2][0]->hs) < FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                    }
                    else if(wave_type(crx[1][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                            wave_type(crx[2][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTER_PART;
                    }
                    else
                    {
                        printf("ERROR: init_EB()\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        printf("Inconsistent wave types\n");
                        clean_up(ERROR);
                    }
                    Rect_blk_edge(icrds,wave).edge[0][0].etype = PARTIAL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][1].etype = PARTIAL_SURF;
              
                    Rect_blk_edge(icrds,wave).edge[0][0].cen[0] = crds_s[0];
                    Rect_blk_edge(icrds,wave).edge[0][0].cen[1] =
                               0.5*(crds_s[1] + Coords(crx[2][0]->pt)[1]);
                    Rect_blk_edge(icrds,wave).edge[0][0].length =
                            sqrt(sqr(crds_s[0]-Coords(crx[2][0]->pt)[0]) +
                                 sqr(crds_s[1] - Coords(crx[2][0]->pt)[1]));
                    Rect_blk_edge(icrds,wave).edge[1][1].cen[1] = crds_e[1];
                    Rect_blk_edge(icrds,wave).edge[1][1].cen[0] =
                               0.5*(crds_e[0]+Coords(crx[1][0]->pt)[0]);
                    Rect_blk_edge(icrds,wave).edge[1][1].length =
                            sqrt(sqr(crds_e[0]-Coords(crx[1][0]->pt)[0]) +
                                 sqr(crds_e[1]-Coords(crx[1][0]->pt)[1]));

                    Rect_blk_edge(icrds,wave).edge[0][1].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][1].length = gr->h[1];
                    Rect_blk_edge(icrds,wave).edge[1][0].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][0].length = gr->h[0];

                    a_frac[1] = 0.5*(Coords(crx[1][0]->pt)[0]-crds_s[0])*
                                    (crds_e[1]-Coords(crx[2][0]->pt)[1])/area;
                    a_frac[0] = 1.0-a_frac[1];
                    Rect_blk_edge(icrds,wave).config = 5;

                    // TMP
                    if(deb_flag)
                        printf("ic[%d %d] crds_e[%9.8g %9.8g] crds_s[%9.8g %9.8g], crx[1] [%9.8g %9.8g],"
                              "crx[2] [%9.8g %9.8g]\n",
                               icrds[0], icrds[1], crds_e[0], crds_e[1],
                               crds_s[0], crds_s[1],
                               Coords(crx[1][0]->pt)[0], Coords(crx[1][0]->pt)[1],
                               Coords(crx[2][0]->pt)[0], Coords(crx[2][0]->pt)[1]);
                }
                else if(nc[0] == 1 && nc[1] == 1 && nc[2] == 0 && nc[3] == 0)
                {
                    if(wave_type(crx[0][0]->hs) < FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[1][0]->hs) < FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                    }
                    else if(wave_type(crx[0][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                            wave_type(crx[1][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTER_PART;
                    }
                    else
                    {
                        printf("ERROR: init_EB()\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        printf("Inconsistent wave types\n");
                        clean_up(ERROR);
                    }
                    Rect_blk_edge(icrds,wave).edge[1][0].etype = PARTIAL_SURF;
                    Rect_blk_edge(icrds,wave).edge[1][1].etype = PARTIAL_SURF;

                    if(Rect_blk_edge(icrds,wave).ctype == PARTIAL)
                    {
                        if(corn_comp[0][0] == comp)
                        {
                            Rect_blk_edge(icrds,wave).edge[1][0].cen[0] = 
                               0.5*(crds_s[0] + Coords(crx[0][0]->pt)[0]);
                            Rect_blk_edge(icrds,wave).edge[1][0].cen[1] = crds_s[1];
                            Rect_blk_edge(icrds,wave).edge[0][0].length =
                                sqrt(sqr(crds_s[0]-Coords(crx[0][0]->pt)[0]) +
                                 sqr(crds_s[1] - Coords(crx[0][0]->pt)[1]));

                            Rect_blk_edge(icrds,wave).edge[1][1].cen[1] = crds_e[1];
                            Rect_blk_edge(icrds,wave).edge[1][1].cen[0] =
                               0.5*(crds_s[0]+Coords(crx[1][0]->pt)[0]);
                            Rect_blk_edge(icrds,wave).edge[1][1].length =
                                sqrt(sqr(crds_s[0]-Coords(crx[1][0]->pt)[0]) +
                                 sqr(crds_s[1]-Coords(crx[1][0]->pt)[1]));
                        }
                        else
                        {
                            Rect_blk_edge(icrds,wave).edge[1][0].cen[0] = 
                               0.5*(crds_e[0] + Coords(crx[0][0]->pt)[0]);
                            Rect_blk_edge(icrds,wave).edge[1][0].cen[1] = crds_s[1];
                            Rect_blk_edge(icrds,wave).edge[0][0].length =
                                sqrt(sqr(crds_e[0]-Coords(crx[0][0]->pt)[0]) +
                                 sqr(crds_s[1] - Coords(crx[0][0]->pt)[1]));

                            Rect_blk_edge(icrds,wave).edge[1][1].cen[1] = crds_e[1];
                            Rect_blk_edge(icrds,wave).edge[1][1].cen[0] =
                               0.5*(crds_e[0]+Coords(crx[1][0]->pt)[0]);
                            Rect_blk_edge(icrds,wave).edge[1][1].length =
                                sqrt(sqr(crds_e[0]-Coords(crx[1][0]->pt)[0]) +
                                 sqr(crds_e[1]-Coords(crx[1][0]->pt)[1]));
                        }
                    }

                    Rect_blk_edge(icrds,wave).edge[0][0].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][0].length = gr->h[1];
                    Rect_blk_edge(icrds,wave).edge[0][1].etype = FULL_SURF;
                    Rect_blk_edge(icrds,wave).edge[0][1].length = gr->h[1];
                    a_frac[0] = 0.5*(Coords(crx[1][0]->pt)[0]-crds_s[0]+
                                     Coords(crx[0][0]->pt)[0]-crds_s[0])*gr->h[1]/area;
                    a_frac[1] = 1.0-a_frac[0];
                    Rect_blk_edge(icrds,wave).config = 6;
                    /*
                    printf("wave type : %d %d\n",
                               wave_type(crx[0][0]->hs), wave_type(crx[1][0]->hs));
                    printf("ERROR: init_EB() processing case \n");
                    printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                    clean_up(ERROR);
                    */
                }
                else if(nc[0] == 1 && nc[1] == 1 && nc[2] == 1 && nc[3] == 1)
                {
                    if(wave_type(crx[0][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[1][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[2][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE &&
                       wave_type(crx[3][0]->hs) >= FIRST_PHYSICS_WAVE_TYPE)
                    {
                        Rect_blk_edge(icrds,wave).ctype = INTERNAL;
                        for(k = 0; k < dim; k++)
                        {
                            Rect_blk_edge(icrds,wave).edge[k][0].etype = FULL_SURF;
                            Rect_blk_edge(icrds,wave).edge[k][1].etype = FULL_SURF;
                            Rect_blk_edge(icrds,wave).edge[k][0].length = gr->h[(k+1)%2];
                            Rect_blk_edge(icrds,wave).edge[k][1].length = gr->h[(k+1)%2];
                        }
                        Rect_blk_edge(icrds,wave).a_frac[0] = 1.0;
                        Rect_blk_edge(icrds,wave).config = 0;
                    }
                    else
                    {
                        if(comp_count != 0)
                        {
                            Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                            if(comp_label[0] == YES)
                                Rect_blk_edge(icrds,wave).edge[1][1].etype = N_SURF;
                            else
                                Rect_blk_edge(icrds,wave).edge[1][1].etype = FULL_SURF;
                            if(comp_label[1] == YES)
                                Rect_blk_edge(icrds,wave).edge[1][0].etype = N_SURF;
                            else
                                Rect_blk_edge(icrds,wave).edge[1][0].etype = FULL_SURF;
                            if(comp_label[2] == YES)
                                Rect_blk_edge(icrds,wave).edge[0][0].etype = N_SURF;
                            else
                                Rect_blk_edge(icrds,wave).edge[0][0].etype = FULL_SURF;
                            if(comp_label[3] == YES)
                                Rect_blk_edge(icrds,wave).edge[0][1].etype = N_SURF;
                            else
                                Rect_blk_edge(icrds,wave).edge[0][1].etype = FULL_SURF;
                            for(k = 0; k < dim; k++)
                            {
                                Rect_blk_edge(icrds,wave).edge[k][0].length = gr->h[(k+1)%2];
                                Rect_blk_edge(icrds,wave).edge[k][1].length = gr->h[(k+1)%2];
                            }
                            if(comp_label[0] == YES && comp_label[1] == YES &&
                               comp_label[2] == YES && comp_label[3] == YES)
                                Rect_blk_edge(icrds,wave).ctype = EXTERNAL;
                        }
                        else
                        {
                            printf("ERROR: init_EB() 1 1 1 1case needs to handle\n");
                            printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                            clean_up(ERROR);
                        }
                    }
                }
                else
                {
                    if(comp_count != 0)
                    {
                        Rect_blk_edge(icrds,wave).ctype = PARTIAL;
                        if(comp_label[0] == YES)
                            Rect_blk_edge(icrds,wave).edge[1][1].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[1][1].etype = FULL_SURF;
                        if(comp_label[1] == YES)
                            Rect_blk_edge(icrds,wave).edge[1][0].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[1][0].etype = FULL_SURF;
                        if(comp_label[2] == YES)
                            Rect_blk_edge(icrds,wave).edge[0][0].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[0][0].etype = FULL_SURF;
                        if(comp_label[3] == YES)
                            Rect_blk_edge(icrds,wave).edge[0][1].etype = N_SURF;
                        else
                            Rect_blk_edge(icrds,wave).edge[0][1].etype = FULL_SURF;
                        for(k = 0; k < dim; k++)
                        {
                            Rect_blk_edge(icrds,wave).edge[k][0].length = gr->h[(k+1)%2];
                            Rect_blk_edge(icrds,wave).edge[k][1].length = gr->h[(k+1)%2];
                        }
                        if(comp_label[0] == YES && comp_label[1] == YES &&
                           comp_label[2] == YES && comp_label[3] == YES)
                            Rect_blk_edge(icrds,wave).ctype = EXTERNAL;
                    }
                    else
                    {
                        printf("ERROR: init_EB() case needs to handle\n");
                        printf("ic[%d %d], nc[%d %d %d %d]\n", j, i, nc[0], nc[1], nc[2], nc[3]);
                        clean_up(ERROR);
                    }
                }
                if(deb_flag == YES)
                {
                    printf("init_EB(),  blk_edge[%d][%d].ctype = %d\n",
                       icrds[0],icrds[1], Rect_blk_edge(icrds,wave).ctype);
                    printf("L, U edge type [%d %d]\n",
                                Rect_blk_edge(icrds,wave).edge[0][0].etype,
                                Rect_blk_edge(icrds,wave).edge[0][1].etype);
                    printf("B, T edge type [%d %d]\n",
                                Rect_blk_edge(icrds,wave).edge[1][0].etype,
                                Rect_blk_edge(icrds,wave).edge[1][1].etype);
                }
            }
        }

        DEBUG_LEAVE(init_EB)
        return status;
}

/* 
storage order in side =[left,right,lower,upper].
*/
LOCAL int set_bdry_cell_edge(
        BLK_EDGE  *blk_edge,
        int       *ic,
        float     *crds,
        COMPONENT comp,
        BDRY_SIDE *side,
        Wave      *wv)
{
        int       i, k, dim = wv->rect_grid->dim, sd;
        float     ecrds[MAXD], coords_on[MAXD];
        float     *dh = wv->rect_grid->h;
        TRI_GRID  *tri_grid = wave_tri_soln(wv)->tri_grid;
        INTERFACE *intfc = tri_grid->grid_intfc;
        float     t[MAXD];
        HYPER_SURF_ELEMENT *hse;
        HYPER_SURF *hs;
        int        No_phy_bdry;

        for(k = 0; k < dim*2; k++)
        {
            if(side[k] != NOT_A_BDRY)
            {
                for(i = 0; i < dim; i++)
                    ecrds[i] = crds[i];

                No_phy_bdry = NO;
                switch(side[k])
                {
                    case LEFT_BDRY:
                        ecrds[0] -= 0.5*dh[0];
                        sd = 0;
                        if(buffered_boundary_type(rect_boundary_type(intfc,0,0)))
                            No_phy_bdry = YES;
                    break;
                    case RIGHT_BDRY:
                        ecrds[0] += 0.5*dh[0];
                        sd = 1;
                        if(buffered_boundary_type(rect_boundary_type(intfc,0,1)))
                            No_phy_bdry = YES;
                    break;
                    case LOWER_BDRY:
                        ecrds[1] -= 0.5*dh[1];
                        sd = 0;
                        if(buffered_boundary_type(rect_boundary_type(intfc,1,0)))
                            No_phy_bdry = YES;
                    break;
                    case UPPER_BDRY:
                        ecrds[1] += 0.5*dh[1];
                        sd = 1;
                        if(buffered_boundary_type(rect_boundary_type(intfc,1,1)))
                            No_phy_bdry = YES;
                    break;
                    default:
                       printf("ERROR: case not implemented in set_cell_edge\n");
                       clean_up(ERROR);
                }
                if (No_phy_bdry == NO)
                {
                    if(!nearest_interface_point(ecrds,comp,tri_grid->grid_intfc,
                        NO_SUBDOMAIN,NULL,coords_on,t,&hse,&hs))
                    {
                        printf("ERROR: set_bdry_cell_edge()\n");
                        printf("could not find the nearest intfc"
                          " point on cell boundary side\n");
                        clean_up(ERROR);
                    }

                    /*
                    if(ic[0] == 0 && ic[1] == 0)
                    {
                        printf("set_bdry_cell_edge: ecrds[%g %g] find curve side[%d], sd = %d\n", 
                        ecrds[0], ecrds[1], k, sd);
                        print_curve(hs->obj.c);
                    }
                    */

                    switch(wave_type(hs))
                    {
                    case NEUMANN_BOUNDARY:
                        blk_edge->edge[k<=1 ? 0:1][sd].etype = N_SURF;
                    break;
                    case DIRICHLET_BOUNDARY:
                        blk_edge->edge[k<=1 ? 0:1][sd].etype = D_SURF;
                    break;
                    default:
                    printf("ERROR: set_bdry_cell_edge()\n");
                    printf("unexpected boundary type for cell [%d %d] on side[%d] sd = %d, sid[k] = %d\n",
                             ic[0], ic[1], k, sd, side[k]);
                    clean_up(ERROR);
                    }
                    assign(blk_edge->edge[k<=1 ? 0:1][sd].cen,ecrds,sizeof(float)*2);
                    blk_edge->edge[k<=1 ? 0:1][sd].length = dh[k];
                }
                else
                {
                    blk_edge->edge[k<=1 ? 0:1][sd].etype = FULL_SURF;
                    blk_edge->edge[k<=1 ? 0:1][sd].length = dh[k];
                    continue;
                }
            }
            else
            {
                for(i = 0; i < dim; i++)
                    ecrds[i] = crds[i];
                if(k == 0)
                {
                    ecrds[0] -= 0.5*dh[0];
                    sd = 0;
                }
                else if(k == 1)
                {
                    ecrds[0] += 0.5*dh[0];
                    sd = 1;
                }
                else if(k == 2)
                {
                    ecrds[1] -= 0.5*dh[1];
                    sd = 0;
                }
                else
                {
                    ecrds[1] += 0.5*dh[1];
                    sd = 1;
                }
                blk_edge->edge[k<=1 ? 0:1][sd].etype = FULL_SURF;
                blk_edge->edge[k<=1 ? 0:1][sd].length = dh[k];
            }
        }
        return YES;
}

/* The physical boundary side is not buffered */
/* nonphyiscal boundary side is buffered*/
LOCAL int cell_on_phy_boundary(
        int        *ic,
        RECT_GRID  *gr,
        BDRY_SIDE  *side)
{
        static const BDRY_SIDE sd[] = { LEFT_BDRY,
                                          RIGHT_BDRY,
                                          LOWER_BDRY,
                                          UPPER_BDRY,
                                          ZMIN_BDRY,
                                          ZMAX_BDRY};
        int    i, dim = gr->dim, on_bd = NO;

        for(i = 0; i < dim*2; i++)
            side[i] = NOT_A_BDRY;
    
        for(i = 0; i < dim; i++)
        {
            if(ic[i] == imin[i]) 
            {
                side[2*i] = sd[2*i];
                on_bd = YES;
            }
            if(ic[i] == imax[i]-1)
            {
                side[2*i+1] = sd[2*i+1];
                on_bd = YES;
            }
        }
        return on_bd;
} 

EXPORT COMPONENT Find_rect_comp_ver2(
        int             *ic,
        Wave            *wave)
{
        int             imax0, imax1;

        imax0 = wave->rect_grid->gmax[0] + wave->rect_grid->ubuf[0];
        imax1 = wave->rect_grid->gmax[1] + wave->rect_grid->ubuf[1];

        if ((ic[0] < -wave->rect_grid->lbuf[0]) || (ic[0] >= imax0) ||
            (ic[1] < -wave->rect_grid->lbuf[1]) || (ic[1] >= imax1))
            return exterior_component(wave_tri_soln(wave)->tri_grid->grid_intfc);
        else
            return Rect_comp(ic,wave);

}               /*Find_rect_comp_ver2*/


// return component in the nbhd of cc
//            cn
//         cw cc ce
//            cs
EXPORT int cell_nbhd_config(
        COMPONENT    cn,
        COMPONENT    cs,
        COMPONENT    cw,
        COMPONENT    ce,
        int          *label)
{
        int    count = 0, i;

        for(i = 0; i < dim+2; i++)
            label[i] = NO;

        if(cn == ext_comp ||
           is_excluded_comp(cn,intfc) ||
           cn == POTS_COMP)
        {
            count++;
            label[0] = YES;
        }
        if(cs == ext_comp ||
           is_excluded_comp(cs,intfc) ||
           cs == POTS_COMP)
        {
            count++;
            label[1] = YES;
        }
        if(cw == ext_comp ||
           is_excluded_comp(cw,intfc) ||
           cw == POTS_COMP)
        {
            count++;
            label[2] = YES;
        }
        if(ce == ext_comp ||
           is_excluded_comp(ce,intfc) ||
           ce == POTS_COMP)
        {
            count++;
            label[3] = YES;
        }
        return count;
}

