#include <ghyp/ghyp.h>
#include <gdecs/vecdecs.h>
#include <ghyp/ghyploc.h>
// #include <gdecs/gdecs.h>

#if defined (PETSC)
#include "petscksp.h"
#endif /* if defined (PETSC) */

#define DEBUG_NPT

/* Local global used for irregular point */
/* For efficiency */
LOCAL   int     gmax[MAXD], vsten_rad, is_vec_method;
LOCAL   int     dim;
LOCAL   int     lbuf[MAXD], ubuf[MAXD];
LOCAL   float   h[MAXD];
LOCAL   float   L[MAXD], VL[MAXD], VU[MAXD];
LOCAL   int     PERIOD_BC = NO;
LOCAL   int     *cell_index = NULL;

LOCAL   COMPONENT       ext_comp;
LOCAL   float           hdir, hndir, dir[MAXD];
LOCAL   GRID_DIRECTION  prev_side, next_side;
LOCAL   INTERFACE       *intfc = NULL;
LOCAL   int             endpt;
LOCAL   int             deb_sten = NO;
// Chemical components related 
LOCAL   float           *mu_chem;
LOCAL   int             n_chem = 0;
/* End of Local global used for irregular point */

LOCAL int          excluded_comp(COMPONENT,INTERFACE*);
LOCAL void         nor_deriv_sweep_3l_comp_segments(int,int*,float*,
                      Wave*,Wave*,Front*,Front*,int*,int,int,float,float*,int*,NS_Field_Stencil*);
LOCAL Vec_NS_Gas*  alloc_ns_vgas(Vec_NS_Gas*,int,int,int);
LOCAL void         include_normal_vari(int,int*,float*,Wave*,Wave*,Front*,Front*,
                      int*,int,int,float,float*,NS_Field_Stencil*);
LOCAL int          g_load_NS_state_vectors(int,int*,Vec_NS_Gas*,int,int,Wave*,Wave*,int*,int);
LOCAL void         upwind_reconstruct_in_dir(int,int*,/*float*,*/Wave*,/*Wave*,*/Front*,Front*,int,int,int*,
                         /*int,int,*/float,float*,int,NS_Field_Stencil*,int);
LOCAL void         set_NS_gas_storage(float**,Vec_NS_Gas*,Upwind_NS*);
LOCAL void         ns2d_unsplit_bct_linear_reconstructor(int,int,Upwind_NS*);
LOCAL void         limiter_limiting(float**,int,int,Upwind_NS*);
LOCAL void         ns2d_normal_vel_half_step(int,int,float,float*,int,int*,int,Upwind_NS*);
LOCAL void         ns2d_transerve_vel_step(int,int,float,float*,int,int*,int,Upwind_NS*);
LOCAL void         limited_vel_slope(int,float*,float*,float*,float*);
LOCAL void         free_ns_vgas(Vec_NS_Gas*);
LOCAL void         press_grad_and_visc(int,int,float,float*,int,int*,int,Upwind_NS*,
                      NS_Field_Stencil*,int*,int);
LOCAL void         include_chem_src(int,int,float,float*,int,int*,int,Upwind_NS*,
                      NS_Field_Stencil*,int*,int);
LOCAL void         irregular_pt_nor_deriv(int,int*,float*,float,Wave*,Wave*,Front*,Front*,NS_Field_Stencil*);
LOCAL NS_Field_Stencil *alloc_NS_field_space(NS_Field_Stencil*,Wave*);
LOCAL void         project_half_step_vel(NS_Field_Stencil*);

/* Functions for irregular cells */
LOCAL   void       NS_reg_grid_state(float*,float,int,int*,int,NS_Field_Stencil*,POINT*,COMPONENT);
LOCAL   COMPONENT  Find_rect_comp_ver1(int,int*,Wave*);
LOCAL   bool       NS_interior_sten_reg(int,int,int,int*,Stencil*);
// LOCAL   COMPONENT  Find_rect_comp_ver2(int*,Wave*);
// LOCAL   COMPONENT  set_NS_stencil_comps_and_crosses(int,int,int*,Stencil*,int);
LOCAL   void       find_outer_ns_states(int,int,Stencil*,CRX_SIDE);
LOCAL   void       copy_ns_states(int,CRX_SIDE,int,int,Stencil*,CRXING*);
LOCAL   void       set_ns_dirichlet_bdry_sten_states(int,HYPER_SURF*,Stencil*,
                                               CRX_SIDE);
LOCAL   void       set_ns_neumann_bdry_sten_states(int,int,int,COMPONENT,CRXING*,
                                             Stencil*,CRX_SIDE);
LOCAL   int        crds_of_ic_ver1(float*,Wave*,int*);
LOCAL   void       find_out_ns_states_in_other_dir(int,int,int,int*,Stencil*);
LOCAL   void       copy_Tstates_ver1(int,float*,int,COMPONENT,Stencil*,CRXING*);
LOCAL   void       set_ns_neumann_bdry_sten_Tstates(int,int,float*,int,COMPONENT,CRXING*,
                                             Stencil*);
LOCAL   void       npt_normal_vari(float*,float,Locstate,const float*,int,int*,int*,NS_Field_Stencil*);
LOCAL   void       clear_vec_ns_gas_set_flags(Vec_NS_Gas*);

LOCAL   void       nor_advec_vel_on_edge(float*,float,Wave*,Wave*,Front*,Front*,
                         COMPONENT,NS_Field_Stencil*);

LOCAL   float      colella_limiting_vel(float,float);
LOCAL   void       upwinding_on_edge(float*,float,Wave*,Wave*,Front*,Front*,
                         COMPONENT,NS_Field_Stencil*);
LOCAL   void       project_vel_star(float,NS_Field_Stencil*);
LOCAL   void       cell_edge_vector_u(float,NS_Field_Stencil*);
// LOCAL   void       comput_final_vel_on_edge(float*,float,Wave*,Wave*,Front*,Front*,
//                          COMPONENT,NS_Field_Stencil*);
#if defined(PETSC)
LOCAL   float     K[2] = {0.5,0.5};
LOCAL   int       comm_size = 1;
LOCAL   void      set_matrix_row(Mat,int,int,int**,int**,NS_Field_Stencil*);
LOCAL   void      set_right_hand_side(Vec,int,int,int**,NS_Field_Stencil*);
// LOCAL   void      create_global_element_list(Front*,int,int,int*,int**,int**);
// LOCAL   int       cell_nbhd_config(COMPONENT,COMPONENT,COMPONENT,COMPONENT,int*);
#endif /* if defined(PETSC) */
LOCAL   void      init_stencil_element2D(BLK_EDGE**,P_CELL2D**,Front*,Wave*,NS_Field_Stencil*,int,float);
LOCAL   void      init_stencil_element2D_ver1(BLK_EDGE**,P_CELL2D**,Front*,Wave*,NS_Field_Stencil*,int,float);
LOCAL   void      init_stencil_element2D_ver2(BLK_EDGE**,P_CELL2D**,Front*,Wave*,NS_Field_Stencil*,int,float);
LOCAL   void      init_stencil_element2D_ver3(Wave*,P_CELL2D**,Front*,Wave*,NS_Field_Stencil*,int,float);
LOCAL   void      regular_fd_sten(P_CELL2D*,float*,float*,int,int,int*,NS_Field_Stencil*,float);
LOCAL   void      regular_fd_sten_ver2(P_CELL2D*,float*,float*,int,int,int*,NS_Field_Stencil*,float,int);
LOCAL   void      bdry_cell_sten(P_CELL2D*,float*,float*,int,int,int*,
                                 NS_Field_Stencil*,ELEMENT_TYPE,float);
LOCAL   void      bdry_cell_sten_ver2(P_CELL2D*,float*,float*,int,int,int*,
                                 NS_Field_Stencil*,ELEMENT_TYPE,float);
LOCAL   void      bdry_cell_sten_ver3(P_CELL2D*,float*,float*,int,int,int*,
                                 NS_Field_Stencil*,ELEMENT_TYPE,float,int);
LOCAL   void      set_matrix_A_and_right_hand_side(Mat,Vec,Wave*,
                       P_CELL2D**,int,int,int**,int**,NS_Field_Stencil*);
// LOCAL   void      adjust_nor_vel_on_edge(NS_Field_Stencil*);
LOCAL   void      adjust_nor_vel_on_edge_ver2(NS_Field_Stencil*);
LOCAL   void      adjust_nor_vel_on_edge_ver3(NS_Field_Stencil*);
LOCAL   void      build_convective_deriv(NS_Field_Stencil*);
LOCAL   void      get_cell_edge_nor_vel(int*,NS_Field_Stencil*,float**,VEL_ST**);
LOCAL   void      partial_surf_sten(P_CELL2D*,float*,float*,int,int*,
                       NS_Field_Stencil*,int,int,BLK_EDGE**,float);
LOCAL   void      grid_den_viscosity(NS_Field_Stencil*,Wave*,Front*);
LOCAL   void      grid_den_viscosity_ver1(NS_Field_Stencil*,Wave*,Front*);
LOCAL   void      grid_den_viscosity_full_cell_ver(NS_Field_Stencil*,Wave*,Front*);

LOCAL   void      diffusion_convection_step(float,NS_Field_Stencil*);
LOCAL   void      set_Crank_Nicolson_sten(float,int*,NS_Field_Stencil*,P_CELL2D*,P_CELL2D*);
// LOCAL   bool      Crank_N_sten_reg(int,Stencil*,int*);
// LOCAL   void      sten_pt_loc_in_dir(int*,int,int,Wave*,int,Stencil*);
// LOCAL   int       velocity_from_boundary(Stencil*,CRXING*,POINT*,COMPONENT,Locstate,float*); 
LOCAL   void      edge_1_over_den(NS_Field_Stencil*,Wave*);
LOCAL   void      half_step_den(NS_Field_Stencil*,Wave*);
LOCAL   void      cell_edge_vector_u_ver2(float,NS_Field_Stencil*);
LOCAL   void      cell_edge_vector_u_ver3(float,NS_Field_Stencil*);
LOCAL   void      cell_edge_vector_u_ver4(float,NS_Field_Stencil*);
LOCAL   void      update_vel_press(float,NS_Field_Stencil*);
LOCAL   void      update_vel_press_ver2(float,NS_Field_Stencil*);
LOCAL   void      free_NS_field_space(NS_Field_Stencil*);
LOCAL   void      NS_set_params_jumps(Vec_NS_Gas*,int,int);
// LOCAL   void      enforce_neumann_BC_on_nor_advec_vel(NS_Field_Stencil*);
LOCAL Chem_Stencil *alloc_Chem_field_sten(Chem_Stencil*,Wave*,int);
// LOCAL void        factorize_vel_on_edge(Wave*,Front*,NS_Field_Stencil*);
LOCAL   bool      g_NS_load_source_vectors(int,int*,float**,Vec_NS_Gas*,Vec_Src*,
                                           int,int,RECT_GRID*);
LOCAL   void      scatter_projection(NS_Field_Stencil*);
LOCAL   void      enforce_BC_on_scalar_field(NS_Field_Stencil*,Front*,Wave*);
LOCAL   void      scatter_diff_conv_vel(NS_Field_Stencil*);
LOCAL   void      fix_ghost_boundary_cells(Wave*,Front*);
/* Debugging flag */
LOCAL   int        deb_npt = NO;

// The implementation follows J. Bell & P. Colella "A Cartesian 
// Grid Projection Method for the Incompressible Euler Equations
// in Complex Geometries"
// and J. Bell "A Second-Order Projection Method for Variable-Density
// Flow", J.C.P. 101, 334-348 (1992).
// The computation of transverse derivative, which corresponds to the vU_y term 
// in Eq. (3.1) is modified to use 3-point monotonically limited central difference.
// If we wish to use upwinding, we need to increase transervse stencil size.

/* THis is a 2D solver actually */
EXPORT int g_FD_NS_solver(
        int        *iperm,
        float      *dh,
        float      dt,
        Wave       *wv,
        Wave       *nwv,
        Front      *fr,
        Front      *nfr,
        COMPONENT  max_comp)
{
	RECT_GRID  *gr = fr->rect_grid;
	int        k, i;
	NS_Field_Stencil *ns_field = NULL;
	// int        edgexmax, edgeymax, edgex_n_segs, edgey_n_segs;
        int        ix, iy;
        int        imin[3], imax[3], idirs[3];
        int        icoords[MAXD], i0min, i0max, vsize, i1;
        float      dir[MAXD];
	int        nrad; 
        Vec_NS_Gas  *vgas = NULL;
        Vec_NS_Gas  *Tvgas[2] = {NULL, NULL};    /* Unsplit, for cross direvative.
                                                  * Tangential upper and lower state.
                                                  */
        Chem_Stencil  *chem_sten = NULL;
        //Source
        Vec_Src       *src = NULL;

	DEBUG_ENTER(g_FD_NS_solver)

        /* Set local globals */
        dim = gr->dim;
        for (i = 0; i < dim; ++i)
        {
            gmax[i] = gr->gmax[i];
            h[i] = gr->h[i];
            L[i] = gr->L[i];
            VL[i] = gr->VL[i];
            VU[i] = gr->VU[i];
            lbuf[i] = max(gr->lbuf[i], 4);
            ubuf[i] = max(gr->ubuf[i], 4);
        }
        intfc = nfr->interf;
        ext_comp = exterior_component(intfc);
        /* END set local globals */

        if(rect_boundary_type(fr->interf,0,0) == SUBDOMAIN_BOUNDARY)
            PERIOD_BC = YES;

        // TMP
        // printf("step = %d, enter g_FD_NS_solver, PERIOD_BC = %d\n", fr->step, PERIOD_BC);
                 
        // Fix ghost cell values For RT problem
        // This work is moved into irregular_pt_nor_deriv.
        // fix_ghost_boundary_cells(wv,fr); 

        if(ns_field == NULL)
	    ns_field = alloc_NS_field_space(ns_field,wv);
        ns_field->n_chem = n_chem;
        if(n_chem != 0)
        {
            ns_field->chem_sten = chem_sten = 
                 alloc_Chem_field_sten(chem_sten,wv,n_chem);
            ns_field->chem_sten->mu_chem = mu_chem;
        }

	/* Temp. set to use 2nd order scheme */
	nrad = (ns_field->npts_sten - ns_field->npts_sten%2)/2;

        ns_field->rect_gr = wv->rect_grid;
        ns_field->fr = fr;
        ns_field->newfr = nfr;
        ns_field->wave = wv;
        ns_field->newwave = nwv;

        // load density and viscosity
        // The density averaged over volume fraction is assigned to Wave.
        // Should use ns_field->wave here.
        // grid_den_viscosity(ns_field,ns_field->newwave,ns_field->fr);

        grid_den_viscosity_ver1(ns_field,ns_field->wave,ns_field->fr);
        // grid_den_viscosity_full_cell_ver(ns_field,ns_field->wave,ns_field->fr);
        if(debugging("scalar_field"))
            enforce_BC_on_scalar_field(ns_field,ns_field->fr,ns_field->wave);
        edge_1_over_den(ns_field, wv);

	// construct advective terms
	for(k = 0; k < dim; k++)
	{
            for (i = 0; i < dim; ++i)
            {
                idirs[i] = iperm[(i+k)%dim];
                dir[i] = 0.0;
            }
            /* Set sweep range */
            for (i = 0; i < dim; ++i)
            {
                imin[i] = 0;
                imax[i] = gr->gmax[idirs[i]];
            }
            for (i = 1; i < dim; ++i)
            {
                if (gr->lbuf[idirs[i]] > 0)
                    imin[i] -= gr->lbuf[idirs[i]];
                if (gr->ubuf[idirs[i]] > 0)
                    imax[i] += gr->ubuf[idirs[i]];
            }
            if (gr->lbuf[idirs[0]] > 0)
                imin[0] -= nrad;
            if (gr->ubuf[idirs[0]] > 0)
                imax[0] += nrad;

            /* If sweep 3 lines together */
            for (i = 1; i < dim; ++i)
            {
                imin[i] += 1;
                imax[i] -= 1;
            }

            dir[idirs[0]] = 1.0;
            i0min = imin[0];        i0max = imax[0];

            vsize = i0max - i0min;
            // ns_field->vgas = alloc_ns_vgas(ns_field->vgas,vsize,dim);
            vgas = alloc_ns_vgas(vgas,vsize,dim,n_chem);
            Tvgas[0] = alloc_ns_vgas(Tvgas[0],vsize,dim,n_chem);
            Tvgas[1] = alloc_ns_vgas(Tvgas[1],vsize,dim,n_chem);
            // Source
            src = muscl_alloc_vsrc(src,vsize,wv); 
            ns_field->vgas = vgas;
            ns_field->Tvgas[0] = Tvgas[0];
            ns_field->Tvgas[1] = Tvgas[1];
            // Source
            ns_field->src = src;   

            icoords[idirs[0]] = 0;  /* value used inside */

	    // printf("sweep[%d] idir[%d %d], imin[1]-imax[1] %d %d, i0min-i0max %d %d\n",
            //        k, idirs[0],idirs[1],imin[1], imax[1], i0min,i0max);

            for(i1 = imin[1]; i1 < imax[1]; i1++)
            {
                icoords[idirs[1]] = i1;
                nor_deriv_sweep_3l_comp_segments(k,iperm,dir,wv,nwv,fr,
                        nfr,icoords,i0min,i0max,dt,dh,idirs,ns_field);
            }

            free_ns_vgas(vgas);
            free_ns_vgas(Tvgas[0]);
            free_ns_vgas(Tvgas[1]);
            vgas = Tvgas[1] = Tvgas[0] = NULL; 
            // Source 
            g_free_vsrc(src);
            src = NULL;            

            // TMP
            // printf("---------------Before irregular_pt_nor_deriv\n");
            // fflush(stdout);

            irregular_pt_nor_deriv(k,iperm,dh,dt,wv,nwv,fr,nfr,ns_field);

            // printf("---------------After irregular_pt_nor_deriv\n");
            // fflush(stdout);
	}
		
        // Compute normal advective velocity on the edge
        nor_advec_vel_on_edge(dh,dt,wv,nwv,fr,nfr,max_comp,ns_field);
        // enforce_neumann_BC_on_nor_advec_vel(ns_field);

        // Upwind velocity and density, chemicals on edge
        upwinding_on_edge(dh,dt,wv,nwv,fr,nfr,max_comp,ns_field);
        // factorize_vel_on_edge(wv,fr,ns_field);

        /* If we increase transverse stencil size, the transverse derivative
         * could be computed at the same time. Although this causes some 
         * overhead in computing, it is easy to implement in case that
         * complex geometry boundary is presented.
        // Compute transverse derivative on the field
        // Compute velocity on edge, include Gp, viscosity, transverse 
        comput_final_vel_on_edge(dh,dt,wv,nwv,fr,nfr,max_comp,ns_field);
        */
 
        /*** DO NOT PROJECT half step VEL **/
        // printf("Before project_half_step_vel\n");
        // MAC Projection on reconstructed velocity
        project_half_step_vel(ns_field);

        // Here, we should update the pressure within the ghost cells.
        // NOTE: This should be implemented like scatter_state
        // For now, since we use periodic boundary (or symmetric) condition in the
        // x-direction, we implement a simple version.
        // Homo. Dirichlet BC for top and bottom boundaries
        // Linearly extrapolate phi for top and bottom boundaries
        scatter_projection(ns_field);
   
        // printf("\n\n********After project_half_step_vel\n");
        // printf("Entered g_FD_NS_solver\n");
	// printf("stop after project_half_step_vel\n");
	// clean_up(0);

        // adjust normal velocity on cell edge according to projected phi
        // adjust_nor_vel_on_edge(ns_field);
        adjust_nor_vel_on_edge_ver2(ns_field);
        // adjust_nor_vel_on_edge_ver3(ns_field);


        // Convective derivative
        build_convective_deriv(ns_field);

        // grid_den_viscosity(ns_field,ns_field->newwave,ns_field->newfr);
        // The density averaged over volume fraction is assigned to NEWWave.

        grid_den_viscosity_ver1(ns_field,ns_field->newwave,ns_field->newfr);
        // grid_den_viscosity_full_cell_ver(ns_field,ns_field->newwave,ns_field->newfr);
        if(debugging("scalar_field"))
            enforce_BC_on_scalar_field(ns_field,ns_field->newfr,ns_field->newwave);
        // Do viscosity here
        edge_1_over_den(ns_field, nwv);

        // \rho^{n+1/2} on cell edge is used for the projection step 
        half_step_den(ns_field, nwv);

        // printf("\n\n ******* Before diffusion_convection_step **************\n\n");
        diffusion_convection_step(dt,ns_field);

        // printf("EXIT after diffusion_convection_step\n");
        // exit(0);

        // Top and bottom ghost cells are set here.
        scatter_diff_conv_vel(ns_field);

        // printf("Before project_vel_star\n");
        project_vel_star(dt,ns_field);

        // Here, we should update the pressure within the ghost cells.
        // For now, since we use periodic boundary condition in the
        // x-direction, we implement a simple version.
        // Homo. Dirichlet BC for top and bottom boundaries.
        // Linearly extrapolate phi for top and bottom boundaries
        scatter_projection(ns_field);

        // update u^star and pressure field
        // update_vel_press(dt,ns_field);
        // Do not add the projection correction
        update_vel_press_ver2(dt,ns_field);

        /**
        // TEST divergence free()
        {
            int ic0[MAXD] = {1,1};
            int icl[MAXD] = {0,1};
            int icr[MAXD] = {2,1};
            int ict[MAXD] = {1,2};
            int icb[MAXD] = {1,0};
            float div, *dh = wv->rect_grid->h;
            div = (Vel(Rect_state(icr,nwv))[0] - Vel(Rect_state(icl,nwv))[0])/(2.0*dh[0]) + 
                  (Vel(Rect_state(ict,nwv))[1] - Vel(Rect_state(icb,nwv))[1])/(2.0*dh[1]);
            printf("div at [%d %d] = %g\n", ic0[0], ic0[1], div);
        }
        **/

        free_NS_field_space(ns_field);

        if(chem_sten != NULL)
        {
            g_React_Diff_solver(iperm,dh,dt,wv,nwv,fr,nfr,max_comp,chem_sten);
            free_these(14, chem_sten->hedge_l_storage, chem_sten->hedge_r_storage,
                       chem_sten->hel_st, chem_sten->her_st, chem_sten->hedge_m_storage,
                       chem_sten->hem_st, chem_sten->vedge_l_storage, chem_sten->vedge_r_storage,
                       chem_sten->vel_st, chem_sten->ver_st, chem_sten->vedge_m_storage,
                       chem_sten->vem_st, chem_sten->con_der_storage, chem_sten->con_der);
            free(chem_sten);
        }

	DEBUG_LEAVE(g_FD_NS_solver)
}

LOCAL void fix_ghost_boundary_cells(
       Wave      *wv,
       Front     *fr)
{
       RECT_GRID *gr = wv->rect_grid;
       int       i, dim = gr->dim, ix, iy;
       int       smin[MAXD], smax[MAXD], ic[MAXD], ic2[MAXD], ic3[MAXD];
       float     nor[MAXD] = {0.0, 1.0};

       for(i = 0; i < dim; i++)
       {
           smin[i] = -gr->lbuf[i];
           smax[i] = gr->gmax[i] + gr->ubuf[i];
       }

       for(ix = smin[0]; ix < smax[0]; ix++)
       {
           // Top side
           for(iy = gr->gmax[1]; iy < smax[1]; iy++)
           {
               ic[0] = ix; ic[1] = iy; 
               ic2[1] = 2*gmax[1] - ic[1]-1;  ic2[0] = ix;
               assign(Rect_state(ic,wv),Rect_state(ic2,wv),fr->sizest);
               g_reflect_state(Rect_state(ic,wv),fr->interf,NULL,NULL,nor);
               // Linear extrapolate pressure, only need to do this
               // for the first ghost cell
               if(iy == gr->gmax[1])
               {
                   ic3[1] = ic2[1]-1;  ic3[0] = ix;
                   Press(Rect_state(ic,wv)) = 2.0*Press(Rect_state(ic2,wv)) - Press(Rect_state(ic3,wv));
               }
           }
           // bottom side
           for(iy = -1; iy >= smin[1]; iy--)
           {
               ic[0] = ix; ic[1] = iy;
               ic2[1] = -ic[1]-1;  ic2[0] = ix;
               assign(Rect_state(ic,wv),Rect_state(ic2,wv),fr->sizest);
               g_reflect_state(Rect_state(ic,wv),fr->interf,NULL,NULL,nor);
               if(iy == -1)
               {
                   ic3[1] = ic2[1]+1;  ic3[0] = ix;
                   Press(Rect_state(ic,wv)) = 2.0*Press(Rect_state(ic2,wv)) - Press(Rect_state(ic3,wv));
                   // 
                   // printf("reset[%d %d] pressure %9.8g\n", ic[0], ic[1], Press(Rect_state(ic,wv)));
               }
           }
       }

}

LOCAL void scatter_diff_conv_vel(
       NS_Field_Stencil  *ns_field)
{
       Wave      *wv = ns_field->wave;   
       Front     *fr = ns_field->fr;   
       RECT_GRID *gr = wv->rect_grid;
       int       i, dim = gr->dim, ix, iy;
       int       smin[MAXD], smax[MAXD], ic[MAXD], ic2[MAXD];

       for(i = 0; i < dim; i++)
       {
           smin[i] = -gr->lbuf[i];
           smax[i] = gr->gmax[i] + gr->ubuf[i];
       }

       if(rect_boundary_type(fr->interf,0,0) == SUBDOMAIN_BOUNDARY)
       {
           for(iy = 0; iy < gmax[1]; iy++)
           {
               for(ix = smin[0]; ix < 0; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = gmax[0]+ic[0];  ic2[1] = iy;
                   Rect_u_star(ic, ns_field).v[0] = Rect_u_star(ic2, ns_field).v[0];
                   Rect_u_star(ic, ns_field).v[1] = Rect_u_star(ic2, ns_field).v[1];
               }
               for(ix = gr->gmax[0]; ix < smax[0]; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = ic[0] - gmax[0];  ic2[1] = iy;
                   Rect_u_star(ic, ns_field).v[0] = Rect_u_star(ic2, ns_field).v[0];
                   Rect_u_star(ic, ns_field).v[1] = Rect_u_star(ic2, ns_field).v[1];
               }
           }
       }
       else
       {
            // Reflection side
           for(iy = 0; iy < gmax[1]; iy++)
           {
               for(ix = smin[0]; ix < 0; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = -ic[0]-1;  ic2[1] = iy;
                   Rect_u_star(ic, ns_field).v[0] = -1.0*Rect_u_star(ic2, ns_field).v[0];
                   Rect_u_star(ic, ns_field).v[1] = Rect_u_star(ic2, ns_field).v[1];
               }
               for(ix = gr->gmax[0]; ix < smax[0]; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = 2*gmax[0] - ic[0]-1;  ic2[1] = iy;
                   Rect_u_star(ic, ns_field).v[0] = -1.0*Rect_u_star(ic2, ns_field).v[0];
                   Rect_u_star(ic, ns_field).v[1] = Rect_u_star(ic2, ns_field).v[1];
                   /**
                   if(iy == 4)
                   {
                       printf("\nIn scatter_projection\n");
                       printf("ic[%d %d] press %9.8g from [%d %d] press %9.8g\n\n",
                               ic[0], ic[1], Rect_phi(ic, ns_field),
                               ic2[0], ic2[1], Rect_phi(ic2, ns_field));
                   }
                   **/
               }
           }
       }

       // Top and bottome side uses homo Dirichlet BC
       // Top and bottome side uses Dirichlet BC
       for(ix = smin[0]; ix < smax[0]; ix++)
       {
           // Top side, set first ghost cell
           // for(iy = gr->gmax[1]; iy < smax[1]; iy++)
           {
               ic[0] = ix; ic[1] = gr->gmax[1];
               ic2[1] = 2*gmax[1] - ic[1]-1;  ic2[0] = ix;

               Rect_u_star(ic, ns_field).v[0] = -1.0*Rect_u_star(ic2, ns_field).v[0];
               Rect_u_star(ic, ns_field).v[1] = -1.0*Rect_u_star(ic2, ns_field).v[1];
               // Rect_u_star(ic, ns_field).v[0] = 1.0*Rect_u_star(ic2, ns_field).v[0];
               // Rect_u_star(ic, ns_field).v[1] = 1.0*Rect_u_star(ic2, ns_field).v[1];
           }
           // bottom side, set first ghost cell
           // for(iy = -1; iy >= smin[1]; iy--)
           {
               ic[0] = ix; ic[1] = -1;
               ic2[1] = -ic[1]-1;  ic2[0] = ix;

               Rect_u_star(ic, ns_field).v[0] = -1.0*Rect_u_star(ic2, ns_field).v[0];
               Rect_u_star(ic, ns_field).v[1] = -1.0*Rect_u_star(ic2, ns_field).v[1];
               // Rect_u_star(ic, ns_field).v[0] = 1.0*Rect_u_star(ic2, ns_field).v[0];
               // Rect_u_star(ic, ns_field).v[1] = 1.0*Rect_u_star(ic2, ns_field).v[1];

               // TMP
               // printf("scatter_diff_conv_vel, ic[%d %d] by [%d %d] vel (%g %g) (%g %g)\n",
               //              ic[0], ic[1], ic2[0], ic2[1], 
               //               Rect_u_star(ic, ns_field).v[0], Rect_u_star(ic, ns_field).v[1],
               //                    Rect_u_star(ic2, ns_field).v[0], Rect_u_star(ic2, ns_field).v[1]);
           }
       }
}

LOCAL void scatter_projection(
       NS_Field_Stencil  *ns_field)
{
       Wave      *wv = ns_field->wave;   
       Front     *fr = ns_field->fr;   
       RECT_GRID *gr = wv->rect_grid;
       int       i, dim = gr->dim, ix, iy;
       int       smin[MAXD], smax[MAXD], ic[MAXD], ic2[MAXD], ic3[MAXD];

       for(i = 0; i < dim; i++)
       {
           smin[i] = -lbuf[i];
           smax[i] = gmax[i] + ubuf[i];
       }
       
       if(rect_boundary_type(fr->interf,0,0) == SUBDOMAIN_BOUNDARY)
       {
           for(iy = 0; iy < gmax[1]; iy++)
           {
               for(ix = smin[0]; ix < 0; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = gmax[0]+ic[0];  ic2[1] = iy;
                   Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
               }
               for(ix = gr->gmax[0]; ix < smax[0]; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = ic[0] - gmax[0];  ic2[1] = iy;
                   Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
               }
           }          
       }
       else
       {
           for(iy = 0; iy < gmax[1]; iy++)
           {
               // reflecting
               for(ix = smin[0]; ix < 0; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = -ic[0]-1;  ic2[1] = iy;
                   Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
               }

               // linearly extrapolate left side
               /**
               ic[0] = -1; ic[1] = iy;
               ic2[0] = -ic[0]-1;  ic2[1] = iy;
               ic3[0] = ic2[0]+1;  ic3[1] = iy;

               // printf("In scatter_projection [%d %d]\n", ic[0], ic[1]);
               Rect_phi(ic, ns_field) = 2.0*Rect_phi(ic2,ns_field) - Rect_phi(ic3,ns_field);

               // linearly extrapolate right side
               ic[0] = gmax[0];               ic[1] = iy;
               ic2[0] = 2*gmax[0] - ic[0]-1;  ic2[1] = iy;
               ic3[0] = ic2[0]-1;             ic3[1] = iy;

               // printf("In scatter_projection [%d %d]\n", ic[0], ic[1]);
               Rect_phi(ic, ns_field) = 2.0* Rect_phi(ic2,ns_field) - Rect_phi(ic3,ns_field);
               **/

               /** Reflecting */
               for(ix = gr->gmax[0]; ix < smax[0]; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = 2*gmax[0] - ic[0]-1;  ic2[1] = iy;
                   Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
                   // if(iy == 4)
                   // {
                   //     printf("\nIn scatter_projection\n");
                   //     printf("ic[%d %d] press %9.8g from [%d %d] press %9.8g\n\n",
                   //             ic[0], ic[1], Rect_phi(ic, ns_field), 
                   //             ic2[0], ic2[1], Rect_phi(ic2, ns_field));
                   // }
               }
           }          
       }

       // Top and bottome side uses homo Dirichlet BC for velocity, 
       // linearly extrapolate phi
       for(ix = smin[0]; ix < smax[0]; ix++)
       {
           // Top side, set first ghost cell
           // for(iy = gr->gmax[1]; iy < smax[1]; iy++)
           if(smax[1] > gr->gmax[1])
           {
               ic[0] = ix; ic[1] = gr->gmax[1];
               ic2[0] = ix, ic2[1] = 2*gmax[1] - ic[1]-1;
               ic3[0] = ix; ic3[1] = ic2[1]-1; 

               Rect_phi(ic, ns_field) = 2.0* Rect_phi(ic2,ns_field) - Rect_phi(ic3,ns_field);
           
               //  TMP
               /**
               if(ic[0] == 0 || ic[1] == 1)
               printf("scatter_diff_conv_vel, ic[%d %d] by [%d %d] [%d %d] phi (%g %g), ans %g\n",
                            ic[0], ic[1], ic2[0], ic2[1], ic3[0], ic3[1], 
                                   Rect_phi(ic2, ns_field), Rect_phi(ic3, ns_field),
                                   Rect_phi(ic, ns_field));
               **/
           }
           // bottom side, set first ghost cell
           // for(iy = -1; iy >= smin[1]; iy--)
           if(smin[1] < 0)
           {
               ic[0] = ix; ic[1] = -1;
               ic2[0] = ix; ic2[1] = -ic[1]-1; 
               ic3[0] = ix; ic3[1] = ic2[1]+1; 

               Rect_phi(ic, ns_field) = 2.0*Rect_phi(ic2,ns_field) - Rect_phi(ic3,ns_field);
               // TMP
               // printf("scatter_diff_conv_vel, ic[%d %d] by [%d %d] vel (%g %g) (%g %g)\n",
               //              ic[0], ic[1], ic2[0], ic2[1],
               //               Rect_u_star(ic, ns_field).v[0], Rect_u_star(ic, ns_field).v[1],
               //                    Rect_u_star(ic2, ns_field).v[0], Rect_u_star(ic2, ns_field).v[1]);
           }
       }
}

LOCAL void enforce_BC_on_scalar_field(
       NS_Field_Stencil  *ns_field,
       Front             *fr,
       Wave              *wv)
{
       RECT_GRID *gr = wv->rect_grid;
       int       i, dim = gr->dim, ix, iy;
       int       smin[MAXD], smax[MAXD], ic[MAXD], ic2[MAXD];

       for(i = 0; i < dim; i++)
       {
           smin[i] = -gr->lbuf[i];
           smax[i] = gr->gmax[i] + gr->ubuf[i];
       }
       
       if(rect_boundary_type(fr->interf,0,0) == SUBDOMAIN_BOUNDARY)
       {
           for(iy = 0; iy < gmax[1]; iy++)
           {
               for(ix = smin[0]; ix < 0; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = gmax[0]+ic[0];  ic2[1] = iy;
                   Dens(Rect_state(ic,wv)) = Dens(Rect_state(ic2,wv));
                   Rect_den(ic,ns_field) = Rect_den(ic2,ns_field); 
                   // Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
               }
               for(ix = gr->gmax[0]; ix < smax[0]; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = ic[0] - gmax[0];  ic2[1] = iy;
                   Dens(Rect_state(ic,wv)) = Dens(Rect_state(ic2,wv));
                   Rect_den(ic,ns_field) = Rect_den(ic2,ns_field); 
                   // Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
               }
           }          
       }
       else
       {
           for(iy = 0; iy < gmax[1]; iy++)
           {
               for(ix = smin[0]; ix < 0; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = -ic[0]-1;  ic2[1] = iy;
                   Dens(Rect_state(ic,wv)) = Dens(Rect_state(ic2,wv));
                   Rect_den(ic,ns_field) = Rect_den(ic2,ns_field); 
                   // Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
               }
               for(ix = gr->gmax[0]; ix < smax[0]; ix++)
               {
                   ic[0] = ix; ic[1] = iy;
                   ic2[0] = 2*gmax[0] - ic[0]-1;  ic2[1] = iy;
                   Dens(Rect_state(ic,wv)) = Dens(Rect_state(ic2,wv));
                   Rect_den(ic,ns_field) = Rect_den(ic2,ns_field); 
                   // Rect_phi(ic, ns_field) = Rect_phi(ic2, ns_field);
                   /**
                   if(iy == 4)
                   {
                       printf("\nIn scatter_projection\n");
                       printf("ic[%d %d] press %9.8g from [%d %d] press %9.8g\n\n",
                               ic[0], ic[1], Rect_phi(ic, ns_field), 
                               ic2[0], ic2[1], Rect_phi(ic2, ns_field));
                   }
                   **/
               }
           }          
       }
}

/***
LOCAL void factorize_vel_on_edge(
       Wave    *wv,
       Front   *fr,
       NS_Field_Stencil  *ns_field)
{
       TRI_GRID        *ntg = wave_tri_soln(wv)->tri_grid;
       BLK_EDGE        **blk_edge = ntg->blk_edge;
       RECT_GRID     *gr = wv->rect_grid;
       int        xmax,ymax;
       int        ix,iy;
       int        dir,j,k,nor;
       int        icoords[MAXD], ic[MAXD];
       float      *coords,coord[MAXD];      // coords is the cell center
                                         // and coord is the cell edge center
       float      *h = gr->h;
       float      rweight = 0.0, dt_coeff = 1.0;

       xmax = gr->gmax[0];
       ymax = gr->gmax[1];

       for (iy = 0;  iy < ymax;  ++iy)
       {
           // for (ix = 0;  ix < xmax-1;  ++ix)
           for (ix = 0;  ix < xmax;  ++ix)
           {
               if(blk_edge[ix][iy].ctype == EXTERNAL)
                   continue;

               if(blk_edge[ix][iy].ctype == PARTIAL)
               {
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           if(blk_edge[ix][iy].edge[dir][j].etype == N_SURF ||
                              blk_edge[ix][iy].edge[dir][j].etype == PARTIAL_SURF)
                           {
                               coord[0] = blk_edge[ix][iy].edge[dir][j].cen[0];
                               coord[1] = blk_edge[ix][iy].edge[dir][j].cen[1];
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == FULL_SURF)
                           {
                               coord[dir] = coords[dir]+ nor*h[dir]/2;
                               coord[(dir+1)%2] = coords[(dir+1)%2];
                           }
                           else
                               continue;

                       }
                   }
               }
           }
       }
}
***/

LOCAL void free_NS_field_space(
        NS_Field_Stencil      *field)
{
        free_these(2,field->u_states[0],field->u_states[1]);
        free_these(2,field->n_vel_edge[0],field->n_vel_edge[1]);
        free_these(2,field->uL_states[0],field->uL_states[1]);
        free_these(2,field->uR_states[0],field->uR_states[1]);
        free(field->Gp);
        free(field->phi);
        free_these(2,field->vlap[0],field->vlap[1]);
        free(field->con_der);
        free(field->rho);
        free(field->mu);
        free(field->u_star);
        free_these(2,field->rho_recip[0],field->rho_recip[1]);

        free_these(2,field->hr_st, field->vr_st);

        Set_free(field,I_to_ijk);
        Set_free(field,IJK_to_I);

        free(field->comps);
        free(field->den_vis); 

        free(field);
}

// update u^star and pressure field
// To finish, the update of the pressure field
LOCAL void update_vel_press(
        float   dt,
        NS_Field_Stencil *ns_field)
{
        COMPONENT  comp;
        COMPONENT  compll;  
        COMPONENT  comprr;  
        int        i, j, k;
        int        ic[MAXD], icl[MAXD], icr[MAXD];
        float      dv, phil, phir;
        Wave       *nwv = ns_field->newwave;
        Wave       *wv = ns_field->wave;
        Front      *nfr = ns_field->newfr;
        TRI_GRID   *ntg = wave_tri_soln(nwv)->tri_grid;
        BLK_EDGE   *blk_edge = ntg->blk_edge;
        RECT_GRID  *gr = nwv->rect_grid;
        Gas_param       **prms_list;
        Locstate   stold, stnew;
        int        *gmax = gr->gmax, nprms;
        float      *dh = gr->h, speed;

        nprms = return_params_list(&prms_list);

        // Update u_x (x-component) of velocity field
        for(i = 1; i < gmax[0]-1; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic[0] = i;
                icl[1] = icr[1] = ic[1] = j;
                icl[0] = ic[0]-1; icr[0] = ic[0]+1;
                comp = Rect_comp(ic,wv);
                compll = Rect_comp(icl,wv);
                comprr = Rect_comp(icr,wv);

                if(comp != ext_comp &&
                   !is_excluded_comp(comp,nfr->interf) &&
                   comp != POTS_COMP)
                {
                    stnew = Rect_state(ic,nwv);

                    if(compll != ext_comp &&
                       !is_excluded_comp(compll,nfr->interf) &&
                       compll != POTS_COMP)
                        phil = Rect_phi(icl,ns_field);
                    else
                        phil = Rect_phi(ic,ns_field);

                    if(comprr != ext_comp && 
                       !is_excluded_comp(comprr,nfr->interf) &&
                       comprr != POTS_COMP)   
                        phir = Rect_phi(icr,ns_field);
                    else
                        phir = Rect_phi(ic,ns_field);
                    Vel(stnew)[0] = Rect_u_star(ic,ns_field).v[0] - 
                           dt*(phir-phil)/(2.0*dh[0]);
                }                
            }
        }        
      
        // left side boundary cell u_x (x-component) of velocity field
        // Here no differences are made between the Dir. boundary and Neumann boundary
        ic[0] = 0; icr[0] = 1;
        for(j = 0; j < gmax[1]; j++)
        {
            icr[1] = icl[1] = ic[1] = j;
            comp = Rect_comp(ic,wv);
            comprr = Rect_comp(icr,wv);
            if(comp != ext_comp &&
               !is_excluded_comp(comp,nfr->interf) &&
               comp != POTS_COMP)
            {
                stnew = Rect_state(ic,nwv);

                if(comprr != ext_comp &&
                   !is_excluded_comp(comprr,nfr->interf) &&
                   comprr != POTS_COMP)
                    phir = Rect_phi(icr,ns_field);
                else
                    phir = Rect_phi(ic,ns_field);

                if(gr->lbuf[0] != 0)
                {  
                    icl[0] = -1;
                    compll = Rect_comp(icl,wv);
                    if(compll != ext_comp &&
                       !is_excluded_comp(compll,nfr->interf) &&
                       compll != POTS_COMP)
                        phil = Rect_phi(icl,ns_field);
                    else
                        phil = Rect_phi(ic,ns_field);
                }
                else
                    phil = Rect_phi(ic,ns_field);
                Vel(stnew)[0] = Rect_u_star(ic,ns_field).v[0] - 
                           dt*(phir-phil)/(2.0*dh[0]);
            }
        }
        // right side boundary cell u_x (x-component) of velocity field
        // Here no differences are made between the Dir. boundary and Neumann boundary
        ic[0] = gmax[0]-1; icl[0] = gmax[0]-2;
        for(j = 0; j < gmax[1]; j++)
        {
            icr[1] = icl[1] = ic[1] = j;
            comp = Rect_comp(ic,wv);
            compll = Rect_comp(icl,wv);
            if(comp != ext_comp &&
               !is_excluded_comp(comp,nfr->interf) &&
               comp != POTS_COMP)
            {
                stnew = Rect_state(ic,nwv);

                if(compll != ext_comp &&
                   !is_excluded_comp(compll,nfr->interf) &&
                   compll != POTS_COMP)
                    phil = Rect_phi(icl,ns_field);
                else
                    phil = Rect_phi(ic,ns_field);

                if(gr->ubuf[0] != 0)
                {  
                    icr[0] = gmax[0];
                    comprr = Rect_comp(icr,wv);
                    if(comprr != ext_comp &&
                       !is_excluded_comp(comprr,nfr->interf) &&
                       comprr != POTS_COMP)
                        phir = Rect_phi(icr,ns_field);
                    else
                        phir = Rect_phi(ic,ns_field);
                }
                else
                    phir = Rect_phi(ic,ns_field);
                Vel(stnew)[0] = Rect_u_star(ic,ns_field).v[0] - 
                           dt*(phir-phil)/(2.0*dh[0]);
            }
        }
        // Update u_y (y-component) of velocity field
        // Update the pressure field
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 1; j < gmax[1]-1; j++)
            {
                icl[0] = icr[0] = ic[0] = i;
                ic[1] = j;
                icl[1] = ic[1]-1; icr[1] = ic[1]+1;
                comp = Rect_comp(ic,wv);
                compll = Rect_comp(icl,wv);
                comprr = Rect_comp(icr,wv);

                if(comp != ext_comp &&
                   !is_excluded_comp(comp,nfr->interf) &&
                   comp != POTS_COMP)
                {
                    stnew = Rect_state(ic,nwv);

                    if(compll != ext_comp &&
                       !is_excluded_comp(compll,nfr->interf) &&
                       compll != POTS_COMP)
                        phil = Rect_phi(icl,ns_field);
                    else
                        phil = Rect_phi(ic,ns_field);

                    if(comprr != ext_comp &&
                       !is_excluded_comp(comprr,nfr->interf) &&
                       comprr != POTS_COMP)
                        phir = Rect_phi(icr,ns_field);
                    else
                        phir = Rect_phi(ic,ns_field);
                    Vel(stnew)[1] = Rect_u_star(ic,ns_field).v[1] -
                           dt*(phir-phil)/(2.0*dh[1]);
                    
                }
            }
        }

        // lower side boundary cell u_y (y-component) of velocity field
        // Here no differences are made between the Dir. boundary and Neumann boundary
        // Update the pressure field
        ic[1] = 0; icr[1] = 1;
        for(i = 0; i < gmax[0]; i++)
        {
            icr[0] = icl[0] = ic[0] = i;
            comp = Rect_comp(ic,wv);
            comprr = Rect_comp(icr,wv);
            if(comp != ext_comp &&
               !is_excluded_comp(comp,nfr->interf) &&
               comp != POTS_COMP)
            {
                stnew = Rect_state(ic,nwv);

                if(comprr != ext_comp &&
                   !is_excluded_comp(comprr,nfr->interf) &&
                   comprr != POTS_COMP)
                    phir = Rect_phi(icr,ns_field);
                else
                    phir = Rect_phi(ic,ns_field);

                if(gr->lbuf[1] != 0)
                {
                    icl[1] = -1;
                    compll = Rect_comp(icl,wv);
                    if(compll != ext_comp &&
                       !is_excluded_comp(compll,nfr->interf) &&
                       compll != POTS_COMP)
                        phil = Rect_phi(icl,ns_field);
                    else
                        phil = Rect_phi(ic,ns_field);
                }
                else
                    phil = Rect_phi(ic,ns_field);
                Vel(stnew)[1] = Rect_u_star(ic,ns_field).v[1] -
                           dt*(phir-phil)/(2.0*dh[1]);
                /**
                if(ic[0] == 0)
                {
                    printf("ic[%d %d], oldv[%14.12f, %14.12f], phil = %14.12f, phir = %14.12f\n",
                        ic[0], ic[1],  Rect_u_star(ic,ns_field).v[0],
                         Rect_u_star(ic,ns_field).v[1], phil, phir);
                }
                **/
            }
        }

        // upper side boundary cell u_y (y-component) of velocity field
        // Here no differences are made between the Dir. boundary and Neumann boundary
        // Update the pressure field
        ic[1] = gmax[1]-1; icl[1] = gmax[1]-2;
        for(i = 0; i < gmax[0]; i++)
        {
            icr[0] = icl[0] = ic[0] = i;
            comp = Rect_comp(ic,wv);
            compll = Rect_comp(icl,wv);
            if(comp != ext_comp &&
               !is_excluded_comp(comp,nfr->interf) &&
               comp != POTS_COMP)
            {
                stnew = Rect_state(ic,nwv);

                if(compll != ext_comp &&
                   !is_excluded_comp(compll,nfr->interf) &&
                   compll != POTS_COMP)
                    phil = Rect_phi(icl,ns_field);
                else
                    phil = Rect_phi(ic,ns_field);

                if(gr->ubuf[1] != 0)
                {
                    icr[1] = gmax[1];
                    comprr = Rect_comp(icr,wv);
                    if(comprr != ext_comp &&
                       !is_excluded_comp(comprr,nfr->interf) &&
                       comprr != POTS_COMP)
                        phir = Rect_phi(icr,ns_field);
                    else
                        phir = Rect_phi(ic,ns_field);
                }
                else
                    phir = Rect_phi(ic,ns_field);
                Vel(stnew)[1] = Rect_u_star(ic,ns_field).v[1] -
                           dt*(phir-phil)/(2.0*dh[1]);
                /**
                if(ic[0] == 0)
                {
                    printf("ic[%d %d], oldv[%14.12f, %14.12f], phil = %14.12f, phir = %14.12f\n",
                        ic[0], ic[1],  Rect_u_star(ic,ns_field).v[0],
                         Rect_u_star(ic,ns_field).v[1], phil, phir);
                }
                **/
            }
        }

        // Update the pressure field
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic[0] = i;
                ic[1] = j;
                comp = Rect_comp(ic,nwv);

                if(comp != ext_comp &&
                   !is_excluded_comp(comp,nfr->interf) &&
                   comp != POTS_COMP)
                {
                    compll = Rect_comp(ic,wv);
                    if(compll == ext_comp ||
                       is_excluded_comp(compll,nfr->interf) &&
                       compll != POTS_COMP)
                    {
                        printf("WARNING: pdate_vel_press\n");
                        printf("new old comp are not consistent\n");
                        printf("compold = %d, compnew = %d. ic[%d %d]\n",
                             compll, comp, ic[0], ic[1]);
                        // clean_up(ERROR);
                        // Should be fixed with geometry
                        Press(Rect_state(ic,wv)) = 1.0;
                    }

                    stnew = Rect_state(ic,nwv);
                    set_type_of_state (stnew,TGAS_STATE); 

                    Press(stnew) = Press(Rect_state(ic,wv)) + Rect_phi(ic,ns_field);
                    for(k = 0; k < nprms; k++)
                    {
                        if(prms_list[k]->eos->_NS_eos == YES && prms_list[k]->comp == comp)
                        {
                            Dens(stnew) = prms_list[k]->dens;
                            Params(stnew) = prms_list[k];
                            break;
                        }
                    }
                    speed = max(fabs(Vel(stnew)[0]), fabs(Vel(stnew)[1]));
                    speed = max(speed, dh[0]);
                    if(fabs(Vel(stnew)[0]) > fabs(Vel(stnew)[1]))
                        k = 0;
                    else
                        k = 1;
                    // TMP
                    printf("ic[%d %d] call set_max_wave_speed %g\n",
                         ic[0], ic[1], speed);
                    set_max_wave_speed(k,speed,stnew,Rect_coords(ic,nwv),nwv);
                }
            }
        }

        if(debugging("NS_vel"))
        {
            printf("VELOCITY OF CELLS--------\n");
            for(j = 0; j < gmax[1]; j++)
            {
                // j = 0;
                printf("Row[%d] :", j);
                for(i = 0; i < gmax[0]; i++)
                {
                    ic[0] = i; ic[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    printf("(%d %g %g) ",
                          ic[0], Vel(stnew)[0], Vel(stnew)[1]);
                }
                ic[0] = gmax[0]-1; ic[1] = j;
                stnew =  Rect_state(ic,nwv);
                // printf("(%d %g %g) ",
                //       ic[0], Vel(stnew)[0], Vel(stnew)[1]);
                printf("\n");

                /**
                j = gmax[1]-1;
                printf("Row[%d] :", j);
                for(i = 0; i < gmax[0]; i++)
                {
                    ic[0] = i; ic[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    printf("(%d %g %g) ",
                          ic[0], Vel(stnew)[0], Vel(stnew)[1]);
                }
                ic[0] = gmax[0]-1; ic[1] = j;
                stnew =  Rect_state(ic,nwv);
                // printf("(%d %g %g) ",
                //       ic[0], Vel(stnew)[0], Vel(stnew)[1]);
                printf("\n");
                **/
            }
            /**
            printf("Pressure OF CELLS--------\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("Row[%d] :", j);
                for(i = 0; i < min(3,gmax[0]); i++)
                {
                    ic[0] = i; ic[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    printf("(%d %g) ",
                          ic[0], Press(stnew));
                }
                ic[0] = gmax[0]-1; ic[1] = j;
                stnew =  Rect_state(ic,nwv);
                    printf("(%d %g) ",ic[0], Press(stnew));
                printf("\n");
            }
            **/
        }
        // TEST
        /**
        if(gmax[0] > 3)
        {
            ic[0] = 3; ic[1] = 1;
            stnew =  Rect_state(ic,nwv);
            printf("ic[%d %d] vel [%9.8g %9.8g] press(%9.8g) at the end of step comp %d\n",
                  ic[0], ic[1], Vel(stnew)[0], Vel(stnew)[1], Press(stnew),
                  Rect_comp(ic,nwv));
            ic[0] = 3; ic[1] = gmax[1]-2;
            stnew =  Rect_state(ic,nwv);
            printf("ic[%d %d] vel [%9.8g %9.8g] press(%9.8g) at the end of step comp %d\n",
                  ic[0], ic[1], Vel(stnew)[0], Vel(stnew)[1], Press(stnew),
                  Rect_comp(ic,nwv));
        }
        {
            ic[0] = 0; ic[1] = gmax[1]-1;
            stnew =  Rect_state(ic,nwv);
            printf("ic[%d %d] vel [%g %g] press(%g) at the end of step\n",
                  ic[0], ic[1], Vel(stnew)[0], Vel(stnew)[1], Press(stnew));
            ic[0] = gmax[0]-1; ic[1] = gmax[1]-1;
            stnew =  Rect_state(ic,nwv);
            printf("ic[%d %d] vel [%g %g] press(%g) at the end of step\n",
                  ic[0], ic[1], Vel(stnew)[0], Vel(stnew)[1], Press(stnew));
        }
        **/
}

// Update u^(n+1) and pressure field
// \phi field comes from 5-point L. (approximate projection).
// Therefore \rho is cell center value.
LOCAL void update_vel_press_ver2(
        float   dt,
        NS_Field_Stencil *ns_field)
{
        int        i, j, k;
        int        ic[MAXD], icl[MAXD], icr[MAXD];
        float      dv, phil, phir, phi, speed;
        Wave       *nwv = ns_field->newwave;
        Wave       *wv = ns_field->wave;
        Front      *nfr = ns_field->newfr;
        TRI_GRID   *ntg = wave_tri_soln(nwv)->tri_grid;
        BLK_EDGE   *blk_edge = ntg->blk_edge;
        RECT_GRID  *gr = nwv->rect_grid;
        float      *dh = gr->h;
        int        *gmax = gr->gmax;
        int        nprms;
        COMPONENT  comp;
        COMPONENT  compll;
        COMPONENT  comprr;
        Gas_param       **prms_list;
        float      rweight;  
        Locstate   stold, stnew;
        float      test;

        nprms = return_params_list(&prms_list);

        // Update the density field
        // averaged density and viscosity are set in
        //  grid_den_viscosity_ver1();

        // Update u_x (x-component) of velocity field
        // \phi is obtained by linear extrapolation if necessary
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic[0] = i;
                ic[1] = j;

                comp = Rect_comp(ic,nwv);

                if( !excluded_comp(comp,nfr->interf))
                {
                    // rweight = 1.0/Dens(Rect_state(ic,nwv));
                    rweight = 1.0/Rect_den(ic,ns_field);
                    icl[1] = icr[1] = ic[1];
                    icl[0] = ic[0]-1; icr[0] = ic[0]+1;
                    compll = Find_rect_comp_ver2(icl,nwv);
                    comprr = Find_rect_comp_ver2(icr,nwv);

                    stnew = Rect_state(ic,nwv);
                    phi = Rect_phi(ic,ns_field);

                    // FOR RT, on regular domain
                    phir = Rect_phi(icr,ns_field);
                    phil = Rect_phi(icl,ns_field);
                   
                    /**** case for handling geometric change
                    if( !excluded_comp(compll,nfr->interf))
                        phil = Rect_phi(icl,ns_field);
                    else if( !excluded_comp(comprr,nfr->interf))
                    {
                        phir = Rect_phi(icr,ns_field);
                        phil = 2.0*phi - phir;
                    }
                    else
                        phil = phi;

                    if( !excluded_comp(comprr,nfr->interf))   
                        phir = Rect_phi(icr,ns_field);
                    else if( !excluded_comp(compll,nfr->interf))
                    {
                        phil = Rect_phi(icl,ns_field);
                        phir = 2.0*phi - phil;
                    }
                    else
                        phir = phi;
                    ***/
                    // test = 0.5*(dt*(phir-phi)*v_line_rhor(icr,ns_field)/dh[0] +  
                    //             dt*(phi-phil)*v_line_rhor(ic,ns_field)/dh[0]);

                    // test = 0.0;

                    // Vel(stnew)[0] = Rect_u_star(ic,ns_field).v[0] - test;

                    Vel(stnew)[0] = Rect_u_star(ic,ns_field).v[0] - 
                           dt*rweight*(phir-phil)/(2.0*dh[0]);

                    // TMP
                    /**
                    if((ic[0] == 0 && ic[1] == 0) ||
                       (ic[0] == 1 && ic[1] == 0)
                      )
                    {
                        printf("IN update_vel_press_ver2, ic[%d %d] x-comp vel\n", ic[0], ic[1]);
                        printf("Rect_u_star[0] = %9.8g, rweight = %9.8g, correction %g\n",
                              Rect_u_star(ic,ns_field).v[0], rweight, test);
                        printf("phir %11.10g, phil %11.10g, phi %11.10g\n",
                              phir, phil, phi);
                    }
                    **/
                }
            }
        }        
      
        // Update u_y (y-component) of velocity field
        // Update the pressure field
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic[0] = i;
                ic[1] = j;
                comp = Rect_comp(ic,nwv);

                if( !excluded_comp(comp,nfr->interf))
                {
                    // rweight = 1.0/Dens(Rect_state(ic,nwv));
                    rweight = 1.0/Rect_den(ic,ns_field);
                    icl[0] = icr[0] = ic[0];
                    icl[1] = ic[1]-1; icr[1] = ic[1]+1;
                    compll = Find_rect_comp_ver2(icl,nwv);
                    comprr = Find_rect_comp_ver2(icr,nwv);

                    stnew = Rect_state(ic,nwv);
                    phi = Rect_phi(ic,ns_field);

                    // FOR RT
                    phil = Rect_phi(icl,ns_field);
                    phir = Rect_phi(icr,ns_field);

                    /***
                    if( !excluded_comp(compll,nfr->interf))
                        phil = Rect_phi(icl,ns_field);
                    else if( !excluded_comp(comprr,nfr->interf))
                    {
                        phir = Rect_phi(icr,ns_field);
                        phil = 2.0*phi - phir;
                    }
                    else
                        phil = phi;

                    if( !excluded_comp(comprr,nfr->interf))
                        phir = Rect_phi(icr,ns_field);
                    else if( !excluded_comp(compll,nfr->interf))
                    {
                        phil = Rect_phi(icl,ns_field);
                        phir = 2.0*phi - phil;
                    }
                    else
                        phir = phi;
                    ***/

                    Vel(stnew)[1] = Rect_u_star(ic,ns_field).v[1] -
                           dt*rweight*(phir-phil)/(2.0*dh[1]);

                    // test = 0.5*(dt*(phir-phi)*h_line_rhor(icr,ns_field)/dh[1] + 
                    //             dt*(phi-phil)*h_line_rhor(ic,ns_field)/dh[1]);

                    // test = 0.0;

                    // Vel(stnew)[1] = Rect_u_star(ic,ns_field).v[1] - test;

                    // TMP
                    /**
                    if((ic[0] == 0 && ic[1] == 3) ||
                       (ic[0] == 1 && ic[1] == 3)
                      )
                    {
                        printf("IN update_vel_press_ver2, ic[%d %d] Y-comp vel\n", ic[0], ic[1]);
                        printf("Rect_u_star[1] = %9.8g, rweight = %9.8g, phir %9.8g, phil %9.8g"
                                    " correction %g\n",
                              Rect_u_star(ic,ns_field).v[1], rweight, phir, phil,
                                    dt*rweight*(phir-phil)/(2.0*dh[1]));
                        { 
                            float test2, test3, test4;
                            test2 = h_line_vel(icr,ns_field).v[1] - dt*(phir-phi)*h_line_rhor(icr,ns_field)/dh[1];
                            test3 = h_line_vel(ic,ns_field).v[1] - dt*(phi-phil)*h_line_rhor(ic,ns_field)/dh[1];
                            test4 = 0.5*(test2+test3);
                        printf("test2 = %g, h_line = %g; test3 = %g, h_line = %g, avg = %g, test_old = %g\n",
                                      dt*(phir-phi)*h_line_rhor(icr,ns_field)/dh[1],
                                      h_line_vel(icr,ns_field).v[1],
                                      dt*(phi-phil)*h_line_rhor(ic,ns_field)/dh[1],
                                      h_line_vel(ic,ns_field).v[1],
                                      test4,
                                      Rect_u_star(ic,ns_field).v[1] - dt*rweight*(phir-phil)/(2.0*dh[1]));
                        }
                    }
                    **/
                }
            }
        }

        // Update the pressure field
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic[0] = i;
                ic[1] = j;
                comp = Rect_comp(ic,nwv);

                if( !is_excluded_comp(comp,nfr->interf))
                {
                    compll = Rect_comp(ic,wv);
                    if( excluded_comp(compll,nfr->interf))
                    {
                        printf("WARNING: pdate_vel_press\n");
                        printf("new old comp are not consistent\n");
                        printf("compold = %d, compnew = %d. ic[%d %d]\n",
                             compll, comp, ic[0], ic[1]);
                        // clean_up(ERROR);
                        // Should be fixed with geometry
                        Press(Rect_state(ic,nwv)) = 0.0;
                    }

                    stnew = Rect_state(ic,nwv);
                    set_type_of_state (stnew,TGAS_STATE); 

                    Press(stnew) = Press(Rect_state(ic,wv)) + Rect_phi(ic,ns_field);
                    // Press(stnew) = Rect_phi(ic,ns_field);
                    // TMP
                    /**
                    if(ic[0] == 8 && (ic[1] == 3 || ic[1] == 2 || ic[1] == 4))
                    {
                        printf("set [%d %d] press, phi %g, old_p %g\n",
                            ic[0], ic[1], Rect_phi(ic,ns_field), Press(Rect_state(ic,wv)));
                    }
                    **/

                    for(k = 0; k < nprms; k++)
                    {
                        if(prms_list[k]->eos->_NS_eos == YES && prms_list[k]->comp == comp)
                        {
                            Dens(stnew) = prms_list[k]->dens;
                            Params(stnew) = prms_list[k];
                            break;
                        }
                    }
                    speed = max(fabs(Vel(stnew)[0]), fabs(Vel(stnew)[1]));
                    speed = max(speed, dh[0]);
                    if(fabs(Vel(stnew)[0]) > fabs(Vel(stnew)[1]))
                        k = 0;
                    else
                        k = 1;
                    set_max_wave_speed(k,speed,stnew,Rect_coords(ic,nwv),nwv);
                }
            }
        }


        // TMP
        // if(deb_sten == YES)
        /**
        {
            ic[0] = 32; ic[1] = 3;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));  
            ic[0] = 33; ic[1] = 3;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));  
            ic[0] = 34; ic[1] = 3;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));  
            ic[0] = 33; ic[1] = 2;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));  
            ic[0] = 33; ic[1] = 4;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));  
        }
        **/

        if(debugging("NS_vel") && NO)
        {
            printf("\n-------------------------\n");
            printf("-----VELOCITY AND PRESSURE OF CELLS--------\n");
            for(j = gmax[1]-1; j > 0; j--)
            {
                printf("Row[%d] :", j);
                for(i = gmax[0]/2-2; i <= gmax[0]/2; i++)
                {
                    // if(i != 0 && i != gmax[0]-1) continue;
                    int ic2[MAXD];
                    Locstate st;
                    ic[0] = i; ic[1] = j;
                    ic2[0] = gmax[0]-i-1; ic2[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    st =  Rect_state(ic2,nwv);
                    printf("(%d %g %g, %g) ",
                          ic[0], Vel(stnew)[0], Vel(stnew)[1], Press(stnew));
                }
                printf("\n");
            }

            /**
            printf("----Pressure OF CELLS--------\n");
            for(j = 0; j < gmax[1]; j++)
            // for(j = 3; j <= 5; j++)
            {
                printf("Row[%d] :", j);
                for(i = 0; i < (gmax[0]); i++)
                {
                    if(i != 0 && i != gmax[0]-1) continue;
                    ic[0] = i; ic[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    printf("(%d %9.8g)",
                          ic[0], Press(stnew));
                }
                printf("\n");
            }
            **/
        }

            /**
            printf("\n-------------------------\n");
            printf("----- VELOCITY DIFF OF CELLS--------\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("Row[%2d] :", j);
                // for(i = 0; i < 5; i++)
                for(i = 0; i < gmax[0]/2; i++)
                {
                    ic[0] = i; ic[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    printf("%d %g %g; ",
                          ic[0],  Vel(stnew)[0], Vel(stnew)[1]);
                }
                printf("\n");
            }

            printf("----Pressure OF CELLS--------\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("Row[%2d] :", j);
                for(i = 0; i < gmax[0]; i++)
                {
                    ic[0] = i; ic[1] = j;
                    stnew =  Rect_state(ic,nwv);
                    printf("%d %g; ",
                          ic[0], Press(stnew));
                }
                printf("\n");
            }
            **/

        // TMP
        /**
        {
            int ic2[MAXD];
            Locstate st;
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
            printf("\n\n---------VELOCITY DIFF OF CELLS---------\n\n");

            ic2[0] = 0; ic2[1] = 18;
            st =  Rect_state(ic2,nwv);
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 1; ic2[1] = 18;
            st =  Rect_state(ic2,nwv);
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 0; ic2[1] = 19;
            st =  Rect_state(ic2,nwv);
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 1; ic2[1] = 19;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 0; ic2[1] = 20;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 1; ic2[1] = 20;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 0; ic2[1] = 21;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 1; ic2[1] = 21;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 0; ic2[1] = 22;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);

            ic2[0] = 1; ic2[1] = 22;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1],
                   Vel(st)[0], Vel(st)[1]);
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
        }
        **/
        // TEST
}


LOCAL void edge_1_over_den(
        NS_Field_Stencil   *ns_field,
        Wave               *wv)
{
        int             smin[MAXD], smax[MAXD];
        TRI_GRID        *ntg;
        BLK_EDGE        *blk_edge;
        Front           *nfr = ns_field->newfr, *fr = ns_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD], icrds[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vold[MAXD], vnew[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh;
        int             vset;
        int             deb_flag = NO;
        RECT_GRID       *gr = wv->rect_grid;

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -gr->lbuf[i];
            smax[i] = gr->gmax[i] + gr->ubuf[i];
        }

        dh = wv->rect_grid->h;
        ntg = wave_tri_soln(wv)->tri_grid;
        blk_edge = ntg->blk_edge;

        // Vertical edge
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0]+1;  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix-1; ic[1] = iy;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,wv);
                comp1 = Rect_comp(ic1,wv);

                // v_line_rhor(ic1,ns_field) = 1.0;
                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                   comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    // v_line_rhor(ic1,ns_field) = 0.5*(1.0/Rect_den(ic,ns_field) 
                    //                                + 1.0/Rect_den(ic1,ns_field));
                    v_line_rhor(ic1,ns_field) = 1.0/(0.5*(Rect_den(ic,ns_field) 
                                                   + Rect_den(ic1,ns_field)));
                    /**
                    if(ix == 3 && (iy == 4 || iy == 5))
                    {
                        printf("edge_1_over_den, v_line_rhor(%d %d) = %9.8g, left %9.8g, right %9.8g\n",
                                    ic1[0], ic1[1], v_line_rhor(ic1,ns_field), 
                                      Rect_den(ic,ns_field), Rect_den(ic1,ns_field)); 
                    }
                    **/
                }
                else if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                        comp != POTS_COMP)
                {
                    v_line_rhor(ic1,ns_field) = (1.0/Rect_den(ic,ns_field));
                }
                else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                        comp1 != POTS_COMP)
                {
                    v_line_rhor(ic1,ns_field) = (1.0/Rect_den(ic1,ns_field));
                }
                else
                {
                    // Cell on the right side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    printf("ERROR: edge_1_over_den, den unset\n");
                    clean_up(ERROR);
                }
                // TMP
                // printf("edge[%d %d] rho = %g\n", ic1[0], ic1[1], 1.0/v_line_rhor(ic1,ns_field));
            }
        }

        // Left vertical edge
        ic[0] = ix = smin[0];
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            ic[1] = iy;
            comp = Rect_comp(ic,wv);
            icrds[0] = ix; icrds[1] = iy;
            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;         

            if(comp == ext_comp || ( is_excluded_comp(comp,nfr->interf)))
            {
                printf("ERROR: edge_1_over_den, den unset 1\n");
                clean_up(ERROR);
            }

            // v_line_rhor(ic,ns_field) = 1.0;
            v_line_rhor(ic,ns_field) = (1.0/Rect_den(ic,ns_field));
        }

        // Right vertical edge
        ix = smax[0]-1;
        ic[0] = smax[0]-1;
        ic1[0] = smax[0];
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[1] = iy; ic1[1] = iy;
            comp = Rect_comp(ic,wv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;         

            if(comp == ext_comp || ( is_excluded_comp(comp,nfr->interf)))
            {
                printf("ERROR: edge_1_over_den, den unset 2\n");
                clean_up(ERROR);
            }

            // printf("update Right vertical edge 1/rho edge ic1[%d %d]\n", ic1[0], ic1[1]);
            v_line_rhor(ic1,ns_field) = (1.0/Rect_den(ic,ns_field));
            // v_line_rhor(ic1,ns_field) = 1.0;
        }

        // Horizontal edge
        for (iy = smin[1]+1;  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix; ic[1] = iy-1;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,wv);
                comp1 = Rect_comp(ic1,wv);

                // printf("on ix, iy [%d %d]\n", ix, iy);
                // h_line_rhor(ic1,ns_field) = 1.0;
                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                   comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    // h_line_rhor(ic1,ns_field) = 0.5*(1.0/Rect_den(ic,ns_field) 
                    //                                 + 1.0/Rect_den(ic1,ns_field));
                    h_line_rhor(ic1,ns_field) = 1.0/(0.5*(Rect_den(ic,ns_field) 
                                                   + Rect_den(ic1,ns_field)));
                }
                else if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                        comp != POTS_COMP)
                {
                    h_line_rhor(ic1,ns_field) = (1.0/Rect_den(ic,ns_field));
                }
                else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                        comp1 != POTS_COMP)
                {
                    h_line_rhor(ic1,ns_field) = (1.0/Rect_den(ic1,ns_field));
                }
                else
                {
                    // Cell on the right side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    printf("ERROR: edge_1_over_den, den unset h-edge\n");
                    clean_up(ERROR);
                }             
            }
        }

        // bottom horizontal edge
        ic[1] = iy = smin[1];
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[0] = ix;
            comp = Rect_comp(ic,wv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            if(comp == ext_comp || ( is_excluded_comp(comp,nfr->interf)))
            {
                printf("ERROR: edge_1_over_den, den unset h-edge 1\n");
                clean_up(ERROR);
            }

            h_line_rhor(ic,ns_field) = (1.0/Rect_den(ic,ns_field));
            // h_line_rhor(ic,ns_field) = 1.0;
        }
       
        // top horizontal edge
        ic[1] = iy = smax[1]-1;
        ic1[1] =  smax[1];
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[0] = ix; ic1[0] = ix;
            comp = Rect_comp(ic,wv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;
            if(comp == ext_comp || ( is_excluded_comp(comp,nfr->interf)))
            {
                printf("ERROR: edge_1_over_den, h-edge den unset 2\n");
                clean_up(ERROR);
            }

            h_line_rhor(ic1,ns_field) = (1.0/Rect_den(ic,ns_field));
            // h_line_rhor(ic1,ns_field) = 1.0;
        }

        // TMP
        if(debugging("NS_vel") && NO)
        {
            printf("\n\n***********************\n");
            printf("In  edge_1_over_den, show 1/rho on horizontal edge\n");            
            for(j = 0; j <= gmax[1]; j++)
            // for(j = 4; j <= 5; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i < min(14,gmax[0]); i++)
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d %g ", i, h_line_rhor(ic, ns_field)); 
                }
                printf("\n");
            }
            printf("In  edge_1_over_den, show 1/rho on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            // for(j = 4; j <= 5; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d %g ", i, v_line_rhor(ic, ns_field));
                }
                printf("\n");
            }
            printf("***********************\n\n");
        }

}

/* 
 This is solve L(dt\phi) = D(u)
*/
/********************
IMPORTANT: The above eqn is solved with the
modifed boundary condition. Set right side boundary's \phi = 0 (outflow side)
(Dirichlet condition for pressure projection)
*******************/
LOCAL void project_vel_star(
        float             dt,
        NS_Field_Stencil  *ns_field)
{
        int             Nx = ns_field->rect_gr->gmax[0];
        int             Ny = ns_field->rect_gr->gmax[1];
        int             indices,ix,iy;
        P_CELL2D        **p_cell;
        // float           **potential;
        int             *n_dist = NULL, myid, ICOORDS[2];
        int             **I_to_ijk;   /* local  i, j, k, local I */
        int             **IJK_to_I;  /* global I, J, K, global I */
        int             NLblocks,ilower, iupper, I, i, j;
        int             ic[MAXD];
        Wave            *newwv = ns_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(newwv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = ns_field->newfr;
        float           norm,min,max,x,y,*coords;

        DEBUG_ENTER(project_vel_star)

#if defined (PETSC)
            /*PETSC Variables*/
        KSP                  ksp;
        Mat                  A;
        Vec                  phi, B, C;
        PetscScalar          *vphi;
        MatNullSpace         nullsp;
        PetscErrorCode       ierr;
        PC                   pc;
        int                  its;
        PetscReal            norm_B, norm_C;
        KSPConvergedReason   reason;
#endif /*defined (PETSC)*/

#if defined (PETSC)

#if defined(__MPI__)
        MPI_Comm_size(PETSC_COMM_WORLD,&comm_size);
#endif /* defined(__MPI__) */
        // ierr = PetscOptionsGetInt(PETSC_NULL,"-n",&n,PETSC_NULL);CHKERRQ(ierr);

        myid = pp_mynode();
        find_Cartesian_coordinates(myid,nfr->pp_grid,ICOORDS);

        matrix(&p_cell,Nx,Ny,sizeof(P_CELL2D));
        vector(&n_dist,comm_size,sizeof(int));

        if(ns_field->alloc.I_to_ijk != YES)
        { 
            matrix(&I_to_ijk,Nx*Ny,2,sizeof( int));
            matrix(&IJK_to_I,(Nx+4),(Ny+4),sizeof( int));

            ns_field->I_to_ijk = I_to_ijk;
            ns_field->alloc.I_to_ijk = YES;
            ns_field->IJK_to_I = IJK_to_I;
            ns_field->alloc.IJK_to_I = YES;
        }
        else
        {
            I_to_ijk = ns_field->I_to_ijk;
            IJK_to_I = ns_field->IJK_to_I;
        }

        NLblocks = 0;
        for(j = 0; j < Ny; j++)
        {
            ic[1] = j;
            // Phis of right side boundary is set to be 0.
            // Now it is handled by the ghost cell
            // for(i = 0; i < Nx-1; i++)
            for(i = 0; i < Nx; i++)
            {
                ic[0] = i;
                if(Rect_comp(ic,newwv) != ext_comp &&
                   !is_excluded_comp(Rect_comp(ic,newwv),nfr->interf) &&
                   Rect_comp(ic,newwv) != POTS_COMP)
                {
                    I_to_ijk[NLblocks][0] = i;
                    I_to_ijk[NLblocks][1] = j;
                    NLblocks++;
                }
            }
        }

#if defined(__MPI__)
        pp_all_gather((POINTER)&NLblocks,sizeof(int),(POINTER)n_dist,
                  sizeof(int)*comm_size);
#else /* defined(__MPI__) */
        assign((POINTER)n_dist,(POINTER)&NLblocks,sizeof(int));
#endif /* defined(__MPI__) */

        ilower = 0;
        iupper = n_dist[0]-1;

        for (i = 1; i <= myid; i++)
        {
            ilower += n_dist[i-1];
            iupper += n_dist[i];
        }

        create_global_element_list(nfr,ilower,iupper,ICOORDS,I_to_ijk,IJK_to_I);

#if defined(__MPI__)
        // MatCreateMPIAIJ(PETSC_COMM_WORLD,iupper-ilower+1,iupper-ilower+1,
        //         PETSC_DECIDE,PETSC_DECIDE,7,PETSC_NULL,7,PETSC_NULL,&A);
#else
        MatCreate(PETSC_COMM_WORLD, &A);
        MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,iupper-ilower+1,iupper-ilower+1);
        MatSetFromOptions(A);
#endif // if defined(__MPI__) //

        VecCreate(PETSC_COMM_WORLD,&B);
        PetscObjectSetName((PetscObject) B, "RHS");
        VecSetSizes(B,(iupper-ilower + 1),PETSC_DECIDE);
        VecSetFromOptions(B); // CHKERRQ(ierr);
        VecDuplicate(B,&phi);
        VecDuplicate(B,&C);

        // Extrapolate cell edge vector field u^*
        // cell_edge_vector_u_ver2(dt, ns_field);
        // Extrapolate cell edge vector field u^* - u^n
        // cell_edge_vector_u_ver3(dt, ns_field);

        // Extrapolate cell edge vector field u^*
        // Based on ver2(), should enforce symmetry
        cell_edge_vector_u_ver4(dt, ns_field);

        // init_stencil_element2D(blk_edge,p_cell,nfr,newwv,ns_field,YES,dt);
        // Phi of left side boundary cells are set to zero
        // init_stencil_element2D_ver1(blk_edge,p_cell,nfr,newwv,ns_field,YES,dt);
        // Phi of right side boundary cells are set to zero
        // init_stencil_element2D_ver2(blk_edge,p_cell,nfr,newwv,ns_field,YES,dt);
        init_stencil_element2D_ver3(newwv,p_cell,nfr,newwv,ns_field,YES,dt);

        if(debugging("NS_vel") && NO)
        {
            int ic[MAXD], ic2[MAXD];
            float temp, linb, linc, lint;
            printf("Right Hand SIDE to Possion--------\n");
            printf("Ny/2 = %d\n", Ny/2);
            for(ix = 0; ix <= Nx/2; ix++)
            {
                printf("Col[%d]: \n", ix);
                iy = 10;
                // for(iy = 0; iy <= Ny/2; iy++)
                // for(iy = 0; iy < 1; iy++)
                {
                    ic[0] = ix; ic[1] = iy;
                    ic2[0] = ix; ic2[1] = Ny-iy-1;
                    // if(fabs(p_cell[ic[0]][ic[1]].r_side - 
                    //         p_cell[ic2[0]][ic2[1]].r_side) >0.00000000001)
                    //     printf("(%d %9.8g, %9.8g) ",
                    //       ic[0], p_cell[ic[0]][ic[1]].r_side,
                    //         p_cell[ic2[0]][ic2[1]].r_side); 
                     
                     printf("(%d %9.8g)\n",
                         ic[1], p_cell[ic[0]][ic[1]].r_side);
                     printf("line-0 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                             p_cell[ix][iy].set[0][0], p_cell[ix][iy].set[1][0],
                             p_cell[ix][iy].set[2][0],
                             p_cell[ix][iy].stencil[0][0],
                             p_cell[ix][iy].stencil[1][0], p_cell[ix][iy].stencil[2][0]);
                     printf("line-1 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                             p_cell[ix][iy].set[0][1], p_cell[ix][iy].set[1][1],
                             p_cell[ix][iy].set[2][1],
                             p_cell[ix][iy].stencil[0][1],
                             p_cell[ix][iy].stencil[1][1], p_cell[ix][iy].stencil[2][1]);
                     printf("line-2 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                             p_cell[ix][iy].set[0][2], p_cell[ix][iy].set[1][2],
                             p_cell[ix][iy].set[2][2],
                             p_cell[ix][iy].stencil[0][2], p_cell[ix][iy].stencil[1][2],
                             p_cell[ix][iy].stencil[2][2]);
                     /**
                     printf("(%d %9.8g)\n",
                         ic2[1], p_cell[ic2[0]][ic2[1]].r_side);
                     printf("line-0 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                             p_cell[ic2[0]][ic2[1]].set[0][0], p_cell[ic2[0]][ic2[1]].set[1][0],
                             p_cell[ic2[0]][ic2[1]].set[2][0],
                             p_cell[ic2[0]][ic2[1]].stencil[0][0],
                             p_cell[ic2[0]][ic2[1]].stencil[1][0], 
                             p_cell[ic2[0]][ic2[1]].stencil[2][0]);
                     printf("line-1 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                             p_cell[ic2[0]][ic2[1]].set[0][1], 
                             p_cell[ic2[0]][ic2[1]].set[1][1],
                             p_cell[ic2[0]][ic2[1]].set[2][1],
                             p_cell[ic2[0]][ic2[1]].stencil[0][1],
                             p_cell[ic2[0]][ic2[1]].stencil[1][1], 
                             p_cell[ic2[0]][ic2[1]].stencil[2][1]);
                     printf("line-2 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                             p_cell[ic2[0]][ic2[1]].set[0][2], 
                             p_cell[ic2[0]][ic2[1]].set[1][2],
                             p_cell[ic2[0]][ic2[1]].set[2][2],
                             p_cell[ic2[0]][ic2[1]].stencil[0][2], 
                             p_cell[ic2[0]][ic2[1]].stencil[1][2],
                             p_cell[ic2[0]][ic2[1]].stencil[2][2]);
                     temp = fabs(p_cell[ic[0]][ic[1]].r_side - p_cell[ic2[0]][ic2[1]].r_side);
                     linb = fabs(p_cell[ix][iy].stencil[0][0] - p_cell[ic2[0]][ic2[1]].stencil[0][2]) +
                            fabs(p_cell[ix][iy].stencil[1][0] - p_cell[ic2[0]][ic2[1]].stencil[1][2]) +
                            fabs(p_cell[ix][iy].stencil[2][0] - p_cell[ic2[0]][ic2[1]].stencil[2][2]);

                     linc = fabs(p_cell[ix][iy].stencil[0][1] - p_cell[ic2[0]][ic2[1]].stencil[0][1]) +
                            fabs(p_cell[ix][iy].stencil[1][1] - p_cell[ic2[0]][ic2[1]].stencil[1][1]) +
                            fabs(p_cell[ix][iy].stencil[2][1] - p_cell[ic2[0]][ic2[1]].stencil[2][1]);

                     lint = fabs(p_cell[ix][iy].stencil[0][2] - p_cell[ic2[0]][ic2[1]].stencil[0][0]) +
                            fabs(p_cell[ix][iy].stencil[1][2] - p_cell[ic2[0]][ic2[1]].stencil[1][0]) +
                            fabs(p_cell[ix][iy].stencil[2][2] - p_cell[ic2[0]][ic2[1]].stencil[2][0]);
                     printf("(%d --- %d %9.8g, %9.8g)\n",
                         ic[1], ic2[1], fabs(temp), fabs( linb + linc + lint));
                     */
                     printf("\n");
                }
                printf("\n");
            }
            printf("\n");
        }

        for (I = ilower; I <= iupper; I++)
        {
            // set_matrix_row(A,I,ilower,I_to_ijk,IJK_to_I,ns_field);
            // set_right_hand_side(B,I,ilower,I_to_ijk,ns_field);
            set_matrix_A_and_right_hand_side(A,B,newwv,
                  p_cell,I,ilower,I_to_ijk,IJK_to_I,ns_field);
        }

        MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
        MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);

        VecAssemblyBegin(B);
        VecAssemblyEnd(B);

        // MatNullSpaceCreate(PETSC_COMM_WORLD,PETSC_TRUE,0,PETSC_NULL,&nullsp);

        KSPCreate(PETSC_COMM_WORLD,&ksp); // CHKERRQ(ierr);
        KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN); // CHKERRQ(ierr);

        KSPGetPC(ksp,&pc);
        PCSetType(pc,PCLU);

#if defined(__MPI__)
        KSPGetPC(ksp,&pc);  //CHKERRQ(ierr);
        PCSetType(pc,PCJACOBI);  //CHKERRQ(ierr);
#endif // if defined(__MPI__) 

        // KSPSetNullSpace(ksp,nullsp);
        // KSPSetType(ksp,KSPBCGS);
        // KSPSetTolerances(ksp,1.e-12,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);  // CHKERRQ(ierr);
        // KSPSetTolerances(ksp,1.e-13,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);  // CHKERRQ(ierr);

        // new, seems very fast, 062607
        // KSPSetType(ksp,KSPLGMRES);
        // KSPSetType(ksp,KSPFGMRES);

        KSPSetTolerances(ksp,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);
        KSPSetComputeSingularValues(ksp, PETSC_TRUE);

        // KSPSetFromOptions(ksp); // CHKERRQ(ierr); 062607
        // KSPSetUp(ksp);

        KSPSolve(ksp,B,phi); // CHKERRQ(ierr);

        if(debugging("NS_vel") && NO)
        {
            KSPView(ksp,PETSC_VIEWER_STDOUT_WORLD);  // CHKERRQ(ierr); 
        }

        KSPGetIterationNumber(ksp,&its);
        KSPGetResidualNorm(ksp,&norm);

        KSPComputeExtremeSingularValues(ksp, &max, &min);
   
        /**
        printf("\n U^star Projection: Number of iteration  is %d;\n",its);
        printf("Residual norm = %6.5e,extreme test SV is [%7.6e ; %7.6e]\n",norm,min,max);
        {
            MatMult(A,phi,C);
            VecAXPY(C,-1.0,B);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            printf("After Petsc solver, norm B = %g, norm (A*x-B) = %g\n",
                         norm_B, norm_C);
        }
        **/

        KSPGetConvergedReason(ksp,&reason);
        if(reason < 0)
        {
            printf("U^star Projection:\n");
            printf("ERROR KSPSolve failed, try to increase number of iterations\n");
            KSPSetTolerances(ksp,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT,(10*NLblocks));
            KSPSolve(ksp,B,phi); // CHKERRQ(ierr);
            // KSPSetType(ksp,KSPLSQR);

            // Test 2
            KSPGetConvergedReason(ksp,&reason);
            {
                PetscReal rtol, atol, dtol;
                int maxits, isize;
                KSPType ksptype;
                KSPGetTolerances(ksp,&rtol,&atol,&dtol,&maxits);
                printf("default rtol = %g, atol = %g, dtol = %g, maxits = %d\n",
                   rtol, atol, dtol, maxits);
                VecGetLocalSize(phi, &isize);
                KSPGetType(ksp,&ksptype);
                printf("vec_size = %d, ksptype = %s\n", isize, ksptype);

                MatMult(A,phi,C);
                VecAXPY(C,-1.0,B);
                VecNorm(C,NORM_2,&norm_C);
                VecNorm(B,NORM_2,&norm_B);
                printf("After Petsc solver, norm B = %g, norm (A*x-B) = %g\n",
                         norm_B, norm_C);
                printf("rtol*B_norm = %g, dtol*B_norm = %g\n",
                   rtol*norm_B, dtol*norm_B);
            }
            KSPGetIterationNumber(ksp,&its);
            KSPGetResidualNorm(ksp,&norm);
            printf("2nd try: Residual norm = %6.5e,"
              "extreme test SV is [%7.6e ; %7.6e], reason %d, its = %d\n",
                   norm,min,max, reason, its);
            if(reason < 0)
            { 
                KSPView(ksp,PETSC_VIEWER_STDOUT_WORLD);
                clean_up(ERROR);
            }
        }

        if(debugging("NS_vel") && NO)
        {
            MatView(A,PETSC_VIEWER_STDOUT_WORLD);
            VecView(B,PETSC_VIEWER_STDOUT_WORLD);
            VecView(phi,PETSC_VIEWER_STDOUT_WORLD);
        }

        for (iy = 0; iy < Ny; iy++)
        {
            /* 030707
            // Phis of right side boundary cells are set to be zero
            ic[0] = Nx-1;
            ic[1] = iy;
            Rect_phi(ic,ns_field) = 0.0;
            */
            // for (ix = 0; ix < Nx-1; ix++)
            for (ix = 0; ix < Nx; ix++)
            {
                indices = IJK_to_I[ix+2][iy+2];
                ic[0] = ix; ic[1] = iy;

                if (indices >= 0)
                {
                   VecGetArray(phi,&vphi);
                   // potential[ix+2][iy+2] = vphi[indices-ilower];
                   Rect_phi(ic,ns_field) = vphi[indices-ilower];
                   // VecGetArray(B,&vphi);

                   // printf("ic[%d %d] phi = %g, b = %g\n",
                   //   ix, iy, potential[ix+2][iy+2], vphi[indices-ilower]);

                   /*
                   if(debugging("NS_vel"))
                       printf("ic[%d %d] phi = %g\n",
                          ix, iy, Rect_phi(ic,ns_field));
                   */

                   // VecRestoreArray(phi,&values);
                   VecRestoreArray(phi,&vphi);
                }
            }
        }

        // TEST
        if(debugging("NS_vel"))
        {
            printf("project_vel_star, right_hand_side\n");
            for(j = 0; j < Ny; j++)
            // for(j = 3; j <= 5; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i < gmax[0]; i++)
                for(i = 0; i <= Nx/2; i++)
                {
                    ic[0] = i; ic[1] = j;
                    printf("%d %g; ", i, p_cell[i][j].r_side);
                }
                printf("\n");
            }
        }

        if(debugging("NS_vel"))
        {
            printf("\n\n-----------After Petsc project_vel_star\n");
            printf("------------- MAC phi--------------\n");
            for(j = 0; j < Ny; j++)
            // for(j = 3; j <= 5; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i < gmax[0]; i++)
                for(i = 0; i < Nx; i++)
                {
                    ic[0] = i; ic[1] = j;
                    printf("(%d %8.7g)", i, Rect_phi(ic, ns_field));
                }
                printf("\n");
            }

        }
        // END TEST

        // TEST
        if(debugging("NS_vel") && NO)
        {
            MatMult(A,phi,C);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            // printf("-------A*phi, print RHS----\n");
            // VecView(B,PETSC_VIEWER_STDOUT_WORLD);
            printf("After Petsc solver, norm B = %g, norm (A*phi) = %g\n",
                         norm_B, norm_C);
        }
        // END TEST        

        VecDestroy(B);
        VecDestroy(phi);
        VecDestroy(C);
        MatDestroy(A);
        KSPDestroy(ksp);
        // MatNullSpaceDestroy(nullsp);

        free(p_cell);
#endif /* if defined (PETSC) */
        if(n_dist != NULL)
            free(n_dist);

        if(debugging("NS_vel") && NO)
        {
            int ic[MAXD] = {1,1};
            int icl[MAXD] = {0,1};
            int icr[MAXD] = {2,1};
            int ict[MAXD] = {1,2};
            int icb[MAXD] = {1,0};
            float lap_phi;
            lap_phi = (Rect_phi(icl,ns_field) + Rect_phi(icr,ns_field) +
                      Rect_phi(ict,ns_field) + Rect_phi(icb,ns_field) - 
                      4.0*Rect_phi(ic,ns_field))*dt;
            printf("Soln laplian phi[%d %d] = %g  (%g %g %g %g) %g\n", 
                  ic[0], ic[1], lap_phi,
                  Rect_phi(icl,ns_field), Rect_phi(icr,ns_field), 
                  Rect_phi(ict,ns_field), Rect_phi(icb,ns_field),
                  Rect_phi(ic,ns_field));
        }
        if(debugging("NS_vel") && NO)
        {
            int ic[MAXD] = {1,0};
            int icl[MAXD] = {0,0};
            int icr[MAXD] = {2,0};
            int ict[MAXD] = {1,1};
            float lap_phi;
            lap_phi = (Rect_phi(icl,ns_field) + Rect_phi(icr,ns_field) +
                      Rect_phi(ict,ns_field)  - 
                      3.0*Rect_phi(ic,ns_field))*dt;
            printf("Soln laplian phi[%d %d] = %g\n", ic[0], ic[1], lap_phi);
        }

        DEBUG_LEAVE(project_vel_star)
}

/* Set vector field (u^*-u^n)/dt on cell edge for projection
 * Some interpolation has not been thought through.
*/
LOCAL void cell_edge_vector_u(
        float             dt,
        NS_Field_Stencil  *ns_field)
{
        int             smin[MAXD], smax[MAXD];
        Wave            *newwv = ns_field->wave;
        Wave            *wv = ns_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(ns_field->newwave)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = ns_field->newfr, *fr = ns_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD], icrds[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vold[MAXD], vnew[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh = wv->rect_grid->h;
        int             vset;

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        // Vertical edge 
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0]+1;  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix-1; ic[1] = iy;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                // printf("on ix, iy [%d %d]\n", ix, iy);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                {
                    for(i = 0; i < dim; i++)
                    {
                        vnew[i] = 0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]);
                        vold[i] = 0.5*(vel(i,Rect_state(ic,wv)) + vel(i,Rect_state(ic1,wv)));
                    }
                    for(i = 0; i < dim; i++)
                        v_line_vel(ic1,ns_field).v[i] = (vnew[i]-vold[i])/dt;
                }
                else 
                {
                    // Cell on the right side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    // Dirichlet boundary type, might need to re-consider
                    if(Rect_blk_edge(icrds,wv).edge[0][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[0][0].etype == PARTIAL_SURF ||
                       Rect_blk_edge(icrds,wv).edge[0][0].etype == FULL_SURF)
                    {
                        // constant extrapolation
                        if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                v_line_vel(ic1,ns_field).v[i] = 
                                 (Rect_u_star(ic,ns_field).v[i]-vel(i,Rect_state(ic,wv)))/dt;
                            }
                        }
                        else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                v_line_vel(ic1,ns_field).v[i] = 
                                 (Rect_u_star(ic1,ns_field).v[i]-vel(i,Rect_state(ic1,wv)))/dt;
                            }
                        }
                        else
                        {
                            printf("ERROR: here[%d %d], what's the case?\n", ix, iy);
                            printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype);
                            clean_up(ERROR);
                        }
                    }
                    else
                    {
                        printf("ERROR: case here[%d %d], what's the case?\n", ix, iy);
                        printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype);
                        clean_up(ERROR);
                    }
                }
            }
        }

        // Left vertical edge
        ic[0] = ix = smin[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic,ns_field).v[i] = 0.0;
            }

            sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);
            
            sten->fr  = fr;
            sten->wave = wv;
            sten->newfr = nfr; sten->newwave = newwv;
            newcrx = oldcrx = NULL;

            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                oldcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,oldcrx,sten->p[-1],
                           comp,sten->worksp[-1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            sten->fr  = nfr;
            sten->wave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                v_line_vel(ic,ns_field).v[i] = (vnew[i] - vold[i])/dt;

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //        ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }
        
        // Right vertical edge
        ic[0] = ix = smax[0]-1;
        ic1[1] = smax[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[0][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic1,ns_field).v[i] = 0.0;
            }

            sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);
            
            sten->fr  = fr;
            sten->wave = wv;
            sten->newfr = nfr; sten->newwave = newwv;
            newcrx = oldcrx = NULL;

            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                oldcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,oldcrx,sten->p[1],
                           comp,sten->worksp[1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            sten->fr  = nfr;
            sten->wave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                v_line_vel(ic1,ns_field).v[i] = (vnew[i] - vold[i])/dt;
        }
        
        // Horizontal edge 
        for (iy = smin[1]+1;  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix; ic[1] = iy-1;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                // printf("on ix, iy [%d %d]\n", ix, iy);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                {
                    for(i = 0; i < dim; i++)
                    {
                        vnew[i] = 0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]);
                        vold[i] = 0.5*(vel(i,Rect_state(ic,wv)) + vel(i,Rect_state(ic1,wv)));
                    }
                    for(i = 0; i < dim; i++)
                        h_line_vel(ic1,ns_field).v[i] = (vnew[i]-vold[i])/dt;
                }
                else 
                {
                    // Cell on the upper side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    // Dirichlet boundary type, might need to re-consider
                    if(Rect_blk_edge(icrds,wv).edge[1][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[1][0].etype == PARTIAL_SURF ||
                       Rect_blk_edge(icrds,wv).edge[1][0].etype == FULL_SURF)
                    {
                        // constant extrapolation
                        if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                h_line_vel(ic1,ns_field).v[i] = 
                                 (Rect_u_star(ic,ns_field).v[i]-vel(i,Rect_state(ic,wv)))/dt;
                            }
                        }
                        else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                h_line_vel(ic1,ns_field).v[i] = 
                                 (Rect_u_star(ic1,ns_field).v[i]-vel(i,Rect_state(ic1,wv)))/dt;
                            }
                        }
                        else
                        {
                            printf("ERROR: horizontal edge here[%d %d], what's the case?\n", ix, iy);
                            printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype);
                            clean_up(ERROR);
                        }
                    }
                    else
                    {
                        printf("ERROR: horizontal case here[%d %d], what's the case?\n", ix, iy);
                        printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype); 
                        printf("edge type %d N_SURF = %d\n", Rect_blk_edge(icrds,wv).edge[1][0].etype,N_SURF);
                        clean_up(ERROR);
                    }
                }
            }
        }

        // bottom horizontal edge
        ic[1] = iy = smin[1];
        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic,ns_field).v[i] = 0.0;
            }

            sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);
            
            sten->fr  = fr;
            sten->wave = wv;
            sten->newfr = nfr; sten->newwave = newwv;
            newcrx = oldcrx = NULL;

            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                oldcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,oldcrx,sten->p[-1],
                           comp,sten->worksp[-1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            sten->fr  = nfr;
            sten->wave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                h_line_vel(ic,ns_field).v[i] = (vnew[i] - vold[i])/dt;

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }

        // top horizontal edge
        ic[1] = iy = smax[1]-1;
        ic1[1] =  smax[1];

        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy; 
            ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[1][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic1,ns_field).v[i] = 0.0;
            }

            sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);
            
            sten->fr  = fr;
            sten->wave = wv;
            sten->newfr = nfr; sten->newwave = newwv;
            newcrx = oldcrx = NULL;

            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                oldcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,oldcrx,sten->p[1],
                           comp,sten->worksp[1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), top boundary, no crossing\n");
                clean_up(ERROR);
            }

            sten->fr  = nfr;
            sten->wave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), top boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                h_line_vel(ic1,ns_field).v[i] = (vnew[i] - vold[i])/dt;

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }

}

/* Set vector field u^* on cell edge for projection
 * Some interpolation has not been thought through.
 *. THe handling of boundary cells is not symmetric. Also,
 * The edge velocities of boundary cells should be zero???
 Cell Case: crxing dir[0,1,2,3] = [1, 0, 0, 1]  
 Below this cell: obstacle.
 The edge velocity of [1][0] (lower side) should be zero ???
*/
LOCAL void cell_edge_vector_u_ver2(
        float             dt,
        NS_Field_Stencil  *ns_field)
{
        int             smin[MAXD], smax[MAXD];
        Wave            *newwv = ns_field->wave;
        Wave            *wv = ns_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(ns_field->newwave)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = ns_field->newfr, *fr = ns_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD], icrds[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vnew[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh = wv->rect_grid->h;
        int             vset;

        if(debugging("NS_vel"))
        {
            printf("--------------Entered cell_edge_vector_u_ver2\n");
            printf("Crank-Nicolson velocity:\n");
            for (iy = 0; iy < gmax[1]; iy++)
            {
                printf("Row-cell[%d]: ", iy);
                for (ix = 0; ix < min(4,gmax[0]); ix++)
                {
                    ic[0] = ix; ic[1] = iy;
                    printf("(%d %g %g) ", ix, Rect_u_star(ic,ns_field).v[0],
                             Rect_u_star(ic,ns_field).v[1]);
                }
                printf("\n");
            }
        }

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        // Vertical edge 
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0]+1;  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix-1; ic[1] = iy;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                // printf("on ix, iy [%d %d]\n", ix, iy);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                {
                    for(i = 0; i < dim; i++)
                    {
                        v_line_vel(ic1,ns_field).v[i] = vnew[i] = 
                          0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]);
                    }
                }
                else 
                {
                    // Cell on the right side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    // Dirichlet boundary type, might need to re-consider
                    if(Rect_blk_edge(icrds,wv).edge[0][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[0][0].etype == PARTIAL_SURF ||
                       Rect_blk_edge(icrds,wv).edge[0][0].etype == FULL_SURF)
                    {
                        // constant extrapolation
                        if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                v_line_vel(ic1,ns_field).v[i] = 
                                   Rect_u_star(ic,ns_field).v[i];
                            }
                        }
                        else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                v_line_vel(ic1,ns_field).v[i] = 
                                   Rect_u_star(ic1,ns_field).v[i];
                            }
                        }
                        else
                        {
                            printf("ERROR: here[%d %d], what's the case?\n", ix, iy);
                            printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype);
                            clean_up(ERROR);
                        }
                    }
                    else
                    {
                        printf("ERROR: case here[%d %d], what's the case?\n", ix, iy);
                        printf("cell type %d %d\n", Rect_blk_edge(icrds,wv).ctype);
                        clean_up(ERROR);
                    }
                }
            }
        }

        // Left vertical edge
        ic[0] = ix = smin[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            // TMP
            // printf("cell_edge_vector_u_ver2, cell[%d %d], left edge_type = %d, N_SURF = %d\n",
            //                 ix, iy, blk_edge[ix][iy].edge[0][0].etype, N_SURF);

            if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);
            
            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                v_line_vel(ic,ns_field).v[i] = vnew[i];

            // printf("cell[%d %d], left side: vnew[%g %g]\n",
            //        ic[0], ic[1], vnew[0], vnew[1]);
        }
        
        // Right vertical edge
        ic[0] = ix = smax[0]-1;
        ic1[0] = smax[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic1[1] = ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[0][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic1,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);
            
            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver2(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                v_line_vel(ic1,ns_field).v[i] = vnew[i];
        }
        
        // Horizontal edge 
        for (iy = smin[1]+1;  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix; ic[1] = iy-1;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                /***
                if(ix == 3 && (iy == 1 || iy == 8 || iy ==9))
                {
                    printf("**************\n");
                    printf("cell_edge_vector_u_ver2 on ix, iy [%d %d], ctype = %d\n", ix, iy,
                                    Rect_blk_edge(icrds,wv).ctype);
                    printf("comp = %d, comp1 = %d [ix][iy].edge[1][0].etype =%d\n", 
                          comp, comp1,Rect_blk_edge(icrds,wv).edge[1][0].etype);
                    printf("******END********\n");
                }
                ***/

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                {
                    for(i = 0; i < dim; i++)
                    {
                        h_line_vel(ic1,ns_field).v[i] = vnew[i] = 
                             0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]);
                    }

                    /*
                    if(ix == 3 && (iy == 1 || iy == 8 || iy == 9))
                    {
                        printf("set ic1[%d %d] h_vel (%g %g) case 0\n",
                             ic1[0], ic1[1], h_line_vel(ic1,ns_field).v[0], 
                             h_line_vel(ic1,ns_field).v[1]);
                    }
                    */
                }
                else 
                {
                    // Cell on the upper side of the edge
                    if(ix == 3 && (iy == 1 || iy == 8 || iy == 9))
                    {
                        printf("blk_edge[%d][%d].ctype = %d, case 1\n",
                             ix, iy, Rect_blk_edge(icrds,wv).ctype);
                    }

                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    /***
                    if(ix == 3 && (iy == 1 || iy == 8 || iy == 9))
                    {
                        printf("set [%d][%d] edge[1][0].etype= %d case 2\n",
                             ix, iy, blk_edge[ix][iy].edge[1][0].etype);
                    }
                    ***/

                    // Dirichlet boundary type, might need to re-consider
                    if(Rect_blk_edge(icrds,wv).edge[1][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[1][0].etype == PARTIAL_SURF ||
                       Rect_blk_edge(icrds,wv).edge[1][0].etype == FULL_SURF)
                    {
                        // constant extrapolation
                        if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                h_line_vel(ic1,ns_field).v[i] = 
                                 Rect_u_star(ic,ns_field).v[i];
                            }
                        }
                        else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                h_line_vel(ic1,ns_field).v[i] = 
                                 Rect_u_star(ic1,ns_field).v[i];
                            }
                        }
                        else
                        {
                            printf("ERROR: horizontal edge here[%d %d], what's the case?\n", ix, iy);
                            printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype);
                            clean_up(ERROR);
                        }
                    }
                    else
                    {
                        printf("ERROR: horizontal case here[%d %d], what's the case?\n", ix, iy);
                        printf("cell type %d\n", Rect_blk_edge(icrds,wv).ctype);
                        printf("edge type %d N_SURF = %d\n", Rect_blk_edge(icrds,wv).edge[1][0].etype,N_SURF);
                        clean_up(ERROR);
                    }

                    /***
                    if(ix == 3 && (iy == 1 || iy == 8 || iy == 9))
                    {
                        printf("[%d][%d] edge[1][0].etype= %d case 2, h_vel (%g %g)\n",
                             ix, iy, blk_edge[ix][iy].edge[1][0].etype,
                             h_line_vel(ic1,ns_field).v[0], h_line_vel(ic1,ns_field).v[1]);
                    }
                    ***/
                }
            }
        }

        // bottom horizontal edge

        ic[1] = iy = smin[1];
        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            ic[0] = ix;
            icrds[0] = ix; icrds[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver2(), left boundary, no crossing\n");
                clean_up(ERROR);
            }


            // Dirichlet 
            for(i = 0; i < dim; i++)
                h_line_vel(ic,ns_field).v[i] = vnew[i];

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }

        // top horizontal edge
        ic[1] = iy = smax[1]-1;
        ic1[1] =  smax[1];

        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic1[0] = ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[1][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic1,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);
            
            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u(), top boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                h_line_vel(ic1,ns_field).v[i] = vnew[i];

            // printf("ic1[%d %d] set h_line_vel as Dirichlet vnew[%g %g]\n",
            //       ic1[0], ic1[1], vnew[0], vnew[1]);

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }


        // TMP
        if(debugging("NS_vel"))
        {
            printf("In cell_edge_vector_u_ver2, show vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i < min(gmax[0],4); i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d, %g %g) ", i,
                        h_line_vel(ic, ns_field).v[0], h_line_vel(ic, ns_field).v[1]);
                }
                printf("\n");
            }
            /***
            printf("In cell_edge_vector_u_ver2, show vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= min(gmax[0],4); i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d, %g %g) ", i,
                       v_line_vel(ic, ns_field).v[0],v_line_vel(ic, ns_field).v[1]);
                }
                printf("\n");
            }
            ***/ 
        }
}


/* Set vector field u^* on cell edge for projection
 * Some interpolation has not been thought through.
*/
LOCAL void cell_edge_vector_u_ver3(
        float             dt,
        NS_Field_Stencil  *ns_field)
{
        int             smin[MAXD], smax[MAXD];
        Wave            *newwv = ns_field->wave;
        Wave            *wv = ns_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(ns_field->newwave)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = ns_field->newfr, *fr = ns_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD], icrds[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vnew[MAXD], vold[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh = wv->rect_grid->h;
        int             vset;
        Locstate        ost1, ost2;

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        // Vertical edge 
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0]+1;  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix-1; ic[1] = iy;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                // printf("on ix, iy [%d %d]\n", ix, iy);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                {
                    for(i = 0; i < dim; i++)
                    {
                        v_line_vel(ic1,ns_field).v[i] = vnew[i] = 
                          0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]) -
                          0.5*( Vel(Rect_state(ic,wv))[i] + Vel(Rect_state(ic1,wv))[i]);
                    }
                }
                else 
                {
                    // Cell on the right side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    // Dirichlet boundary type, might need to re-consider
                    if(Rect_blk_edge(icrds,wv).edge[0][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[0][0].etype == PARTIAL_SURF ||
                       Rect_blk_edge(icrds,wv).edge[0][0].etype == FULL_SURF)
                    {
                        // constant extrapolation
                        if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                v_line_vel(ic1,ns_field).v[i] = 
                                   Rect_u_star(ic,ns_field).v[i] -
                                   Vel(Rect_state(ic,wv))[i];
                            }
                        }
                        else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                v_line_vel(ic1,ns_field).v[i] = 
                                   Rect_u_star(ic1,ns_field).v[i]-
                                   Vel(Rect_state(ic1,wv))[i];
                            }
                        }
                        else
                        {
                            printf("ERROR: here[%d %d], what's the case?\n", ix, iy);
                            clean_up(ERROR);
                        }
                    }
                    else
                    {
                        printf("ERROR: case here[%d %d], what's the case?\n", ix, iy);
                        clean_up(ERROR);
                    }
                }
            }
        }

        // Left vertical edge
        ic[0] = ix = smin[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy; 
            ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            // TMP
            // printf("cell_edge_vector_u_ver2, cell[%d %d], left edge_type = %d, N_SURF = %d\n",
            //                 ix, iy, blk_edge[ix][iy].edge[0][0].etype, N_SURF);

            if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);

            sten->fr  = fr;
            sten->wave = wv;
            sten->newwave = wv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), left boundary, no crossing, old\n");
                clean_up(ERROR);
            }
            
            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                v_line_vel(ic,ns_field).v[i] = vnew[i]-vold[i];

            // printf("cell[%d %d], left side: vnew[%g %g]\n",
            //        ic[0], ic[1], vnew[0], vnew[1]);
        }
        
        // Right vertical edge
        ic[0] = ix = smax[0]-1;
        ic1[0] = smax[0];
        prev_side = WEST;
        next_side = EAST;
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            icrds[0] = ix; icrds[1] = iy;
            ic1[1] = ic[1] = iy;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[0][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    v_line_vel(ic1,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);

            sten->fr  = fr;
            sten->wave = wv;
            sten->newwave = wv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), left boundary, no crossing old\n");
                clean_up(ERROR);
            }
            
            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), left boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                v_line_vel(ic1,ns_field).v[i] = vnew[i]-vold[i];
        }
        
        // Horizontal edge 
        for (iy = smin[1]+1;  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix; ic[1] = iy-1;
                ic1[0] = ix; ic1[1] = iy;
                comp = Rect_comp(ic,newwv);
                comp1 = Rect_comp(ic1,newwv);

                // printf("on ix, iy [%d %d]\n", ix, iy);

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                {
                    for(i = 0; i < dim; i++)
                    {
                        h_line_vel(ic1,ns_field).v[i] = vnew[i] = 
                             0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]) -
                             0.5*( Vel(Rect_state(ic,wv))[i] - Vel(Rect_state(ic1,wv))[i] );
                    }
                }
                else 
                {
                    // Cell on the upper side of the edge
                    if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                        continue;

                    // Dirichlet boundary type, might need to re-consider
                    if(Rect_blk_edge(icrds,wv).edge[1][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    else if(Rect_blk_edge(icrds,wv).edge[1][0].etype == PARTIAL_SURF ||
                       Rect_blk_edge(icrds,wv).edge[1][0].etype == FULL_SURF)
                    {
                        // constant extrapolation
                        if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                h_line_vel(ic1,ns_field).v[i] = 
                                 Rect_u_star(ic,ns_field).v[i] - 
                                 Vel(Rect_state(ic,wv))[i];
                            }
                        }
                        else if(comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                        {
                            for(i = 0; i < dim; i++)
                            {
                                h_line_vel(ic1,ns_field).v[i] = 
                                 Rect_u_star(ic1,ns_field).v[i] - 
                                 Vel(Rect_state(ic1,wv))[i];
                            }
                        }
                        else
                        {
                            printf("ERROR: horizontal edge here[%d %d], what's the case?\n", ix, iy);
                            clean_up(ERROR);
                        }
                    }
                    else
                    {
                        printf("ERROR: horizontal case here[%d %d], what's the case?\n", ix, iy);
                        clean_up(ERROR);
                    }
                }
            }
        }

        // bottom horizontal edge

        ic[1] = iy = smin[1];
        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy; 
            ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);

            sten->fr  = fr;
            sten->wave = wv;
            sten->newwave = wv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), bottom boundary, no crossing old\n");
                clean_up(ERROR);
            }

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[-1] != 0)
            {
                newcrx = sten->crx[-1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), left boundary, no crossing\n");
                clean_up(ERROR);
            }


            // Dirichlet 
            for(i = 0; i < dim; i++)
                h_line_vel(ic,ns_field).v[i] = vnew[i] - vold[i];

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }

        // top horizontal edge
        ic[1] = iy = smax[1]-1;
        ic1[1] =  smax[1];

        prev_side = SOUTH;
        next_side = NORTH;
        for (ix = smin[0];  ix < smax[0];  ++ix)
        {
            icrds[0] = ix; icrds[1] = iy;  
            ic1[0] = ic[0] = ix;
            comp = Rect_comp(ic,newwv);

            if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                 continue;

            // Neumann boundary type, set to 0.0
            if(Rect_blk_edge(icrds,wv).edge[1][1].etype == N_SURF)
            {
                for(i = 0; i < dim; i++)
                    h_line_vel(ic1,ns_field).v[i] = 0.0;
                continue;
            }

            sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);
            
            sten->fr  = fr;
            sten->wave = wv;
            sten->newwave = wv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vold);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), top boundary, no crossing, old\n");
                clean_up(ERROR);
            }

            sten->fr  = nfr;
            sten->wave = newwv;
            sten->newwave = newwv;
            ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
            if(sten->nc[1] != 0)
            {
                newcrx = sten->crx[1][0];
                vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
            }
            else
            {
                printf("ERROR: cell_edge_vector_u_ver3(), top boundary, no crossing\n");
                clean_up(ERROR);
            }

            // Dirichlet 
            for(i = 0; i < dim; i++)
                h_line_vel(ic1,ns_field).v[i] = vnew[i] - vold[i];

            // printf("ic1[%d %d] set h_line_vel as Dirichlet vnew[%g %g]\n",
            //       ic1[0], ic1[1], vnew[0], vnew[1]);

            // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
            //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
        }


        // TMP
        if(debugging("NS_vel"))
        {
            printf("In cell_edge_vector_u_ver3, show vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i < gmax[0]; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d, %g %g) ", i,
                        h_line_vel(ic, ns_field).v[0], h_line_vel(ic, ns_field).v[1]);
                }
                printf("\n");
            }
            printf("In cell_edge_vector_u_ver3, show vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d, %g %g) ", i,
                       v_line_vel(ic, ns_field).v[0],v_line_vel(ic, ns_field).v[1]);
                }
                printf("\n");
            }
        }
}



LIB_LOCAL int  velocity_from_boundary(
        Stencil    *sten,
        CRXING     *crx,
        POINT      *pt,
        COMPONENT  comp,
        Locstate   st,
        float      *v)
{
        HYPER_SURF  *hs;
        Front       *fr = sten->fr;
        Wave        *wave = sten->wave;
        int         i;

        hs = crx->hs;
        if(wave_type(hs) == NEUMANN_BOUNDARY)
        {
            if(fr->neumann_bdry_state)
            {
                if ((*fr->neumann_bdry_state)(Coords(pt),comp,
                     crx->pt,hs,fr,(POINTER)wave,st))
                {
                    for(i = 0; i < dim; i++)
                        v[i] = vel(i,st);
                    return YES;
                }
                else
                {
                    // printf("ERROR: velocity_from_boundary, NEUMANN\n");
                    // print_curve(Curve_of_hs(hs));
                    // clean_up(ERROR);
                }
            }
            for(i = 0; i < dim; i++)
               v[i] = 0.0;
            return YES;
        }  
        if(wave_type(hs) == DIRICHLET_BOUNDARY)
        {
             evaluate_dirichlet_boundary_state(Coords(pt),hs,fr,wave,
                                              st);
             for(i = 0; i < dim; i++)
                v[i] = vel(i,st);
             return YES;
        }
        else
        {
             printf("ERROR: velocity_from_boundary, can not handle\n");
             printf("wave type %d comp = %d\n", wave_type(hs), comp);
             print_general_vector("point crds",Coords(pt),dim,"\n");
             print_curve(Curve_of_hs(hs));
             print_crxings(crx, NO);
             geomview_intfc_plot2d("gview_plot",fr->step+1000,
                    fr->interf,fr->rect_grid);
             geomview_intfc_plot2d("gview_plot",fr->step+2000,
                    wave_tri_soln(wave)->tri_grid->grid_intfc,fr->rect_grid);
             clean_up(ERROR);
        }

        for(i = 0; i < dim; i++)
            v[i] =  vel(i,Rect_state(sten->icoords[0],wave));
        return NO;
}


LIB_LOCAL void sten_pt_loc_in_dir(
        int     *ic, 
        int     dir,
        int     end,
        Wave    *wv,
        int     on_cell_edge,
        Stencil *sten)
{
        int     j, l;
        int     **icoords;
        float   *dh = wv->rect_grid->h;
        float   half = 0.0;

        icoords = sten->icoords;

        for(j = 0; j < dim; j++)
            icoords[0][j] = ic[j];

        for (j = -end; j <= end; ++j)
            icoords[j][dir] = ic[dir] + j;

        for (j = 0; j < dim; ++j)
        {
            if(on_cell_edge == YES)
                half = 0.5*dh[j];

            if (j == dir)
            {
                Coords(sten->p[0])[dir] =
                   cell_center(ic[dir],dir,wv->rect_grid);
                for (l = 1; l <= end; ++l)
                {
                    Coords(sten->p[-l])[dir] =
                         Coords(sten->p[0])[dir]-l*dh[j] + half;
                    Coords(sten->p[l])[dir] =
                         Coords(sten->p[0])[dir]+l*dh[j] - half;
                }
            }
            else
            {
                for (l = -end; l <= end; ++l)
                    Coords(sten->p[l])[j] = Rect_coords(ic,wv)[j];
            }
        }
}

/* 
 * grid_den_viscosity: exactly count the area of partial 
 * (including physical boundary) cells, and the density and viscosity
 * is weighted by the area.  
*/
LOCAL void grid_den_viscosity(
        NS_Field_Stencil  *ns_field,
        Wave              *wv,
        Front             *fr)
{
        int             smin[MAXD], smax[MAXD];
        // Wave            *newwv = ns_field->newwave;
        // TRI_GRID        *ntg = wave_tri_soln(ns_field->newwave)->tri_grid;
        TRI_GRID        *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        // Front           *nfr = ns_field->newfr;
        int             ix, iy, ic[MAXD], i, j, k, i1, icrds[MAXD];
        Gas_param       **prms_list;
        int             nprms;
        COMPONENT       *comps, comp, comp1;
        float           **den_vis, *a_frac;
        int             deb_flag = NO;
        RECT_GRID       *gr = wv->rect_grid;

        nprms = return_params_list(&prms_list);
        vector(&comps, nprms, sizeof(COMPONENT));
        matrix(&den_vis,nprms,2,sizeof(float));
 
        k = 0;
        for (i=0; i<nprms; i++)
        {
            if(prms_list[i]->eos->_NS_eos == YES)
            {
                comps[k] = prms_list[i]->comp;
                den_vis[k][0] = prms_list[i]->dens;
                den_vis[k][1] = prms_list[i]->shear_visc;
                k++;
            }
        }
        ns_field->n_eos = k;
        ns_field->comps = comps;
        ns_field->den_vis = den_vis;

        /*
        for(i = 0; i < k; i++)
        {
            printf("param[%d] dens %g, viscosity %g, comp %d\n", i, 
                prms_list[i]->dens, prms_list[i]->shear_visc, prms_list[i]->comp);
        }
        */

        for(i = 0; i < dim; i++) 
        {
            smin[i] = -gr->lbuf[i];
            smax[i] = gr->gmax[i] + gr->ubuf[i];
        }

        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                ic[0] = ix; ic[1] = iy;

                /***
                if(ix == 3 && (iy == 1 || iy == 8))
                    deb_flag = YES;
                else
                    deb_flag = NO;
                ***/
                
                Rect_den(ic,ns_field) = 0.0; 
                Rect_visc(ic,ns_field) = 0.0; 

                if(Rect_blk_edge(ic,wv).ctype == EXTERNAL)
                    continue;
                comp = Rect_comp(ic,wv);
                a_frac = Rect_blk_edge(ic,wv).a_frac;

                if(Rect_blk_edge(ic,wv).ctype == PARTIAL)
                {
                    for(i = 0; i < k; i++)
                    {
                        if(comp == comps[i])
                            break;
                    }
                    if(Rect_blk_edge(ic,wv).config == 0 ||
                       Rect_blk_edge(ic,wv).config == 1 ||
                       Rect_blk_edge(ic,wv).config == 4 ||
                       Rect_blk_edge(ic,wv).config == 5)
                    {
                        Rect_den(ic,ns_field) = den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[0]; 
                        Rect_visc(ic,ns_field) = den_vis[i][1]*Rect_blk_edge(ic,wv).a_frac[0]; 

                        /*
                        if(deb_flag == YES)
                        {
                            printf("case 0, ic[%d %d] a_frac = %9.8g\n", ix, iy, 
                                den_vis[i][0]*blk_edge[ix][iy].a_frac[0]);
                        }
                        */
                    }
                    else if(Rect_blk_edge(ic,wv).config == 3)
                    {
                        Rect_den(ic,ns_field) = den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[1]; 
                        Rect_visc(ic,ns_field) = den_vis[i][1]*Rect_blk_edge(ic,wv).a_frac[1]; 
                        if(deb_flag == YES)
                        {
                            printf("case 1, ic[%d %d] a_frac = %9.8g\n", ix, iy,
                                    den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[1]);
                        }

                    }
                    else if(Rect_blk_edge(ic,wv).config == 2 ||
                            Rect_blk_edge(ic,wv).config == 6)
                    {
                        if(Rect_blk_edge(ic,wv).corn_comp[0][0] == comp)
                        {
                            Rect_den(ic,ns_field) = den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[0]; 
                            Rect_visc(ic,ns_field) = den_vis[i][1]*Rect_blk_edge(ic,wv).a_frac[0]; 
                            if(deb_flag == YES)
                            {
                                printf("case 2, ic[%d %d] a_frac = %9.8g\n", ix, iy,
                                    den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[0]);
                            }
                        }  
                        else
                        {
                            Rect_den(ic,ns_field) = den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[1]; 
                            Rect_visc(ic,ns_field) = den_vis[i][1]*Rect_blk_edge(ic,wv).a_frac[1]; 
                            if(deb_flag == YES)
                            {
                                printf("case 3, ic[%d %d] a_frac = %9.8g\n", ix, iy,
                                    den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[1]);
                            }
                        }
                    }
                    else
                    {
                        printf("ERROR: update_den_viscosity handle case: PARTIAL config %d\n",
                             Rect_blk_edge(ic,wv).config);
                        printf("IN update_den_viscosity, [%d][%d].ctype = %d\n",
                           ix, iy, Rect_blk_edge(ic,wv).ctype);
                        for(i = 0; i < 2; i++)
                        {
                            for(j = 0; j < 2; j++)
                            {
                                printf("blk->edge[%d][%d]->type = %d ,",
                                 i, j, Rect_blk_edge(ic,wv).edge[i][j].etype);
                            }
                            printf("\n");
                        }
                        clean_up(ERROR);
                    }
                }
                else if(Rect_blk_edge(ic,wv).ctype == INTERNAL)
                {
                    for(i = 0; i < k; i++)
                    {
                        if(comp == comps[i])
                        {
                            Rect_den(ic,ns_field) = den_vis[i][0];
                            Rect_visc(ic,ns_field) = den_vis[i][1];
                            break;
                        } 
                    }
                }
                else if(Rect_blk_edge(ic,wv).ctype == INTER_PART)
                {
                    if(Rect_blk_edge(ic,wv).config == 1 ||
                       Rect_blk_edge(ic,wv).config == 3)    
                    {
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][0] == comps[i])
                                break;
                        }
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][0] == comps[i1])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                    }
                    else if(Rect_blk_edge(ic,wv).config == 4)
                    {
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][1] == comps[i])
                                break;
                        }
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][1] == comps[i1])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                    }
                    else if(Rect_blk_edge(ic,wv).config == 5)
                    {
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][1] == comps[i1])
                                break;
                        }
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][1] == comps[i])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                    }
                    else if(Rect_blk_edge(ic,wv).config == 2 ||
                            Rect_blk_edge(ic,wv).config == 6)
                    {
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][0] == comps[i])
                                break;
                        }
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][1] == comps[i1])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                    }
                    else
                    {
                        printf("ERROR: update_den_viscosity handle case INTER_PART: config %d\n",
                                 Rect_blk_edge(ic,wv).config);
                        printf("IN update_den_viscosity, [%d][%d].ctype = %d\n",
                           ix, iy, Rect_blk_edge(ic,wv).ctype);
                        for(i = 0; i < 2; i++)
                        {
                            for(j = 0; j < 2; j++)
                            {
                                printf("blk->edge[%d][%d]->type = %d ,",
                                 i, j, Rect_blk_edge(ic,wv).edge[i][j].etype);
                            }
                            printf("\n");
                        }
                        clean_up(ERROR);
                    }
                }
                else
                {
                    printf("ERROR: grid_den_viscosity handle case:\n");
                    printf("IN grid_den_viscosity, [%d][%d].ctype = %d\n",
                       ix, iy, Rect_blk_edge(ic,wv).ctype);
                    for(i = 0; i < 2; i++)
                    {
                        for(j = 0; j < 2; j++)
                        {
                            printf("blk->edge[%d][%d]->type = %d ,",
                             i, j, Rect_blk_edge(ic,wv).edge[i][j].etype);
                        }
                        printf("\n");
                    }
                    clean_up(ERROR);
                }
                // DEBUG TMP
                /**
                {
                    printf("IN grid_den_viscosity, [%d][%d] comp %d, den = %g, mu = %g\n",
                       ix, iy, comp, Rect_den(ic,ns_field), Rect_visc(ic,ns_field));
                }
                **/
            }
        }

        // free(comps);
        // free(den_vis); 

        if(debugging("NS_vel"))
        {
            printf("-----Cell Density-------------\n");
            for (iy = smin[1];  iy < smax[1];  ++iy)
            {
                printf("Row[%d] dens: ",iy);
                for (ix = smin[0];  ix < min(4,smax[0]);  ++ix)
                {
                    ic[0] = ix; ic[1] = iy;
                    printf("(%d, %9.8g) ", ix, Rect_den(ic,ns_field));
                }
                printf("\n");
            }
        }
}

/*
* Crank-Nicolson to compute u^{*}
*/
LOCAL void diffusion_convection_step(
        float             dt,
        NS_Field_Stencil  *ns_field)
{
        P_CELL2D        **p_cellx, **p_celly; // used for laplacian
        int             Nx = ns_field->rect_gr->gmax[0];
        int             Ny = ns_field->rect_gr->gmax[1];
        Wave            *newwv = ns_field->newwave;
        Wave            *wv = ns_field->wave;
        Front           *newfr = ns_field->newfr;
        Front           *fr = ns_field->fr;
        TRI_GRID        *ntg = wave_tri_soln(newwv)->tri_grid;
        TRI_GRID        *tg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        int             ix, iy, ic[MAXD], i, j;
        float           *dh = wv->rect_grid->h;

        int             *n_dist = NULL, myid, ICOORDS[2];
        int             NLblocks,ilower, iupper, I;
        int             **I_to_ijk = ns_field->I_to_ijk;   /* local  i, j, k, local I */
        int             **IJK_to_I = ns_field->IJK_to_I;  /* global I, J, K, global I */
        int             idxm[1],  sten_index[3], indices;
        float           norm,min,max,x,y,*coords;
        float           val[1], values[3][3];

#if defined (PETSC)
            /*PETSC Variables*/
        KSP                  ksp;
        Mat                  A;
        Vec                  phi, B, C;
        PetscScalar          *vphi;
        MatNullSpace         nullsp;
        PC                   pc;
        int                  its;
        PetscReal            norm_B, norm_C;
        KSPConvergedReason   reason;
#endif /*defined (PETSC)*/

        // printf("Crank-Nicolson step dt = %g\n", dt);
        // dt = 0.0001;

#if defined (PETSC)
#if defined(__MPI__)
        MPI_Comm_size(PETSC_COMM_WORLD,&comm_size);
#endif /* defined(__MPI__) */

        myid = pp_mynode();
        find_Cartesian_coordinates(myid,newfr->pp_grid,ICOORDS);
        vector(&n_dist,comm_size,sizeof(int));

        // NEW
        if(ns_field->alloc.I_to_ijk != YES)
        {
            matrix(&I_to_ijk,Nx*Ny,2,sizeof( int));
            matrix(&IJK_to_I,(Nx+4),(Ny+4),sizeof( int));

            ns_field->I_to_ijk = I_to_ijk;
            ns_field->alloc.I_to_ijk = YES;
            ns_field->IJK_to_I = IJK_to_I;
            ns_field->alloc.IJK_to_I = YES;
        }
        else
        {
            I_to_ijk = ns_field->I_to_ijk;
            IJK_to_I = ns_field->IJK_to_I;
        }
        // END NEW

        NLblocks = 0;
        for(j = 0; j < Ny; j++)
        {
            ic[1] = j;
            for(i = 0; i < Nx; i++)
            {
                ic[0] = i;
                if(Rect_comp(ic,newwv) != ext_comp &&
                   !is_excluded_comp(Rect_comp(ic,newwv),newfr->interf) &&
                   Rect_comp(ic,newwv) != POTS_COMP)
                {
                    I_to_ijk[NLblocks][0] = i;
                    I_to_ijk[NLblocks][1] = j;
                    NLblocks++;
                }
            }
        }

#if defined(__MPI__)
            pp_all_gather((POINTER)&NLblocks,sizeof(int),(POINTER)n_dist,
                  sizeof(int)*comm_size);
#else /* defined(__MPI__) */
            assign((POINTER)n_dist,(POINTER)&NLblocks,sizeof(int));
#endif /* defined(__MPI__) */

        ilower = 0;
        iupper = n_dist[0]-1;

        for (i = 1; i <= myid; i++)
        {
            ilower += n_dist[i-1];
            iupper += n_dist[i];
        }

        create_global_element_list(newfr,ilower,iupper,ICOORDS,I_to_ijk,IJK_to_I);

#endif /* if defined (PETSC) */

        matrix(&p_cellx,Nx,Ny,sizeof(P_CELL2D));
        matrix(&p_celly,Nx,Ny,sizeof(P_CELL2D));

        for(iy = 0; iy < Ny; iy++)
        {
            for(ix = 0; ix < Nx; ix++)
            {
                ic[0] = ix; ic[1] = iy; 
                for(i = 0; i < 3; ++i)
                    for(j = 0; j < 3; ++j)
                {
                    {
                       p_celly[ix][iy].set[i][j] = p_cellx[ix][iy].set[i][j] = NO;
                       p_celly[ix][iy].stencil[i][j] = p_cellx[ix][iy].stencil[i][j] = 0.0;
                    }
                }
                p_celly[ix][iy].r_side = p_cellx[ix][iy].r_side = 0.0;

                if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                   continue;
                // finite difference stencil

                //         printf("IN diffusion_convection_step::::::::::::\n");
                //         printf("set ic[%d %d] p_cellx = %10.9g, p_celly = %10.9g\n",
                //           ix,iy,  p_cellx[ix][iy].r_side, p_celly[ix][iy].r_side);

                // printf("IN diffusion_convection_step, set ic[%d %d], comp %d %d\n", ic[0], ic[1],
                //                     Rect_comp(ic,wv), Rect_comp(ic,newwv));

                set_Crank_Nicolson_sten(dt, ic, ns_field,&p_cellx[ix][iy], &p_celly[ix][iy]);
                if(debugging("NS_vel") && NO)
                {
                    // if((ic[1] == 11 && (ic[0] == 9 || ic[0] == 10)) || 
                    //    (ic[1] == 12 && (ic[0] == 9 || ic[0] == 10))
                    //   )
                    if(ic[1] == 11)
                    {
                        printf("IN diffusion_convection_step::::::::::::\n");
                        printf("ic[%d %d] p_celly = %11.10g\n",
                          ix,iy,  p_celly[ix][iy].r_side);
                        printf("line-0 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ix][iy].set[0][0], 
                             p_celly[ix][iy].set[1][0],
                             p_celly[ix][iy].set[2][0],
                             p_celly[ix][iy].stencil[0][0],
                             p_celly[ix][iy].stencil[1][0], 
                             p_celly[ix][iy].stencil[2][0]);
                        printf("line-1 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ix][iy].set[0][1], 
                             p_celly[ix][iy].set[1][1],
                             p_celly[ix][iy].set[2][1],
                             p_celly[ix][iy].stencil[0][1],
                             p_celly[ix][iy].stencil[1][1], 
                             p_celly[ix][iy].stencil[2][1]);
                        printf("line-2 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ix][iy].set[0][2], 
                             p_celly[ix][iy].set[1][2],
                             p_celly[ix][iy].set[2][2],
                             p_celly[ix][iy].stencil[0][2], 
                             p_celly[ix][iy].stencil[1][2],
                             p_celly[ix][iy].stencil[2][2]);

                        printf("ic[%d %d] p_cellx = %11.10g\n",
                          ix,iy,  p_cellx[ix][iy].r_side);
                        printf("line-0 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_cellx[ix][iy].set[0][0],
                             p_cellx[ix][iy].set[1][0],
                             p_cellx[ix][iy].set[2][0],
                             p_cellx[ix][iy].stencil[0][0],
                             p_cellx[ix][iy].stencil[1][0],
                             p_cellx[ix][iy].stencil[2][0]);
                        printf("line-1 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_cellx[ix][iy].set[0][1],
                             p_cellx[ix][iy].set[1][1],
                             p_cellx[ix][iy].set[2][1],
                             p_cellx[ix][iy].stencil[0][1],
                             p_cellx[ix][iy].stencil[1][1],
                             p_cellx[ix][iy].stencil[2][1]);
                        printf("line-2 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_cellx[ix][iy].set[0][2],
                             p_cellx[ix][iy].set[1][2],
                             p_cellx[ix][iy].set[2][2],
                             p_cellx[ix][iy].stencil[0][2],
                             p_cellx[ix][iy].stencil[1][2],
                             p_cellx[ix][iy].stencil[2][2]);

                    }
                }
                // TMP, have not considered viscosity yet
                // Rect_u_star(ic,ns_field).v[0] = p_cellx[ix][iy].r_side;
                // Rect_u_star(ic,ns_field).v[1] = p_celly[ix][iy].r_side;
                // TMP, have not considered viscosity yet
            }
        }

        /**
        {
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
            printf("\n\n---------Advection-Diffusion step Y vel -----\n\n");
            for (iy = 0; iy < Ny; iy++)
            {
                printf("Row[%d] ", iy);
                // for (ix = 0; ix < 3; ix++)
                for (ix = 0; ix <= Nx/2; ix++)
                {
                    int ic2[MAXD];
                    ic[0] = ix; ic[1] = iy;
                    ic2[0] = Nx-ix-1; ic2[1] = iy;
                    // printf("(%d %g %g)", ix, Rect_u_star(ic,ns_field).v[0] +
                    //            Rect_u_star(ic2,ns_field).v[0],
                    //           Rect_u_star(ic,ns_field).v[1]-
                    //           Rect_u_star(ic2,ns_field).v[1]);
                    printf("(%d %g)", ix,
                              Rect_u_star(ic,ns_field).v[1]);
                }
                printf("\n");
            }
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
        }

        free(p_cellx); free(p_celly);
        return;
        **/

        // TMP
        /***
        for(iy = 0; iy < Ny; iy++)
        {
            int ic2[MAXD];
            for(ix = 0; ix < Nx/2; ix++)
            {
                ic[0] = ix; ic[1] = iy;
                ic2[0] = Nx-ix-1; ic2[1] = iy;

                        printf("IN diffusion_convection_step::::::::::::\n");
                        printf("ic[%d %d] p_celly = %11.10g\n",
                          ix,iy,  p_celly[ix][iy].r_side);
                        printf("line-0 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ix][iy].set[0][0],
                             p_celly[ix][iy].set[1][0],
                             p_celly[ix][iy].set[2][0],
                             p_celly[ix][iy].stencil[0][0],
                             p_celly[ix][iy].stencil[1][0],
                             p_celly[ix][iy].stencil[2][0]);
                        printf("line-1 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ix][iy].set[0][1],
                             p_celly[ix][iy].set[1][1],
                             p_celly[ix][iy].set[2][1],
                             p_celly[ix][iy].stencil[0][1],
                             p_celly[ix][iy].stencil[1][1],
                             p_celly[ix][iy].stencil[2][1]);
                        printf("line-2 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ix][iy].set[0][2],
                             p_celly[ix][iy].set[1][2],
                             p_celly[ix][iy].set[2][2],
                             p_celly[ix][iy].stencil[0][2],
                             p_celly[ix][iy].stencil[1][2],
                             p_celly[ix][iy].stencil[2][2]);

                        printf("ic[%d %d] p_celly = %11.10g\n",
                          ic2[0],ic2[1],  p_celly[ic2[0]][ic2[1]].r_side);
                        printf("line-0 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ic2[0]][ic2[1]].set[0][0],
                             p_celly[ic2[0]][ic2[1]].set[1][0],
                             p_celly[ic2[0]][ic2[1]].set[2][0],
                             p_celly[ic2[0]][ic2[1]].stencil[0][0],
                             p_celly[ic2[0]][ic2[1]].stencil[1][0],
                             p_celly[ic2[0]][ic2[1]].stencil[2][0]);
                        printf("line-1 [%d %d %d], [%10.9g %10.9g %10.9g]\n",
                             p_celly[ic2[0]][ic2[1]].set[0][1],
                             p_celly[ic2[0]][ic2[1]].set[1][1],
                             p_celly[ic2[0]][ic2[1]].set[2][1],
                             p_celly[ic2[0]][ic2[1]].stencil[0][1],
                             p_celly[ic2[0]][ic2[1]].stencil[1][1],
                             p_celly[ic2[0]][ic2[1]].stencil[2][1]);
                        printf("line-2 [%d %d %d], [%10.9g %10.9g %10.9g]\n\n",
                             p_celly[ix][iy].set[0][2],
                             p_celly[ix][iy].set[1][2],
                             p_celly[ix][iy].set[2][2],
                             p_celly[ix][iy].stencil[0][2],
                             p_celly[ix][iy].stencil[1][2],
                             p_celly[ix][iy].stencil[2][2]);
            }
        }
        ***/
        /***
        // TMP, have not considered viscosity yet
        // if(debugging("NS_vel"))
        {
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
            printf("\n\n---------Advection-Diffusion step vel DIFF-----\n\n");
            // for (iy = 0; iy < Ny; iy++)
            for (iy = 0; iy < Ny; iy++)
            {
                printf("Row[%d]: ", iy);
                for (ix = 0; ix < Nx/2; ix++)
                {
                    int ic2[MAXD];
                    ic[0] = ix; ic[1] = iy;
                    ic2[0] = Nx-ix-1; ic2[1] = iy;
                    printf("(%d %g %g)", ix, Rect_u_star(ic,ns_field).v[0] +
                               Rect_u_star(ic2,ns_field).v[0],
                              Rect_u_star(ic,ns_field).v[1]-
                              Rect_u_star(ic2,ns_field).v[1]);
                }
                printf("\n");
            }
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
        }
        free(p_cellx); free(p_celly);
        return;
        // TMP, have not considered viscosity yet

        if(debugging("NS_vel"))
        {
            printf("\n\nRight hand of crank_Nicolson-Y::::\n\n");
            // for(iy = 0; iy < Ny; iy++)
            for(iy = 3; iy < 4; iy++)
            {
                printf("Row[%d]: ", iy);
                for(ix = 0; ix < Nx; ix++)
                {
                    ic[0] = ix; ic[1] = iy; 
                    printf("(%d %12.11g) ", ix, p_celly[ix][iy].r_side);
                }
                printf("\n");
            }
        }
        ***/

#if defined(__MPI__)
#else
        MatCreate(PETSC_COMM_WORLD, &A);
        MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,iupper-ilower+1,iupper-ilower+1);
        MatSetFromOptions(A);
#endif // if defined(__MPI__) //

        VecCreate(PETSC_COMM_WORLD,&B);
        VecSetSizes(B,(iupper-ilower + 1),PETSC_DECIDE);
        VecSetFromOptions(B);
        VecDuplicate(B,&phi);
        VecDuplicate(B,&C);

        for (I = ilower; I <= iupper; I++)
        {
            // set_matrix_row(A,I,ilower,I_to_ijk,IJK_to_I,ns_field);
            // set_right_hand_side(B,I,ilower,I_to_ijk,ns_field);
            // set_matrix_A_and_right_hand_side(A,B,blk_edge,
            //       p_cell,I,ilower,I_to_ijk,IJK_to_I,ns_field);
            for(i = 0; i < dim; i++)
                ic[i] = I_to_ijk[I-ilower][i];
            idxm[0] = I;
            ix = ic[0]; iy = ic[1];
            if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                continue;

            // TMP
            /*
            printf("Work on cell %d %d, set[%d %d %d]\n", ic[0], ic[1],
                 p_cellx[ic[0]][ic[1]].set[1][0],
                 p_cellx[ic[0]][ic[1]].set[1][1],
                 p_cellx[ic[0]][ic[1]].set[1][2]);
            */

            // values[2] = values[0] = -0.5*dt/sqr(dh[0])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field);
            // values[1] = dt/sqr(dh[0])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field)+1.0;
            for(i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                    values[i][j] = p_cellx[ic[0]][ic[1]].stencil[i][j];
            }
            for( i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                { 
                    sten_index[i] = IJK_to_I[ix+i+1][iy+j+1];
                    // NEW for PERIOD_BC in the x direction
                    if(PERIOD_BC == YES && ix == 0 && i == 0 && p_cellx[ix][iy].set[i][j] == YES)
                    {
                        sten_index[i] = IJK_to_I[ix+i+gmax[0]+1][iy+j+1];
                    }
                    else if(PERIOD_BC == YES && ix == gmax[0]-1 && i == 2 && p_cellx[ix][iy].set[i][j] == YES)
                    {
                        sten_index[i] = IJK_to_I[ix+i-gmax[0]+1][iy+j+1];
                    }

                    if(p_cellx[ic[0]][ic[1]].set[i][j] == YES)
                    {
                        MatSetValues(A,1,idxm,1,&sten_index[i],
                            &values[i][j],INSERT_VALUES);
                    // TMP
                    // printf("cell ic[%d %d] set %d\n", ic[0], ic[1], sten_index[i]);
                    }
                }
            }
            val[0] = p_cellx[ic[0]][ic[1]].r_side;
            VecSetValues(B,1,idxm,val,INSERT_VALUES);
        }

        MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
        MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);

        VecAssemblyBegin(B);
        VecAssemblyEnd(B);

        // MatNullSpaceCreate(PETSC_COMM_WORLD,PETSC_TRUE,0,PETSC_NULL,&nullsp);
        KSPCreate(PETSC_COMM_WORLD,&ksp);
        KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN);
        KSPGetPC(ksp,&pc);
#if defined(__MPI__)
        PCSetType(pc,PCJACOBI);
#endif // if defined(__MPI__)

        // KSPSetNullSpace(ksp,nullsp);
        // KSPSetType(ksp,KSPBCGS);
        // KSPSetTolerances(ksp,1.e-12,PETSC_DEFAULT,PETSC_DEFAULT,700);
        KSPSetTolerances(ksp,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);
        KSPSetComputeSingularValues(ksp, PETSC_TRUE);

        KSPSetFromOptions(ksp);
        // KSPSetUp(ksp);

        KSPSolve(ksp,B,phi);
        KSPGetIterationNumber(ksp,&its);
        KSPGetResidualNorm(ksp,&norm);

        KSPComputeExtremeSingularValues(ksp, &max, &min);
        /**
        printf("\n X-component of velocity:\n"
            " Number of iteration  is %d,"
            "residual norm = %6.5e,extreme test SV is [%7.6e ; %7.6e]\n",its,norm,min,max);
        **/ 
        KSPGetConvergedReason(ksp,&reason);
        if(reason < 0)
        {
            printf("ERROR: diffusion_convection_step X-component of velocity:\n");
            printf("Ksp diverges\n");
            clean_up(ERROR);
        }        

        // MatView(A,PETSC_VIEWER_STDOUT_WORLD);
        // VecView(B,PETSC_VIEWER_STDOUT_WORLD);

        for (iy = 0; iy < Ny; iy++)
        {
            for (ix = 0; ix < Nx; ix++)
            {
                indices = IJK_to_I[ix+2][iy+2];
                ic[0] = ix; ic[1] = iy;

                if (indices >= 0)
                {
                   VecGetArray(phi,&vphi);
                   Rect_u_star(ic,ns_field).v[0] = vphi[indices-ilower];
                   // Rect_phi(ic,ns_field) = vphi[indices-ilower];

                   // printf("ic[%d %d] x-comp vel = %g\n",
                   //    ix, iy, vphi[indices-ilower]);
                   VecRestoreArray(phi,&vphi);
                }
            }
        }

        if(debugging("NS_vel"))
        {
            MatMult(A,phi,C);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            // printf("-------A*phi, print RHS----\n");
            // VecView(B,PETSC_VIEWER_STDOUT_WORLD);
            printf("After CN-X_component, norm B = %g, norm (A*phi) = %g\n",
                         norm_B, norm_C);
        }
        // END TEST

        VecDestroy(B);
        VecDestroy(phi);
        VecDestroy(C);
        MatDestroy(A);
        KSPDestroy(ksp);
        // MatNullSpaceDestroy(nullsp);

        /* Compute y-component of the velocity field */
#if defined(__MPI__)
#else
        MatCreate(PETSC_COMM_WORLD, &A);
        MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,iupper-ilower+1,iupper-ilower+1);
        MatSetFromOptions(A);
#endif // if defined(__MPI__) //

        VecCreate(PETSC_COMM_WORLD,&B);
        VecSetSizes(B,(iupper-ilower + 1),PETSC_DECIDE);
        VecSetFromOptions(B);
        VecDuplicate(B,&phi);
        VecDuplicate(B,&C);

        for (I = ilower; I <= iupper; I++)
        {
            for(i = 0; i < dim; i++)
                ic[i] = I_to_ijk[I-ilower][i];
            idxm[0] = I;
            ix = ic[0]; iy = ic[1];
            if(Rect_blk_edge(ic,newwv).ctype == EXTERNAL)
                continue;

            // values[2] = values[0] = -0.5*dt/sqr(dh[1])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field);
            // values[1] = dt/sqr(dh[1])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field)+1.0;
            for(i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                    values[i][j] = p_celly[ic[0]][ic[1]].stencil[i][j];
            }
            for( i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                {
                    sten_index[i] = IJK_to_I[ix+i+1][iy+j+1];
                    // NEW for PERIOD_BC in the x direction
                    if(PERIOD_BC == YES && ix == 0 && i == 0 && p_celly[ix][iy].set[i][j] == YES)
                    {
                        sten_index[i] = IJK_to_I[ix+i+gmax[0]+1][iy+j+1];
                    }
                    else if(PERIOD_BC == YES && ix == gmax[0]-1 && i == 2 && p_celly[ix][iy].set[i][j] == YES)
                    {
                        sten_index[i] = IJK_to_I[ix+i-gmax[0]+1][iy+j+1];
                    }

                    if(p_celly[ic[0]][ic[1]].set[i][j] == YES)
                    {
                        MatSetValues(A,1,idxm,1,&sten_index[i],
                            &values[i][j],INSERT_VALUES);
                    }
                }
            }
            val[0] = p_celly[ic[0]][ic[1]].r_side;
            VecSetValues(B,1,idxm,val,INSERT_VALUES);
        }

        MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
        MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);

        VecAssemblyBegin(B);
        VecAssemblyEnd(B);

        // MatNullSpaceCreate(PETSC_COMM_WORLD,PETSC_TRUE,0,PETSC_NULL,&nullsp);
        KSPCreate(PETSC_COMM_WORLD,&ksp);
        KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN);
        KSPGetPC(ksp,&pc);
#if defined(__MPI__)
        PCSetType(pc,PCJACOBI);
#endif // #if defined(__MPI__)

        // KSPSetNullSpace(ksp,nullsp);
        // KSPSetType(ksp,KSPBCGS);
        // KSPSetTolerances(ksp,1.e-12,PETSC_DEFAULT,PETSC_DEFAULT,700);
        KSPSetTolerances(ksp,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);
        KSPSetComputeSingularValues(ksp, PETSC_TRUE);

        KSPSetFromOptions(ksp);
        KSPSetUp(ksp);

        KSPSolve(ksp,B,phi);
        KSPGetIterationNumber(ksp,&its);
        KSPGetResidualNorm(ksp,&norm);

        KSPComputeExtremeSingularValues(ksp, &max, &min);
        /**
        printf("\n Y-component of velocity:\n"
            " Number of iteration  is %d,"
            "residual norm = %6.5e,extreme test SV is [%7.6e ; %7.6e]\n",its,norm,min,max);
        **/
        KSPGetConvergedReason(ksp,&reason);
        if(reason < 0)
        {
            printf("ERROR: diffusion_convection_step Y-component of velocity:\n");
            printf("Ksp diverges\n");
            clean_up(ERROR);
        }

        for (iy = 0; iy < Ny; iy++)
        {
            for (ix = 0; ix < Nx; ix++)
            {
                indices = IJK_to_I[ix+2][iy+2];
                ic[0] = ix; ic[1] = iy;

                if (indices >= 0)
                {
                   VecGetArray(phi,&vphi);
                   Rect_u_star(ic,ns_field).v[1] = vphi[indices-ilower];
                   // Rect_phi(ic,ns_field) = vphi[indices-ilower];

                   // printf("ic[%d %d] phi = %g, b = %g\n",
                   //   ix, iy, potential[ix+2][iy+2], vphi[indices-ilower]);
                   VecRestoreArray(phi,&vphi);
                }
            }
        }

        // TEST
        if(debugging("NS_vel"))
        {
            MatMult(A,phi,C);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            // printf("-------A*phi, print RHS----\n");
            // VecView(B,PETSC_VIEWER_STDOUT_WORLD);
            printf("After CN-Y_component, norm B = %g, norm (A*phi) = %g\n",
                         norm_B, norm_C);
        }
        // END TEST

        VecDestroy(B);
        VecDestroy(phi);
        VecDestroy(C);
        MatDestroy(A);
        KSPDestroy(ksp);
        // MatNullSpaceDestroy(nullsp);

        if(debugging("NS_vel"))
        {
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
            printf("\n\n---------Advection-Diffusion step vel-----\n\n");
            for (iy = 0; iy < Ny; iy++)
            {
                printf("Row[%d] ", iy);
                // for (ix = 0; ix < 3; ix++)
                for (ix = 0; ix <= Nx/2; ix++)
                {
                    int ic2[MAXD];
                    ic[0] = ix; ic[1] = iy;
                    ic2[0] = Nx-ix-1; ic2[1] = iy;
                    // printf("(%d %g %g)", ix, Rect_u_star(ic,ns_field).v[0] +
                    //            Rect_u_star(ic2,ns_field).v[0],
                    //           Rect_u_star(ic,ns_field).v[1]-
                    //           Rect_u_star(ic2,ns_field).v[1]);
                    printf("(%d %g %g)", ix, Rect_u_star(ic,ns_field).v[0],
                              Rect_u_star(ic,ns_field).v[1]);
                }
                printf("\n");
            }
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
        }

        if(debugging("NS_vel"))
        {
            printf("\n\nRight hand DIFF of crank_Nicolson::::\n\n");
            for(iy = 0; iy < Ny; iy++)
            {
                printf("Row[%d]: ", iy);
                for(ix = 0; ix < Nx/2; ix++)
                {
                    int ic2[MAXD];
                    ic[0] = ix; ic[1] = iy;
                    ic2[0] = Nx-ix-1; ic2[1] = iy;
                    printf("(%d %12.11g %12.11g) ", ix, p_celly[ix][iy].r_side-
                                            p_celly[ic2[0]][ic2[1]].r_side,
                                  p_cellx[ix][iy].r_side+
                                            p_cellx[ic2[0]][ic2[1]].r_side);
                }
                printf("\n");
            }
        }

        // TMP
        /**
        {
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
            printf("\n\n---------Advection-Diffusion step vel-----\n\n");
            int ic2[MAXD];
            ic2[0] = 9; ic2[1] = 12;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1], 
                   Rect_u_star(ic2,ns_field).v[0], Rect_u_star(ic2,ns_field).v[1]);
            ic2[0] = 10; ic2[1] = 12;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1], 
                   Rect_u_star(ic2,ns_field).v[0], Rect_u_star(ic2,ns_field).v[1]);
            ic2[0] = 9; ic2[1] = 11;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1], 
                   Rect_u_star(ic2,ns_field).v[0], Rect_u_star(ic2,ns_field).v[1]);
            ic2[0] = 10; ic2[1] = 11;
            printf("ic[%d %d] Vel: (%11.10g %11.10g)\n", ic2[0], ic2[1], 
                   Rect_u_star(ic2,ns_field).v[0], Rect_u_star(ic2,ns_field).v[1]);
            printf("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
        }
        **/

        free(p_cellx); free(p_celly);
        free(n_dist);
}

LOCAL void set_Crank_Nicolson_sten(
        float     dt,
        int       *ic,
        NS_Field_Stencil *ns_field,
        P_CELL2D  *p_cellx,
        P_CELL2D  *p_celly)
{
        Wave      *newwv = ns_field->newwave;
        Wave      *wv = ns_field->wave;
        TRI_GRID  *tg = wave_tri_soln(wv)->tri_grid;
        TRI_GRID  *ntg = wave_tri_soln(newwv)->tri_grid;
        Front     *fr = ns_field->fr, *newfr = ns_field->newfr;
        COMPONENT comp, ncomp, nncomp;
        int       dir, ndir, j, k, l, end = 1; // 2nd order central diff
        static    Stencil *sten = NULL, *nsten;
        int       **icoords, **nicoords; 
        int       iperm[2] = {0, 1}; // No need to permute
        int       label_storex[3], *labelx, label_storey[3], *labely, irregx, irregy;
        float     *dh = wv->rect_grid->h;
        BLK_EDGE  *blk_edge = ntg->blk_edge;
        float     u_xx, u_yy, v_xx, v_yy; // they are actually 
                              // u_xx + u_yy, and v_xx + v_yy respectively. 
        int       debug_flag = NO;
        P_CELL2D  *p_cell[2];
        float     values[3][3];
        float     src[MAXD] = {0.0, 0.0}, *crds;

        labelx = label_storex + 1;
        labely = label_storey + 1;
        comp = Rect_comp(ic,wv);

        p_cell[0] = p_cellx;
        p_cell[1] = p_celly;
        dir = 0; ndir = 1;

        // TMP
        /**
        if(((ic[0] == 0 && ic[1] == 0) ||
            (ic[0] == 1 && ic[1] == 0)) 
           )
        {
            printf("\n\n|||||||------set_Crank_Nicolson_sten on ic[%d %d]\n", ic[0], ic[1]);
            // printf("comp = %d\n", comp);
            // printf("is exterior_comp %d, excluded comp %d\n", is_exterior_comp(comp, fr->interf),
            //                  is_excluded_comp(comp,fr->interf));
            debug_flag = YES;
            deb_npt = YES;
        }
        **/

        if (sten == NULL)
        {
            sten = alloc_stencil(3,fr); 
            nsten = alloc_stencil(3,fr); 
        }

        icoords = sten->icoords; 
        sten->st[0] = Rect_state(ic,wv);

        // TMP, should look from the intfc geomtry
        if(Rect_comp(ic,wv) != Rect_comp(ic,newwv))
        {
            if(is_excluded_comp(Rect_comp(ic,wv), fr->interf))
            {
                sten->st[0] = sten->worksp[0]; 
                hyp_solution(Rect_coords(ic,wv), Rect_comp(ic,newwv), NULL, UNKNOWN_SIDE,
                               fr,wv,sten->st[0], NULL);
                if(debug_flag == YES)
                {
                    printf("hyp_solution, new comp %d\n", Rect_comp(ic,newwv));
                    g_print_state(sten->st[0]);
                }
            }
        }
        // END TMP

        // Compute source
        if (source_terms_exist() == YES)
        {
            if(is_gravity() == YES)
            {
                float time;
                time = fr->time + 0.5*dt;
                crds = Rect_coords(ic,wv); 
                src[0] += gravity(crds,time)[0];
                src[1] += gravity(crds,time)[1];
            }
        }
         

        sten->fr  = fr;
        sten->wave = wv;
        sten->newfr = newfr; sten->newwave = newwv;

        nsten->fr  = fr;
        nsten->wave = wv;
        nsten->newfr = newfr; nsten->newwave = newwv;
        nicoords = nsten->icoords; 
        nsten->st[0] = Rect_state(ic,wv);

        // TMP, should look from the intfc geomtry
        /***
        if(Rect_comp(ic,wv) != Rect_comp(ic,newwv))
        {
            if(is_excluded_comp(Rect_comp(ic,wv), fr->interf))
            {
                nsten->st[0] = nsten->worksp[0];
                hyp_solution(Rect_coords(ic,wv), Rect_comp(ic,newwv), NULL, UNKNOWN_SIDE,
                               fr,wv,nsten->st[0], NULL);
            }
        }
        ***/
        // END TMP

        for(j = 0; j < dim; j++)
            nicoords[0][j] = icoords[0][j] = ic[j];

        for (j = -end; j <= end; ++j)
        {
            icoords[j][dir] = ic[dir] + j;       
            icoords[j][(dir+1)%dim] = ic[(dir+1)%dim];
        }
        for (j = -end; j <= end; ++j)
        {
            nicoords[j][dir] = ic[dir];       
            nicoords[j][(dir+1)%dim] = ic[(dir+1)%dim] + j;
        }
            
        for (j = 0; j < dim; ++j)
        {
            if (j == dir)
            {
                Coords(nsten->p[0])[dir] = Coords(sten->p[0])[dir] = 
                         cell_center(ic[dir],dir,wv->rect_grid);
                for (l = 1; l <= end; ++l)
                {
                    Coords(sten->p[-l])[dir] = 
                             Coords(sten->p[0])[dir]-l*dh[j];
                    Coords(sten->p[l])[dir]  = 
                             Coords(sten->p[0])[dir]+l*dh[j];
                }
                for (l = 1; l <= end; ++l)
                {
                    Coords(nsten->p[-l])[dir] =
                             Coords(sten->p[0])[dir];
                    Coords(nsten->p[l])[dir]  =
                             Coords(sten->p[0])[dir];
                }
            }
            else
            {
                for (l = -end; l <= end; ++l)
                    Coords(sten->p[l])[j] = Rect_coords(ic,wv)[j];
                for (l = -end; l <= end; ++l)
                    Coords(nsten->p[l])[j] = Rect_coords(ic,wv)[j] + l*dh[j];
            }
        }

        prev_side = WEST;
        next_side = EAST;
        ncomp = set_NS_stencil_comps_and_crosses(dir,dir,iperm,sten,end);

        prev_side = SOUTH;
        next_side = NORTH;
        ncomp = set_NS_stencil_comps_and_crosses(ndir,ndir,iperm,nsten,end);

        /***
        if(debug_flag == YES)
        {
            for (l = -end; l <= end; ++l)
            {
                printf("dir[%d] ic[%d %d], comp = %d coords(%g %g)\n", 
                     dir, icoords[l][0], icoords[l][1], sten->comp[l],
                      Coords(sten->p[l])[0],  Coords(sten->p[l])[1]); 
            }

            for (l = -end; l <= end; ++l)
            {
                printf("nextdir[%d] ic[%d %d], comp = %d coords(%g %g)\n", 
                     ndir, nicoords[l][0], nicoords[l][1], nsten->comp[l],
                      Coords(nsten->p[l])[0],  Coords(nsten->p[l])[1]); 
            }
            printf("\n");
        }
        ***/

        for (l = -end; l <= end; ++l)
            labelx[l] = YES;
        irregx = Crank_N_sten_reg(dir,sten,labelx,ns_field);

        if(debug_flag == YES)
            printf("after Crank_N_sten_reg, labelx\n");

        for (l = -end; l <= end; ++l)
        {
            if(labelx[l] == YES)
            {
                p_cell[dir]->set[l+end][1] = YES;
                p_cell[ndir]->set[l+end][1] = YES;
            }
        }

        for (l = -end; l <= end; ++l)
            labely[l] = YES;
        irregy = Crank_N_sten_reg(ndir,nsten,labely,ns_field);

        if(debug_flag == YES)
            printf("after Crank_N_sten_reg, labely\n");

        for (l = -end; l <= end; ++l)
        {
            if(labely[l] == YES)
            { 
                p_cell[dir]->set[1][l+end] = YES;
                p_cell[ndir]->set[1][l+end] = YES;
            }
        }

        // TMP, should be fixed with geometry
        p_cell[dir]->set[1][1] = p_cell[ndir]->set[1][1] = YES; 

        // TMP
        /***
        if(debug_flag == YES)
        {
            printf("set crank ic[%d %d] dir %d, set[%d %d %d]\n",
                 ic[0], ic[1], dir, labelx[-1], labelx[0], labelx[1]);
            printf("set crank ic[%d %d] nextdir %d, set[%d %d %d]\n",
                 ic[0], ic[1], ndir, labely[-1], labely[0], labely[1]);
            printf("set crank ic[%d %d] dir %d, set[%d %d %d]\n",
                ic[0], ic[1], ndir, p_cell[ndir]->set[1][0], 
                 p_cell[ndir]->set[1][1],
                 p_cell[ndir]->set[1][2]);
        }
        ***/

        values[0][1] = values[2][1] = -0.5*dt/sqr(dh[dir])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field);
        values[1][2] = values[1][0] = -0.5*dt/sqr(dh[ndir])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field);
        values[1][1] = dt/sqr(dh[dir])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field)+ 
                       dt/sqr(dh[ndir])/Rect_den(ic,ns_field)*Rect_visc(ic,ns_field) + 1.0;

        if(debug_flag == YES) 
        {
            printf("scr[%d %d] = %g %g, irregx = %d, irregy = %d\n", 
                       ic[0], ic[1],  src[0], src[1], irregx, irregy);
            printf("after comput values for (%d %d)\n", ic[0], ic[1]);
            // printf("sten->st [%d %d %d]\n",
            //    sten->st[-1], sten->st[0], sten->st[1]);
            // printf("nsten->st [%d %d %d]\n",
            //    nsten->st[-1], nsten->st[0], nsten->st[1]);
            // printf("LINE0-values %g \n", values[1][0]);
            // printf("LINE1-values %g %g %g\n", values[0][1], values[1][1], values[2][1]);
            // printf("LINE2-values %g %g %g\n", values[1][2]);
                        
        }

        // nvalues[2] = nvalues[0] = -0.5*dt/sqr(dh[ndir])/Rect_den(ic,ns_field);
        // nvalues[1] = dt/sqr(dh[ndir])/Rect_den(ic,ns_field)+1.0;
        //
        u_xx = Rect_visc(ic,ns_field)*(vel(dir, sten->st[1]) - 2.0*vel(dir, sten->st[0]) + 
                     vel(dir, sten->st[-1]))/sqr(dh[dir]);
        u_yy = Rect_visc(ic,ns_field)*(vel(dir, nsten->st[1]) - 2.0*vel(dir, nsten->st[0]) + 
                     vel(dir, nsten->st[-1]))/sqr(dh[ndir]);

        if(debug_flag == YES) 
        {
            printf("after comput u_xx, u_yy for (%d %d) = %12.11g, %12.11g\n", ic[0], ic[1], u_xx, u_yy);
            printf("for u_xx[dir = %d] vel(%g %g %g)\n", dir,
                 vel(dir, sten->st[-1]), vel(dir, sten->st[0]), vel(dir, sten->st[1]));
            printf("for u_yy[dir = %d] vel(%g %g %g)\n", ndir,
                 vel(dir, nsten->st[-1]), vel(dir, nsten->st[0]), vel(dir, nsten->st[1]));
        }

        v_xx = Rect_visc(ic,ns_field)*(vel(ndir, sten->st[1]) - 2.0*vel(ndir, sten->st[0]) + 
                     vel(ndir, sten->st[-1]))/sqr(dh[dir]);
        v_yy = Rect_visc(ic,ns_field)*(vel(ndir, nsten->st[1]) - 2.0*vel(ndir, nsten->st[0]) + 
                     vel(ndir, nsten->st[-1]))/sqr(dh[ndir]);
            // TEMP 11/28/06, have not considered viscosity yet
            // u_xx = 0.0;

        if(debug_flag == YES) 
        { 
            printf("after comput v_xx, v_yy for (%d %d) = %12.11g, %12.11g\n", ic[0], ic[1], v_xx, v_yy);
            printf("for v_xx[dir = %d] vel(%g %g %g)\n", dir,
                 vel(ndir, sten->st[-1]), vel(ndir, sten->st[0]), vel(ndir, sten->st[1]));
            printf("for v_yy[dir = %d] vel(%g %g %g)\n\n", ndir,
                 vel(ndir, nsten->st[-1]), vel(ndir, nsten->st[0]), vel(ndir, nsten->st[1]));
        }

        p_cell[dir]->r_side = dt/Rect_den(ic,ns_field)*
                                  (0.5*(u_xx+u_yy) - Rect_gradP(ic,ns_field)[dir])
                   - dt*Rect_con_der(ic,ns_field).v[dir] + vel(dir,sten->st[0])
                   + dt*src[0];

        p_cell[ndir]->r_side = dt/Rect_den(ic,ns_field)*
                                  (0.5*(v_xx+v_yy) - Rect_gradP(ic,ns_field)[ndir])
                   - dt*Rect_con_der(ic,ns_field).v[ndir] + vel(ndir,sten->st[0])
                   + dt*src[1];

        // TMP
        if(debug_flag == YES)
        {
            float exp_u, exp_v;
            // printf("set crank ic[%d %d] dir %d, dt = %g,"
            //            " dt*Rect_con_der(ic,ns_field).v[dir] = %12.11g, dt/rho(-p) = %12.11g\n",
            //      ic[0],ic[1], dir, dt, dt*Rect_con_der(ic,ns_field).v[dir], dt/Rect_den(ic,ns_field)*
            //             (0.5*(u_xx+u_yy) - Rect_gradP(ic,ns_field)[dir]));
            printf("set crank ic[%d %d] ndir %d, dt = %g,"
                       " -dt*Rect_con_der(ic,ns_field).v[dir] = %12.11g, dt/rho(-p) = %12.11g\n",
                 ic[0],ic[1], ndir, dt, -dt*Rect_con_der(ic,ns_field).v[ndir], dt/Rect_den(ic,ns_field)*
                        (0.5*(v_xx+v_yy) - Rect_gradP(ic,ns_field)[ndir]));

            printf("den %12.11g, v_xx, v_yy(%12.11g, %12.11g) Rect_gradP[1] %12.11g\n",
                       Rect_den(ic,ns_field), v_xx, v_yy, Rect_gradP(ic,ns_field)[ndir]);

            printf("Rect_con_der (%12.11g, %12.11g), Rect_gradP(%12.11g, %12.11g), vel1 = %g\n",
                         Rect_con_der(ic,ns_field).v[dir], Rect_con_der(ic,ns_field).v[ndir], 
                            Rect_gradP(ic,ns_field)[dir], Rect_gradP(ic,ns_field)[ndir],
                        vel(ndir,sten->st[0]));

            exp_u = dt/Rect_den(ic,ns_field)*
                                  ((u_xx+u_yy) - Rect_gradP(ic,ns_field)[dir])
                   - dt*Rect_con_der(ic,ns_field).v[dir] + vel(dir,sten->st[0])
                   + dt*src[0];
            exp_v = dt/Rect_den(ic,ns_field)*
                                  ((v_xx+v_yy) - Rect_gradP(ic,ns_field)[ndir])
                   - dt*Rect_con_der(ic,ns_field).v[ndir] + vel(ndir,sten->st[0])
                   + dt*src[1];
            printf(" explicit step velocity = (%g %g)\n", exp_u, exp_v);
        }

        for (l = 0; l < 3; ++l)
        {
            p_cell[dir]->stencil[1][l] = values[1][l];
            p_cell[dir]->stencil[l][1] = values[l][1];
            p_cell[ndir]->stencil[1][l] = values[1][l];
            p_cell[ndir]->stencil[l][1] = values[l][1];
        }

        // FOR RT problem, X-direction, reflecting BC or Periodic BC, 
        // Y-direction, Neumann BC.
        
        if(ic[0] == 0)
        {
            if(PERIOD_BC == NO)
            {
                p_cell[dir]->set[0][1] = NO;
                p_cell[dir]->stencil[1][1] += -1.0*values[0][1];
                p_cell[ndir]->set[0][1] = NO;
                p_cell[ndir]->stencil[1][1] += 1.0*values[0][1];
            }
            else
            {
                p_cell[dir]->set[0][1] = YES;
                p_cell[ndir]->set[0][1] = YES;
            }
        }
        if(ic[0] == gmax[0]-1)
        {
            if(PERIOD_BC == NO)
            {
                p_cell[dir]->set[2][1] = NO;
                p_cell[dir]->stencil[1][1] += -1.0*values[2][1];
                p_cell[ndir]->set[2][1] = NO;
                p_cell[ndir]->stencil[1][1] += 1.0*values[2][1];
            }
            else
            {
                p_cell[dir]->set[2][1] = YES;
                p_cell[ndir]->set[2][1] = YES;
            }
        }
        if(ic[1] == 0)
        {
            p_cell[dir]->set[1][0] = NO;
            p_cell[dir]->stencil[1][1] += -1.0*values[1][0];
            p_cell[ndir]->set[1][0] = NO;
            p_cell[ndir]->stencil[1][1] += -1.0*values[1][0];
        }
        if(ic[1] == gmax[1]-1)  
        {
            p_cell[dir]->set[1][2] = NO;
            p_cell[dir]->stencil[1][1] += -1.0*values[1][2];
            p_cell[ndir]->set[1][2] = NO;
            p_cell[ndir]->stencil[1][1] += -1.0*values[1][2];
        }

        if(debug_flag == YES)
            deb_npt = NO;

        /**
        printf("set crank ic[%d %d] dir %d, set[%g %g %g], right %g\n",
                ic[0], ic[1], dir, values[0], 
                     values[1], values[2], p_cell[dir]->r_side);
        printf("before comput u_xx for (%d %d)\n",
                 ic[0], ic[1]);
        printf("print state 1\n");
        g_verbose_print_state(sten->st[1]);
        printf("print state 0\n");
        g_verbose_print_state(sten->st[0]);
        printf("print state -1\n");
        g_verbose_print_state(sten->st[-1]);
        */
}

/* Look at one direction, for computation of
 * u_xx. Check if meet the boundary only.
 */
LIB_LOCAL bool Crank_N_sten_reg(
        int       idir,
        Stencil   *sten,
        int       *label,
        NS_Field_Stencil *ns_field)
{
        Front        *fr = sten->fr;
        Wave         *wave = sten->wave;
        int          **icoords = sten->icoords, icrds[MAXD];
        COMPONENT    new_comp = sten->newcomp, *comp = sten->comp;
        int          i, j, k, indx, irregular = NO;
        CRX_SIDE    side[2] = {PREV, NEXT};
        CRXING      *crx;
        HYPER_SURF  *hs;
        size_t      sizest = fr->sizest;
        int         end; 
        RECT_GRID   *gr = wave->rect_grid;
        Locstate    st;
        
        end = sten->npts/2;

        for (i = -end; i <= end; ++i)
        {
            if(comp[i] == ext_comp ||
               is_excluded_comp(comp[i],fr->interf))
            {
                label[i] = NO;
                irregular =  YES;
            }
        }

        for (i = 1; i <= end; ++i)
        {
            for(j = 0; j < 2; j++)  
            {
                indx = i*side[j];

                // TMP
                if(idir == 0 && deb_npt == YES)
                {
                    printf("IN Crank_N_sten_reg, indx %d, j = %d, nc = %d\n",
                              indx, j, sten->nc[indx]); 
                }

                // Simple version
                sten->st[indx] = sten->worksp[indx];
                set_type_of_state(sten->st[indx],TGAS_STATE);
                Set_params(sten->st[indx], sten->st[0]);

                if(idir == 1)
                {
                    // Top and bottom, homo dirichlet BC
                    if(icoords[indx][1]  == -1)
                    {
                        icrds[0] = icoords[indx][0];
                        icrds[1] = icoords[indx][1] + 1;
                        st = Rect_state(icrds,wave);
                        Vel(sten->st[indx])[0] = -1.0*Vel(st)[0];
                        Vel(sten->st[indx])[1] = -1.0*Vel(st)[1];

                        label[indx] = NO;
                        irregular = YES;
                    }
                    else if(icoords[indx][1]  == gmax[1])
                    {
                        icrds[0] = icoords[indx][0];
                        icrds[1] = icoords[indx][1] - 1;
                        st = Rect_state(icrds,wave);
                        Vel(sten->st[indx])[0] = -1.0*Vel(st)[0];
                        Vel(sten->st[indx])[1] = -1.0*Vel(st)[1];
                        label[indx] = NO;
                        irregular = YES;
                    }
                    else
                    {
                        sten->st[indx] = Rect_state(icoords[indx],wave);
                    }
                }
                else
                {
                    // Left and right, reflecting BC
                    if(PERIOD_BC == YES)
                    { 
                        sten->st[indx] = Rect_state(icoords[indx],wave);
                        if(icoords[indx][0]  == -1)
                        {
                            label[indx] = NO;
                            irregular = YES;
                        }
                        else if(icoords[indx][0]  == gmax[0])
                        {
                            label[indx] = NO;
                            irregular = YES;
                        }
                        /**
                        // if(icoords[indx][0]  == -1 && icoords[indx][1] == 108)
                        if(icoords[indx][0]  == -1 && icoords[indx][1] == 90)
                        {
                            icrds[0] = icoords[indx][0]+1;
                            icrds[1] = icoords[indx][1];
                            st = Rect_state(icrds,wave);
                            printf("label indx[%d] = %d\n", indx, label[indx]);
                            printf("In Crank_N_sten_reg ic[%d %d] state",icoords[indx][0], icoords[indx][1]);
                            verbose_print_state("indx state",sten->st[indx]);
                            printf("In Crank_N_sten_reg ic[%d %d] state",icrds[0], icrds[1]);
                            verbose_print_state("icrds state",st);

                            icrds[0] = fr->rect_grid->gmax[0]-1;
                            icrds[1] = icoords[indx][1];
                            st = Rect_state(icrds,wave);
                            printf("In Crank_N_sten_reg ic[%d %d] state",icrds[0], icrds[1]);
                            verbose_print_state("right ic state",st);
                        } 
                        **/
                    }
                    else
                    {
                        if(icoords[indx][0]  == -1)
                        {
                            icrds[0] = icoords[indx][0]+1;
                            icrds[1] = icoords[indx][1];
                            st = Rect_state(icrds,wave);
                            Vel(sten->st[indx])[0] = -1.0*Vel(st)[0];
                            Vel(sten->st[indx])[1] = Vel(st)[1];
                            label[indx] = NO;
                            irregular = YES;
                        }
                        else if(icoords[indx][0]  == gmax[0])
                        {
                            icrds[0] = icoords[indx][0]-1;
                            icrds[1] = icoords[indx][1];
                            st = Rect_state(icrds,wave);
                            Vel(sten->st[indx])[0] = -1.0*Vel(st)[0];
                            Vel(sten->st[indx])[1] = Vel(st)[1];
                            label[indx] = NO;
                            irregular = YES;
                        }
                        else
                        {
                            sten->st[indx] = Rect_state(icoords[indx],wave);
                        }
                    }
                }

                // TMP
                if(deb_npt == YES)
                {
                    printf("IN Crank_N_sten_reg, indx %d dir = %d, velocity (%g %g)\n",
                              indx, idir, Vel(sten->st[indx])[0], Vel(sten->st[indx])[1]);
                }

                if(sten->nc[indx] != 0)
                {
                    crx = sten->crx[indx][0];
                    hs = crx->hs;
                    if(wave_type(hs) == NEUMANN_BOUNDARY)
                    {
                        label[indx] = N_SURF;
                        irregular = YES;
                    }
                    if(wave_type(hs) == DIRICHLET_BOUNDARY)
                    {
                        label[indx] = D_SURF;
                        irregular = YES;
                    }
                }

                /***
                if(sten->nc[indx] != 0)
                {
                    crx = sten->crx[indx][0];
                    hs = crx->hs;    
                    if(wave_type(hs) == NEUMANN_BOUNDARY)
                    {
                        label[indx] = N_SURF;
                        irregular = YES;
                    }
                    if(wave_type(hs) == DIRICHLET_BOUNDARY)
                    {
                        label[indx] = D_SURF;
                        irregular = YES;
                    }
                    if(wave_type(hs) == NEUMANN_BOUNDARY)
                    {
                        if(fr->neumann_bdry_state)  
                        {
                            if ((*fr->neumann_bdry_state)(Coords(sten->p[indx]),new_comp,
                                    crx->pt,hs,fr,(POINTER)wave,sten->worksp[indx]))
                            {
                                sten->st[indx] = sten->worksp[indx];
                                for (k = 0; k < dim; ++k)
                                    Coords(sten->p[indx])[k] = Coords(crx->pt)[k];
                                assign(left_state(sten->p[indx]),sten->st[indx],sizest);
                                assign(right_state(sten->p[indx]),sten->st[indx],sizest);
                            }
                            else
                            {
                                printf("ERROR: Crank_N_sten_reg, NEUMANN\n");
                                clean_up(ERROR);
                            }
                        }
                    }
                    if(wave_type(hs) == DIRICHLET_BOUNDARY)
                    {
                        sten->st[indx] = sten->worksp[indx];
                        evaluate_dirichlet_boundary_state(Coords(sten->p[indx]),hs,fr,wave,
                                                     sten->st[indx]);
                        assign(left_state(sten->p[indx]),sten->st[indx],sizest);
                        assign(right_state(sten->p[indx]),sten->st[indx],sizest);
                    }
                    else if(wave_type(hs) == CONTACT)
                    {
                        sten->st[indx] = Rect_state(icoords[indx],wave);
                        ///  TMP
                        if(idir == 1 && deb_npt == YES)
                        {
                            printf("Crank_N_sten_reg _0, idir == 1, indx = %d, ic[%d %d], end = %d\n",
                                        indx, icoords[indx][0], icoords[indx][1], end);
                        }
                    }
                }
                else
                {
                    sten->st[indx] = Rect_state(icoords[indx],wave);
                    ///  TMP
                    if(idir == 1 && deb_npt == YES)
                    {
                        printf("Crank_N_sten_reg, idir == 1, indx = %d, ic[%d %d], end = %d\n",
                                        indx, icoords[indx][0], icoords[indx][1], end);
                    }
                }
                ***/
            }
        }
        return irregular;
}

LOCAL void build_convective_deriv(
        NS_Field_Stencil  *ns_field)
{
        int        i, j, k, dim;
        int        ic[MAXD], icx[MAXD], icy[MAXD];
        float      **dv;
        VEL_ST     **edgeu;
        Wave       *wv = ns_field->newwave;
        TRI_GRID   *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE   *blk_edge = ntg->blk_edge;
        float      *dh = wv->rect_grid->h;
        int        *gmax = wv->rect_grid->gmax;

        matrix(&dv,2,2,sizeof(float));
        matrix(&edgeu,2,2,sizeof(VEL_ST));

        dim = wv->rect_grid->dim;
        // If treat ic[] as cell index, the edge is the lower side of the cell
        for(i = 0; i < gmax[0]; i++)
        {
            ic[0] = i;
            for(j = 0; j < gmax[1]; j++)
            {
                ic[1] = j;
                if(Rect_blk_edge(ic,wv).ctype == PARTIAL ||
                   Rect_blk_edge(ic,wv).ctype == INTERNAL ||
                   Rect_blk_edge(ic,wv).ctype == INTER_PART)
                {
                    get_cell_edge_nor_vel(ic,ns_field,dv,edgeu);
                    for(k = 0; k < dim; k++)
                    {
                        Rect_con_der(ic,ns_field).v[k] = 0.5/dh[0]*(dv[0][0] + dv[0][1])*
                                (edgeu[0][1].v[k] - edgeu[0][0].v[k])
                                + 0.5/dh[1]*(dv[1][0] + dv[1][1])*
                                (edgeu[1][1].v[k] - edgeu[1][0].v[k]);
                    }
                    if(ns_field->n_chem != 0)
                    {
                        icx[0] = icy[0] = ic[0];
                        icx[1] = icy[1] = ic[1];
                        icx[0]++; icy[1]++;
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            Rect_chem_der(ic,ns_field->chem_sten)[k] = 
                                  1.0/dh[0]*(dv[0][1]*v_line_chem(icx,ns_field->chem_sten)[k] - 
                                             dv[0][0]*v_line_chem(ic,ns_field->chem_sten)[k])
                                + 1.0/dh[1]*(dv[1][1]*h_line_chem(icy,ns_field->chem_sten)[k] -
                                             dv[1][0]*h_line_chem(ic,ns_field->chem_sten)[k]);
                        }
                    }

                    // TMP
                    /**
                    if(ic[0] == 0 && (ic[1] == 19 || ic[1] == 18))
                    {
                        k = 1;
                        printf("-----------------, build_convective_deriv,,,,,,,\n");
                        printf("cell[%d %d] vel[%g %g] dv[%g %g, %g %g]\n", 
                                ic[0], ic[1],
                                Rect_con_der(ic,ns_field).v[0],
                                Rect_con_der(ic,ns_field).v[1], dv[0][0], dv[0][1],
                                dv[1][0], dv[1][1]);
                        printf("ul[%g %g], ur[%g %g], ub[%g %g], ut[%g %g] result v1 = %g\n",
                                 edgeu[0][0].v[0], edgeu[0][0].v[1],
                                 edgeu[0][1].v[0], edgeu[0][1].v[1],
                                 edgeu[1][0].v[0], edgeu[1][0].v[1],
                                 edgeu[1][1].v[0], edgeu[1][1].v[1],
                                0.5/dh[0]*(dv[0][0] + dv[0][1])*
                                (edgeu[0][1].v[k] - edgeu[0][0].v[k])
                                + 0.5/dh[1]*(dv[1][0] + dv[1][1])*
                                (edgeu[1][1].v[k] - edgeu[1][0].v[k]));
                        printf("-------END-------\n");
                    }
                    **/
                }
            }
        }


        if(debugging("NS_vel"))
        {
            printf("---------build_convective_deriv-----\n");
            // for(j = 0; j < gmax[1]; j++)
            for(j = 11; j < 13; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i < gmax[0]; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf("(%d %g %g)", i, Rect_con_der(ic,ns_field).v[0], Rect_con_der(ic,ns_field).v[1]);
                }
                printf("\n");
            }
            printf("END ---------build_convective_deriv----- END\n");
        }

        free(dv); free(edgeu);
}


LOCAL void get_cell_edge_nor_vel(
        int              *ic,
        NS_Field_Stencil *ns_field,
        float            **dv,
        VEL_ST           **eu)
{
        int        i, j, k;
        int        ic2[MAXD], dim;
        Wave       *wv = ns_field->newwave;
        TRI_GRID   *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE   *blk_edge = ntg->blk_edge;

        dim = wv->rect_grid->dim;
 
        ic2[0] = ic[0];
        ic2[1] = ic[1];
        for(i = 0; i < 2; i++)
        {
            for(j = 0; j < 2; j++)
                dv[i][j] = 0.0;
        }

        if(Rect_blk_edge(ic,wv).ctype == PARTIAL ||
           Rect_blk_edge(ic,wv).ctype == INTERNAL ||
           Rect_blk_edge(ic,wv).ctype == INTER_PART)
        {
            // If treat ic[] as cell index, the edge is the lower side of the cell
            // if(blk_edge[ic[0]][ic[1]].edge[1][0].etype == FULL_SURF ||
            //    blk_edge[ic[0]][ic[1]].edge[1][0].etype == PARTIAL_SURF)

            dv[1][0] = n_vel_h_edge(ic, ns_field); 
            for(i = 0; i < dim; i++)
                eu[1][0].v[i] = h_line_vel(ic,ns_field).v[i];

            // if(blk_edge[ic[0]][ic[1]].edge[1][1].etype == FULL_SURF ||
            //    blk_edge[ic[0]][ic[1]].edge[1][1].etype == PARTIAL_SURF)
            {
                ic2[1]++;
                dv[1][1] = n_vel_h_edge(ic2, ns_field); 
                for(i = 0; i < dim; i++)
                    eu[1][1].v[i] = h_line_vel(ic2,ns_field).v[i];
                ic2[1]--;
            }
            // left side of cell edge
            // if(blk_edge[ic[0]][ic[1]].edge[0][0].etype == FULL_SURF ||
            //    blk_edge[ic[0]][ic[1]].edge[0][0].etype == PARTIAL_SURF)

            dv[0][0] = n_vel_v_edge(ic, ns_field);
            for(i = 0; i < dim; i++)
                eu[0][0].v[i] = v_line_vel(ic,ns_field).v[i];
            // if(blk_edge[ic[0]][ic[1]].edge[0][1].etype == FULL_SURF ||
            //    blk_edge[ic[0]][ic[1]].edge[0][1].etype == PARTIAL_SURF)
            {
                ic2[0]++;
                dv[0][1] = n_vel_v_edge(ic2, ns_field);
                for(i = 0; i < dim; i++)
                    eu[0][1].v[i] = v_line_vel(ic2,ns_field).v[i];
            }
        }
}

// Adjust normal velocity on cell edges,
// Physical boundary cell edges are excluded.
/***
LOCAL void adjust_nor_vel_on_edge(
        NS_Field_Stencil  *ns_field)
{
        int        i, j, k;
        int        ic[MAXD], ic2[MAXD];
        float      dv;
        Wave       *wv = ns_field->newwave;
        TRI_GRID   *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE   **blk_edge = ntg->blk_edge;
        float      *dh = wv->rect_grid->h;
        int        *gmax = wv->rect_grid->gmax;
        float      rweight;

        // ic[] is edge index. compute on horizontal edge
        // If treat ic[] as cell index, the edge is the lower side of the cell
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic2[0] = ic[0] = i;
                ic2[1] = ic[1] = j;
                if(blk_edge[ic[0]][ic[1]].ctype == PARTIAL ||
                   blk_edge[ic[0]][ic[1]].ctype == INTERNAL ||
                   blk_edge[ic[0]][ic[1]].ctype == INTER_PART)
                {
                    if(blk_edge[ic[0]][ic[1]].edge[1][0].etype == FULL_SURF ||
                       blk_edge[ic[0]][ic[1]].edge[1][0].etype == PARTIAL_SURF)
                    {
                        if(blk_edge[ic[0]][ic[1]].comp_label[1][0] == NO)
                        {
                            ic2[1]--;
                            rweight = h_line_rhor(ic, ns_field);
                            dv = (Rect_phi(ic,ns_field)-Rect_phi(ic2,ns_field))/dh[1]*rweight;
                            n_vel_h_edge(ic, ns_field) -= dv; 
                        }
                    }
                }
            }
        }
        // If treat ic[] as cell index, the edge is the upper side of the cell
        for(i = 0; i < gmax[0]; i++)
        {
            ic2[0] = ic[0] = i;
            ic2[1] = ic[1] = gmax[1]-1;
            if(blk_edge[ic[0]][ic[1]].ctype == PARTIAL ||
               blk_edge[ic[0]][ic[1]].ctype == INTERNAL ||
               blk_edge[ic[0]][ic[1]].ctype == INTER_PART)
            {
                if(blk_edge[ic[0]][ic[1]].edge[1][1].etype == FULL_SURF ||
                   blk_edge[ic[0]][ic[1]].edge[1][1].etype == PARTIAL_SURF)
                {
                    if(blk_edge[ic[0]][ic[1]].comp_label[1][1] == NO)
                    {
                        ic2[1]++;
                        rweight = h_line_rhor(ic2, ns_field);
                        dv = (Rect_phi(ic2,ns_field)-Rect_phi(ic,ns_field))/dh[1]*rweight;
                        n_vel_h_edge(ic2, ns_field) -= dv; 
                    }
                }
            }
        }

        // ic[] is edge index. compute on vertical edge
        // If treat ic[] as cell index, the edge is the left side of the cell
        for(i = 0; i < gmax[0]; i++)
        {
            for(j = 0; j < gmax[1]; j++)
            {
                ic2[0] = ic[0] = i;
                ic2[1] = ic[1] = j;
                if(blk_edge[ic[0]][ic[1]].ctype == PARTIAL ||
                   blk_edge[ic[0]][ic[1]].ctype == INTERNAL ||
                   blk_edge[ic[0]][ic[1]].ctype == INTER_PART)
                {
                    if(blk_edge[ic[0]][ic[1]].edge[0][0].etype == FULL_SURF ||
                       blk_edge[ic[0]][ic[1]].edge[0][0].etype == PARTIAL_SURF)
                    {
                        if(blk_edge[ic[0]][ic[1]].comp_label[0][0] == NO)
                        {
                            ic2[0]--;
                            rweight = v_line_rhor(ic, ns_field);
                            dv = (Rect_phi(ic,ns_field)-Rect_phi(ic2,ns_field))/dh[0]*rweight;
                            n_vel_v_edge(ic, ns_field) -= dv;
                            // tmp
                            // printf("normal vel edge[%d][%d] = %g  (phi %g %g,) ic2[%d %d], dv = %g\n", 
                            //          ic[0],ic[1],n_vel_v_edge(ic, ns_field),
                            //           Rect_phi(ic,ns_field),
                            //           Rect_phi(ic2,ns_field), ic2[0], ic2[1], dv);
                        }
                    }
                }
            }
        }

        // If treat ic[] as cell index, the edge is the right side of the cell
        for(j = 0; j < gmax[1]; j++)
        {
            ic2[0] = ic[0] = gmax[0]-1;
            ic2[1] = ic[1] = j;
            if(blk_edge[ic[0]][ic[1]].ctype == PARTIAL ||
               blk_edge[ic[0]][ic[1]].ctype == INTERNAL ||
               blk_edge[ic[0]][ic[1]].ctype == INTER_PART)
            {
                if(blk_edge[ic[0]][ic[1]].edge[0][1].etype == FULL_SURF ||
                   blk_edge[ic[0]][ic[1]].edge[0][1].etype == PARTIAL_SURF)
                {
                    if(blk_edge[ic[0]][ic[1]].comp_label[0][1] == NO)
                    {
                        ic2[0]++;
                        rweight = v_line_rhor(ic2, ns_field);
                        dv = (Rect_phi(ic2,ns_field)-Rect_phi(ic,ns_field))/dh[0]*rweight;
                        n_vel_v_edge(ic2, ns_field) -= dv;
                    }
                }
            }
        }

        // TMP
        if(debugging("NS_vel"))
        {
            printf("In adjust_nor_vel_on_edge, show vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i < gmax[0]; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d, %g) ", i, n_vel_h_edge(ic, ns_field));            
                }
                printf("\n");
            }
            printf("In adjust_nor_vel_on_edge, show vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d, %g) ", i, n_vel_v_edge(ic, ns_field));
                }
                printf("\n");
            }
        }
}
***/

// Solve D^{MAC} (1/\rho^n) G^{MAC} \phi = D^{MAC}U^{n+1/2} for \phi
// With finite volume discretization
/********************
IMPORTANT: The above eqn is solved with the
modifed boundary condition. Set right side boundary's \phi = 0 (outflow)
(Dirichlet condition for pressure projection)
*******************/
LOCAL void project_half_step_vel(
        NS_Field_Stencil *ns_field)
{
        int             Nx = ns_field->rect_gr->gmax[0];
        int             Ny = ns_field->rect_gr->gmax[1];
        int             indices,ix,iy;
        P_CELL2D        **p_cell;
        // float           **potential;
        int             *n_dist, myid, ICOORDS[2];
        int             **I_to_ijk;   /* local  i, j, k, local I */
        int             **IJK_to_I;  /* global I, J, K, global I */
        int             NLblocks,ilower, iupper, I, i, j;
        int             ic[MAXD];
        Wave            *newwv = ns_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(newwv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = ns_field->newfr;
        float           norm,min,max,x,y,*coords;

#if defined (PETSC)
            /*PETSC Variables*/
        KSP                  ksp;
        Mat                  A;
        Vec                  phi, B, C;
        PetscScalar          *vphi;
        MatNullSpace         nullsp;
        int                  its;
        PetscReal            norm_B, norm_C;
        KSPConvergedReason   reason;
        PC                   pc; 
#endif /*defined (PETSC)*/

#if defined (PETSC)

#if defined(__MPI__)
        MPI_Comm_size(PETSC_COMM_WORLD,&comm_size);
#endif /* defined(__MPI__) */

        // printf("Entered  project_half_step_vel dt %g = %d\n", ns_field->fr->dt, comm_size);

        matrix(&p_cell,Nx,Ny,sizeof(P_CELL2D));
        // matrix(&potential,Nx+4,Ny+4,sizeof(float));

        // init_stencil_element2D(ntg->blk_edge,p_cell,nfr,newwv,ns_field,NO,0.0);
        // Phi of right side boundary cells are set to zero
        // init_stencil_element2D_ver2(ntg->blk_edge,p_cell,nfr,newwv,ns_field,NO,0.0);
        init_stencil_element2D_ver3(newwv,p_cell,nfr,newwv,ns_field,NO,0.0);

        myid = pp_mynode();
        find_Cartesian_coordinates(myid,nfr->pp_grid,ICOORDS);

        vector(&n_dist,comm_size,sizeof(int));
        matrix(&I_to_ijk,Nx*Ny,2,sizeof( int));
        matrix(&IJK_to_I,(Nx+4),(Ny+4),sizeof( int));

        ns_field->I_to_ijk = I_to_ijk;
        ns_field->alloc.I_to_ijk = YES;
        ns_field->IJK_to_I = IJK_to_I;
        ns_field->alloc.IJK_to_I = YES;

        // count interior blks, which support
        // 5 point laplacian stencil
        // ext_comp = exterior_component(nfr->interf);
        NLblocks = 0;
        for(j = 0; j < Ny; j++)
        // for(j = 0; j < Ny-1; j++)
        {
            ic[1] = j;
            // for(i = 0; i < Nx-1; i++)
            for(i = 0; i < Nx; i++)
            {
                ic[0] = i;
                if(Rect_comp(ic,newwv) != ext_comp &&
                   !is_excluded_comp(Rect_comp(ic,newwv),nfr->interf) &&
                   Rect_comp(ic,newwv) != POTS_COMP) 
                {
                    I_to_ijk[NLblocks][0] = i;
                    I_to_ijk[NLblocks][1] = j;
                    NLblocks++;
                }
            } 
        }

        // TMP 
        /**
        printf("project_half_step_vel:: NLblocks = %d\n", NLblocks);
        if(cell_index == NULL)
            vector(&cell_index, Ny*Nx, sizeof(int));
        for(i = 0; i < Ny*Nx; i++)
            cell_index[i] = -1000000000;
        **/
        
          
#if defined(__MPI__)
        pp_all_gather((POINTER)&NLblocks,sizeof(int),(POINTER)n_dist,
              sizeof(int)*comm_size);
#else /* defined(__MPI__) */
        assign((POINTER)n_dist,(POINTER)&NLblocks,sizeof(int));
#endif /* defined(__MPI__) */
     
        // printf("n_dist[0] = %d\n", n_dist[0]);

        ilower = 0;
        iupper = n_dist[0]-1;

        for (i = 1; i <= myid; i++)
        {
            ilower += n_dist[i-1];
            iupper += n_dist[i];
        }

        // TMP printf("Enter project_half_step_vel ilower = %d, iupper = %d\n", ilower, iupper);

        ns_field->ilower = ilower;
        ns_field->iupper = iupper;

        // Need to create global elements here, how???
        create_global_element_list(nfr,ilower,iupper,ICOORDS,I_to_ijk,IJK_to_I);

#if defined(__MPI__)
        // MatCreateMPIAIJ(PETSC_COMM_WORLD,iupper-ilower+1,iupper-ilower+1,
        //         PETSC_DECIDE,PETSC_DECIDE,7,PETSC_NULL,7,PETSC_NULL,&A);
#else
        MatCreate(PETSC_COMM_WORLD, &A);
        MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,iupper-ilower+1,iupper-ilower+1);
        MatSetFromOptions(A);
#endif // if defined(__MPI__) //

        VecCreate(PETSC_COMM_WORLD,&B);
        PetscObjectSetName((PetscObject) B, "RHS");
        VecSetSizes(B,(iupper-ilower + 1),PETSC_DECIDE);
        VecSetFromOptions(B);
        VecDuplicate(B,&phi);
        VecDuplicate(B,&C);

        // printf("ilower = %d, iupper = %d\n", ilower, iupper);
        for (I = ilower; I <= iupper; I++)
        {
            // set_matrix_row(A,I,ilower,I_to_ijk,IJK_to_I,ns_field);
            // set_right_hand_side(B,I,ilower,I_to_ijk,ns_field);
            set_matrix_A_and_right_hand_side(A,B,newwv,
                  p_cell,I,ilower,I_to_ijk,IJK_to_I,ns_field);
        }
        // TMP
        // for(i = 0; i < Ny*Nx; i++)
        //     printf("cell_index[%d] = %d\n", i, cell_index[i]);

        MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
        MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);

        VecAssemblyBegin(B);
        VecAssemblyEnd(B);

        // MatNullSpaceCreate(PETSC_COMM_WORLD,PETSC_TRUE,0,PETSC_NULL,&nullsp);
        KSPCreate(PETSC_COMM_WORLD,&ksp);
        KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN);
 
        KSPGetPC(ksp,&pc);
        PCSetType(pc,PCLU);

#if defined(__MPI__)
        KSPGetPC(ksp,&pc);  //CHKERRQ(ierr);
        PCSetType(pc,PCJACOBI);  //CHKERRQ(ierr);
#endif // if defined(__MPI__)
 
        // new, seems very fast, 062607
        // KSPSetType(ksp,KSPLGMRES);
        // KSPSetType(ksp,KSPLSQR);
        // KSPSetType(ksp,KSPFGMRES);

        // KSPSetNullSpace(ksp,nullsp);
        // KSPSetType(ksp,KSPBCGS);
        // Old tolerance = 1.e-12
        // KSPSetTolerances(ksp,1.e-13,PETSC_DEFAULT,PETSC_DEFAULT,700);
        KSPSetTolerances(ksp,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);
        // KSPSetTolerances(ksp,1.0e-7,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);
        KSPSetComputeSingularValues(ksp, PETSC_TRUE);

        // try not to use KSPSetFromOptions(ksp); 062607
        // KSPSetFromOptions(ksp);

        // KSPSetUp(ksp);

        KSPSolve(ksp,B,phi);
        KSPGetIterationNumber(ksp,&its);
        KSPGetResidualNorm(ksp,&norm);

        KSPComputeExtremeSingularValues(ksp, &max, &min);
        KSPGetConvergedReason(ksp,&reason);

        // TMP
        /*
        printf("\n project_half_step_vel:::: Number of iteration  is %d, converge ?= %d\n", 
                its, reason); 
        printf("residual norm = %6.5e,extreme test SV is [%7.6e ; %7.6e]\n",norm,min,max);
        {
            // KSPView(ksp,PETSC_VIEWER_STDOUT_WORLD); 
            MatMult(A,phi,C);
            VecAXPY(C,-1.0,B);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            printf("After Petsc solver, norm B = %g, norm (A*x-B) = %g\n",
                     norm_B, norm_C);
        }
        */

        if(reason < 0)
        {
            printf("ERROR: project_half_step_vel\n");
            printf("ERROR KSPSolve failed, try to increase number of iterations\n");
            KSPSetTolerances(ksp,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT,(10*NLblocks));
            // KSPSetType(ksp,KSPLSQR);
            // KSPSetType(ksp,KSPLGMRES);

            KSPSolve(ksp,B,phi); // CHKERRQ(ierr);
            // Test 2
            KSPGetConvergedReason(ksp,&reason);
            {
                PetscReal rtol, atol, dtol;
                int maxits, isize;
                KSPType ksptype;
                KSPGetTolerances(ksp,&rtol,&atol,&dtol,&maxits);
                printf("default rtol = %g, atol = %g, dtol = %g, maxits = %d\n",
                   rtol, atol, dtol, maxits);
                VecGetLocalSize(phi, &isize);
                KSPGetType(ksp,&ksptype);
                printf("vec_size = %d, ksptype = %s\n", isize, ksptype);

                MatMult(A,phi,C);
                VecAXPY(C,-1.0,B);
                VecNorm(C,NORM_2,&norm_C);
                VecNorm(B,NORM_2,&norm_B);
                printf("After Petsc solver, norm B = %g, norm (A*x-B) = %g\n",
                         norm_B, norm_C);
                printf("rtol*B_norm = %g, dtol*B_norm = %g\n",
                   rtol*norm_B, dtol*norm_B);
            }
            KSPGetIterationNumber(ksp,&its);
            KSPGetResidualNorm(ksp,&norm);
            printf("2nd try: Residual norm = %6.5e,"
              "extreme test SV is [%7.6e ; %7.6e], reason %d, its = %d\n",
                   norm,min,max, reason, its);
            KSPView(ksp,PETSC_VIEWER_STDOUT_WORLD); 
            if(reason < 0)
                clean_up(ERROR);
        }

        // printf("\n Number of iteration  is %d,"
        //   "residual norm = %6.5e,extreme test SV is [%7.6e ; %7.6e]\n",its,norm,min,max);

        if(debugging("NS_vel") && NO)
        {
            MatView(A,PETSC_VIEWER_STDOUT_WORLD);
            VecView(B,PETSC_VIEWER_STDOUT_WORLD);
        }

        for (iy = 0; iy < Ny; iy++)
        {
            // Phis of right side boundary cells are set to be zero
            /**
            if(iy == Ny -1)
            {
                for (ix = 0; ix < Nx; ix++)
                {
                    ic[0] = ix;
                    ic[1] = iy;
                    Rect_phi(ic,ns_field) = 0.0;
                }
                continue;
            }
            **/

            // for (ix = 0; ix < Nx-1; ix++)
            for (ix = 0; ix < Nx; ix++)
            {
                indices = IJK_to_I[ix+2][iy+2];
                ic[0] = ix; ic[1] = iy;

                if (indices >= 0)
                {
                    VecGetArray(phi,&vphi);
                    // potential[ix+2][iy+2] = vphi[indices-ilower];
                    Rect_phi(ic,ns_field) = vphi[indices-ilower];
                    // VecGetArray(B,&vphi);

                    // printf("ic[%d %d] phi = %g, b = %g\n", 
                    //   ix, iy, potential[ix+2][iy+2], vphi[indices-ilower]);
                    // if(debugging("NS_vel"))
                    //     printf("ic[%d %d] phi = %g\n", 
                    //        ix, iy, Rect_phi(ic,ns_field));
                    // VecRestoreArray(phi,&values);
                    VecRestoreArray(phi,&vphi);
                }
            }
        }

        // TMP
        // if(deb_sten == YES)
        /***
        {
            ic[0] = 32; ic[1] = 3;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));
            ic[0] = 33; ic[1] = 3;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));
            ic[0] = 34; ic[1] = 3;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));
            ic[0] = 33; ic[1] = 2;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));
            ic[0] = 33; ic[1] = 4;
            printf("[%d %d] phi = %g\n",
                     ic[0], ic[1], Rect_phi(ic,ns_field));
        }
        ***/ 

        // TEST
        if(debugging("NS_vel"))
        {
            printf("Petsc solver half step, right_hand_side-\n");
            for(j = 0; j < Ny; j++)
            // for(j = 10; j < 13; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i < gmax[0]; i++)
                for(i = 0; i <= Nx/2; i++)
                {
                    int ii, jj;
                    jj = j; ii = Nx-i-1;
                    // if(i != 0 && i != Nx-1) continue;
 
                    ic[0] = i; ic[1] = j;
                    // printf("(%d %g)", i, fabs(p_cell[i][j].r_side)
                    //         - fabs(p_cell[ii][jj].r_side));
                    printf("(%d %g)", i, p_cell[i][j].r_side);
                }
                printf("\n");
            }
        }
        // END TEST

        // TEST
        if(debugging("NS_vel"))
        {
            int ic2[MAXD];
            MatMult(A,phi,C);
            VecNorm(C,NORM_2,&norm_C);
            VecNorm(B,NORM_2,&norm_B);
            // printf("-------A*phi, print RHS----\n");
            // VecView(B,PETSC_VIEWER_STDOUT_WORLD);
            printf("After Petsc solver half step, norm B = %g, norm (A*phi) = %g\n",
                         norm_B, norm_C);
            printf("------------- MAC phi------------\n");
            // for(j = 0; j < Ny; j++)
            for(j = 107; j < 111; j++)
            {
                printf("row%2d: ", j);
                for(i = 0; i < 3; i++)
                // for(i = 0; i < Nx; i++)
                {
                    ic2[1] = j; ic2[0] = Nx-i-1;
                    // if(i != 0 && i != Nx-1) continue;
                    ic[0] = i; ic[1] = j;
                    // printf("(%d %9.8g)", i, Rect_phi(ic, ns_field)-Rect_phi(ic2, ns_field));
                    printf("%d %g; ", i, Rect_phi(ic, ns_field));
                }
                for(i = 47; i < Nx; i++)
                {
                    ic[0] = i; ic[1] = j;
                    printf("%d %g; ", i, Rect_phi(ic, ns_field));
                }
                printf("\n");
            }

        }
        // END TEST

        VecDestroy(B);
        VecDestroy(phi);
        VecDestroy(C);
        MatDestroy(A);
        KSPDestroy(ksp);
        // MatNullSpaceDestroy(nullsp);

        // printf("Before free these\n");

        free(p_cell);
        // free(potential);
        // free_these(3,n_dist, I_to_ijk, IJK_to_I);
        free(n_dist);

#endif /*defined (PETSC)*/
}
/*
LOCAL int cell_sup_5pt_lap_sten(
        int   *ic,
        Wave  *wv)
{
        int i, dim = wv->rect_grid->dim;

        int icl[MAXD] = {0,1};
        int icr[MAXD] = {2,1};
        int ict[MAXD] = {1,2};
        int icb[MAXD] = {1,0};

        for(i = 0; i < dim; i++)
        {
        }
        
}
*/

/* set projection stencil to solve
 * DG\phi = DU before construction of convective derivatives.
 * \phi on the left boundary is set to be zero.
 * This is to use Dirichlet boundary condition on the left side
 */
LOCAL void init_stencil_element2D(
         BLK_EDGE      **blk_edge,
         P_CELL2D      **p_cell,
         Front         *fr,
         Wave          *wv,
         NS_Field_Stencil *ns_field,
         int           add_time,
         float         dt)
{
       RECT_GRID     *gr = wv->rect_grid;
       int        xmax,ymax;
       int        ix,iy;
       int        i,j,nor;
       int        icoords[2], ic[2];
       float      *coords,coord[2];      // coords is the cell center
                                         // and coord is the cell edge center
       float      *h = gr->h;
       float      rweight = 0.0, dt_coeff = 1.0;

       xmax = gr->gmax[0];
       ymax = gr->gmax[1];

       if(add_time == YES)
           dt_coeff = dt;

       for (iy = 0;  iy < ymax;  ++iy)
       {
           for (ix = 1;  ix < xmax;  ++ix)
           {
               icoords[0] = ix; icoords[1] = iy;
               coords = Rect_coords(icoords,wv);

               for(i = 0; i < 3; ++i)
                  for(j = 0; j < 3; ++j)
               {
                  {
                     p_cell[ix][iy].set[i][j] = NO;
                     p_cell[ix][iy].stencil[i][j] = 0.0;
                  }
               }

               p_cell[ix][iy].r_side = 0.0;
               // TMP
               /*
               printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                   ix, iy, blk_edge[ix][iy].ctype);
               for(i = 0; i < 2; i++)
               {
                   for(j = 0; j < 2; j++)
                   {
                       printf("blk->edge[%d][%d]->type = %d ,", 
                           i, j, blk_edge[ix][iy].edge[i][j].etype);
                   }
                   printf("\n");
               }
               */
               /**
               if(ix == 3 && (iy == 1 || iy == 8))
                   deb_sten = YES;
               else  
                   deb_sten = NO;
               **/

               if(blk_edge[ix][iy].ctype == EXTERNAL)
                   continue; 

               if(blk_edge[ix][iy].ctype == PARTIAL)
               {
                   // cells only have four edges
                   for(i = 0; i < 2; ++i)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           if(blk_edge[ix][iy].edge[i][j].etype == D_SURF ||
                              blk_edge[ix][iy].edge[i][j].etype == N_SURF ||
                              blk_edge[ix][iy].edge[i][j].etype == PARTIAL_SURF)
                           {
                               coord[0] = blk_edge[ix][iy].edge[i][j].cen[0];
                               coord[1] = blk_edge[ix][iy].edge[i][j].cen[1];
                           }
                           else if(blk_edge[ix][iy].edge[i][j].etype == FULL_SURF)
                           {
                               coord[i] = coords[i]+ nor*h[i]/2; coord[(i+1)%2] = coords[(i+1)%2];
                           }
                           else
                               continue;
 
                           ic[0] = ix; ic[1] = iy;
                           if(i == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].edge[i][j].etype == FULL_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[i][j] == NO) 
                                   regular_fd_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                       icoords,ns_field,rweight);
                               else
                               {
                                   bdry_cell_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                        icoords,ns_field,blk_edge[ix][iy].edge[i][j].etype,rweight);
                                   // printf("WARNING: init_stencil_element2D, full_edge,"
                                   //     " but the outter side is body cell[%d %d]\n",
                                   //             ix, iy);
                                   // clean_up(ERROR);
                               }
                           }
                           else if(blk_edge[ix][iy].edge[i][j].etype == D_SURF ||
                                   blk_edge[ix][iy].edge[i][j].etype == N_SURF)
                           {
                               bdry_cell_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[i][j].etype,rweight);
                           }
                           else if(blk_edge[ix][iy].edge[i][j].etype == PARTIAL_SURF)
                           {
                               bdry_cell_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[i][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, partial edge 0,"
                               //     " implementing\n");
                               // clean_up(ERROR);
                           }
                           else
                           {
                               printf("ERROR: init_stencil_element2D, partial edge,"
                                   " implementing\n");
                               clean_up(ERROR);
                           }
                       }
                   }
 
                   /*
                   for(i = 0; i < 3; i++)
                   {
                       for(j = 0; j < 3; j++)
                       {
                           printf("p_cellstencil[%d][%d] = %g,", i, j, p_cell[ix][iy].stencil[i][j]);
                       }
                       printf("\n");
                   }
                   */
 
               }
               else if(blk_edge[ix][iy].ctype == INTERNAL)
               {
                   // cells only have four edges
                   for(i = 0; i < 2; ++i)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           coord[i] = coords[i]+ nor*h[i]/2; coord[(i+1)%2] = coords[(i+1)%2];

                           ic[0] = ix; ic[1] = iy;
                           if(i == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].comp_label[i][j] == NO)
                           {
                               regular_fd_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                          icoords,ns_field,rweight);
                           }
                           else
                           {
                               bdry_cell_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[i][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //     " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                       }
                   }
               }
               else if(blk_edge[ix][iy].ctype == INTER_PART)
               {
                   // cells only have four edges
                   for(i = 0; i < 2; ++i)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           coord[i] = coords[i]+ nor*h[i]/2; coord[(i+1)%2] = coords[(i+1)%2];

                           ic[0] = ix; ic[1] = iy;
                           if(i == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].comp_label[i][j] == NO)
                           {
                               regular_fd_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                       icoords,ns_field,rweight);
                           }
                           else
                           {
                               bdry_cell_sten(&(p_cell[ix][iy]),coord,h,i,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[i][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //     " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                       }
                   }                   
                   /***
                   printf("ERROR: init_stencil_element2D handle case INTER_PART:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, blk_edge[ix][iy].ctype);
                   for(i = 0; i < 2; i++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             i, j, blk_edge[ix][iy].edge[i][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
                   ***/
               }
               else
               {
                   printf("ERROR: init_stencil_element2D handle case:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, blk_edge[ix][iy].ctype);
                   for(i = 0; i < 2; i++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             i, j, blk_edge[ix][iy].edge[i][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
               }
           }
       }

}

LOCAL void partial_surf_sten(
            P_CELL2D   *p_cell,
            float      *coord,
            float      *h,
            int        dir,
            int        *ic,
            NS_Field_Stencil  *ns_field,
            int        i,      // direction
            int        j,      //side
            BLK_EDGE   **blk_edge,
            float      rweight)
{
           int       eic[MAXD];
           float     nv, dh;

           dh = h[(i+1)%2];

           assign(eic,ic,sizeof(int)*2);
           if(i == 0)
           {
               if(dir == 1)
                   eic[0]++;
               nv = v_line_vel(eic,ns_field).v[0];
           }
           else
           {
               if(dir == 1)
                   eic[1]++;
               nv = h_line_vel(eic,ns_field).v[1];
           }

           p_cell->stencil[1][1] -= 1.0;
           p_cell->set[1][1] = YES;
           p_cell->r_side += blk_edge[ic[0]][ic[1]].edge[i][j].length*nv*dir;

           // TMP
           // if(ic[0] == 0 && ic[1] == 0)
           /**
           {
               printf("Partial surf\n");
               printf("ic[%d %d] dir[%d] nor = %d, add to r_side = %g, partial_surf_sten\n",
                       ic[0], ic[1], i, dir, blk_edge[ic[0]][ic[1]].edge[i][j].length*nv*dir);
           }
           **/

           if(blk_edge[ic[0]][ic[1]].comp_label[i][j] == NO)
           {
               if(i == 0)
               {
                   // phi in ic[0][*] is set to zero
                   if(ic[0] != 1 || (ic[0] == 1 && dir != -1))
                   {
                      p_cell->set[1+dir][1] = YES;
                      p_cell->stencil[1+dir][1] += blk_edge[ic[0]][ic[1]].edge[i][j].length/dh*rweight;
                   }
               }
               else
               {
                   p_cell->set[1][1+dir] = YES;
                   p_cell->stencil[1][1+dir] += blk_edge[ic[0]][ic[1]].edge[i][j].length/dh*rweight;
               }
           }
           else
           {
               if(i == 0)
                   p_cell->set[1+dir][1] = NO;
               else
                   p_cell->set[1][1+dir] = NO;
           }
}

/* finite volume discretization of 
 * DG\phi = DU, here \phi is a scalar,
 * U is vector, D is divergence operator,
 * and G is gradient.
 * regular_fd_sten(), bdry_cell_sten();
 * DG = L_{compact}; 
 * Note: the standard 9-point laplacian is 
 * replaced by a compact 5-point Laplacian
 */

LOCAL void bdry_cell_sten(
           P_CELL2D  *p_cell,
           float     *coord,
           float     *h,
           int       i,
           int       dir,
           int       *ic,
           NS_Field_Stencil *ns_field,
           ELEMENT_TYPE  etype,
           float       rweight)
{
           int       eic[MAXD];
           float     nv;

           assign(eic,ic,sizeof(int)*2);
           if(i == 0)
           {
               if(dir == 1)
                   eic[0]++;
               nv = v_line_vel(eic,ns_field).v[0];
           }
           else
           {
               if(dir == 1)
                   eic[1]++;
               nv = h_line_vel(eic,ns_field).v[1];
           }

           // Normal vel. should be dependent on the
           // boundary type, this is handled already.
           // We assume correct boundary vel is
           // stored in ns_field.
           // NEUMANN u_normal = 0; DIRICHLET u_nor = u_bdry.
           // p_cell->r_side += h[(i+1)%2]*nv*dir; 
           // The boundary condition for phi is : (n \cdot \grad phi = 0)

           // TMP
           /**
           if(ic[0] == 0 && ic[1] == 0)
           {
               printf("ic[%d %d] dir[%d] nor = %d, add to r_side = %g, bdry_cell_sten\n",
                       ic[0], ic[1], i, dir, h[(i+1)%2]*nv*dir);
           }
           **/

           // Original 
           // p_cell->r_side += h[(i+1)%2]*nv*dir; 
           // NEW, it's equivalent to reflect normal vel
           p_cell->r_side += h[(i+1)%2]*nv*dir; 


           // TMP
           if(deb_sten == YES)
           {
               printf("--bdry_cell_sten ic[%d %d] dir[%d], nor %d, add to r_side %9.8g\n",
                       ic[0], ic[1], i, dir, h[(i+1)%2]*nv*dir);
           }

           // left hand side moves to right
           // p_cell->r_side -= h[(i+1)%2]*nv*dir;
           // printf("bdry_cell_sten ic[%d %d], direction = %d, side %d, (r_side = %g)\n",
           //          ic[0], ic[1], i, dir, h[(i+1)%2]*nv*dir);

           // Original
           // p_cell->stencil[1][1] -= 1.0;
           // p_cell->set[1][1] = YES;

           // NEW
           // p_cell->stencil[1][1] -= 1.0;
           // p_cell->set[1][1] = YES;

           if(i == 0)
           {
               p_cell->set[1+dir][1] = NO;
               // NEW
               // p_cell->stencil[1-dir][1] += 1.0;
               // p_cell->set[1-dir][1] = YES;
           }
           else
           {
               p_cell->set[1][1+dir] = NO;
               // NEW
               // p_cell->stencil[1][1-dir] += 1.0;
               // p_cell->set[1][1-dir] = YES;
           }

           /**
           if(etype == D_SURF)
           {
           }
           else if(etype == N_SURF)
           {
           }
           **/
}


LOCAL void regular_fd_sten(
           P_CELL2D  *p_cell,
           float     *coord,
           float     *h,
           int       i,
           int       nor,
           int       *ic,
           NS_Field_Stencil *ns_field,
           float     rweight)
{
           int       eic[MAXD];
           float     nv;

           assign(eic,ic,sizeof(int)*2);
           if(i == 0)
           {
               if(nor == 1)
                   eic[0]++;
               nv = v_line_vel(eic,ns_field).v[0];
           }
           else
           {
               if(nor == 1)
                   eic[1]++;
               nv = h_line_vel(eic,ns_field).v[1];
           }
           p_cell->r_side += h[(i+1)%2]*nv*nor;

           // TMP
           if(deb_sten == YES)
           {
               printf("--regular_fd_sten ic[%d %d] dir[%d], nor %d, add to r_side %9.8g\n",
                       ic[0], ic[1], i, nor, h[(i+1)%2]*nv*nor);
           }

           p_cell->stencil[1][1] -= 1.0*rweight;
           p_cell->set[1][1] = YES;
 
           if(i == 0)
           {
               // NOTE: \phi of ic[0][*] is set to 0
               if(ic[0] != 1 || (ic[0] == 1 && nor != -1))
               {
                   p_cell->set[1+nor][1] = YES;
                   p_cell->stencil[1+nor][1] += 1.0*rweight;
               }
               if(deb_sten == YES)
               {
                   printf("--regular_fd_stenx-dir set[%d][%d] coeff add= %9.8g\n",
                             1+nor,1, rweight);
               }
           }
           else
           {
               p_cell->set[1][1+nor] = YES;
               p_cell->stencil[1][1+nor] += 1.0*rweight;

               if(deb_sten == YES)
               {
                   printf("--regular_fd_stenx-dir set[%d][%d] coeff add = %9.8g\n",
                             1, 1+nor, rweight);
               }
           }
}

#if defined (PETSC)
LIB_LOCAL   void            create_global_element_list(
        Front           *fr,
        int             ilower,
        int             iupper,
        int             *ICOORDS,
        int             **I_to_ijk,
        int             **IJK_to_I)
{
        int             **bfs,**bfr;
        int             Nx = fr->rect_grid->gmax[0];
        int             Ny = fr->rect_grid->gmax[1];
        int             NNx = fr->pp_grid->gmax[0];
        int             NNy = fr->pp_grid->gmax[1];
        int             I, ix, iy;
        int             icoords[2];
        int             send,recv;
        int             IIX = ICOORDS[0];
        int             IIY = ICOORDS[1];  // subdomain icoords 
        int             his_id, tag;

#if defined(__MPI__)
        MPI_Status      status;
#endif // #if defined(__MPI__) //

        for (iy = 0; iy < Ny + 4; iy++)
            for (ix = 0; ix < Nx + 4; ix++)
              {
                IJK_to_I[ix][iy] = -1;
              }
        for (I = ilower; I <= iupper; I++)
        {
            ix = I_to_ijk[I-ilower][0];
            iy = I_to_ijk[I-ilower][1];
            IJK_to_I[ix+2][iy+2] = I;
        }

        // printf("in create_global_element_list, NNx = %d, NNy = %d\n",
        //        NNx, NNy);

          //subroutine to form local mappings//

        matrix(&bfs,2,Ny,sizeof(int));
        matrix(&bfr,2,Ny,sizeof(int));

        if (IIX == 0) // left boundary is the domain boundary //
        { 
            for (iy = 0; iy < Ny; iy++)
               IJK_to_I[0][iy+2] = IJK_to_I[1][iy+2] = -1;
        }

        if (IIX == NNx-1) // right boundary is the domain boundary //
        {
            for (iy = 0; iy < Ny; iy++)
               IJK_to_I[Nx+3][iy+2] = IJK_to_I[Nx+2][iy+2] = -1;
        }

        if(IIX%2 == 0)
        {
            //first pass to  right and also receive righ side//
            if (IIX < NNx-1)
            {
                ICOORDS[0] += 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[0] -= 1;

                for (iy = 0; iy < Ny; iy++)
                {
                    bfs[0][iy] = IJK_to_I[Nx][iy+2];
                    bfs[1][iy] = IJK_to_I[Nx+1][iy+2];
                }

                tag = 8;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //

                tag = 9;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else 
                assign(*bfr,*bfs,sizeof(int)*2*Ny);
#endif // if defined(__MPI__) //


                for (iy = 0; iy < Ny; iy++)
                {
                    IJK_to_I[Nx+2][iy+2] = bfr[0][iy];
                    IJK_to_I[Nx+3][iy+2] = bfr[1][iy];
                }
             }
             //then pass to left and receive left side//
             if (IIX > 0)
             {
                 ICOORDS[0] -= 1;
                 his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                 ICOORDS[0] += 1;

                 for (iy = 0; iy < Ny; iy++)
                 {
                    bfs[0][iy] = IJK_to_I[2][iy+2];
                    bfs[1][iy] = IJK_to_I[3][iy+2];
                 }
                 tag = 10;
#if defined(__MPI__)
                 MPI_Send(*bfs,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
                 tag = 11;
#if defined(__MPI__)
                 MPI_Recv(*bfr,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                 assign(*bfr,*bfs,sizeof(int)*2*Ny);
#endif // if defined(__MPI__) //

                 for (iy = 0; iy < Ny; iy++)
                 {
                    IJK_to_I[0][iy+2] = bfr[0][iy];
                    IJK_to_I[1][iy+2] = bfr[1][iy];
                 }
             }
        }
        if(IIX%2 != 0)
        {
            //first receive left side and pass to left//
            if (IIX > 0)
            {
                ICOORDS[0] -= 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[0] += 1;
                tag = 8;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                for (iy = 0; iy < Ny; iy++)
                {
                   bfr[0][iy] = IJK_to_I[2][iy+2];
                   bfr[1][iy] = IJK_to_I[3][iy+2];
                }
#endif // if defined(__MPI__) //

                for (iy = 0; iy < Ny; iy++)
                {
                   IJK_to_I[0][iy+2] = bfr[0][iy];
                   IJK_to_I[1][iy+2] = bfr[1][iy];
                }
                for (iy = 0; iy < Ny; iy++)
                {
                   bfs[0][iy] = IJK_to_I[2][iy+2];
                   bfs[1][iy] = IJK_to_I[3][iy+2];
                }
                tag = 9;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
             }
               //then receive right side and pass to right//
             if (IIX < NNx-1)
             {
                ICOORDS[0] += 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[0] -= 1;

                tag = 10;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                for (iy = 0; iy < Ny; iy++)
                {
                   bfr[0][iy] = IJK_to_I[Nx][iy+2];
                   bfr[1][iy] = IJK_to_I[Nx+1][iy+2];
                }
#endif // if defined(__MPI__) //

                for (iy = 0; iy < Ny; iy++)
                {
                   IJK_to_I[Nx+2][iy+2] = bfr[0][iy];
                   IJK_to_I[Nx+3][iy+2] = bfr[1][iy];
                }
                for (iy = 0; iy < Ny; iy++)
                {
                   bfs[0][iy] = IJK_to_I[Nx][iy+2];
                   bfs[1][iy] = IJK_to_I[Nx+1][iy+2];
                }

                tag = 11;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Ny,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
             }
        }
        free_these(2,bfs,bfr);
        // y direction //

        matrix(&bfs,2,Nx,sizeof(int));
        matrix(&bfr,2,Nx,sizeof(int));

        if (IIY == 0) // front boundary is the domain boundary //
            for (ix = 0; ix < Nx; ix++)
              IJK_to_I[ix+2][0] = IJK_to_I[ix+2][1] = -1;

        if (IIY == NNy-1) // back boundary is the domain boundary //
            for (ix = 0; ix < Nx; ix++)
              IJK_to_I[ix+2][Ny+2] = IJK_to_I[ix+2][Ny+3] = -1;

        if (IIY%2 == 0)
        {
            if (IIY < NNy-1)
            {
                ICOORDS[1] += 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[1] -= 1;

                for (ix = 0; ix < Nx; ix++)
                {
                    bfs[0][ix] = IJK_to_I[ix+2][Ny];
                    bfs[1][ix] = IJK_to_I[ix+2][Ny+1];
                }
                tag = 12;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //

                tag = 13;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                assign(*bfr,*bfs,sizeof(int)*2*Nx);
#endif // if defined(__MPI__) //

                for (ix = 0; ix < Nx; ix++)
                {
                    IJK_to_I[ix+2][Ny+2] = bfr[0][ix];
                    IJK_to_I[ix+2][Ny+3] = bfr[1][ix];
                }
            }
            if (IIY > 0)
            {
                ICOORDS[1] -= 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[1] += 1;

                for (ix = 0; ix < Nx; ix++)
                {
                    bfs[0][ix] = IJK_to_I[ix+2][2];
                    bfs[1][ix] = IJK_to_I[ix+2][3];
                }

                tag = 14;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
                tag = 15;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                assign(*bfr,*bfs,sizeof(int)*2*Nx);
#endif // if defined(__MPI__) //

                for (ix = 0; ix < Nx; ix++)
                {
                    IJK_to_I[ix+2][0] = bfr[0][ix];
                    IJK_to_I[ix+2][1] = bfr[1][ix];
                }
            }
        }
        if (IIY%2 != 0)
        {
            if (IIY > 0)
            {
                ICOORDS[1] -= 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[1] += 1;

                tag = 12;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                for (ix = 0; ix < Nx; ix++)
                {
                    bfr[0][ix] = IJK_to_I[ix+2][2];
                    bfr[1][ix] = IJK_to_I[ix+2][3];
                }
#endif // if defined(__MPI__) //

                for (ix = 0; ix < Nx; ix++)
                {
                    IJK_to_I[ix+2][0] = bfr[0][ix];
                    IJK_to_I[ix+2][1] = bfr[1][ix];
                }
                for (ix = 0; ix < Nx; ix++)
                {
                    bfs[0][ix] = IJK_to_I[ix+2][2];
                    bfs[1][ix] = IJK_to_I[ix+2][3];
                }

                tag = 13;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
             }
             if (IIY < NNy-1)
             {
                ICOORDS[1] += 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[1] -= 1;

                tag = 14;
#if defined(__MPI__)
                MPI_Recv(*bfr,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                for (ix = 0; ix < Nx; ix++)
                {
                    bfr[0][ix] = IJK_to_I[ix+2][Ny];
                    bfr[1][ix] = IJK_to_I[ix+2][Ny+1];
                }
#endif // if defined(__MPI__) //

                for (ix = 0; ix < Nx; ix++)
                {
                    IJK_to_I[ix+2][Ny+2] = bfr[0][ix];
                    IJK_to_I[ix+2][Ny+3] = bfr[1][ix];
                }
                for (ix = 0; ix < Nx; ix++)
                {
                    bfs[0][ix] = IJK_to_I[ix+2][Ny];
                    bfs[1][ix] = IJK_to_I[ix+2][Ny+1];
                }
                tag = 15;
#if defined(__MPI__)
                MPI_Send(*bfs,2*Nx,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
             }
         }

         free_these(2,bfs,bfr);

                   //routine to send and receive corner part//
        if(IIY < 1)
        {
            if(IIX < fr->pp_grid->gmax[0]-1)
            {
                  //pass and recv the upper right corner//

                 ICOORDS[0] += 1; ICOORDS[1] += 1;
                 his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                 ICOORDS[0] -= 1; ICOORDS[1] -= 1;

                 if(his_id >= 0 && his_id < comm_size)
                 {
                    tag = 16;
                    send = IJK_to_I[Nx+1][Ny+1];
#if defined(__MPI__)
                    MPI_Send(&send,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //

                    tag = 17;
#if defined(__MPI__)
                    MPI_Recv(&recv,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                    recv = send;
#endif // if defined(__MPI__) //
                    IJK_to_I[Nx+2][Ny+2] = recv;
                 }
            }
            if(IIX > 0)
            {
                      //pass and recv the upper left corner//

                ICOORDS[0] -= 1; ICOORDS[1] += 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[0] += 1; ICOORDS[1] -= 1;

                if(his_id >= 0 && his_id < comm_size)
                {
                   tag = 18;
                   send = IJK_to_I[2][Ny+1];
#if defined(__MPI__)
                   MPI_Send(&send,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //

                   tag = 19;
#if defined(__MPI__)
                   MPI_Recv(&recv,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                   recv = send;
#endif // if defined(__MPI__) //
                   IJK_to_I[1][Ny+2] = recv;
                }
            }
        }
        else if(IIY >= 1)
        {
            if(IIX > 0)
            {
                  //recv and pass the lower left corner//

                 ICOORDS[0] -= 1; ICOORDS[1] -= 1;
                 his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                 ICOORDS[0] += 1; ICOORDS[1] += 1;

                 if(his_id >= 0 && his_id < comm_size)
                 {
                    tag = 16;
#if defined(__MPI__)
                    MPI_Recv(&recv,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                    recv = IJK_to_I[2][2];
#endif // if defined(__MPI__) //
                    IJK_to_I[1][1] = recv;

                    tag = 17;
                    send = IJK_to_I[2][2];
#if defined(__MPI__)
                    MPI_Send(&send,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
                 }
            }
            if(IIX < fr->pp_grid->gmax[0]-1)
            {
                       //recv and pass the lower right corner//
                ICOORDS[0] += 1; ICOORDS[1] -= 1;
                his_id = domain_id(ICOORDS,fr->pp_grid->gmax,2);
                ICOORDS[0] -= 1; ICOORDS[1] += 1;

                if(his_id >= 0 && his_id < comm_size)
                {
                   tag = 18;
#if defined(__MPI__)
                   MPI_Recv(&recv,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD,&status);
#else
                   recv = IJK_to_I[Nx+1][2];
#endif // if defined(__MPI__) //

                   IJK_to_I[Nx+2][1] = recv;

                   tag = 19;
                   send = IJK_to_I[Nx+1][2];
#if defined(__MPI__)
                   MPI_Send(&send,1,MPI_INT,his_id,tag,PETSC_COMM_WORLD);
#endif // if defined(__MPI__) //
                }
            }
        }
        return;
}

LOCAL   void  set_matrix_A_and_right_hand_side(
        Mat             A,
        Vec             B,
        Wave            *wv_blk,
        P_CELL2D        **p_cell,
        int             I,
        int             ilower,
        int             **I_to_ijk,
        int             **IJK_to_I,
        NS_Field_Stencil *ns_field)
{
        Front           *nfr = ns_field->newfr;
        Wave            *nwv = ns_field->newwave;
        float           hxs, hys;
        int             i,j,ix,iy;
        int             icoords[MAXD],idxm[1];
        int             icfar[MAXD];
        COMPONENT       cc, cn, cs, cw, ce, cfar;
        int             comp_label[5], comp_count;
        float           values[7],*coords;
        int             cols[7];
        int             sten_index[3][3];

        idxm[0] = I;

        for(i = 0; i < dim; i++)
            icoords[i] = I_to_ijk[I-ilower][i];
        ix = icoords[0];
        iy = icoords[1];

        // tmp
        /**
        if(ix == 0 && iy == 0)
        {
            printf(" set cell[%d %d] type = %d\n",
                 ix, iy, Rect_blk_edge(icoords,wv_blk).ctype);
            for(i = 0; i < 2; i++)
            {
                for(j = 0; j < 2; j++)
                {
                    printf("blk->edge[%d][%d]->type = %d ,", i, j, 
                             Rect_blk_edge(icoords,wv_blk).edge[i][j].etype);
                }    
                printf("\n");
            }

            for(i = 0; i < 3; i++)
            {
                for(j = 0; j < 3; j++)
                {
                    printf("stencil[%d][%d] = %g (%d),", i, j, p_cell[ix][iy].stencil[i][j],
                            p_cell[ix][iy].set[i][j]);
                }
                printf("\n");
            }
            printf("right side = %g\n", p_cell[ix][iy].r_side);
        }
        **/

        if(Rect_blk_edge(icoords,wv_blk).ctype == EXTERNAL)
            return;

        
        if(p_cell[ix][iy].set[1][1] == NO)
        {
            printf("ERROR: set_matrix_A_and_right_hand_side, ic[%d][%d] [1][1] not set\n",
                              ix, iy);
            printf("cell type = %d, comp[%d]\n", Rect_blk_edge(icoords,wv_blk).ctype,
                     Rect_comp(icoords,nwv));
            printf("line-0 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                    p_cell[ix][iy].set[0][0], p_cell[ix][iy].set[1][0],
                    p_cell[ix][iy].set[2][0],
                    p_cell[ix][iy].stencil[0][0],
                    p_cell[ix][iy].stencil[1][0], p_cell[ix][iy].stencil[2][0]);
            printf("line-1 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                    p_cell[ix][iy].set[0][1], p_cell[ix][iy].set[1][1],
                    p_cell[ix][iy].set[2][1],
                    p_cell[ix][iy].stencil[0][1],
                    p_cell[ix][iy].stencil[1][1], p_cell[ix][iy].stencil[2][1]);
            printf("line-2 [%d %d %d], [%9.8g %9.8g %9.8g]\n",
                    p_cell[ix][iy].set[0][2], p_cell[ix][iy].set[1][2],
                    p_cell[ix][iy].set[2][2],
                    p_cell[ix][iy].stencil[0][2], p_cell[ix][iy].stencil[1][2],
                    p_cell[ix][iy].stencil[2][2]);
            printf("comp_label of blk_edge[%d][%d] = [%d %d %d %d]\n",
                   ix, iy,
                   Rect_blk_edge(icoords,wv_blk).comp_label[0][0], 
                   Rect_blk_edge(icoords,wv_blk).comp_label[0][1], 
                   Rect_blk_edge(icoords,wv_blk).comp_label[1][0], 
                   Rect_blk_edge(icoords,wv_blk).comp_label[1][1]);
            printf("blk_edge[%d][%d].edge[][].etype = %d, %d %d %d\n",
                     ix, iy, 
               Rect_blk_edge(icoords,wv_blk).edge[0][0].etype,
               Rect_blk_edge(icoords,wv_blk).edge[0][1].etype,
               Rect_blk_edge(icoords,wv_blk).edge[1][0].etype,
               Rect_blk_edge(icoords,wv_blk).edge[1][1].etype);
            clean_up(ERROR);
        }

        for(i = 0; i < 3; ++i)
        {
            for(j = 0; j < 3; ++j)
            {
                sten_index[i][j] = IJK_to_I[ix+i+1][iy+j+1];
                // NEW for PERIOD_BC in the x direction
                if(PERIOD_BC == YES && ix == 0 && i == 0 && p_cell[ix][iy].set[i][j] == YES)
                {
                    sten_index[i][j] = IJK_to_I[ix+i+gmax[0]+1][iy+j+1];
                }
                else if(PERIOD_BC == YES && ix == gmax[0]-1 && i == 2 && p_cell[ix][iy].set[i][j] == YES)
                {
                    sten_index[i][j] = IJK_to_I[ix+i-gmax[0]+1][iy+j+1];
                }

                if(p_cell[ix][iy].set[i][j] == YES)
                {
                    MatSetValues(A,1,idxm,1,&sten_index[i][j],
                        &(p_cell[ix][iy].stencil[i][j]),INSERT_VALUES);

                    // TMP
                    /**
                    if(PERIOD_BC == YES && ix == 0 && i == 0)
                    {
                        printf("case 0\n");
                        printf("0 cell[%d %d] set sten on [%d %d] with indx %d\n",
                             ix, iy, I_to_ijk[IJK_to_I[ix+i+gmax[0]+1][iy+j+1]-ilower][0],
                                     I_to_ijk[IJK_to_I[ix+i+gmax[0]+1][iy+j+1]-ilower][1], 
                             sten_index[i][j]); 
                        cell_index[sten_index[i][j]] = sten_index[i][j];
                    }
                    else if(PERIOD_BC == YES && ix == gmax[0]-1 && i == 2)
                    {
                        printf("case 1\n");
                        printf("1 cell[%d %d] set sten on [%d %d] with indx %d\n",
                            ix, iy, 
                            I_to_ijk[IJK_to_I[ix+i-gmax[0]+1][iy+j+1]-ilower][0],
                            I_to_ijk[IJK_to_I[ix+i-gmax[0]+1][iy+j+1]-ilower][1], sten_index[i][j]);
                        cell_index[sten_index[i][j]] = sten_index[i][j];
                    }
                    else
                    {
                        printf("case 2\n");
                        printf("2 cell[%d %d] set sten on [%d %d] with indx %d\n",
                           ix, iy, 
                           I_to_ijk[IJK_to_I[ix+i+1][iy+j+1]-ilower][0],
                           I_to_ijk[IJK_to_I[ix+i+1][iy+j+1]-ilower][1], sten_index[i][j]);
                        cell_index[sten_index[i][j]] = sten_index[i][j];
                    }
                    **/ 
                }
            }
        }

        if(debugging("NS_vel"))
        {
            /*
            // if(ix == 9 && (iy == 0 || iy == 9))
            {
            printf("ic[%d %d] sten, right_side = %9.8g\n", ix, iy, p_cell[ix][iy].r_side);
            printf("line-0 [%d %d %d], [%9.8g %9.8g %9.8g]\n", 
                    p_cell[ix][iy].set[0][0], p_cell[ix][iy].set[1][0], 
                    p_cell[ix][iy].set[2][0],
                    p_cell[ix][iy].stencil[0][0], 
                    p_cell[ix][iy].stencil[1][0], p_cell[ix][iy].stencil[2][0]);
            printf("line-1 [%d %d %d], [%9.8g %9.8g %9.8g]\n", 
                    p_cell[ix][iy].set[0][1], p_cell[ix][iy].set[1][1], 
                    p_cell[ix][iy].set[2][1],
                    p_cell[ix][iy].stencil[0][1], 
                    p_cell[ix][iy].stencil[1][1], p_cell[ix][iy].stencil[2][1]);
            printf("line-2 [%d %d %d], [%9.8g %9.8g %9.8g]\n", 
                    p_cell[ix][iy].set[0][2], p_cell[ix][iy].set[1][2], 
                    p_cell[ix][iy].set[2][2],
                    p_cell[ix][iy].stencil[0][2], p_cell[ix][iy].stencil[1][2], 
                    p_cell[ix][iy].stencil[2][2]);
            printf("\n");
            }
            */ 
        }

        values[0] = p_cell[ix][iy].r_side;
        VecSetValues(B,1,idxm,values,INSERT_VALUES);
}

// The cell edge length has not been accounted for yet
LOCAL   void            set_right_hand_side(
        Vec             B,
        int             I,
        int             ilower,
        int             **I_to_ijk,
        NS_Field_Stencil *ns_field)
{
        int             i,ix,iy,indices[1], ic[MAXD];
        int             eicw[MAXD], eice[MAXD], eicn[MAXD], eics[MAXD]; // edge index of cell ic[]
        float           values[1];
        Front           *nfr = ns_field->newfr;
        Wave            *nwv = ns_field->newwave;
        float           hx = nfr->rect_grid->h[0];
        float           hy = nfr->rect_grid->h[1];
        float           hxy, vw, ve, vn, vs;
        COMPONENT       cc;
        float           *coords;

        hxy = hx*hy;

        indices[0]=I;

        ic[0] = ix = I_to_ijk[I-ilower][0];
        ic[1] = iy = I_to_ijk[I-ilower][1];

        cc = Rect_comp(ic,nwv);
        if(cc == ext_comp ||
           is_excluded_comp(cc,intfc))
            return;

        eicw[0] = ic[0]; eicw[1] = ic[1]; 
        eice[0] = ic[0]+1; eice[1] = ic[1]; 
        eicn[0] = ic[0]; eicn[1] = ic[1]+1; 
        eics[0] = ic[0]; eics[1] = ic[1]; 
 
        vw = v_line_vel(eicw,ns_field).v[0];
        ve = v_line_vel(eice,ns_field).v[0];
        vn = h_line_vel(eicn,ns_field).v[1];
        vs = h_line_vel(eics,ns_field).v[1];

        values[0] = (ve - vw)/hx + (vn - vs)/hy;
        VecSetValues(B,1,indices,values,INSERT_VALUES);
        
}

LOCAL   void            set_matrix_row(
        Mat             A,
        int             I,
        int             ilower,
        int             **I_to_ijk,
        int             **IJK_to_I,
        NS_Field_Stencil *ns_field)
{
        Front           *nfr = ns_field->newfr;
        Wave            *nwv = ns_field->newwave;
        float           hx = nfr->rect_grid->h[0];
        float           hy = nfr->rect_grid->h[1];
        float           hxs, hys;
        int             i,ix,iy;
        int             icoords[MAXD],idxm[1], icn[MAXD], ics[MAXD], icw[MAXD], ice[MAXD];
        int             icfar[MAXD];
        COMPONENT       cc, cn, cs, cw, ce, cfar;
        int             comp_label[5], comp_count;
        float           values[7],*coords;
        int             cols[7];

        idxm[0] = I;
        hxs = hx*hx;
        hys = hy*hy;
 
        for(i = 0; i < dim; i++)
        {
            icn[i] = ics[i] = icw[i] = ice[i] = 
                  icoords[i] = I_to_ijk[I-ilower][i];
        }
        ix = icoords[0];
        iy = icoords[1];

        // if(ix == 15 && iy == 23)
        //     printf("debug ic[%d %d]\n", ix, iy);

        icn[1]++; ics[1]--;
        ice[0]++; icw[0]--;

        cc = Rect_comp(icoords,nwv);
        cn = Find_rect_comp_ver2(icn,nwv);
        cs = Find_rect_comp_ver2(ics,nwv);
        cw = Find_rect_comp_ver2(icw,nwv);
        ce = Find_rect_comp_ver2(ice,nwv);

        if(cc == ext_comp ||
           is_excluded_comp(cc,intfc))
            return;

        comp_count = cell_nbhd_config(cn, cs, cw, ce, comp_label);

        /*
        printf("ic = %d %d, case %d\n", icoords[0], icoords[1], comp_count);
        for(i = 0; i < 4; i++)
            printf("comp_label[%d] = %d\n", i, comp_label[i]);
        */

        if(comp_count == 0)
        {
            // printf("ERROR: in set_matrix_row(), case %d need to be handled\n",
            //          comp_count);
            // printf("ic = %d %d\n", icoords[0], icoords[1]);

            cols[0] = I;
            cols[1] = IJK_to_I[ix+1][iy+2];
            cols[2] = IJK_to_I[ix+3][iy+2];
            cols[3] = IJK_to_I[ix+2][iy+1];
            cols[4] = IJK_to_I[ix+2][iy+3];

            values[0] = -2.0/hxs - 2.0/hys;
            values[1] = 1.0/hxs;
            values[2] = 1.0/hxs;
            values[3] = 1.0/hys;
            values[4] = 1.0/hys;

            MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
        }
        else if(comp_count == 1)
        {
            /*
            printf("ic = %d %d\n", icoords[0], icoords[1]);
            printf("WARNING: in set_matrix_row(), case %d need to be handled\n",
                     comp_count);
            for(i = 0; i < 4; i++)
                printf("comp_label[%d] = %d\n", i, comp_label[i]);
            */

            if(comp_label[0] == YES)
            {
                // north side
                for(i = 0; i < dim; i++)
                    icfar[i] = ics[i];
                icfar[1]--;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc))
                {
                    printf("ERROR: in set_matrix_row(), north side new case\n");
                    clean_up(ERROR);
                }
                values[0] = -2.0/hxs + 1.0/hys;
                values[1] = -2.0/hys;
                values[2] = 1.0/hys;
                values[3] = 1.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+1];
                cols[2] = IJK_to_I[ix+2][iy];
                cols[3] = IJK_to_I[ix+3][iy+2];
                cols[4] = IJK_to_I[ix+1][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else if(comp_label[1] == YES)
            {
                // south side
                for(i = 0; i < dim; i++)
                    icfar[i] = icn[i];
                icfar[1]++;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc))
                {
                    printf("ERROR: in set_matrix_row(), south side new case\n");
                    clean_up(ERROR);
                }
                values[0] = -2.0/hxs + 1.0/hys;
                values[1] = -2.0/hys;
                values[2] = 1.0/hys;
                values[3] = 1.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+3];
                cols[2] = IJK_to_I[ix+2][iy+4];
                cols[3] = IJK_to_I[ix+1][iy+2];
                cols[4] = IJK_to_I[ix+3][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else if(comp_label[2] == YES)
            {
                // west side
                for(i = 0; i < dim; i++)
                    icfar[i] = ice[i];
                icfar[0]++;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc))
                {
                    printf("ERROR: in set_matrix_row(), west side new case\n");
                    clean_up(ERROR);
                }
                values[0] = -2.0/hys + 1.0/hxs;
                values[1] = 1.0/hys;
                values[2] = 1.0/hys;
                values[3] = -2.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+3];
                cols[2] = IJK_to_I[ix+2][iy+1];
                cols[3] = IJK_to_I[ix+3][iy+2];
                cols[4] = IJK_to_I[ix+4][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else if(comp_label[3] == YES)
            {
                // east side
                for(i = 0; i < dim; i++)
                    icfar[i] = icw[i];
                icfar[0]--;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc))
                {
                    printf("ERROR: in set_matrix_row(), east side new case\n");
                    clean_up(ERROR);
                }
                values[0] = -2.0/hys + 1.0/hxs;
                values[1] = 1.0/hys;
                values[2] = 1.0/hys;
                values[3] = -2.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+3];
                cols[2] = IJK_to_I[ix+2][iy+1];
                cols[3] = IJK_to_I[ix+1][iy+2];
                cols[4] = IJK_to_I[ix][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else
            {
                printf("ic = %d %d\n", icoords[0], icoords[1]);
                for(i = 0; i < 4; i++)
                    printf("comp_label[%d] = %d\n", i, comp_label[i]);
                printf("ERROR: in set_matrix_row(), case %d need to be handled\n",
                     comp_count);
                clean_up(ERROR);
            }
        }
        else if(comp_count == 2)
        {
            int         icfar2[MAXD];
            COMPONENT   cfar2;

            if(comp_label[1] == YES && comp_label[2] == YES)
            {
                // west and south side
                for(i = 0; i < dim; i++)
                {
                    icfar[i] = ice[i];
                    icfar2[i] = icn[i];
                }
                icfar[0]++; icfar2[1]++;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                cfar2 = Find_rect_comp_ver2(icfar2,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc) ||
                   cfar2 == ext_comp ||
                   is_excluded_comp(cfar2,intfc))
                {
                    printf("ERROR: in set_matrix_row(), west and south side new case\n");
                    clean_up(ERROR);
                }
                values[0] = 1.0/hys + 1.0/hxs;
                values[1] = -2.0/hys;
                values[2] = 1.0/hys;
                values[3] = -2.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+3];
                cols[2] = IJK_to_I[ix+2][iy+4];
                cols[3] = IJK_to_I[ix+3][iy+2];
                cols[4] = IJK_to_I[ix+4][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else if(comp_label[1] == YES && comp_label[3] == YES)
            {
                // east and south side
                for(i = 0; i < dim; i++)
                {
                    icfar[i] = icw[i];
                    icfar2[i] = icn[i];
                }
                icfar[0]--; icfar2[1]++;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                cfar2 = Find_rect_comp_ver2(icfar2,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc) ||
                   cfar2 == ext_comp ||
                   is_excluded_comp(cfar2,intfc))
                {
                    printf("ERROR: in set_matrix_row(), east and south side new case\n");
                    printf("icfar[%d %d] comp = %d\n", icfar[0], icfar[1], cfar);
                    printf("icfar2[%d %d] comp2 = %d\n", icfar2[0], icfar2[1], cfar2);
                    clean_up(ERROR);
                }
                values[0] = 1.0/hys + 1.0/hxs;
                values[1] = -2.0/hys;
                values[2] = 1.0/hys;
                values[3] = -2.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+3];
                cols[2] = IJK_to_I[ix+2][iy+4];
                cols[3] = IJK_to_I[ix+1][iy+2];
                cols[4] = IJK_to_I[ix][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else if(comp_label[0] == YES && comp_label[3] == YES)
            {
                // north and east side
                for(i = 0; i < dim; i++)
                {
                    icfar[i] = icw[i];
                    icfar2[i] = ics[i];
                }
                icfar[0]--; icfar2[1]--;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                cfar2 = Find_rect_comp_ver2(icfar2,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc) ||
                   cfar2 == ext_comp ||
                   is_excluded_comp(cfar2,intfc))
                {
                    printf("ERROR: in set_matrix_row(), north and east side new case\n");
                    printf("icfar[%d %d] comp = %d\n", icfar[0], icfar[1], cfar);
                    printf("icfar2[%d %d] comp2 = %d\n", icfar2[0], icfar2[1], cfar2);
                    clean_up(ERROR);
                }
                values[0] = 1.0/hys + 1.0/hxs;
                values[1] = -2.0/hys;
                values[2] = 1.0/hys;
                values[3] = -2.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+1];
                cols[2] = IJK_to_I[ix+2][iy];
                cols[3] = IJK_to_I[ix+1][iy+2];
                cols[4] = IJK_to_I[ix][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else if(comp_label[0] == YES && comp_label[2] == YES)
            {
                // north and west side
                for(i = 0; i < dim; i++)
                {
                    icfar[i] = ice[i];
                    icfar2[i] = ics[i];
                }
                icfar[0]++; icfar2[1]--;
                cfar = Find_rect_comp_ver2(icfar,nwv);
                cfar2 = Find_rect_comp_ver2(icfar2,nwv);
                if(cfar == ext_comp ||
                   is_excluded_comp(cfar,intfc) ||
                   cfar2 == ext_comp ||
                   is_excluded_comp(cfar2,intfc))
                {
                    printf("ERROR: in set_matrix_row(), north and east side new case\n");
                    printf("icfar[%d %d] comp = %d\n", icfar[0], icfar[1], cfar);
                    printf("icfar2[%d %d] comp2 = %d\n", icfar2[0], icfar2[1], cfar2);
                    clean_up(ERROR);
                }
                values[0] = 1.0/hys + 1.0/hxs;
                values[1] = -2.0/hys;
                values[2] = 1.0/hys;
                values[3] = -2.0/hxs;
                values[4] = 1.0/hxs;
                cols[0] = I;
                cols[1] = IJK_to_I[ix+2][iy+1];
                cols[2] = IJK_to_I[ix+2][iy];
                cols[3] = IJK_to_I[ix+3][iy+2];
                cols[4] = IJK_to_I[ix+4][iy+2];
                MatSetValues(A,1,idxm,5,cols,values,INSERT_VALUES);
            }
            else 
            {
                printf("ic = %d %d\n", icoords[0], icoords[1]);
                for(i = 0; i < 4; i++)
                    printf("comp_label[%d] = %d\n", i, comp_label[i]);
                printf("ERROR: in set_matrix_row(), case %d need to be handled, here\n",
                     comp_count);
                printf("cc = %d, cn = %d, cs = %d, cw = %d, ce = %d\n",
                         cc, cn, cs, cw, ce);
                clean_up(ERROR);
            }
        }
        else
        {
            printf("ic = %d %d\n", icoords[0], icoords[1]);
            for(i = 0; i < 4; i++)
                printf("comp_label[%d] = %d\n", i, comp_label[i]);
            printf("ERROR: in set_matrix_row(), case %d need to be handled, here\n",
                 comp_count);
            printf("cc = %d, cn = %d, cs = %d, cw = %d, ce = %d\n",
                     cc, cn, cs, cw, ce);
            clean_up(ERROR);
        }
}

// return component in the nbhd of cc
//            cn
//         cw cc ce
//            cs
/**
LOCAL int cell_nbhd_config(
        COMPONENT    cn, 
        COMPONENT    cs, 
        COMPONENT    cw, 
        COMPONENT    ce, 
        int          *label)
{
        int    count = 0, i;

        for(i = 0; i < dim+2; i++)
            label[i] = NO;

        if(cn == ext_comp ||
           is_excluded_comp(cn,intfc)) 
        {
            count++;
            label[0] = YES; 
        }
        if(cs == ext_comp ||
           is_excluded_comp(cs,intfc)) 
        {
            count++;
            label[1] = YES; 
        }
        if(cw == ext_comp ||
           is_excluded_comp(cw,intfc)) 
        {
            count++;
            label[2] = YES; 
        }
        if(ce == ext_comp ||
           is_excluded_comp(ce,intfc)) 
        {
            count++;
            label[3] = YES; 
        }
        return count;
}
**/
#endif /* if defined (PETSC) */

// Compute velocity on edge, include Gp, viscosity, transverse 
/*
LOCAL void comput_final_vel_on_edge(
        float      *dh,
        float      dt,
        Wave       *wv,
        Wave       *nwv,
        Front      *fr,
        Front      *nfr,
        COMPONENT  max_comp,
        NS_Field_Stencil *ns_field)
{
        int        i, j, k;
        int        *gmax, ic[MAXD];
        float      vn;
        
        gmax = wv->rect_grid->gmax;
        
         
}
*/

// Only compute for the interior cells.
// Upwind \hat{U} based on normal advective velocity on the edge
LOCAL void upwinding_on_edge(
        float      *dh,
        float      dt,
        Wave       *wv,
        Wave       *nwv,
        Front      *fr,
        Front      *nfr,
        COMPONENT  max_comp,
        NS_Field_Stencil *ns_field)
{
        int        i, j, k;
        int        *gmax, ic[MAXD];
        float      vn;
        
        gmax = wv->rect_grid->gmax;

        // ic[] is edge index. First compute on horizontal edge
        for(i = 0; i < gmax[0]; i++)
        {
            ic[0] = i; 
            for(j = 0; j <= gmax[1]; j++)
            {
                ic[1] = j; 
                vn = n_vel_h_edge(ic, ns_field);
                if(vn > 0.0)
                {
                    for(k = 0; k < dim; k++)
                        h_line_vel(ic,ns_field).v[k] = h_line_vel_L(ic,ns_field).v[k];
                    if(ns_field->n_chem != 0) 
                    {
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            h_line_chem(ic,ns_field->chem_sten)[k] = 
                                   h_line_chem_L(ic,ns_field->chem_sten)[k];
                        }
                    }
                }
                else if (0.0 == vn)
                {
                    for(k = 0; k < dim; k++)
                        h_line_vel(ic,ns_field).v[k] = 0.5*(h_line_vel_L(ic,ns_field).v[k] + 
                                                            h_line_vel_R(ic,ns_field).v[k]);
                    if(ns_field->n_chem != 0)
                    {
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            h_line_chem(ic,ns_field->chem_sten)[k] = 0.5*(
                                   h_line_chem_L(ic,ns_field->chem_sten)[k]+
                                   h_line_chem_R(ic,ns_field->chem_sten)[k]);
                        }
                    }
                }
                else
                {
                    for(k = 0; k < dim; k++)
                        h_line_vel(ic,ns_field).v[k] = h_line_vel_R(ic,ns_field).v[k];
                    if(ns_field->n_chem != 0)
                    {
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            h_line_chem(ic,ns_field->chem_sten)[k] =
                                   h_line_chem_R(ic,ns_field->chem_sten)[k];
                        }
                    }
                }

                // TMP
                /***
                if(ic[0] == 0 && (ic[1] == 9 || ic[1] == 1))
                {
                    printf("upwinding_on_edge: ic[%d %d] h_line_vel.v[0] = %10.9g, vn= %g vn == 0.0 ? %d\n",
                                    ic[0], ic[1], h_line_vel(ic,ns_field).v[0], vn, vn == 0.0);
                    printf("h_line_vel_L.v[0] = %10.9g, h_line_vel_R.v[0] = %10.9g\n",
                                 h_line_vel_L(ic,ns_field).v[0], h_line_vel_R(ic,ns_field).v[0]);
                }
                ***/
            }
        }
        // ic[] is edge index. Now compute on vertical edge
        for(i = 0; i <= gmax[0]; i++)
        {
            ic[0] = i;
            for(j = 0; j < gmax[1]; j++)
            {
                ic[1] = j;
                vn = n_vel_v_edge(ic, ns_field);

                //  TMP
                /**
                if(i == 0 || i == gmax[0])
                {
                    if(j == 3)
                    {
                        printf("upwinding_on_edge, vertical edge[%d %d], vn = %9.8g  =0.0? %d\n",
                             i, j, vn, vn == 0.0);
                    } 
                }
                **/
               
                if(vn > 0.0)
                {
                    for(k = 0; k < dim+1; k++)
                        v_line_vel(ic,ns_field).v[k] = v_line_vel_L(ic,ns_field).v[k];
                    if(ns_field->n_chem != 0)
                    {
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            v_line_chem(ic,ns_field->chem_sten)[k] =
                                   v_line_chem_L(ic,ns_field->chem_sten)[k];
                        }
                    }
                }
                else if (0.0 == vn)
                {
                    for(k = 0; k < dim; k++)
                        v_line_vel(ic,ns_field).v[k] = 0.5*(v_line_vel_L(ic,ns_field).v[k] + 
                                                            v_line_vel_R(ic,ns_field).v[k]);
                    // TMP
                    /**
                    if((ic[0] == 1) && ic[1] == 0)
                    {
                        printf("\n\n*******************\n");
                        printf("In upwinding_on_edge, v_line[%d %d] vel (%g %g), L(%g %g). R(%g %g)\n",
                                 ic[0], ic[1],v_line_vel(ic,ns_field).v[0], v_line_vel(ic,ns_field).v[1],
                                    v_line_vel_L(ic,ns_field).v[0], v_line_vel_L(ic,ns_field).v[1],
                              v_line_vel_R(ic,ns_field).v[0], v_line_vel_R(ic,ns_field).v[1]); 
                    }
                    **/
                    if(ns_field->n_chem != 0)
                    {
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            v_line_chem(ic,ns_field->chem_sten)[k] = 0.5*(
                                   v_line_chem_L(ic,ns_field->chem_sten)[k] + 
                                   v_line_chem_R(ic,ns_field->chem_sten)[k]);
                        }
                    }
                }
                else
                {
                    for(k = 0; k < dim; k++)
                        v_line_vel(ic,ns_field).v[k] = v_line_vel_R(ic,ns_field).v[k];
                    if(ns_field->n_chem != 0)
                    {
                        for(k = 0; k < ns_field->n_chem; k++)
                        {
                            v_line_chem(ic,ns_field->chem_sten)[k] = 
                                   v_line_chem_R(ic,ns_field->chem_sten)[k];
                        }
                    }
                }
            }
        }

        // TMP
        if(debugging("NS_vel") && NO)
        {
            printf("\n\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n\n");
            printf("upwinding_on_edge*****vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            // for(j = 10; j <= 13; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                // for(i = 9; i < min(12,gmax[0]); i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d, %g %g;", i, 
                        h_line_vel(ic, ns_field).v[0], h_line_vel(ic, ns_field).v[1]);
                        // h_line_vel(ic, ns_field).v[2]);
                }
                /**
                ic[0] = gmax[0]-1; ic[1] = j;
                printf(" (%d, %g %g) ", ic[0],
                        h_line_vel(ic, ns_field).v[0], h_line_vel(ic, ns_field).v[1]);
                */
                printf("\n");
            }
            printf("upwinding_on_edge*****show vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            // for(j = 10; j <= 13; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                // for(i = 9; i <= min(12,gmax[0]); i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d, %g %g;", i, 
                       v_line_vel(ic, ns_field).v[0],v_line_vel(ic, ns_field).v[1]);
                       // v_line_vel(ic, ns_field).v[2]);
                }
                /*
                ic[0] = gmax[0]; ic[1] = j;
                printf(" (%d, %g %g)", ic[0],
                      v_line_vel(ic, ns_field).v[0],v_line_vel(ic, ns_field).v[1]);
                */
                printf("\n");
            }
            printf("\n\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n\n");
        }
}

// Only compute for the interior cells
// normal advective velocity on the cell edge
LOCAL void nor_advec_vel_on_edge(
        float      *dh,
        float      dt,
        Wave       *wv,
        Wave       *nwv,
        Front      *fr,
        Front      *nfr,
        COMPONENT  max_comp,
        NS_Field_Stencil *ns_field)
{
        int        i, j, k;
        int        *gmax, ic[MAXD];
        float      vl, vr;
        
        gmax = wv->rect_grid->gmax;
        
        // ic[] is edge index. First compute on horizontal edge
        for(i = 0; i < gmax[0]; i++)
        {
            ic[0] = i; 
            for(j = 0; j <= gmax[1]; j++)
            {
                ic[1] = j; 
                vl = h_line_vel_L(ic, ns_field).v[1];
                vr = h_line_vel_R(ic, ns_field).v[1];
 
                // FOR RT problem: Apply Neumann BC here
                if(j == 0 || j == gmax[1])
                {
                    vl = h_line_vel_L(ic, ns_field).v[1] = 0.0;
                    vr = h_line_vel_R(ic, ns_field).v[1] = 0.0;
                }

                n_vel_h_edge(ic, ns_field) = colella_limiting_vel(vl, vr);
                // n_vel_h_edge(ic, ns_field) = 0.5*(vl + vr);

                /**
                if((i == 9 && (j == 2 || j == 8)))
                {
                    printf("colella_limiting (%d %d, vl, vr %10.9g %10.9g) ", ic[0],ic[1],vl, vr);
                    printf("nor_v_h_edge = %10.9g\n", colella_limiting_vel(vl, vr));
                }
                **/
            }
        }

        // Compute on vertical edge
        for(i = 0; i <= gmax[0]; i++)
        {
            ic[0] = i; 
            for(j = 0; j < gmax[1]; j++)
            {
                ic[1] = j; 
                vl = v_line_vel_L(ic, ns_field).v[0];
                vr = v_line_vel_R(ic, ns_field).v[0];
                // TMP
                // printf("(%d %d, vl, vr %g %g) ", ic[0],ic[1],vl, vr);

                n_vel_v_edge(ic, ns_field) = colella_limiting_vel(vl, vr);
                // n_vel_v_edge(ic, ns_field) = 0.5*(vl + vr);

                // printf("nor_v_v_edge = %g\n", colella_limiting_vel(vl, vr));

                //  TMP
                /**
                if(i == 1 && j == 0)
                {
                    printf("nor_advec_vel_on_edge, vertical edge[%d %d], vn %9.8g,"
                                     " vl-vr(%14.13g, %14.13g), vl + vr = %14.13g 100.0*MACH_EPS %14.13g\n",
                             i, j, colella_limiting_vel(vl, vr), vl, vr,
                             vl + vr, 5000.0*MACH_EPS);
                }
                **/
                // END TMP
            }
        }

        // TMP
        if(debugging("NS_vel"))
        {
            printf("nor_advec_vel_on_edge******normal vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            // for(j = 3; j <= 5; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                // for(i = 0; i < gmax[0]/2; i++)
                {
                    int ic2[MAXD];
                    ic[0] = i; 
                    ic[1] = j; 
                    ic2[0] = gmax[0]-i-1;
                    ic2[1] = j;
                    // printf(" (%d %g)", i, fabs(n_vel_h_edge(ic, ns_field) - 
                    //                            n_vel_h_edge(ic2, ns_field)));
                    printf(" (%d %g)", i, (n_vel_h_edge(ic, ns_field)));
                }
                printf("\n");
            }
            printf("nor_advec_vel_on_edge*****normal vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            // for(j = 10; j <= 13; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i <= gmax[0]; i++)
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    int ic2[MAXD];
                    ic[0] = i; 
                    ic[1] = j; 
                    ic2[0] = gmax[0]-i;
                    ic2[1] = j;
                    // printf(" (%d %g)", i, fabs(n_vel_v_edge(ic, ns_field) + 
                    //                            n_vel_v_edge(ic2, ns_field)));
                    printf(" (%d %g)", i, n_vel_v_edge(ic, ns_field));
                }
                // ic[0] = gmax[0]; ic[1] = j;
                // printf(" (%d, %g) ", i, n_vel_v_edge(ic, ns_field));
                printf("\n");
            }
        }
}

// in axis direction, lower side is L, upper side is R
LOCAL float colella_limiting_vel(
        float     vl,
        float     vr)
{
        if(isnan(vl) || isnan(vr))
        {
            printf("ERROR: in colella_limiting_vel, nan vl %g, vr %g\n",vl, vr);
            clean_up(ERROR);
        }

        // return 0.5*(vl+vr);

        // To enforce symmetry
        /**
        if(fabs(vl + vr) < 5000.0*MACH_EPS)
        {
            // printf("WARNING: in colella_limiting_vel, vl + vr < %16.15f\n",10.0*MACH_EPS);
            // printf("vl = %15.14f, vr = %15.14f\n", vl, vr);
            return 0.0;
        }
        **/

        if(vl > 0.0 && (vl + vr) > 0.0)
            return vl;
        if(vl <= 0.0 && (vr >= 0.0 || (vl + vr) ==0.0) )
            return 0.0;
        if(vr < 0.0 && (vl + vr) < 0.0)
            return vr;
        if(vr <= 0.0 && (vl >= 0.0 || (vl + vr) ==0.0) )
            return 0.0;

        printf("ERROR: in colella_limiting_vel, unexpected case\n");
        printf("vl = %14.12f, vr = %14.12f\n", vl, vr);
        clean_up(ERROR);
}


LOCAL NS_Field_Stencil *alloc_NS_field_space(
	NS_Field_Stencil      *field,
	Wave                  *wv)
{
	int       dim, nrad, i;
	RECT_GRID *gr = wv->rect_grid;
	NS_Field_Stencil     *ns_field;
        int        edgexmax, edgeymax, edgex_n_segs, edgey_n_segs;
        int        *gmax, ix, iy;

	dim = gr->dim;

        if(field == NULL)
	{
	    scalar(&ns_field, sizeof(NS_Field_Stencil));
            zero_scalar(ns_field,sizeof(NS_Field_Stencil));

	    /* Temp. set to use 2nd order scheme */
            ns_field->npts_sten = 5;

	    nrad = (ns_field->npts_sten - ns_field->npts_sten%2)/2;

            switch(dim)
            {
            case 2:
                edgexmax = gr->gmax[0]+max(gr->lbuf[0],4)+max(gr->ubuf[0],4);
                edgeymax = gr->gmax[1]+max(gr->lbuf[1],4)+max(gr->ubuf[1],4);
                edgex_n_segs = edgexmax*(edgeymax+1);
                edgey_n_segs = edgeymax*(edgexmax+1);

                ns_field->nnx = edgexmax;
                ns_field->nny = edgeymax;

                // storage for reconstructed half time step vel on cell edge
                vector(&(ns_field->u_states[0]),edgex_n_segs,sizeof(VEL_ST));
                vector(&(ns_field->u_states[1]),edgey_n_segs,sizeof(VEL_ST));

                ns_field->cg_u_states[0] = ns_field->u_states[0] +
                    edgexmax*lbuf[1] + lbuf[0];
                ns_field->cg_u_states[1] = ns_field->u_states[1] +
                    (edgexmax+1)*lbuf[1] + lbuf[0];
                ns_field->alloc.u_states = YES;

                // storage for edge flags
                // vector(&(ns_field->edge_type[0]),edgex_n_segs,sizeof(EDGE_TYPE));
                // vector(&(ns_field->edge_type[1]),edgey_n_segs,sizeof(EDGE_TYPE));

                // ns_field->cg_edge_type[0] = ns_field->edge_type[0] +
                //     edgexmax*gr->lbuf[1] + gr->lbuf[0];
                // ns_field->cg_edge_type[1] = ns_field->edge_type[1] +
                //     (edgexmax+1)*gr->lbuf[1] + gr->lbuf[0];
                // ns_field->alloc.edge_type = YES;

                // storage for saving normal component of velocity on cell edge
                vector(&(ns_field->n_vel_edge[0]),edgex_n_segs,sizeof(float));
                vector(&(ns_field->n_vel_edge[1]),edgey_n_segs,sizeof(float));
                ns_field->cg_n_vel_edge[0] = ns_field->n_vel_edge[0] +
                    edgexmax*lbuf[1] + lbuf[0];
                ns_field->cg_n_vel_edge[1] = ns_field->n_vel_edge[1] +
                    (edgexmax+1)*lbuf[1] + lbuf[0];
                ns_field->alloc.n_vel_edge = YES;
                
                // storage for velocities on edge
                vector(&(ns_field->uL_states[0]),edgex_n_segs,sizeof(VEL_ST));
                vector(&(ns_field->uL_states[1]),edgey_n_segs,sizeof(VEL_ST));

                ns_field->cg_uL_states[0] = ns_field->uL_states[0] +
                    edgexmax*lbuf[1] + lbuf[0];
                ns_field->cg_uL_states[1] = ns_field->uL_states[1] +
                    (edgexmax+1)*lbuf[1] + lbuf[0];
                ns_field->alloc.uL_states = YES;

                vector(&(ns_field->uR_states[0]),edgex_n_segs,sizeof(VEL_ST));
                vector(&(ns_field->uR_states[1]),edgey_n_segs,sizeof(VEL_ST));

                ns_field->cg_uR_states[0] = ns_field->uR_states[0] +
                    edgexmax*lbuf[1] + lbuf[0];
                ns_field->cg_uR_states[1] = ns_field->uR_states[1] +
                    (edgexmax+1)*lbuf[1] + lbuf[0];
                ns_field->alloc.uR_states = YES;

                vector(&(ns_field->Gp),edgexmax*edgeymax,sizeof(Gradient_Var));
                ns_field->cg_Gp = ns_field->Gp + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.Gp = YES;
                vector(&(ns_field->phi),edgexmax*edgeymax,sizeof(float));
                ns_field->cg_phi = ns_field->phi + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.phi = YES;

                vector(&(ns_field->vlap[0]),edgexmax*edgeymax,sizeof(float));
                vector(&(ns_field->vlap[1]),edgexmax*edgeymax,sizeof(float));
                ns_field->cg_vlap[0] = ns_field->vlap[0] + edgexmax*lbuf[1] + lbuf[0];
                ns_field->cg_vlap[1] = ns_field->vlap[1] + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.vlap = YES; 

                // convective derivative
                vector(&(ns_field->con_der),edgexmax*edgeymax,sizeof(VEL_ST));
                ns_field->cg_con_der = ns_field->con_der + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.con_der = YES;

                // Density and viscosity
                vector(&(ns_field->rho),edgexmax*edgeymax,sizeof(float));
                ns_field->cg_rho = ns_field->rho + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.rho = YES;
                vector(&(ns_field->mu),edgexmax*edgeymax,sizeof(float));
                ns_field->cg_mu = ns_field->mu + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.mu = YES;
                // U^star velocity 
                vector(&(ns_field->u_star),edgexmax*edgeymax,sizeof(VEL_ST));
                ns_field->cg_u_star = ns_field->u_star + edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.u_star = YES;

                // storage for 1/rho on cell edge
                vector(&(ns_field->rho_recip[0]),edgex_n_segs,sizeof(float));
                vector(&(ns_field->rho_recip[1]),edgey_n_segs,sizeof(float));

                ns_field->cg_rho_recip[0] = ns_field->rho_recip[0] +
                    edgexmax*lbuf[1] + lbuf[0];
                ns_field->cg_rho_recip[1] = ns_field->rho_recip[1] +
                    (edgexmax+1)*lbuf[1] + lbuf[0];
                ns_field->alloc.rho_recip = YES;
                
                vector(&(ns_field->hr_st),
                             edgex_n_segs,sizeof(float)); 
                ns_field->cg_hr_st = ns_field->hr_st +
                                edgexmax*lbuf[1] + lbuf[0];
                ns_field->alloc.hr_st = YES;

                vector(&(ns_field->vr_st),
                             edgey_n_segs,sizeof(float));
                ns_field->cg_vr_st = ns_field->vr_st +
                                (edgexmax+1)*lbuf[1] + lbuf[0];
                ns_field->alloc.vr_st = YES;
            break;
            default:
                printf("ERROR: in g_FD_NS_solver, implement alloc sten for dim %d\n",dim);
                clean_up(ERROR);
            }
	    return ns_field;
	}	
	return field;
}

LOCAL Chem_Stencil *alloc_Chem_field_sten(
	Chem_Stencil      *field,
	Wave              *wv,
        int               n_chem)
{
	int       dim, i;
	RECT_GRID *gr = wv->rect_grid;
	Chem_Stencil     *chem_sten;
        int        edgexmax, edgeymax, edgex_n_segs, edgey_n_segs;
        int        *gmax, *lbuf, *ubuf, ix, iy;
        int smin[MAXD], smax[MAXD];
        float  **state;
        byte   *storage;

	dim = gr->dim;

        if(field == NULL)
	{
            scalar(&(chem_sten),sizeof(Chem_Stencil));
            
	    /* Temp. set to use 2nd order scheme */
            chem_sten->n_chem = n_chem;

            switch(dim)
            {
            case 2:
                edgexmax = gr->gmax[0]+lbuf[0]+ubuf[0];
                edgeymax = gr->gmax[1]+lbuf[1]+ubuf[1];
                edgex_n_segs = edgexmax*(edgeymax+1);
                edgey_n_segs = edgeymax*(edgexmax+1);

                chem_sten->nnx = edgexmax;
                chem_sten->nny = edgeymax;

                for(i = 0; i < dim; i++)
                {
                    smin[i] = -lbuf[i];
                    smax[i] = gr->gmax[i] + ubuf[i];
                }
                chem_sten->nnx = edgexmax;
                chem_sten->nny = edgeymax;

                vector(&(chem_sten->hedge_l_storage),
                                 edgex_n_segs,sizeof(float)*n_chem); 
                vector(&(chem_sten->hedge_r_storage),
                                 edgex_n_segs,sizeof(float)*n_chem); 
                vector(&(chem_sten->hel_st),edgex_n_segs,sizeof(float*));
                chem_sten->cg_hel_st = chem_sten->hel_st +
                               edgexmax*lbuf[1] + lbuf[0];
                vector(&(chem_sten->her_st),edgex_n_segs,sizeof(float*));
                chem_sten->cg_her_st = chem_sten->her_st +
                                edgexmax*lbuf[1] + lbuf[0];

                    // h-edge middle state
                vector(&(chem_sten->hedge_m_storage),
                                     edgex_n_segs,sizeof(float)*n_chem); 
                vector(&(chem_sten->hem_st),edgex_n_segs,sizeof(float*));
                chem_sten->cg_hem_st = chem_sten->hem_st +
                                edgexmax*lbuf[1] + lbuf[0];

                state = chem_sten->cg_hel_st;
                storage = chem_sten->hedge_l_storage;
                for(iy = smin[1]; iy <= smax[1]; iy++)
                {
                    for(ix = smin[0]; ix < smax[0]; ix++)
                    {
                        *state = (float*)storage; 
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }
                state = chem_sten->cg_her_st;
                storage = chem_sten->hedge_r_storage;
                for(iy = smin[1]; iy <= smax[1]; iy++)
                {
                    for(ix = smin[0]; ix < smax[0]; ix++)
                    {
                        *state = (float*)storage; 
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }
                state = chem_sten->cg_hem_st;
                storage = chem_sten->hedge_m_storage;
                for(iy = smin[1]; iy <= smax[1]; iy++)
                {
                    for(ix = smin[0]; ix < smax[0]; ix++)
                    {
                        *state = (float*)storage; 
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }

                    // storage on vertical edge
                vector(&(chem_sten->vedge_l_storage),
                                     edgey_n_segs,sizeof(float)*n_chem);
                vector(&(chem_sten->vedge_r_storage),
                                     edgey_n_segs,sizeof(float)*n_chem);
                vector(&(chem_sten->vel_st),edgey_n_segs,sizeof(float*));
                vector(&(chem_sten->ver_st),edgey_n_segs,sizeof(float*));

                chem_sten->cg_vel_st = chem_sten->vel_st +
                            (edgexmax+1)*lbuf[1] + lbuf[0];
                chem_sten->cg_ver_st = chem_sten->ver_st +
                                (edgexmax+1)*lbuf[1] + lbuf[0];

                    // v-edge middle state
                vector(&(chem_sten->vedge_m_storage),
                                     edgey_n_segs,sizeof(float)*n_chem);
                vector(&(chem_sten->vem_st),edgey_n_segs,sizeof(float*));
                chem_sten->cg_vem_st = chem_sten->vem_st +
                                (edgexmax+1)*lbuf[1] + lbuf[0];
                  
                state = chem_sten->cg_vel_st;
                storage = chem_sten->vedge_l_storage;
                for(iy = smin[1]; iy < smax[1]; iy++)
                {
                    for(ix = smin[0]; ix <= smax[0]; ix++)
                    {
                        *state = (float*)storage;
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }
                state = chem_sten->cg_ver_st;
                storage = chem_sten->vedge_r_storage;
                for(iy = smin[1]; iy < smax[1]; iy++)
                {
                    for(ix = smin[0]; ix <= smax[0]; ix++)
                    {
                        *state = (float*)storage;
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }
                state = chem_sten->cg_vem_st;
                storage = chem_sten->vedge_m_storage;
                for(iy = smin[1]; iy < smax[1]; iy++)
                {
                    for(ix = smin[0]; ix <= smax[0]; ix++)
                    {
                        *state = (float*)storage;
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }
                // convective derivatives
                vector(&(chem_sten->con_der_storage),edgexmax*edgeymax,
                                    sizeof(float)*n_chem);
                vector(&(chem_sten->con_der),edgexmax*edgeymax,sizeof(float*));
                chem_sten->cg_con_der = chem_sten->con_der + edgexmax*lbuf[1] + lbuf[0];

                state = chem_sten->con_der;
                storage = chem_sten->con_der_storage;
                for(iy = smin[1]; iy < smax[1]; iy++)
                {
                    for(ix = smin[0]; ix < smax[0]; ix++)
                    {
                        *state = (float*)storage;
                        state++;
                        storage += sizeof(float)*n_chem;
                    }
                }

            break;
            default:
                printf("ERROR: in alloc_Chem_field_sten, implement for dim %d\n",dim);
                clean_up(ERROR);
            }
	    return chem_sten;
	}	
	return field;
}

LOCAL void irregular_pt_nor_deriv(
        int             swp_num,
        int             *iperm,
        float           *dh,
        float           dt,
        Wave            *wv,
        Wave            *newwv,
        Front           *fr,
        Front           *newfr,
        NS_Field_Stencil  *ns_field)
{
        POINT           Basep;
        int             i, idirs[MAXD];
        int             imin[3], imax[3];
        static  Stencil *sten = NULL;   // stencil for hyperbolic should work as well.
        static  int     num_pts;
        static  int     **icoords = NULL;
        RECT_GRID       *gr = wv->rect_grid;
        COMPONENT       max_comp = max_component(fr->interf);

        /* SET LOCAL GLOBAL */
        // dim = gr->dim;
        // intfc = fr->interf;
        // ext_comp = exterior_component(intfc);

        for (i = 0; i < dim; ++i)
            idirs[i] = iperm[(i+swp_num)%dim];

        /* SET LOCAL GLOBAL */
        hdir = dh[idirs[0]];     /* grid spacing in sweep direction */
        hndir = dh[idirs[1]];    /* in the other direction */

        if (sten == NULL)
        {
            /* This assumes that the stencil size is constant. */
            num_pts = ns_field->npts_sten;
            endpt = stencil_radius(ns_field); // = 2

            sten = alloc_stencil(num_pts,fr);
#if defined(TWOD) || defined(THREED)
            icoords = sten->icoords;
#endif /* defined(TWOD) || defined(THREED) */
        }

        switch (idirs[0])
        {
        case 0: /*X SWEEP */
            prev_side = WEST;
            next_side = EAST;
            if (debugging("irregular_pt"))
                (void) printf("prev_side = WEST, next_side = EAST\n");
            break;
        case 1: /*Y SWEEP */
            prev_side = SOUTH;
            next_side = NORTH;
            if (debugging("irregular_pt"))
                (void) printf("prev_side = SOUTH, next_side = NORTH\n");
            break;
        case 2: /* Z SWEEP */
            prev_side = LOWER;
            next_side = UPPER;
            if (debugging("irregular_pt"))
                (void) printf("prev_side = LOWER, next_side = UPPER\n");
            break;
        }

        ns_field->pt_sten = sten;

        sten->fr   = fr;        sten->newfr   = newfr;
        sten->wave = wv;      sten->newwave = newwv;
        sten->swap_wv = NULL;

        // Set sweep limits
        set_sweep_limits(wv,swp_num,idirs,imin,imax);
        if (gr->lbuf[idirs[1]] > 0)
            imin[1] += 1;
        if (gr->ubuf[idirs[1]] > 0)
            imax[1] -= 1;

        switch (dim)
        {
#if defined(TWOD)
        case 2:
        {
            int i0, i1;
            int i0max, i1max;
            int i0min, i1min;

            i0min = imin[0];    i0max = imax[0];
            i1min = imin[1];    i1max = imax[1];
            for (i1 = i1min; i1 < i1max; ++i1)
            {
                // set_static_coords(icoords,&Basep,idirs[1],i1,gr);

                // set set_static_coords 
                for (i = -endpt; i <= endpt; ++i)
                    icoords[i][idirs[1]] = i1;
                COORDS(Basep)[idirs[1]] = cell_center(i1,idirs[1],gr);

                sten->reg_stencil = NO;
                for (i0 = i0min; i0 < i0max; ++i0)
                {
                    NS_reg_grid_state(dh,dt,swp_num,iperm,
                                          i0,ns_field,&Basep,max_comp);
                }
            }
            break;
        }
#endif /* defined(TWOD) */
        }

        // reset pointers
        ns_field->pt_sten = NULL;
}

// Compute mix cell, and set edge types
LOCAL   void NS_reg_grid_state(
        float           *dh,
        float           dt,
        int             swp_num,
        int             *iperm,
        int             is,
        NS_Field_Stencil  *ns_field,
        POINT           *basep,
        COMPONENT       max_comp)
{
        Wave            *wave = ns_field->pt_sten->wave;
        Wave            *newwave = ns_field->pt_sten->newwave;
        Front           *fr = ns_field->pt_sten->fr;
        Front           *newfr = ns_field->pt_sten->newfr;
        COMPONENT       *comp;
        COMPONENT       new_comp;     /* mid comp wrt newfr */
        HYPER_SURF      *hs;
        int             i, j, index, idir = iperm[swp_num];
        int             **icoords = ns_field->pt_sten->icoords, ic[MAXD];
        Stencil         *sten = ns_field->pt_sten;

#if defined(DEBUG_NPT)
        static  char fname[3][11] = {"x_npt()","y_npt()","z_npt()"};
#endif /* if defined(DEBUG_NPT) */

        for (i = -endpt; i <= endpt; ++i)
            icoords[i][idir] = is + i;
        sten->prev_reg_stencil = sten->reg_stencil;

#if defined(DEBUG_NPT)
        // if((icoords[0][0] == 61 && icoords[0][1] == 12)
        /**
        if((
            (icoords[0][0] == 0 && icoords[0][1] == 0) ) &&
            (idir == 0 || idir == 1)
           )
        {
            (void) printf("\n%s - ",fname[idir]);
            print_int_vector("icoords[0] = ",icoords[0],dim,"\n");
            deb_npt = YES;
            // add_to_debug("g_neumann");   
        }
        else
        {
            deb_npt = NO;
            // remove_from_debug("g_neumann");   
        }
        **/
#endif /* if defined(DEBUG_NPT) */

                /* Find components and crosses*/

        new_comp = set_NS_stencil_comps_and_crosses(idir,swp_num,iperm,sten,endpt);
        comp = sten->comp;

#if defined(DEBUG_NPT)
        if (deb_npt)
        {
            int side;
            (void) printf("\n%s - ",fname[idir]);
            print_int_vector("icoords[0] = ",icoords[0],dim,"\n");
            (void) printf("comps: ");
            for (i = -endpt; i <= endpt; ++i)
                (void) printf("comp[%d] %d ",i,comp[i]);
            (void) printf("new_comp %d\n",new_comp);
            for(side = 0; side <= 1; side++)
            {
                for (i = -1; i <= 1; ++i)
                    printf("Tcomp[%d][%d] %d ",side, i,sten->Tcomp[side][i]);
                printf("\n");
            }
        }
#endif /* if defined(DEBUG_NPT) */

        if (NS_interior_sten_reg(is,idir,swp_num,iperm,sten))
        {
            // set edge types here
            if(icoords[0][0] >=0 && icoords[0][0] < ns_field->rect_gr->gmax[0] &&
               icoords[0][1] >=0 && icoords[0][1] < ns_field->rect_gr->gmax[1])
            {
                for(i = 0; i < dim; i++)
                    ic[i] = icoords[0][i];

                if(idir == 0)
                {
                    // look x-axis direction
                    // v_edge_flag(icoords[0],ns_field) = REG_E;
                    ic[0]++;
                    // v_edge_flag(ic,ns_field) = REG_E;
                }
                else if(idir == 1)
                {
                    // look y-axis direction
                    // h_edge_flag(icoords[0],ns_field) = REG_E;
                    ic[1]++;
                    // h_edge_flag(ic,ns_field) = REG_E;
                }
            }

            if(deb_npt == YES)
                goto jump_into_deb;
   
            return;  /*Don't overwrite vector solution*/
        }
        else
            sten->prev_reg_stencil = NO;

jump_into_deb:

        for (j = 0; j < dim; ++j)
        {
            if (j == idir)
            {
                Coords(sten->p[0])[idir] = cell_center(is,idir,wave->rect_grid);
                for (i = 1; i <= endpt; ++i)
                {
                    Coords(sten->p[-i])[idir] = Coords(sten->p[0])[idir]-i*hdir;
                    Coords(sten->p[i])[idir]  = Coords(sten->p[0])[idir]+i*hdir;
                }
            }
            else
            {
                for (i = -endpt; i <= endpt; ++i)
                    Coords(sten->p[i])[j] = Coords(basep)[j];
            }
        }
        sten->reg_stencil = YES;
        if (RegionIsFlowSpecified(Rect_state(icoords[0],newwave),
                                  Rect_state(icoords[0],wave),
                                  Coords(sten->p[0]),new_comp,comp[0],fr))
        {
            sten->reg_stencil = NO;
            return;
        }

        if ((new_comp > max_comp) && (swp_num > 0))
        {
            sten->reg_stencil = NO;
            assign(Rect_state(icoords[0],newwave),
                   Rect_state(icoords[0],wave),fr->sizest);
            return;
        }

        if (new_comp > max_comp)
        {
            sten->reg_stencil = NO;
            nearest_intfc_state_and_pt(Coords(sten->p[0]),new_comp,newfr,fr,
                                       sten->worksp[0],Coords(sten->p[0]),&hs);
            assign(Rect_state(icoords[0],newwave),sten->worksp[0],fr->sizest);
            return;
        }

        if ((new_comp  == POTS_COMP))
        {
            sten->reg_stencil = NO;
            assign(Rect_state(icoords[0],newwave),
                   Rect_state(icoords[0],wave),fr->sizest);
            return;
        }

                /* Find mid state */
        if (equivalent_comps(comp[0],new_comp,newfr->interf) == YES)
            sten->st[0] = Rect_state(icoords[0],wave);
        else
        {
            /**
            nearest_intfc_state_and_pt(Coords(sten->p[0]),new_comp,fr,
                                       newfr,sten->worksp[0],
                                       Coords(sten->p[0]),&hs);
            sten->st[0] = sten->worksp[0];
            sten->reg_stencil = NO;
            **/ 
            // Use velocity at the grid point
            // Do not see crxings of CONTACT CURVE
            nearest_intfc_state_and_pt(Coords(sten->p[0]),new_comp,fr,
                                       newfr,sten->worksp[0],
                                       Coords(sten->p[0]),&hs);
            sten->st[0] = sten->worksp[0];
            sten->reg_stencil = NO;

            assign(left_state(sten->p[0]),sten->st[0],fr->sizest);
            assign(sten->st[0],Rect_state(icoords[0],wave),fr->sizest);
            Set_params(sten->st[0],left_state(sten->p[0]));
            // Dens(sten->st[0]) = Params(sten->st[0])->dens;
        }
        assign(left_state(sten->p[0]),sten->st[0],fr->sizest);
        assign(right_state(sten->p[0]),sten->st[0],fr->sizest);

        find_outer_ns_states(is,idir,sten,PREV);
        find_outer_ns_states(is,idir,sten,NEXT);

	find_out_ns_states_in_other_dir(is,idir,swp_num,iperm,sten);

#if defined(DEBUG_NPT)
        if (deb_npt)
        {
            int side;
            (void) printf("STENCIL: "); print_Stencil(sten);
            for (i = -endpt; i <= endpt; ++i)
            {
                (void) printf("state[%d]:       ",i);
                (*fr->print_state)(sten->st[i]);
            }
            
            for(side = 0; side <= 1; side++)
            {
                i = -1;
                for (i = -1; i <= 1; ++i)
                {
                    printf("--Tcomp[%d][%d] %d ",side, i,sten->Tcomp[side][i]);
                    printf("\n");
                    (*fr->print_state)(sten->Tst[side][i]);
                }
            }
        }
#endif /* if defined(DEBUG_NPT) */
	
        index = (wave->num_point_sources) ?
            is_source_block(wave,newfr->interf,new_comp,icoords[0]) : -1;

        // set edge types here
        /***
        if(icoords[0][0] >=0 && icoords[0][0] < ns_field->rect_gr->gmax[0] &&
           icoords[0][1] >=0 && icoords[0][1] < ns_field->rect_gr->gmax[1])
        {
            for(i = 0; i < dim; i++)
                ic[i] = icoords[0][i];

            if(idir == 0)
            {
                ic[0]++;
            }
            else if(idir == 1)
            {
                ic[1]++;
            }
        }
        ***/ 

        
        // TMP
#if defined(DEBUG_NPT)
        if(deb_npt)
            printf("Before enter npt_normal_vari\n");
#endif /* if defined(DEBUG_NPT) */

        /**
        printf("irregular point enter npt_normal_vari\n");
        (void) printf("\n%s - ",fname[idir]);
        print_int_vector("icoords[0] = ",icoords[0],dim,"\n");
        **/

        npt_normal_vari(dh,dt,Rect_state(icoords[0],newwave),dir,
              swp_num,iperm,&index,ns_field);
}

LOCAL void npt_normal_vari(
        float           *dh,
        float           dt,
        Locstate        ans,
        const float     *dir,
        int             swp_num,
        int             *iperm,
        int             *index,
        NS_Field_Stencil    *ns_field)
{
        Stencil         *sten = ns_field->pt_sten;
        Front           *fr = sten->fr;
        Front           *newfr = sten->newfr;
        RECT_GRID       *gr = fr->rect_grid;
        Locstate        st, *state;
        Wave            *wave = sten->wave;
        Wave            *newwave = sten->newwave;
        float           *rho, *p;
        float           *m[MAXD], *v[MAXD];
        float           **coords;
#if !defined(UNRESTRICTED_THERMODYNAMICS)
        float           *vacuum_dens;
        float           *min_pressure;
        float           *min_energy;
#endif /* !defined(UNRESTRICTED_THERMODYNAMICS) */
#if defined(ROTATIONAL_SYMMETRY)
        static float    alpha;
#endif /* defined(ROTATIONAL_SYMMETRY) */
        int             i, j, side;
        int             dim = gr->dim;
        int             idirs[MAXD];
        static Vec_NS_Gas  *vst = NULL;
        // static Vec_NS_Src  *src = NULL;
        static Vec_NS_Gas  *Tvst[2] = {NULL, NULL};
        static int         npts;
        // Source
        static Vec_Src  *src = NULL;
        // static float    **Q = NULL;

        if (is_obstacle_state(sten->st[0]))
        {
            g_obstacle_state(ans,sten->fr->sizest);
            return;
        }

        if(vst == NULL)
        {
            npts = ns_field->npts_sten;

            vst = alloc_ns_vgas(vst,npts,dim,ns_field->n_chem);
            Tvst[0] = alloc_ns_vgas(Tvst[0],npts,dim,ns_field->n_chem);
            Tvst[1] = alloc_ns_vgas(Tvst[1],npts,dim,ns_field->n_chem);
            src = muscl_alloc_vsrc(src,npts,wave);
        }

        clear_vec_ns_gas_set_flags(vst);
        clear_vec_ns_gas_set_flags(Tvst[0]);
        clear_vec_ns_gas_set_flags(Tvst[1]);

        if (RegionIsFlowSpecified(ans,sten->st[0],Coords(sten->p[0]),
                    sten->newcomp,sten->newcomp,fr))
            return;

        for (i = 0; i < dim; ++i)
        {
            idirs[i] = iperm[(i+swp_num)%dim];
            m[i] = vst->m[i];
            v[i] = vst->v[i];
        }
        rho = vst->rho;
        state = vst->state;
        coords = vst->coords;
        p = vst->p;

        for (i = 0; i < npts; ++i)
        {
            state[i] = st = sten->ststore[i];
            coords[i] = Coords(sten->pstore[i]);
            rho[i] = Dens(st);
            p[i] = Press(st);
            for (j = 0; j < dim; ++j)
                v[j][i] = Vel(st)[j];
        }

        Vec_NS_Gas_field_set(vst,rho) = YES;
        Vec_NS_Gas_field_set(vst,p) = YES;
        Vec_NS_Gas_field_set(vst,v) = YES;
        Vec_NS_Gas_field_set(vst,state) = YES;
        Vec_NS_Gas_field_set(vst,coords) = YES;

        for(side = 0; side <= 1; side++)
        {
            for (i = 0; i < dim; ++i)
                v[i] = Tvst[side]->v[i];
            rho = Tvst[side]->rho;
            state = Tvst[side]->state;
            p = Tvst[side]->p;
            for(i = -1; i <= 1; i++)
            {
                state[i+2] = st = sten->Tst[side][i];
                rho[i+2] = Dens(st);
                p[i+2] = Press(st);
                for (j = 0; j < dim; ++j)
                    v[j][i+2] = Vel(st)[j];
            }
            Vec_NS_Gas_field_set(Tvst[side],rho) = YES;
            Vec_NS_Gas_field_set(Tvst[side],p) = YES;
            Vec_NS_Gas_field_set(Tvst[side],v) = YES;
            Vec_NS_Gas_field_set(Tvst[side],state) = YES;
            Vec_NS_Gas_field_set(Tvst[side],coords) = YES;
        }

        ns_field->vgas = vst;
        ns_field->Tvgas[0] = Tvst[0];
        ns_field->Tvgas[1] = Tvst[1];
        //Source
        ns_field->src = src; 

        upwind_reconstruct_in_dir(swp_num,iperm,
           wave,fr,newfr,0,npts,sten->icoords[0],dt,dh,dim,ns_field,0);

        ns_field->vgas = NULL;
        ns_field->Tvgas[0] = NULL;
        ns_field->Tvgas[1] = NULL;
        //Source
        ns_field->src = NULL; 
}


LOCAL void find_out_ns_states_in_other_dir(
        int             is,
        int             idir,
        int             swp_num,
        int             *iperm,
        Stencil         *sten)
{
        Wave       *wave = sten->wave;
        Front      *fr = sten->fr;
        Front      *newfr = sten->newfr;
        COMPONENT  **tcomp = sten->Tcomp, new_comp = sten->newcomp;
        int        **icoords = sten->icoords, ic[MAXD], ictmp[MAXD];
        float      crds[MAXD];
        int        i, j, indx, indir;
        int        in_domain, face;
        CRXING     *crx;
        HYPER_SURF *hs, *Ths;
        CRXING     *new_crx[MAX_NUM_CRX];
        CRXING     *old_crx[MAX_NUM_CRX];
        int        side, sign;
        GRID_DIRECTION  *npgdir, *pgdir, gdir[2][2] = {SOUTH, NORTH,
                                      WEST, EAST};
        static Locstate  testst = NULL; 
        RECT_GRID  *gr = wave->rect_grid; 

        int        error = NO;

        if(testst == NULL)
            alloc_state(fr->interf,&testst,fr->sizest);

        indir = iperm[(swp_num+1)%dim];
        if(iperm[swp_num%dim] == 0)
            npgdir = gdir[1];
        else
            npgdir = gdir[0];

        if(indir == 0)
            pgdir = gdir[1];
        else
            pgdir = gdir[0];

        crx = NULL;

        /* For the purpose when this orthogonal. dir.
         * passes the passive boundary, see copy_Tstates()
         */
        for(side = 0; side <=1; side++)
        {
            for(j = -1; j <= 1; j++)
            {
                assign(sten->Tworksp[side][j], sten->st[j], fr->sizest);
                sten->Tset[side][j] = NO;
            }
        }

        if(deb_npt)
        {
            printf("Entered find_out_ns_states_in_other_dir()\n");
        }

        /* First check in the orthogonal direction. */
        for(side = 0; side <= 1; side++)
        {
            if(side == 0) sign = -1;
            else sign = 1;
            for(j = -1; j <= 1; j++)
            {
                for (i = 0; i < dim; ++i)
                    ic[i] = icoords[0][i];
                ic[idir] += j; ic[indir] += sign*1;

                in_domain = crds_of_ic_ver1(crds,sten->wave,ic);

                // Use velocity at the grid point
                // Do not see crxings of CONTACT CURVE
                if(tcomp[side][j] != ext_comp &&
                   !is_excluded_comp(tcomp[side][j],fr->interf))
                {
                    if(in_domain == NO)
                    {
                        printf("ERROR find_out_states_in_other_dir\n");
                        printf("NOT in domain ic:[");
                        for(i = 0; i < dim; i++)
                            printf("%3d ", ic[i]);
                        printf("] Normal dir[%d]\n", iperm[swp_num%dim]);
                        printf("Compute ic:[");
                        for(i = 0; i < dim; i++)
                            printf("%3d ", icoords[0][i]);
                        clean_up(ERROR);
                    }

                    sten->Tst[side][j] = sten->Tworksp[side][j];
                    assign(sten->Tst[side][j],Rect_state(ic,wave),fr->sizest);
                    Set_params(sten->Tst[side][j],sten->st[0]);
                    // Dens(sten->Tst[side][j]) = Params(sten->st[0])->dens;
                    sten->Tset[side][j] = YES;
                    continue;
                }

                sten->reg_stencil = NO;

                if(deb_npt)
                {
                    printf("find_out_ns_states_in_other_dir"
                      " side %d of state, crx = %d, ic[%d %d], in_domain = %d\n",
                       side, crx, ic[0], ic[1], in_domain);
                }

                // For RT problem with top-bottom Homo. Dirichlet BC
                if(in_domain == NO)
                {
                    int     ic2[MAXD], ic3[MAXD];
                    float   nor[MAXD] = {0.0, 1.0}, vn, tmp_vel[MAXD];

                    if(deb_npt)
                    {
                       printf("test ic[%d %d] side %d to reflect in_domain %d\n",
                          ic[0], ic[1], side, in_domain); 
                    }
                    if(ic[1] < 0)
                    {
                        ic2[1] = -ic[1]-1; ic2[0] = ic[0];
                        assign(testst,Rect_state(ic2,wave),fr->sizest);
                        g_reflect_state(testst,fr->interf,NULL,NULL,nor);
                        // Use No slip BC
                        vn = 0.0;
                        for (i = 0; i < dim; i ++)
                        {
                            tmp_vel[i] = Vel(testst)[i];
                            vn += tmp_vel[i] * nor[i];
                        }
                        for (i = 0; i < dim; i ++)
                            Vel(testst)[i] = (-1.0)*Vel(testst)[i];
                        zero_normal_velocity(testst,nor,dim);
                        for (i = 0; i < dim; i ++)
                            Vel(testst)[i] += vn * nor[i];
                        // END Use No slip BC

                        if(ic[1] == -1)
                        {
                            ic3[1] = ic2[1]+1;  ic3[0] = ic[0];
                            Press(testst) = 2.0*Press(Rect_state(ic2,wave)) - Press(Rect_state(ic3,wave));
                        }
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        assign(sten->Tst[side][j],testst,fr->sizest);
                        Set_params(sten->Tst[side][j],sten->st[0]);
                        sten->Tset[side][j] = YES;

                        // TMP
                        /**
                        if(deb_npt == YES)
                        {
                            printf("inernal cell ic2[%d %d]\n", ic2[0], ic2[1]);
                            printf("inernal cell ic3[%d %d]\n", ic3[0], ic3[1]);
                            g_verbose_print_state(Rect_state(ic2,wave));
                            g_verbose_print_state(testst);
                        }
                        **/

                        continue;
                    }
                    else if(ic[1] >= gmax[1])
                    {
                        ic2[1] = 2*gmax[1] - ic[1]-1; ic2[0] = ic[0];
                        assign(testst,Rect_state(ic2,wave),fr->sizest);
                        g_reflect_state(testst,fr->interf,NULL,NULL,nor);
                        // Use No slip BC
                        vn = 0.0;
                        for (i = 0; i < dim; i ++)
                        {
                            tmp_vel[i] = Vel(testst)[i];
                            vn += tmp_vel[i] * nor[i];
                        }
                        for (i = 0; i < dim; i ++)
                            Vel(testst)[i] = (-1.0)*Vel(testst)[i];
                        zero_normal_velocity(testst,nor,dim);
                        for (i = 0; i < dim; i ++)
                            Vel(testst)[i] += vn * nor[i];
                        // END Use No slip BC

                        if(ic[1] == gmax[1])
                        {
                            ic3[1] = ic2[1]-1;  ic3[0] = ic[0];
                            Press(testst) = 2.0*Press(Rect_state(ic2,wave)) - Press(Rect_state(ic3,wave));
                        }
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        assign(sten->Tst[side][j],testst,fr->sizest);
                        Set_params(sten->Tst[side][j],sten->st[0]);
                        sten->Tset[side][j] = YES;
                        continue;
                    }
                }


                // TMP, the code below has not been tested yet.
                printf("ERROR: find_out_ns_states_in_other_dir, test the node\n");
                clean_up(ERROR);

                // Is it better to use the closest one?
                // old_crx[0] = __Rect_crossing(icoords[j],pgdir[side],
                //                wave_tri_soln(sten->wave)->tri_grid);
                old_crx[0] = NULL;
                crossings_in_direction(old_crx, icoords[j],pgdir[side],
                     wave_tri_soln(sten->wave)->tri_grid);

                // TMP
                /**
                if(deb_npt)
                {
                    printf("orth. direction find [%d][%d] of stencil state, in_domain %d,"
                             " find crx = %d, pgdir[side] = %d\n",
                           ic[0], ic[1], in_domain, old_crx[0], pgdir[side]);
                }
                **/

                if(old_crx[0] != NULL)
                {
                    /* There is a cross berween icoords[j] and its upper line one */
                    crossings_in_direction(new_crx, icoords[j],pgdir[side],
                         wave_tri_soln(sten->newwave)->tri_grid);

                    hs = old_crx[0]->hs;
                    if ((new_crx[0] == NULL) ||
                        (wave_type(hs) != wave_type(new_crx[0]->hs)))
                    {
                       /* crx has moved out of stencil OR
                        * another curve has moved between indx and the
                        * old crx over the course of the time step. */

                        // copy_Tstates(j,crds,side,new_comp,sten,crx);
                        if(new_crx[0] != NULL)
                        {
                            copy_Tstates_ver1(j,crds,side,new_comp,sten,new_crx[0]);
                        }
                        else
                            copy_Tstates_ver1(j,crds,side,new_comp,sten,old_crx[0]);
                    }
                    if ((wave_type(hs) == NEUMANN_BOUNDARY) &&
                        (fr->neumann_bdry_state))
                    {
                        set_ns_neumann_bdry_sten_Tstates(j,side,crds,idir,new_comp,
                                                 old_crx[0],sten);
                        // Seems interior  NEUMANN_BOUNDARY produces instability
                        // sten->Tst[side][j] = sten->Tworksp[side][j];
                        // hyp_solution(crds,new_comp,hs,UNKNOWN_SIDE,fr,wave,sten->Tst[side][j],NULL);
                        // TMP
                        /**
                        if(deb_npt)
                        {
                            printf("orth. dir. find_out_ns_states_in_other_dir: neumann_bdry_sten\n");
                            printf("ic[%d][%d] in_domain %d, pgdir[side] = %d\n",
                                ic[0], ic[1], in_domain, pgdir[side]);
                            (*fr->print_state)(sten->Tst[side][j]);
                        }
                        **/
                    }
                    else if(wave_type(hs) == DIRICHLET_BOUNDARY)
                    {
                        // set_dirichlet_bdry_sten_states(i,hs,sten,side);
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        sten->Tset[side][j] = YES;
                        evaluate_dirichlet_boundary_state(crds,hs,sten->fr,
                              sten->wave,sten->Tst[side][j]);

                        // TMP
                        /**
                        if(deb_npt)
                        {
                            printf("orth. dir. find_out_ns_states_in_other_dir: dirichlet_boundary_state\n");
                            printf("ic[%d][%d] in_domain %d, pgdir[side] = %d\n",
                                ic[0], ic[1], in_domain, pgdir[side]);
                            (*fr->print_state)(sten->Tst[side][j]);
                        }
                        **/
                    }
                    else
                    {
                        /* crx is interior */
                        copy_Tstates_ver1(j,crds,side,new_comp,sten,old_crx[0]);
                    }
                }
                else
                {
                    if(in_domain == YES)
                    {
                        copy_Tstates_ver1(j,crds,side,new_comp,sten,NULL);
                        /*
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        nearest_intfc_state_and_pt(crds,new_comp,sten->fr,
                               sten->newfr,sten->Tst[side][j],crds,&Ths);
                        sten->Tset[side][j] = YES;
                        */
                    }
                    // copy_Tstates(j,crds,side,new_comp,sten,NULL);
                }
            }
        }

        face = 0;
        /* Check in the sweep direction */
        for(j = -1; j <= 1; j+=2)
        {
            for(side = 0; side <= 1; side++)
            {
                if(side == 0) sign = -1;
                else sign = 1;
                for (i = 0; i < dim; ++i)
                    ictmp[i] = ic[i] = icoords[0][i];
                ic[indir] += sign*1;
                ictmp[indir] += sign*1; ictmp[idir] += j;

                if(sten->Tset[side][j] == YES)
                    continue;

                in_domain = crds_of_ic_ver1(crds,sten->wave,ictmp);

                if(deb_npt)
                {
                    printf("Sweep-direction ictmp[%d][%d] ic[%d][%d] in_domain %d\n",
                           ictmp[0], ictmp[1], ic[0], ic[1], in_domain);
                }

                // Use velocity at the grid point
                // Do not see crxings of CONTACT CURVE
                if(tcomp[side][j] != ext_comp &&
                   !is_excluded_comp(tcomp[side][j],fr->interf))
                {
                    if(in_domain == NO)
                    {
                        printf("ERROR find_out_states_in_other_dir, sweep\n");
                        printf("NOT in domain ic:[");
                        for(i = 0; i < dim; i++)
                            printf("%3d ", ic[i]);
                        printf("] Normal dir[%d]\n", iperm[swp_num%dim]);
                        printf("Compute ic:[");
                        for(i = 0; i < dim; i++)
                            printf("%3d ", icoords[0][i]);
                        clean_up(ERROR);
                    }
                    sten->Tst[side][j] = sten->Tworksp[side][j];
                    assign(sten->Tst[side][j],Rect_state(ictmp,wave),fr->sizest);
                    Set_params(sten->Tst[side][j],sten->st[0]);
                    // Dens(sten->Tst[side][j]) = Params(sten->st[0])->dens;
                    sten->Tset[side][j] = YES;
                    continue;
                }

                // For RT problem with top-bottom Homo. Dirichlet BC
                if(in_domain == NO)
                {
                    int     ic2[MAXD], ic3[MAXD];
                    float   nor[MAXD] = {0.0, 1.0};
                    // printf("test ic[%d %d] side %d to reflect in_domain %d\n",
                    //       ictmp[0], ictmp[1], side, in_domain);
                    if(ictmp[1] < 0)
                    {
                        ic2[1] = -ictmp[1]-1; ic2[0] = ictmp[0];
                        assign(testst,Rect_state(ic2,wave),fr->sizest);
                        g_reflect_state(testst,fr->interf,NULL,NULL,nor);
                        if(ictmp[1] == -1)
                        {
                            ic3[1] = ic2[1]+1;  ic3[0] = ictmp[0];
                            Press(testst) = 2.0*Press(Rect_state(ic2,wave)) - Press(Rect_state(ic3,wave));
                        }
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        assign(sten->Tst[side][j],testst,fr->sizest);
                        Set_params(sten->Tst[side][j],sten->st[0]);
                        sten->Tset[side][j] = YES;
                        continue;
                    }
                    else if(ictmp[1] >= gmax[1])
                    {
                        ic2[1] = 2*gmax[1] - ictmp[1]-1; ic2[0] = ictmp[0];
                        assign(testst,Rect_state(ic2,wave),fr->sizest);
                        g_reflect_state(testst,fr->interf,NULL,NULL,nor);
                        if(ictmp[1] == gmax[1])
                        {
                            ic3[1] = ic2[1]-1;  ic3[0] = ictmp[0];
                            Press(testst) = 2.0*Press(Rect_state(ic2,wave)) - Press(Rect_state(ic3,wave));
                        }
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        assign(sten->Tst[side][j],testst,fr->sizest);
                        Set_params(sten->Tst[side][j],sten->st[0]);
                        sten->Tset[side][j] = YES;
                        continue;
                    }
                }

                old_crx[0] = NULL;
                crossings_in_direction(old_crx, ic,npgdir[face],
                     wave_tri_soln(sten->wave)->tri_grid);

                /**
                if(deb_npt)
                {
                    printf("Sweep-direction [%d][%d] of stencil state, in_domain %d,"
                             " find crx = %d, pgdir[side] = %d\n",
                           ic[0], ic[1], in_domain, old_crx[0], npgdir[face]);
                }
                **/

                if(old_crx[0] != NULL)
                {
                    /* There is a cross berween icoords[j] and its upper line one */

                    hs = old_crx[0]->hs;
                    crossings_in_direction(new_crx, ic,npgdir[face],
                         wave_tri_soln(sten->newwave)->tri_grid);
                    if ((new_crx[0] == NULL) ||
                        (wave_type(hs) != wave_type(new_crx[0]->hs)))
                    {
                       /* crx has moved out of stencil OR
                        * another curve has moved between indx and the
                        * old crx over the course of the time step. */

                        copy_Tstates_ver1(j,crds,side,new_comp,sten,old_crx[0]);
                    }
                    if ((wave_type(hs) == NEUMANN_BOUNDARY) &&
                        (fr->neumann_bdry_state))
                    {
                        set_ns_neumann_bdry_sten_Tstates(j,side,crds,idir,new_comp,
                                                  old_crx[0],sten);
                        // sten->Tst[side][j] = sten->Tworksp[side][j];
                        // hyp_solution(crds,new_comp,hs,UNKNOWN_SIDE,fr,wave,sten->Tst[side][j],NULL);
                        // TMP
                        /**
                        if(deb_npt == YES)
                        {
                            printf("Tst[%d][%d] set by neumann[%d,%d] in %s\n",
                                  side,j,ic[0],ic[1], grid_dir(npgdir[face]));
                        }
                        **/
                    }
                    else if(wave_type(hs) == DIRICHLET_BOUNDARY)
                    {
                        sten->Tst[side][j] = sten->Tworksp[side][j];
                        sten->Tset[side][j] = YES;
                        evaluate_dirichlet_boundary_state(crds,hs,sten->fr,
                              sten->wave,sten->Tst[side][j]);
                        // TMP
                        /**
                        if(deb_npt == YES)
                        {
                            printf("Tst[%d][%d] set by dirichlet_bdry[%d,%d] in %s\n",
                                  side,j,ic[0],ic[1], grid_dir(npgdir[face]));
                        }
                        **/
                    }
                    else
                    {
                        /* crx is interior */
                        copy_Tstates_ver1(j,crds,side,new_comp,sten,old_crx[0]);
                    }
                }
            }
            face++;
        }

        /* Last way to set the unset point state */
        for(side = 0; side <= 1; side++)
        {
            if(side == 0) sign = -1;
            else sign = 1;
            for(j = -1; j <= 1; j++)
            {
                if(sten->Tset[side][j] == YES)
                    continue;
                for (i = 0; i < dim; ++i)
                    ic[i] = icoords[0][i];
                ic[idir] += j; ic[indir] += sign*1;
                in_domain = crds_of_ic_ver1(crds,sten->wave,ic);
                copy_Tstates_ver1(j,crds,side,new_comp,sten,NULL);
            }
        }

        /* Enforce symmetry at Neumann bdry of the domain. This is when upper and 
           lower boundaries are Neumann boundaries */
        if(debugging("enf_N_bdry_symm"))
        {
            float tmpnor[MAXD];
            for(side = 0; side <= 1; side++)
            {
                for(j = -1; j <= 1; j++)
                {
                    if(side == 0) sign = -1;
                    else sign = 1;
                    for (i = 0; i < dim; ++i)
                        ictmp[i] = ic[i] = icoords[0][i];
                    ic[indir] += sign*1;
                    ictmp[indir] += sign*1; ictmp[idir] += j;
                    in_domain = crds_of_ic_ver1(crds,sten->wave,ictmp);

                    // TMP
                    if(deb_npt)
                    {
                        printf("enf_N_bdry_symm dir(%d) ictmp[%d][%d] ic[%d][%d] in_domain %d\n",
                               idir,ictmp[0], ictmp[1], ic[0], ic[1], in_domain);
                        printf("-lbuf[1] = %d, rect_boundary_type(fr->interf,1,0) = %d, N_BDRY = %d\n",
                                     -lbuf[1], rect_boundary_type(fr->interf,1,0), NEUMANN_BOUNDARY);
                        fprint_boundary_type(stdout,"",rect_boundary_type(fr->interf,1,0),"\n",fr->interf);
                    }

                    if(in_domain == NO && (ictmp[1] < -gr->lbuf[1])  
                       && (rect_boundary_type(fr->interf,1,0) == NEUMANN_BOUNDARY ||
                           rect_boundary_type(fr->interf,1,0) == MIXED_TYPE_BOUNDARY))
                    {
                        tmpnor[0] = 0.0; tmpnor[1] = 1.0;
                        if(idir == 0) 
                        {
                            // assign(sten->Tworksp[0][j],sten->Tst[1][j],fr->sizest);    
                            // printf("Sweep-direction(%d) ictmp[%d][%d] ic[%d][%d] in_domain %d crds(%g %g)\n",
                            //    idir,ictmp[0], ictmp[1], ic[0], ic[1], in_domain, crds[0], crds[1]);

                            assign(sten->Tworksp[0][j],sten->st[j],fr->sizest);    
                            reflect_state(sten->Tworksp[0][j],fr->interf,NULL,NULL,tmpnor);
                            sten->Tst[0][j] = sten->Tworksp[0][j];
                        }
                        else
                        {
                            // printf("Sweep-direction(%d) ictmp[%d][%d] ic[%d][%d] in_domain %d crds(%g %g) 1\n",
                            //    idir,ictmp[0], ictmp[1], ic[0], ic[1], in_domain, crds[0], crds[1]);
                            assign(sten->Tworksp[side][j],sten->Tst[side][0],fr->sizest);    
                            reflect_state(sten->Tworksp[side][j],fr->interf,NULL,NULL,tmpnor);
                            sten->Tst[side][j] = sten->Tworksp[side][j];
                            /**
                            if(deb_npt)
                                printf("enter reflect_state, dir(1),side %d, j %d,"
                                             " assigned vel (%g %g)from (%g %g)\n",
                                          side,j, Vel(sten->Tst[side][j])[0],
                                          Vel(sten->Tst[side][j])[1],
                                          Vel(sten->Tst[side][0])[0],
                                          Vel(sten->Tst[side][0])[1]);
                            **/
                        }
                    }
                    if(in_domain == NO && (ictmp[1] >= (gmax[1] + gr->ubuf[1])) &&
                        (rect_boundary_type(fr->interf,1,1) == NEUMANN_BOUNDARY ||
                         rect_boundary_type(fr->interf,1,1) == MIXED_TYPE_BOUNDARY))
                    {
                        tmpnor[0] = 0.0; tmpnor[1] = 1.0;
                        if(idir == 0)
                        {
                            // printf("Sweep-direction(%d) ictmp[%d][%d] ic[%d][%d] in_domain %d crds(%g %g) 2\n",
                            //    idir,ictmp[0], ictmp[1], ic[0], ic[1], in_domain, crds[0], crds[1]);
                            assign(sten->Tworksp[1][j],sten->st[j],fr->sizest);    
                            reflect_state(sten->Tworksp[1][j],fr->interf,NULL,NULL,tmpnor);
                            sten->Tst[1][j] = sten->Tworksp[1][j];
                        }
                        else
                        {
                            // printf("Sweep-direction(%d) ictmp[%d][%d] ic[%d][%d] in_domain %d crds(%g %g) 3\n",
                            //    idir,ictmp[0], ictmp[1], ic[0], ic[1], in_domain, crds[0], crds[1]);
                            assign(sten->Tworksp[side][j],sten->Tst[side][0],fr->sizest);    
                            reflect_state(sten->Tworksp[side][j],fr->interf,NULL,NULL,tmpnor);
                            sten->Tst[side][j] = sten->Tworksp[side][j];
                        }
                    }
                }
            }
        }

        for(side = 0; side <=1; side++)
        {
            for(j = -1; j <= 1; j++)
            {
                if(sten->Tset[side][j] == NO)
                {
                    error = YES;
                    break;
                }
            }
        }

        if(error == YES)
        {
            for(side = 0; side <=1; side++)
            {
                for(j = -1; j <= 1; j++)
                {
                    if(sten->Tset[side][j] == NO)
                    {
                        printf("ERROR find_out_states_in_other_dir, Tst[%d][%d] not set\n",
                            side, j);
                    }
                }
            }
            printf("Normal dir[%d]\n", iperm[swp_num%dim]);
            print_int_vector("icoords[0] = ",icoords[0],dim,"\n");
            (void) printf("\n comps: ");
            for (i = -endpt; i <= endpt; ++i)
                (void) printf("comp[%d] %d ",i,sten->comp[i]);
            (void) printf("new_comp %d\n",new_comp);
            for(side = 0; side <= 1; side++)
            {
                for (i = -1; i <= 1; ++i)
                    printf("Tcomp[%d][%d] %d ",side, i,sten->Tcomp[side][i]);
                printf("\n");
            }
            (void) printf("STENCIL: "); print_Stencil(sten);
            clean_up(ERROR);
        }
}

/*
*                       set_ns_neumann_bdry_sten_Tstates():
*
*       Sets stencil states using reflection through a Neumann boundary.
*       Is it complete? Compare with set_neumann_bdry_sten_states().
*/

LOCAL   void    set_ns_neumann_bdry_sten_Tstates(
        int       i,
        int       side,
        float     *crds,
        int       idir,
        COMPONENT new_comp,
        CRXING    *nbdry,
        Stencil   *sten)
{
        HYPER_SURF *hs = nbdry->hs;
        Wave       *wave = sten->wave;
        Front      *fr = sten->fr;
        CRXING     *rcrx;
        float      coords[MAXD], coordsref[MAXD], n[MAXD];
        int        j, indx, ri, dim = fr->rect_grid->dim;
        size_t     sizest = fr->sizest;

        if ((*fr->neumann_bdry_state)(crds,new_comp,nbdry->pt,hs,fr,
                                (POINTER)wave,sten->Tworksp[side][i]))
        {
            sten->Tst[side][i] = sten->Tworksp[side][i];
            sten->Tset[side][i] = YES;

            // just set it to be zero, 031107
            // for(j = 0; j < dim; j++)
            //     Vel(sten->Tst[side][i])[j] = 0.0;
        }
        else
        {
            printf("ERROR set_neumann_bdry_sten_ns_Tstates, the state is not set\n");
            clean_up(ERROR);
        }
}



/*      copy_Tstates()
 *  Copy single state for the other direction.
 */
LOCAL   void copy_Tstates_ver1(
        int        start,
        float      *coords,
        int        side,
        COMPONENT  new_comp,
        Stencil    *sten,
        CRXING     *cross)
{
        int             i, j, in_one;
        size_t          sizest = sten->fr->sizest;
        HYPER_SURF      *Ths;
        if(cross != NULL)
        {
            sten->Tst[side][start] = state_with_comp(cross->pt,cross->hs,new_comp);

            // Have not allocated point for it yet.
            if(sten->Tst[side][start] == NULL)
            {
                if(is_passive_boundary(cross->hs))
                    sten->Tst[side][start] = sten->Tworksp[side][start];
                else
                {
                    printf("ERROR: copy_Tstates_ver1, Tst[%d][%d] NULL\n", side, start);
                    printf("center icoords[%d, %d] comp = %d\n", 
                            sten->icoords[0][0], sten->icoords[0][1], new_comp);
                    print_vector("Coords", 2, coords, "%g ");
                    // print_hypersurface(cross->hs);
                    // clean_up(ERROR);
                    goto find_intfc;
                }
            }
            // Have not allocated point for it yet.
            sten->Tset[side][start] = YES;
        }
        else
        {
find_intfc:
            sten->Tst[side][start] = sten->Tworksp[side][start];
            nearest_intfc_state_and_pt(coords,new_comp,sten->fr,
                    sten->newfr,sten->Tst[side][start],coords,&Ths);

            if((wave_type(Ths) == NEUMANN_BOUNDARY) &&
                        (sten->fr->neumann_bdry_state))
            {
                if ((*sten->fr->neumann_bdry_state)(coords,new_comp,NULL,Ths,sten->fr,
                                (POINTER)(sten->wave),sten->Tworksp[side][start]))
                {
                    sten->Tst[side][start] = sten->Tworksp[side][start];
                }
                else
                {
                    printf("ERROR in copy_Tstates_ver1\n");
                    printf("ERROR set_neumann_bdry_sten_Tstates, the state is not set\n");
                    clean_up(ERROR);
                }
            }
            else if(wave_type(Ths) == DIRICHLET_BOUNDARY)
            {
                sten->Tst[side][start] = sten->Tworksp[side][start];
                evaluate_dirichlet_boundary_state(coords,Ths,sten->fr,
                      sten->wave,sten->Tst[side][start]);
            }
            sten->Tset[side][start] = YES;
        }
}

LOCAL int crds_of_ic_ver1(
        float       *crds,
        Wave        *wave,
        int         *ic)
{
        int         i, in_domain = YES;
        RECT_GRID   *rgr = wave->rect_grid;

        for(i = 0; i < dim; i++)
        {
            if((ic[i] < -rgr->lbuf[i]) || 
                  (ic[i] >= (rgr->gmax[i] + rgr->ubuf[i])))
            in_domain = NO;
        }
        if(in_domain == YES)
        {
            for (i = 0; i < dim; ++i)
                crds[i] = Rect_coords(ic,wave)[i];
            return in_domain;
        }

        for (i = 0; i < dim; ++i)
           crds[i] = rgr->L[i] +
                (ic[i]+0.5)*rgr->h[i];
        return in_domain;
}


/*
*                       find_outer_ns_states():
*       a duplicate of find_outer_states(), merge later
*
*       Loads appropriate states for the outer indices of a stencil on either
*       the prev or next sides.  If an interface is crossed, an appropriate
*       state is picked off the interface, and copied out to the end of the
*       stencil (cf copy_states).  Neumann boundaries are also handled
*       appropriately (cf set_neumann_bdry_sten_states).
*/

LOCAL   void find_outer_ns_states(
        int      is,
        int      idir,
        Stencil  *sten,
        CRX_SIDE side)
{
        Wave       *wave = sten->wave;
        Front      *fr = sten->fr;
        Front      *newfr = sten->newfr;
        COMPONENT  *comp = sten->comp, new_comp = sten->newcomp;
        int        **icoords = sten->icoords;
        int        i, indx;
        CRXING     *crx;
        HYPER_SURF *hs;
        CRXING     *new_crx[MAX_NUM_CRX];
        RECT_GRID  *gr = wave->rect_grid;
        int        ic[MAXD], j, in_domain;
        float      crds[MAXD];
        static Locstate  st = NULL;

        if(st == NULL)
            alloc_state(fr->interf,&st,fr->sizest); 

        crx = NULL;
        for  (i = 1; i <= endpt; ++i)
        {
            indx = i * side;
            /** Change to use velocity at the grid point
            if (equivalent_comps(comp[indx],new_comp,newfr->interf) == YES)
            {
                sten->st[indx] = Rect_state(icoords[indx],wave);
                continue;
            }
            **/ 
            // Use velocity at the grid point
            // Do not see crxings of CONTACT CURVE
            if(comp[indx] != ext_comp &&
               !is_excluded_comp(comp[indx],fr->interf))
            {
                sten->st[indx] = sten->worksp[indx];    
                assign(sten->st[indx],Rect_state(icoords[indx],wave),fr->sizest);
                Set_params(sten->st[indx],sten->st[0]);
                // Dens(sten->st[indx]) = Params(sten->st[0])->dens;
                continue;
            }

            sten->reg_stencil = NO;
            crx = sten->crx[indx][0];

            // To fix ghost_boundary_cells for the Top and bottom boundary
            // for RT problem
            for (j = 0; j < dim; j++)
                ic[j] = icoords[indx][j];
            in_domain = crds_of_ic_ver1(crds,sten->wave,ic);

            if(deb_npt)
            {
                printf("find indx[%d] side %d of state, crx = %d, ic[%d %d], in_domain = %d, endpt %d\n",
                       indx, side, crx, ic[0], ic[1], in_domain, endpt);
            }

            if(in_domain == NO)
            {
                int     ic2[MAXD], ic3[MAXD];
                float   nor[MAXD] = {0.0, 1.0},  vn, tmp_vel[MAXD];

                if(deb_npt)
                    printf("test ic[%d %d] side %d to reflect in_domain %d\n",
                          ic[0], ic[1], side, in_domain);
                if(ic[1] < 0)
                {
                    ic2[1] = -ic[1]-1; ic2[0] = ic[0];
                    assign(st,Rect_state(ic2,wave),fr->sizest);
                    g_reflect_state(st,fr->interf,NULL,NULL,nor);

                    // Use No slip BC for lower and upper BC
                    vn = 0.0;
                    for (j = 0; j < dim; j ++)
                    {
                        tmp_vel[j] = Vel(st)[j];
                        vn += tmp_vel[j] * nor[j];
                    }
                    for (j = 0; j < dim; j ++)
                        Vel(st)[j] = (-1.0)*Vel(st)[j];
                    zero_normal_velocity(st,nor,dim);
                    for (j = 0; j < dim; j ++)
                        Vel(st)[j] += vn * nor[j];
                    // END Use No slip BC for lower and upper BC
                
                    if(ic[1] == -1)
                    {
                        ic3[1] = ic2[1]+1;  ic3[0] = ic[0];
                        Press(st) = 2.0*Press(Rect_state(ic2,wave)) - Press(Rect_state(ic3,wave));
                    }
                    sten->st[indx] = sten->worksp[indx];    
                    assign(sten->st[indx],st,fr->sizest);
                    Set_params(sten->st[indx],sten->st[0]);
                    continue;
                }
                else if(ic[1] >= gr->gmax[1])
                {
                    ic2[1] = 2*gmax[1] - ic[1]-1; ic2[0] = ic[0];
                    assign(st,Rect_state(ic2,wave),fr->sizest);
                    g_reflect_state(st,fr->interf,NULL,NULL,nor);

                    // Use No slip BC for lower and upper BC
                    vn = 0.0;
                    for (j = 0; j < dim; j ++)
                    {
                        tmp_vel[j] = Vel(st)[j];
                        vn += tmp_vel[j] * nor[j];
                    }
                    for (j = 0; j < dim; j ++)
                        Vel(st)[j] = (-1.0)*Vel(st)[j];
                    zero_normal_velocity(st,nor,dim);
                    for (j = 0; j < dim; j ++)
                        Vel(st)[j] += vn * nor[j];
                    // END Use No slip BC for lower and upper BC

                    if(ic[1] == gmax[1])
                    {
                        ic3[1] = ic2[1]-1;  ic3[0] = ic[0];
                        Press(st) = 2.0*Press(Rect_state(ic2,wave)) - Press(Rect_state(ic3,wave));
                    }
                    sten->st[indx] = sten->worksp[indx];    
                    assign(sten->st[indx],st,fr->sizest);
                    Set_params(sten->st[indx],sten->st[0]);
                    continue;
                }
            }

            // printf("find indx[%d] side %d of state, crx = %d, ic[%d %d], in_domain = %d\n",
            //            indx, side, crx, ic[0], ic[1], in_domain);

            if (crx != NULL)
            {
                /*there is a cross between indx and in_one*/

                (void) crossings_in_direction(new_crx,icoords[(i-1)*side],
                                       (side == PREV) ? prev_side : next_side,
                                       wave_tri_soln(sten->newwave)->tri_grid);

                hs = crx->hs;

                if(deb_npt)
                {
                    printf("new_crx[0] = %d\n", new_crx[0]);
                }  

                if ((new_crx[0] == NULL) ||
                    (wave_type(hs) != wave_type(new_crx[0]->hs)))
                {
                   /* crx has moved out of stencil OR
                    * another curve has moved between indx and the
                    * old crx over the course of the time step. */

                    copy_ns_states(i,side,endpt,new_comp,sten,crx);
                    if(deb_npt)
                        printf("entered copy_ns_states\n"); 
                    // return;
                    break;
                }
                if(deb_npt)
                    printf("wave_type(hs) == NEUMANN = %d\n",
                           (wave_type(hs) == NEUMANN_BOUNDARY));
                if ((wave_type(hs) == NEUMANN_BOUNDARY) &&
                    (fr->neumann_bdry_state))
                {
                    set_ns_neumann_bdry_sten_states(i,is,idir,new_comp,
                                                 crx,sten,side);
                    // return;
                    break;
                }
                else if (wave_type(hs) == DIRICHLET_BOUNDARY)
                {
                    set_ns_dirichlet_bdry_sten_states(i,hs,sten,side);
                    // return;
                    break;
                }
                else
                {
                    /* crx is interior */

                    copy_ns_states(i,side,endpt,new_comp,sten,crx);
                    // return;
                    break;
                }
            }
            else
            {
                float              *coords;
                static const float OFFSET = 0.01; /*TOLERANCE*/

                coords = Coords(sten->p[indx]);
                if (is+indx < -gr->lbuf[idir])
                    coords[idir] = VL[idir] + hdir*OFFSET;
                else if (is+indx >= gmax[idir]+gr->ubuf[idir])
                    coords[idir] = VU[idir] - hdir*OFFSET;
                sten->st[indx] = sten->worksp[indx];
                if (!nearest_crossing_state_with_comp(icoords[indx],
                        Coords(sten->p[indx]),new_comp,
                        wave_tri_soln(sten->newwave)->tri_grid,
                        &sten->st[indx]))
                {
                    nearest_intfc_state_and_pt(coords,new_comp,
                                           sten->fr,sten->newfr,
                                           sten->st[indx],
                                           Coords(sten->p[indx]),
                                           &sten->hs[indx]);
                }
                copy_ns_states(i,side,endpt,new_comp,sten,crx);
                // return;
                break;
            }
        }

        /**
        if(debugging("enf_N_bdry_symm"))
        {
            float tmpnor[MAXD], crds[MAXD];
            int   ic[MAXD], j, in_domain;
  
            for (i = 1; i <= endpt; ++i)
            {
                indx = i * side;
                for (j = 0; j < dim; j++)
                    ic[j] = icoords[indx][j];
                in_domain = crds_of_ic_ver1(crds,sten->wave,ic);
                if(deb_npt)
                {
                    printf("test ic[%d %d] side %d to reflect in_domain %d\n",
                              ic[0], ic[1], side, in_domain);
                }
                // Here, use first-order so st[-2] = st[-1] = reflect(st[0])
                if(in_domain == NO && (ic[1] < -lbuf[1])
                   && (rect_boundary_type(fr->interf,1,0) == NEUMANN_BOUNDARY ||
                       rect_boundary_type(fr->interf,1,0) == MIXED_TYPE_BOUNDARY))
                {
                    tmpnor[0] = 0.0; tmpnor[1] = 1.0;
                    sten->st[-1] = sten->worksp[-1];
                    assign(sten->st[-1],sten->st[0],fr->sizest);
                    reflect_state(sten->st[-1],fr->interf,NULL,NULL,tmpnor);
                    sten->st[-2] = sten->worksp[-2];
                    assign(sten->st[-2],sten->st[0],fr->sizest);
                    reflect_state(sten->st[-2],fr->interf,NULL,NULL,tmpnor);
                    break;
                }
                if(in_domain == NO && (ic[1] >= (gmax[1] + ubuf[1])) &&
                   (rect_boundary_type(fr->interf,1,1) == NEUMANN_BOUNDARY ||
                    rect_boundary_type(fr->interf,1,1) == MIXED_TYPE_BOUNDARY))
                {
                    tmpnor[0] = 0.0; tmpnor[1] = 1.0;
                    sten->st[1] = sten->worksp[1];
                    assign(sten->st[1],sten->st[0],fr->sizest);
                    reflect_state(sten->st[1],fr->interf,NULL,NULL,tmpnor);
                    sten->st[2] = sten->worksp[2];
                    assign(sten->st[2],sten->st[0],fr->sizest);
                    reflect_state(sten->st[2],fr->interf,NULL,NULL,tmpnor);
                    break;
                }
            } 
        } 
        **/
}               /*end find_outer_ns_states*/

/*
*                       copy_ns_states():
*
*       We have crossed a non-reflecting curve and are ready to simply copy an
*       appropriate state out to the end of the stencil.  If the cross is not
*       NULL, then it is assumed that the start state has not been set, and
*       an appropriate state is loaded and then copied.
*/

LOCAL   void copy_ns_states(
        int      start,
        CRX_SIDE side,
        int      endpt,
        int      new_comp,
        Stencil  *sten,
        CRXING   *cross)
{
        int             i, j, indx, in_one;
        size_t          sizest = sten->fr->sizest;

        if (cross != NULL)
        {
            indx = start*side;
            sten->st[indx] = state_with_comp(cross->pt,cross->hs,new_comp);
            if (sten->st[indx] != NULL)
            {
                for (j = 0; j < dim; ++j)
                    Coords(sten->p[indx])[j] = Coords(cross->pt)[j];
                assign(left_state(sten->p[indx]),sten->st[indx],sizest);
                assign(right_state(sten->p[indx]),sten->st[indx],sizest);
            }
            else
            {
                sten->st[indx] = sten->worksp[indx];
                nearest_intfc_state_and_pt(Coords(sten->p[indx]),new_comp,
                                           sten->fr,sten->newfr,sten->st[indx],
                                           Coords(sten->p[indx]),
                                           &sten->hs[indx]);
            }
        }

        for (i = start+1; i <= endpt; ++i)
        {
            indx = i*side;
            in_one = indx - side;
            sten->st[indx] = sten->st[in_one];
            for (j = 0; j < dim; ++j)
                Coords(sten->p[indx])[j] = Coords(sten->p[in_one])[j];
            assign(left_state(sten->p[indx]),sten->st[indx],sizest);
            assign(right_state(sten->p[indx]),sten->st[indx],sizest);
        }
        return;
}               /*end copy_ns_states*/



/*
*                       set_ns_dirichlet_bdry_sten_states():
*
*       Sets stencil states through a Dirichlet boundary.
*/

LOCAL   void    set_ns_dirichlet_bdry_sten_states(
        int        i,
        HYPER_SURF *hs,
        Stencil    *sten,
        CRX_SIDE   side)
{
        Wave            *wave = sten->wave;
        Front           *fr = sten->fr;
        int             indx;
        size_t          sizest = fr->sizest;

        for (; i <= endpt; ++i)
        {
            indx = i * side;
            sten->st[indx] = sten->worksp[indx];
            evaluate_dirichlet_boundary_state(Coords(sten->p[indx]),hs,fr,wave,
                                                     sten->st[indx]);
            assign(left_state(sten->p[indx]),sten->st[indx],sizest);
            assign(right_state(sten->p[indx]),sten->st[indx],sizest);
        }
}               /*end set_ns_dirichlet_bdry_sten_states*/

/*
*                       set_neumann_bdry_sten_states():
*
*       Sets stencil states using reflection through a Neumann boundary.
*/

LOCAL   void    set_ns_neumann_bdry_sten_states(
        int       i,
        int       is,
        int       idir,
        COMPONENT new_comp,
        CRXING    *nbdry,
        Stencil   *sten,
        CRX_SIDE  side)
{
        HYPER_SURF *hs = nbdry->hs;
        Wave       *wave = sten->wave;
        Front      *fr = sten->fr;
        CRXING     *rcrx;
        float      coords[MAXD], coordsref[MAXD], n[MAXD];
        int        j, indx, ri;
        size_t     sizest = fr->sizest;

        for (; i <= endpt; ++i)
        {
            indx = i * side;
            // TMP
            /**
            if(deb_npt)
            {
                printf("---Enter set_ns_neumann_bdry_sten_states indx = %d, side %d pt[%g %g]\n",
                         indx, side, Coords(sten->p[indx])[0], Coords(sten->p[indx])[1]);
                add_to_debug("g_neumann");
            }
            **/
 
            if ((*fr->neumann_bdry_state)(Coords(sten->p[indx]),new_comp,
                                          nbdry->pt,hs,fr,
                                          (POINTER)wave,sten->worksp[indx]))
            {
                sten->st[indx] = sten->worksp[indx];
                for (j = 0; j < dim; ++j)
                    Coords(sten->p[indx])[j] = Coords(nbdry->pt)[j];
                assign(left_state(sten->p[indx]),sten->st[indx],sizest);
                assign(right_state(sten->p[indx]),sten->st[indx],sizest);
                // TMP
                // if(deb_npt)
                //     remove_from_debug("g_neumann");
            }
            else
            {
                if (sten->crx[indx][0] != nbdry)
                    copy_ns_states(i,side,endpt,new_comp,sten,NULL);
                else
                    copy_ns_states(i,side,endpt,new_comp,sten,nbdry);
                return;
            }

            if (is_bdry_hs(hs))
            {
                /* TODO:
                 * Currently this section of set_neumann_bdry_sten_states() is
                 * only implemented for rectangular Neumann boundaries
                 * as indicated by the above test.  Interior
                 * or oblique boundaries do not use this boundary
                 * state function. */

                /* Have set the first state, now need to check for another
                 * cross on the other side of the Neumann boundary. Some
                 * technical issues arise due to the way neumann_bdry_state()
                 * reflects coords.
                 */

                if (i == endpt)
                    break;
                ri = (side == PREV) ? -indx - 2*is - 1 :
                                      -indx + 2*(gmax[idir] - is) - 1;
                if (side == PREV)
                {
                    rcrx = (ri >= 0) ? sten->crx[ri+1][0] :
                                       sten->crx[ri][sten->nc[ri]-1];
                }
                else if (side == NEXT)
                {
                    rcrx = (ri <= 0) ? sten->crx[ri-1][0] :
                                       sten->crx[ri][sten->nc[ri]-1];
                }
                if (rcrx != NULL)
                {
                    indx = (++i) * side;
                    for (j = 0; j < dim; ++j)
                        coords[j] = Coords(rcrx->pt)[j];
                    if ((reflect_pt_about_Nbdry(coords,coordsref,n,new_comp,
                                                nbdry->hs,fr))
                        &&
                        (*fr->neumann_bdry_state)(coordsref,new_comp,nbdry->pt,
                                                  hs,fr,(POINTER)wave,
                                                  sten->worksp[indx]))
                    {
                        sten->st[indx] = sten->worksp[indx];
                        for (j = 0; j < dim; ++j)
                            Coords(sten->p[indx])[j] = coordsref[i];
                        assign(left_state(sten->p[indx]),sten->st[indx],sizest);
                        assign(right_state(sten->p[indx]),sten->st[indx],sizest);
                    }
                    copy_ns_states(i,side,endpt,new_comp,sten,NULL);
                    return;
                }
            }
        }
}               /*end set_ns_neumann_bdry_sten_states*/

/*
*                       set_NS_stencil_comps_and_crosses():
*
*       This function loads the components of each point and any crosses
*       that occur in the interior of the stencil.  For the crosses,
*       each stencil entry contains the list of crosses lying on its interior
*       side, and these lists are ordered towards the outer points.
*       Thus sten->crx[0] is meaningless, and sten->crx[1] contains the list
*       of crosses (if any) in order FROM point 0 TO point 1.  Periodic
*       boundaries are essentially invisible (cf Find_rect_comp).
*/

LIB_LOCAL   COMPONENT set_NS_stencil_comps_and_crosses(
        int             idir,
        int             swp_num,
        int             *iperm,
        Stencil         *sten,
        int             end)
{
        Wave            *wave = sten->wave;
        Wave            *newwave = sten->newwave;
        int             **icoords = sten->icoords, ic[MAXD];
        TRI_GRID        *tg = wave_tri_soln(wave)->tri_grid;
        int             i, *nc = sten->nc;
        CRXING          ***crx = sten->crx;
        COMPONENT       *comp = sten->comp;

        int             indir, j;
        GRID_DIRECTION  *pgdir, gdir[2][2] = {NORTH, SOUTH,
                                      EAST, WEST};

        sten->newcomp = Rect_comp(icoords[0],newwave);
        comp[0] = Rect_comp(icoords[0],wave);
        crx[0][0] = NULL;
        nc[0] = 0;
        sten->hs[0] = NULL;/*TODO REMOVE*/

        for (i = 1; i <= end; ++i)
        {
            sten->hs[-i] = sten->hs[i] = NULL;/*TODO REMOVE*/
            comp[-i] = Find_rect_comp_ver1(idir,icoords[-i],wave);
            nc[-i] = crossings_in_direction(crx[-i],icoords[-i+1],prev_side,tg);
            comp[i] = Find_rect_comp_ver1(idir,icoords[i],wave);
            nc[i] = crossings_in_direction(crx[i],icoords[i-1],next_side,tg);
        }

        indir = iperm[(swp_num+1)%dim];
        if(indir == 0)
            pgdir = gdir[1];
        else
            pgdir = gdir[0];

        for(j = -1; j <= 1; j++)
        {
            for (i = 0; i < dim; ++i)
                ic[i] = icoords[0][i];
            ic[idir] += j;
            ic[indir] += 1;
            sten->Tcomp[1][j] = Find_rect_comp_ver2(ic,wave);
            /**
            if(deb_npt == YES)
            {
                printf("Tcomp[%d][%d] on ic[%d %d]\n",
                           1, j, ic[0], ic[1]);
            }
            **/
        }
        for(j = -1; j <= 1; j++)
        {
            for (i = 0; i < dim; ++i)
                ic[i] = icoords[0][i];
            ic[idir] += j;
            ic[indir] -= 1;
            sten->Tcomp[0][j] = Find_rect_comp_ver2(ic,wave);
            /**
            if(deb_npt == YES)
            {
                printf("Tcomp[%d][%d] on ic[%d %d]\n",
                           0, j, ic[0], ic[1]);
            }
            **/
        }

        return sten->newcomp;
}               /*set_NS_stencil_comps_and_crosses*/

/**
LOCAL COMPONENT Find_rect_comp_ver2(
        int             *ic,
        Wave            *wave)
{
        int             imax0, imax1;

        imax0 = gmax[0] + ubuf[0];
        imax1 = gmax[1] + ubuf[1];

        if ((ic[0] < -lbuf[0]) || (ic[0] >= imax0) ||
            (ic[1] < -lbuf[1]) || (ic[1] >= imax1))
            return ext_comp;
        else
            return Rect_comp(ic,wave);

}
**/
/*Find_rect_comp_ver2*/


/*
*                       NS_interior_sten_reg():
*
*       If called after a vector solver, decides whether the stencil at a given
*       point was regular for the vector solver.  In other words, decides
*       whether solution needs to be recomputed at a given point.
*       For the unsplit muscl, for updating the states in one direction, the stencil is like:
*
*       x      X      X      X      x
*       .      |      |      |      .
*       .      |      |      |      .
*       X------X------X------X------X
*       -2     -1     0      1      2
*       .      |      |      |      .
*       .      |      |      |      .
*       x      X      X      X      x
*
*      Capital Xs are the points actually used to update point[0]. Small xs are not
*      used, but they are checked. Because of the way sweep_3l_comp_segments()
*      checks the component.
*/

LOCAL   bool NS_interior_sten_reg(
        int             is,
        int             idir,
        int             swp_num,
        int             *iperm,
        Stencil         *sten)
{
        Wave            *wave = sten->wave;
        Front           *newfr = sten->newfr;
        int             sign, i, j, indir, side;
        int             **icoords = sten->icoords, icrds[MAXD], ictmp[MAXD];
        COMPONENT       cmp;
        COMPONENT       new_comp = sten->newcomp, *comp = sten->comp;
        CRXING          *cross;
        GRID_DIRECTION  *npgdir, *pgdir, gdir[2][2] = {NORTH, SOUTH,
                                      EAST, WEST};

        if (ComponentIsFlowSpecified(new_comp,newfr))
            return NO;

        for (i = -endpt; i <= endpt; ++i)
        {
            if ((equivalent_comps(comp[i],new_comp,newfr->interf) == NO) ||
                (sten->nc[i] != 0))
            {
                return NO;
            }
        }

        if(iperm[swp_num%dim] == 0)
            npgdir = gdir[1];
        else
            npgdir = gdir[0];

        indir = iperm[(swp_num+1)%dim];
        if(indir == 0)
            pgdir = gdir[1];
        else
            pgdir = gdir[0];

        for (i = 0; i < dim; ++i)
            ictmp[i] = icrds[i] = icoords[0][i];

        /* Check the components of the upper and lower rows */
        for(side = 0; side <=1; side++)
        {
            if(side == 0)
                ictmp[indir] = icrds[indir] - 1;
            else
                ictmp[indir] = icrds[indir] + 1;
            for (i = -endpt; i <= endpt; ++i)
            {
                ictmp[idir] = is - i;
                cmp = Find_rect_comp_ver2(ictmp,wave);
                if (equivalent_comps(cmp,new_comp,newfr->interf) == NO)
                    return NO;
            }
            for (i = 0; i < dim; ++i)
                ictmp[i] = icoords[0][i];
        }

        /* Check the crossings in the other direction */
        for(i = -endpt; i <= endpt; i++)
        {
            icrds[idir] = is - i;
            // nearest_crossing2d
            if(__Rect_crossing(icrds,pgdir[0],wave)!= NULL ||
               __Rect_crossing(icrds,pgdir[1],wave)!= NULL)
            {
                /*
                printf("Catched in dir[%d] ndir[%d], ic[%d, %d] griddir[%d, %d]\n",
                  idir, indir, icoords[0][0], icoords[0][1], pgdir[0], pgdir[1]);
                */
                return NO;
            }
        }

        /* Check the crossings in the sweep direction */
        for(j = -1; j <= 1; j+=2)
        {
            for(side = 0; side <= 1; side++)
            {
                if(side == 0) sign = -1;
                else sign = 1;
                for (i = 0; i < dim; ++i)
                    icrds[i] = icoords[0][i];
                icrds[indir] += sign*1;
                if(__Rect_crossing(icrds,npgdir[0],wave)!= NULL ||
                   __Rect_crossing(icrds,npgdir[1],wave)!= NULL)
                {
                    return NO;
                }
            }
        }

        if (endpt >= vsten_rad)
            return YES;

        /* Should add ?? */
        for (i = 0; i < dim; ++i)
            icrds[i] = icoords[0][i];

        for (i = endpt+1; i <= vsten_rad; ++i)
        {
            icrds[idir] = is - i;
            cmp = Find_rect_comp_ver1(idir,icrds,wave);
            cross = __Rect_crossing(icrds,next_side,wave);
            if ((equivalent_comps(cmp,new_comp,newfr->interf) == NO) ||
                (cross != NULL))
                return NO;

            icrds[idir] = is + i;
            cmp = Find_rect_comp_ver1(idir,icrds,wave);
            cross = __Rect_crossing(icrds,prev_side,wave);
            if ((equivalent_comps(cmp,new_comp,newfr->interf) == NO) ||
                (cross != NULL))
                return NO;
        }
        return YES;
}

/*
*                       Find_rect_comp_ver1():
*
*       Finds the component of a given point using the macro Rect_comp.
*/

LOCAL COMPONENT Find_rect_comp_ver1(
        int             idir,
        int             *icoords,
        Wave            *wave)
{
        int             is, imax;
        RECT_GRID       *gr = wave->rect_grid;

        is = icoords[idir];
        imax = gr->gmax[idir] + gr->ubuf[idir];

        if ((is < -gr->lbuf[idir]) || (is >= imax))
            return ext_comp;
        else
            return Rect_comp(icoords,wave);

}               /*Find_rect_comp_ver1*/



LOCAL void nor_deriv_sweep_3l_comp_segments(
        int             swp_num,
        int             *iperm,
        float           *dir,
        Wave            *wv,
        Wave            *newwv,
        Front           *fr,
        Front           *newfr,
        int             *icoords,
        int             i0min,
        int             i0max,
        float           dt,
        float           *dh,
        int             *idirs,
        NS_Field_Stencil      *ns_field)
{
        COMPONENT       comp;
        int             seg_min,seg_max;
        int             seg_min0,seg_max0;
        int             seg_min1,seg_max1;
        int             icrds1[MAXD], icrds0[MAXD];

        seg_min = i0min;

        /*** We only need to sweep one line at a time  ***/
        while (seg_min != i0max)
        {
            icoords[idirs[0]] = seg_min;
            icrds0[idirs[0]] = icrds1[idirs[0]] = seg_min;
            icrds0[idirs[1]] = icoords[idirs[1]] - 1;
            icrds1[idirs[1]] = icoords[idirs[1]] + 1;

            comp = Rect_comp(icoords,wv);
            if(Rect_comp(icoords,wv) != Rect_comp(icrds0,wv) ||
               Rect_comp(icoords,wv) != Rect_comp(icrds1,wv) ||
               Rect_comp(icrds0,wv) != Rect_comp(icrds1,wv))
            {
                seg_min++;
                continue;
            }

            for (seg_max = seg_min+1; seg_max < i0max; seg_max++)
            {
                icoords[idirs[0]] = seg_max;
                if (comp != Rect_comp(icoords,wv))
                    break;
            }

            for (seg_max0 = seg_min+1; seg_max0 < i0max; seg_max0++)
            {
                icrds0[idirs[0]] = seg_max0;
                if (comp != Rect_comp(icrds0,wv))
                    break;
            }
            for (seg_max1 = seg_min+1; seg_max1 < i0max; seg_max1++)
            {
                icrds1[idirs[0]] = seg_max1;
                if (comp != Rect_comp(icrds1,wv))
                    break;
            }

            seg_max = min(min(seg_max0, seg_max1), seg_max);
            icoords[idirs[0]] = seg_max;

            // TMP
            /*
            {
                int ic[MAXD], i;
                ic[idirs[1]] = icoords[idirs[1]];
                for(i = seg_min; i <= seg_max; i++)
                {
                    ic[idirs[0]] = i;
                    printf("sweep ic[%d %d] comp = %d, newcomp = %d\n",
                        ic[0], ic[1], Rect_comp(ic,wv), Rect_comp(ic,newwv));
                }
            }
            */
            // printf("Entered include_normal_vari, seg_min<->seg_max[%d %d]\n",
            //             seg_min, seg_max);
            include_normal_vari(swp_num,iperm,dir,wv,newwv,fr,newfr,
                         icoords,seg_min,seg_max,dt,dh,ns_field);

            seg_min = seg_max;
        }

        /***
        while (seg_min != i0max)
        {
            icoords[idirs[0]] = seg_min;
            comp = Rect_comp(icoords,wv);
            for (seg_max = seg_min+1; seg_max < i0max; seg_max++)
            {
                icoords[idirs[0]] = seg_max;
                if (comp != Rect_comp(icoords,wv))
                    break;
            }

            include_normal_vari(swp_num,iperm,dir,wv,newwv,fr,newfr,
                         icoords,seg_min,seg_max,dt,dh,ns_field);

            seg_min = seg_max;
        }
        ***/
}

LOCAL void include_normal_vari(
        int             swp_num,
        int             *iperm,
        float           *dir,
        Wave            *wv,
        Wave            *newwv,
        Front           *fr,
        Front           *newfr,
        int             *icoords,
        int             imin,
        int             imax,
        float           dt,
        float           *dh,
        NS_Field_Stencil      *ns_field)
{
        Vec_NS_Gas      *vst = ns_field->vgas;
        Vec_NS_Gas      **Tvst = ns_field->Tvgas;
        int             idir = iperm[swp_num];
        int             dim = wv->rect_grid->dim;
        int             nrad;
        int             vsize, i, icrds[MAXD];
        int             j, idirs[MAXD];

#if defined(ROTATIONAL_SYMMETRY)
        static  float   alpha;
        static  bool    first = YES;
#endif /* defined(ROTATIONAL_SYMMETRY) */

        nrad = (ns_field->npts_sten - ns_field->npts_sten%2)/2;
        vsize = imax - imin;

        if (vsize < wv->npts_vsten)
            return;

        if (g_load_NS_state_vectors(swp_num,iperm,vst,0,vsize,
                   wv,newwv,icoords,imin) == CONSTANT_IN_TIME)
            return;

        for (i = 0; i < dim; ++i)
        {
            idirs[i] = iperm[(i+swp_num)%dim];
            icrds[i] = icoords[i];
        }

        // TMP
        // printf("Entered include_normal_vari, icrds[%d %d] imin-imax[%d  %d]\n", 
        //                icrds[idirs[0]], icrds[idirs[1]], imin, imax);

        icrds[idirs[1]] = icoords[idirs[1]] - 1;
        if (g_load_NS_state_vectors(swp_num,iperm,Tvst[0],0,vsize,
                   wv,newwv,icrds,imin) == CONSTANT_IN_TIME)
            return;
    
        icrds[idirs[1]] = icoords[idirs[1]] + 1;
        if (g_load_NS_state_vectors(swp_num,iperm,Tvst[1],0,vsize,
                   wv,newwv,icrds,imin) == CONSTANT_IN_TIME)
            return;

        // upwind_reconstruct_in_dir(swp_num,iperm,dir,
        //    wv,newwv,fr,newfr,0,vsize,icoords,imin,imax,dt,dh,dim,ns_field,imin);
        upwind_reconstruct_in_dir(swp_num,iperm,
           wv,fr,newfr,0,vsize,icoords,dt,dh,dim,ns_field,imin);
} 

LOCAL void upwind_reconstruct_in_dir(
        int             swp_num,
        int             *iperm,
        Wave            *wv,
        Front           *fr,
        Front           *newfr,
        int             offset,
        int             vsize,
        int             *icrds,
        float           dt,
        float           *dh,
        int             dim,
        NS_Field_Stencil      *ns_field,
        int             pbuf)
{
        static Upwind_NS  *upwind = NULL;
        int               i, j, nvar_u;
        float             ***worksp, **source;
        int               start, end;
        int               idirs[MAXD];
        int               ic2[MAXD], icoords[MAXD];
        int               kmin, kmax, k; 
        // Source
        Vec_Src           *src = ns_field->src;
        Vec_NS_Gas        *vst = ns_field->vgas;
        bool              is_src;

        for (i = 0; i < dim; ++i)
        {
            idirs[i] = iperm[(i+swp_num)%dim];
            ic2[i] = icrds[i];
            icoords[i] = icrds[i];
        }

        if(upwind == NULL || vsize > upwind->vsize || dim != upwind->dim)
        {
            if(upwind == NULL)
            {
                scalar(&upwind,sizeof(Upwind_NS));
                zero_scalar(upwind,sizeof(Upwind_NS));
            }
            else
            {
                // free memory space here
                free(upwind->worksp);
                free(upwind->u);
                // NEW
                Set_free(&(upwind->VLst),v);
                Set_free(&(upwind->VRst),v);
                Set_free(&(upwind->VMst),v);
                // Chemical component
                Set_free(&(upwind->VLst),chem);
                Set_free(&(upwind->VRst),chem);
                Set_free(&(upwind->VMst),chem);
                Set_free(upwind,mu_chem);
                // Source 
                Set_free(upwind,source);
                Set_free(upwind,gravity);
            }

            nvar_u = 0;
            for(i = 0; i < dim; i++)
                upwind->index.v[i] = nvar_u++;
            upwind->index.density = nvar_u++;
            upwind->index.pressure = nvar_u++;
            // Chemical components
            for(i = 0; i < ns_field->vgas->n_chem; i++)
                upwind->index.chem[i] = nvar_u++; 
                 
            // v[], \rho, p, some are not needed anyway.
            upwind->worksp_len = 5;
            upwind->nfloats = dim + 2;
            // now include chemicals
            upwind->nfloats += ns_field->vgas->n_chem;

            upwind->dim = dim; 

            tri_array(&(upwind->worksp),upwind->worksp_len,upwind->nfloats,vsize,FLOAT);
            worksp = upwind->worksp;

            vector(&(upwind->u),upwind->nfloats,sizeof(float*));
         
            upwind->backward = worksp[0];
            upwind->central = worksp[1];
            upwind->forward = worksp[2];

            upwind->du = worksp[2];

            upwind->uL = worksp[0];
            upwind->uR = worksp[1];
            upwind->uM = worksp[2];

            // NEW, chemical component
            upwind->VLst.n_chem = ns_field->vgas->n_chem;
            upwind->VRst.n_chem = ns_field->vgas->n_chem;
            upwind->VMst.n_chem = ns_field->vgas->n_chem;
            set_NS_gas_storage(upwind->uL,&(upwind->VLst),upwind);
            set_NS_gas_storage(upwind->uR,&(upwind->VRst),upwind);
            set_NS_gas_storage(upwind->uM,&(upwind->VMst),upwind);
            // chemical diffusion coefficient
            if(ns_field->vgas->n_chem != 0)
                VECTOR(upwind,mu_chem,ns_field->vgas->n_chem,sizeof(float));
            // Source
            if (source_terms_exist() == YES)
                VECTOR(upwind,source,upwind->nfloats,sizeof(float*));
            if (is_gravity() == YES)
                MATRIX(upwind,gravity,dim,vsize,FLOAT);
            upwind->source_sum = worksp[3];
        }
        upwind->sten_rad = (ns_field->npts_sten - ns_field->npts_sten%2)/2;

        // set_NS_gas_storage(upwind->uL,&(upwind->VLst),upwind);
        // set_NS_gas_storage(upwind->uR,&(upwind->VRst),upwind);
        // set_NS_gas_storage(upwind->uM,&(upwind->VMst),upwind);

        upwind->vst = ns_field->vgas; 
        upwind->Tvst[0] = ns_field->Tvgas[0];
        upwind->Tvst[1] = ns_field->Tvgas[1];
        upwind->src = src;

        upwind->offset = offset; 
        upwind->vsize = vsize;

        upwind->u[upwind->index.density] = ns_field->vgas->rho + offset;
        upwind->u[upwind->index.pressure] = ns_field->vgas->p + offset;
        for (i = 0; i < dim; ++i)
            upwind->u[upwind->index.v[i]] = ns_field->vgas->v[i] + offset;
        // chemical component
        for (i = 0; i < ns_field->vgas->n_chem; ++i)
            upwind->u[upwind->index.chem[i]] = ns_field->vgas->chem[i] + offset;

        // Source
        if ((src != NULL) && ((source = upwind->source) != NULL))
        {
            source[upwind->index.density] = src->mass + offset;
            for (i=0; i < dim; ++i)
                source[upwind->index.v[i]] = src->mom[i] + offset;
            source[upwind->index.pressure] = src->energy + offset;
        }
        if (upwind->gravity != NULL)
        {
            float time = fr->time + 0.5*dt;
            float **coords = vst->coords + offset;
            float **g = upwind->gravity + offset;

            for (j = 0; j < vsize; ++j)
            {
                for(i = 0; i < dim; i++)
                    g[i][j] = gravity(coords[j],time)[i];
                // TMP
                // printf("gravity[%g %g] coords[%g %g]\n", 
                //            g[0][j], g[1][j], coords[j][0], coords[j][1]);
            }
        }
 
        is_src = g_NS_load_source_vectors(swp_num, iperm, upwind->gravity,
                         vst,src,offset,vsize,fr->rect_grid);
        upwind->is_src = is_src; 
        
#if defined(DEBUG_NPT)
        if(deb_npt == YES)
            printf("before  ns2d_unsplit_bct_linear_reconstructor\n");
#endif /* if defined(DEBUG_NPT) */

        /* Compute the linear reconstruction slope for velocity, density, chemical comp */
        start = upwind->sten_rad - 1;
        end = vsize - start;
        ns2d_unsplit_bct_linear_reconstructor(start,end,upwind);

#if defined(DEBUG_NPT)
        if(deb_npt == YES)
            printf("before  ns2d_normal_vel_half_step\n");
#endif /* if defined(DEBUG_NPT) */

        /* Half time step velocity expansion in normal direction */
        /* Only the normal component of the velocity is counted here */
        start = upwind->sten_rad;
        end = vsize - upwind->sten_rad + 1;
        // end = vsize - start;
        ns2d_normal_vel_half_step(start,end,dt,dh,swp_num,iperm,dim,upwind);

#if defined(DEBUG_NPT)
        if(deb_npt == YES)
            printf("before  ns2d_transerve_vel_step\n");
#endif /* if defined(DEBUG_NPT) */

        /* The transerve derivative is included here */
        /* Using monotonically limited central difference */
        /* if use Colella upwind difference, it should be computed later */
        start = upwind->sten_rad - 1;
        end =  vsize - upwind->sten_rad + 1;

        ns2d_transerve_vel_step(start,end,dt,dh,swp_num,iperm,dim,upwind);

#if defined(DEBUG_NPT)
        if(deb_npt == YES)
            printf("before  press_grad_and_visc, range [%d  %d]\n", 
               upwind->sten_rad -1, vsize - upwind->sten_rad + 1);
#endif /* if defined(DEBUG_NPT) */

        // Add pressure gredient contribution to half step velocity
        // start = upwind->sten_rad - 1;
        start = upwind->sten_rad - 1;
        end = vsize - upwind->sten_rad + 1;
        press_grad_and_visc(start,end,dt,dh,swp_num,iperm,dim,upwind,ns_field,icoords,pbuf);

        // Chemical components
        if(upwind->vst->n_chem != 0)
            include_chem_src(start,end,dt,dh,swp_num,iperm,dim,upwind,ns_field,icoords,pbuf); 

#if defined(DEBUG_NPT)
        if(deb_npt == YES)
            printf("After press_grad_and_visc\n");
#endif /* if defined(DEBUG_NPT) */

        // The valid reconstructed state should be within this range
        // start = upwind->sten_rad; 
        // end = vsize - upwind->sten_rad + 1; 

        // index for chemical component
        kmin = dim + 2;
        kmax = dim + 2 + upwind->vst->n_chem;

        if(iperm[swp_num] == 0)
        {
#if defined(DEBUG_NPT)
            if(deb_npt == YES)
            {
                if(ns_field->pt_sten != NULL)
	            print_int_vector("set state for : ",ns_field->pt_sten->icoords[0],dim,"\n");
            }
#endif /* if defined(DEBUG_NPT) */
            if(ns_field->pt_sten != NULL)
            {
                start = upwind->sten_rad; 
                end = vsize - upwind->sten_rad; 
                for (i = 0; i < dim; ++i)
                    icoords[i] = ns_field->pt_sten->icoords[0][i];
                for(j = 0; j < dim+1; j++)
                {
                    v_line_vel_L(icoords,ns_field).v[j] =  upwind->uL[j][start];
                    v_line_vel_R(icoords,ns_field).v[j] =  upwind->uR[j][start];

                    // TMP
                    if(isnan(v_line_vel_L(icoords,ns_field).v[j]) ||
                       isnan(v_line_vel_R(icoords,ns_field).v[j]))
                    {
                        printf("ERROR: upwind_reconstruct_in_dir, icoords[%d %d] dir = %d nan %g %g\n",
                             icoords[0], icoords[1], j, v_line_vel_L(icoords,ns_field).v[j],
                             v_line_vel_R(icoords,ns_field).v[j]); 
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*fr->print_state)(ns_field->pt_sten->st[i]);
                        }

                        clean_up(ERROR);
                    } 

                }
                // Chemical component
                if(upwind->vst->n_chem != 0)
                {
                    for(k = kmin, j =0; k < kmax; k++, j++)
                    {
                        v_line_chem_L(icoords,ns_field->chem_sten)[j] = upwind->uL[k][start];
                        v_line_chem_R(icoords,ns_field->chem_sten)[j] = upwind->uR[k][start];
                    }
                }

                icoords[idirs[0]]++;
                for(j = 0; j < dim+1; j++)
                {
                    v_line_vel_L(icoords,ns_field).v[j] =  upwind->uL[j][end];
                    v_line_vel_R(icoords,ns_field).v[j] =  upwind->uR[j][end];
                    // TMP
                    if(isnan(v_line_vel_L(icoords,ns_field).v[j]) ||
                       isnan(v_line_vel_R(icoords,ns_field).v[j]))
                    {
                        printf("ERROR: 1 upwind_reconstruct_in_dir, icoords[%d %d] dir = %d nan %g %g\n",
                             icoords[0], icoords[1], j, v_line_vel_L(icoords,ns_field).v[j],
                             v_line_vel_R(icoords,ns_field).v[j]); 
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                        clean_up(ERROR);
                    } 

                }
                // Chemical component
                if(upwind->vst->n_chem != 0)
                {
                    for(k = kmin, j =0; k < kmax; k++, j++)
                    {
                        v_line_chem_L(icoords,ns_field->chem_sten)[j] = upwind->uL[k][end];
                        v_line_chem_R(icoords,ns_field->chem_sten)[j] = upwind->uR[k][end];
                    }
                }

                return;
            }

            for (i = start; i < end; ++i)
            {
                icoords[idirs[0]] = i + pbuf;
                ic2[idirs[0]] = i + pbuf;
#if defined(DEBUG_NPT)
                if(deb_npt == YES)
                {
	            printf("ic[%d %d] ", icoords[0], icoords[1]);
                    for(j = 0; j < dim; j++)
                    {
                        printf("uL[%d] = %g, uR[%d] = %g; ",
                              j,  upwind->uL[j][i], j, upwind->uR[j][i]);    
                    }
                    printf("\n");
                }
#endif /* if defined(DEBUG_NPT) */
                for(j = 0; j < dim+1; j++)
                {
                    v_line_vel_L(icoords,ns_field).v[j] =  upwind->uL[j][i];
                    v_line_vel_R(icoords,ns_field).v[j] =  upwind->uR[j][i];
                    // TMP
                    if(isnan(v_line_vel_L(icoords,ns_field).v[j]) ||
                       isnan(v_line_vel_R(icoords,ns_field).v[j]))
                    {
                        printf("ERROR: 2 upwind_reconstruct_in_dir, icoords[%d %d] dir = %d nan %g %g\n",
                             icoords[0], icoords[1], j, v_line_vel_L(icoords,ns_field).v[j],
                             v_line_vel_R(icoords,ns_field).v[j]); 
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                        clean_up(ERROR);
                    } 

                    // TMP
                    /**
                    if(icoords[0] == 12 && icoords[1] == 11)
                    {
                        printf("In upwind_reconstruct_in_dir() (3) v_line[%d %d] L(%g %g). R(%g %g)\n",
                                 icoords[0], icoords[1],
                                    v_line_vel_L(icoords,ns_field).v[0], v_line_vel_L(icoords,ns_field).v[1],
                              v_line_vel_R(icoords,ns_field).v[0], v_line_vel_R(icoords,ns_field).v[1]);
                    }
                    **/
                }
                // Chemical component
                if(upwind->vst->n_chem != 0)
                {
                    for(k = kmin, j =0; k < kmax; k++, j++)
                    {
                        v_line_chem_L(icoords,ns_field->chem_sten)[j] = upwind->uL[k][i];
                        v_line_chem_R(icoords,ns_field->chem_sten)[j] = upwind->uR[k][i];
                    }
                }
            }
        }
        else
        {
            if(ns_field->pt_sten != NULL)
            {
                start = upwind->sten_rad; 
                end = vsize - upwind->sten_rad; 
                for (i = 0; i < dim; ++i)
                    icoords[i] = ns_field->pt_sten->icoords[0][i];

#if defined(DEBUG_NPT)
                if(deb_npt == YES)
                {
                    print_int_vector("set state for : ",ns_field->pt_sten->icoords[0],dim,"\n");
                    printf("In direction[%d], start-end[%d %d]\n", idirs[0], start, end);
                }
#endif // if defined(DEBUG_NPT)

                for(j = 0; j < dim+1; j++)
                {
                    h_line_vel_L(icoords,ns_field).v[j] =  upwind->uL[j][start];
                    h_line_vel_R(icoords,ns_field).v[j] =  upwind->uR[j][start];
                    // TMP
                    if(isnan(h_line_vel_L(icoords,ns_field).v[j]) ||
                       isnan(h_line_vel_R(icoords,ns_field).v[j]))
                    {
                        printf("ERROR: 3 upwind_reconstruct_in_dir, icoords[%d %d] dir = %d nan %g %g\n",
                             icoords[0], icoords[1], j, h_line_vel_L(icoords,ns_field).v[j],
                             h_line_vel_R(icoords,ns_field).v[j]); 
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                        clean_up(ERROR);
                    } 
                }
                // Chemical component
                if(upwind->vst->n_chem != 0)
                {
                    for(k = kmin, j =0; k < kmax; k++, j++)
                    {
                        h_line_chem_L(icoords,ns_field->chem_sten)[j] = upwind->uL[k][start];
                        h_line_chem_R(icoords,ns_field->chem_sten)[j] = upwind->uR[k][start];
                    }
                }

                if(deb_npt == YES)
                {
                    printf("-----------h_line_vel[%d %d] L[%10.9g %10.9g] R[%10.9g %10.9g]\n",
                           icoords[0], icoords[1], h_line_vel_L(icoords,ns_field).v[0],
                            h_line_vel_L(icoords,ns_field).v[1],
                            h_line_vel_R(icoords,ns_field).v[0],
                            h_line_vel_R(icoords,ns_field).v[1]);
                }

                /**
                if(icoords[0] == 9 && icoords[1] == 8)
                {
                   printf("h_line_vel_L, R set by ip[%d %d] L[%10.9g %10.9g] R[%10.9g %10.9g]\n", 
                           ns_field->pt_sten->icoords[0][0],
                            ns_field->pt_sten->icoords[0][1],
                              h_line_vel_L(icoords,ns_field).v[0],
                            h_line_vel_L(icoords,ns_field).v[1],
                            h_line_vel_R(icoords,ns_field).v[0],
                            h_line_vel_R(icoords,ns_field).v[1]);
                }
                **/

                icoords[idirs[0]]++;
                for(j = 0; j < dim+1; j++)
                {
                    h_line_vel_L(icoords,ns_field).v[j] =  upwind->uL[j][end];
                    h_line_vel_R(icoords,ns_field).v[j] =  upwind->uR[j][end];
                    // TMP
                    if(isnan(h_line_vel_L(icoords,ns_field).v[j]) ||
                       isnan(h_line_vel_R(icoords,ns_field).v[j]))
                    {
                        printf("ERROR: 4 upwind_reconstruct_in_dir, icoords[%d %d] dir = %d nan %g %g\n",
                             icoords[0], icoords[1], j, h_line_vel_L(icoords,ns_field).v[j],
                             h_line_vel_R(icoords,ns_field).v[j]); 
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                        clean_up(ERROR);
                    } 
                }
                // Chemical component
                if(upwind->vst->n_chem != 0)
                {
                    for(k = kmin, j =0; k < kmax; k++, j++)
                    {
                        h_line_chem_L(icoords,ns_field->chem_sten)[j] = upwind->uL[k][end];
                        h_line_chem_R(icoords,ns_field->chem_sten)[j] = upwind->uR[k][end];
                    }
                }


                if(deb_npt == YES)
                {
                    printf("---------h_line_vel[%d %d] L[%10.9g %10.9g] R[%10.9g %10.9g]\n",
                           icoords[0], icoords[1], h_line_vel_L(icoords,ns_field).v[0],
                            h_line_vel_L(icoords,ns_field).v[1],
                            h_line_vel_R(icoords,ns_field).v[0],
                            h_line_vel_R(icoords,ns_field).v[1]);
                }
                /**
                if(icoords[0] == 9 && icoords[1] == 8)
                {
                   printf("2 h_line_vel_L, R set by ip[%d %d] L[%10.9g %10.9g] R[%10.9g %10.9g]\n", 
                         ns_field->pt_sten->icoords[0][0],
                            ns_field->pt_sten->icoords[0][1],
                            h_line_vel_L(icoords,ns_field).v[0],
                            h_line_vel_L(icoords,ns_field).v[1],
                            h_line_vel_R(icoords,ns_field).v[0],
                            h_line_vel_R(icoords,ns_field).v[1]);
                }
                **/ 
                return;
            }

            for (i = start; i < end; ++i)
            {
                icoords[idirs[0]] = i + pbuf;
                for(j = 0; j < dim+1; j++)
                {
                    h_line_vel_L(icoords,ns_field).v[j] =  upwind->uL[j][i];
                    h_line_vel_R(icoords,ns_field).v[j] =  upwind->uR[j][i];
                    // TMP
                    /**
                    if(icoords[0] == 2 && icoords[1] == 9)
                    {
                        printf("WARNING: h_line_vel[%d %d] L[1]-R[1][%g %g] by regular\n",
                           icoords[0], icoords[1], h_line_vel_L(icoords,ns_field).v[1],
                            h_line_vel_R(icoords,ns_field).v[1]);
                    }
                    if(icoords[0] == 9 && icoords[1] == 8)
                    {
                       printf("3 h_line_vel_L, R set by ip[%d %d]\n", icoords[0],
                            icoords[1]);
                    }
                    **/
                    // TMP
                    if(isnan(h_line_vel_L(icoords,ns_field).v[j]) ||
                       isnan(h_line_vel_R(icoords,ns_field).v[j]))
                    {
                        printf("ERROR: 5 upwind_reconstruct_in_dir, icoords[%d %d] dir = %d nan %g %g\n",
                             icoords[0], icoords[1], j, h_line_vel_L(icoords,ns_field).v[j],
                             h_line_vel_R(icoords,ns_field).v[j]); 
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                        clean_up(ERROR);
                    } 
                }
                // Chemical component
                if(upwind->vst->n_chem != 0)
                {
                    for(k = kmin, j =0; k < kmax; k++, j++)
                    {
                        h_line_chem_L(icoords,ns_field->chem_sten)[j] = upwind->uL[k][i];
                        h_line_chem_R(icoords,ns_field->chem_sten)[j] = upwind->uR[k][i];
                    }
                }


                if(deb_npt == YES)
                {
                    printf("h_line_vel[%d %d] L[1]-R[1][%g %g]\n",
                           icoords[0], icoords[1], h_line_vel_L(icoords,ns_field).v[1],
                            h_line_vel_R(icoords,ns_field).v[1]);
                }
            }
        }
}

LOCAL bool g_NS_load_source_vectors(
        int                swp_num,
        int                *iperm,
        float              **grav,
        Vec_NS_Gas         *vst,
        Vec_Src            *src,
        int                offset,
        int                vs,
        RECT_GRID          *gr)
{
        int                i, j;
        int                dim = gr->dim;
        float              *v_src[SMAXD];
        float              *v_src0, *v_srci;
        float              *v0, *rho, *E, *mom[MAXD];

        if (src == NULL)
            return NO;

        for (i = 0; i < dim; ++i)
            v_src[i] = src->mom[i] + offset;

        for (i = 0; i < dim; ++i)
        {
            v_srci = v_src[i];
            for (j = 0; j < vs; ++j)
                v_srci[j] = 0.0;
        }

        rho = vst->rho + offset;

        // should add gravity .
        if (grav != NULL)
        {
            float  **g = grav + offset;
            for(i = 0; i < dim; i++)
            {
                v_srci = v_src[i];
                for (j = 0; j < vs; ++j)
                    // v_srci[j] += g[i][j]/rho[j];
                    v_srci[j] += g[i][j];
            }
        }

        return YES;
}


LOCAL void press_grad_and_visc(
        int         start,
        int         end,
        float       dt,
        float       *dh,
        int         swp_num,
        int         *iperm,
        int         dim,
        Upwind_NS   *upwind,
        NS_Field_Stencil   *ns_field,
        int         *icoords,
        int         pbuf)
{
        int             i, j, k, idirs[MAXD];
        int             kmax, sten_rad = 0;
        float           **u, **uL, **uR;
        float           *uk, *uLk, *uRk;
        float           dtdni, dn, dp[MAXD], diff[MAXD];
        Locstate        s1, s2, s3;
        Vec_NS_Gas      *vst = upwind->vst;
        Vec_NS_Gas      **Tvst = upwind->Tvst;
        int             ic[MAXD];
        float           v0[MAXD], v1[MAXD], v2[MAXD];
        float           mu = 0.0, den = 0.0;
        COMPONENT       comp;
        RECT_GRID       *gr = ns_field->wave->rect_grid;  

        for (j = 0; j < dim; ++j)
        {
            idirs[j] = iperm[(j+swp_num)%dim];
            ic[j] = icoords[j];
        }

        kmax = dim;
        u = upwind->u;
        uL = upwind->uL;
        uR = upwind->uR;

        // Find the density and viscosity associated with the comp
        // Find the stencil radius for irregular points
        if(ns_field->pt_sten != NULL)
        {
            sten_rad = upwind->sten_rad;

            /**
            comp = ns_field->pt_sten->newcomp;
            for(i = 0; i < ns_field->n_eos; i++)
            {
                if(comp == ns_field->comps[i])
                {
                    den = ns_field->den_vis[i][0];
                    mu = ns_field->den_vis[i][1];
                }
            }
            **/
        }

        // TMP
        if(deb_npt)
        {
            printf("Enter press_grad_and_visc, ic[%d %d], start = %d, end = %d, pbuf %d, sten_rad = %d\n",
                 icoords[0], icoords[1], start, end, pbuf, sten_rad);
        }
  
        for(j = start; j < end; j++)
        {
            s1 = Tvst[0]->state[j];
            s2 = vst->state[j];
            s3 = Tvst[1]->state[j];

            for(i = 0; i < dim; i++)
            {
                v0[i] = Tvst[0]->v[i][j];
                v1[i] = vst->v[i][j];
                v2[i] = Tvst[1]->v[i][j];
            }

            if(ns_field->pt_sten != NULL)
            {
                icoords[idirs[0]] = ic[idirs[0]] + j - sten_rad; 
            }
            else
                icoords[idirs[0]] = j + pbuf;

            // Find the density and mu of the cell            
            if((icoords[0] < -gr->lbuf[0] || icoords[0] >=gr->gmax[0]+gr->ubuf[0]  ||
                icoords[1] < -gr->lbuf[1] || icoords[1] >=gr->gmax[1]+gr->ubuf[1]) ||
               (Rect_comp(icoords,ns_field->newwave) == ext_comp ||
                is_excluded_comp(Rect_comp(icoords,ns_field->newwave),
                ns_field->newfr->interf)) ||
                Rect_comp(icoords,ns_field->newwave) == POTS_COMP)
            {
                if(ns_field->pt_sten != NULL)
                {
                    comp = ns_field->pt_sten->newcomp;
                    for(i = 0; i < ns_field->n_eos; i++)
                    {
                        if(comp == ns_field->comps[i])
                        {
                            den = ns_field->den_vis[i][0];
                            mu = ns_field->den_vis[i][1];
                            break;
                        }
                    }
                    // TMP
                    if(i == ns_field->n_eos)
                    {
                        printf("ERROR: press_grad_and_visc, point outside domain, sten i = %d\n", i);
                        printf("ic[%d %d], start = %d, end = %d, pbuf %d, sten_rad = %d\n",
                         icoords[0], icoords[1], start, end, pbuf, sten_rad);
                        printf("comp = %d, newcomp %d\n", 
                             Rect_comp(icoords,ns_field->newwave), ns_field->pt_sten->newcomp);
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*ns_field->fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                        clean_up(ERROR);
                    }
                }
                else if(Rect_comp(icoords,ns_field->newwave) == POTS_COMP)
                {
                    // s2 = vst->state[j];
                    den = Params(s2)->dens;
                    mu = Params(s2)->shear_visc;
                }
                else
                {
                    printf("ERROR: press_grad_and_visc, point outside domain,\n");
                    printf("ic[%d %d], start = %d, end = %d, pbuf %d, sten_rad = %d\n",
                         icoords[0], icoords[1], start, end, pbuf, sten_rad);
                    printf("comp = %d\n", Rect_comp(icoords,ns_field->newwave));
                    clean_up(ERROR);
                }
            }
            else
            {
                mu = Rect_visc(icoords,ns_field);
                den = Rect_den(icoords,ns_field);
                if(fabs(den) < 0.00000001)
                {
                    printf("WARNING: press_grad_and_visc, ic[%d %d] den = %g  ns_field->pt_sten = %d\n",
                        icoords[0], icoords[1], den, ns_field->pt_sten);
                    /**
                    printf("comp = %d crds (%g %g)\n", 
                          Rect_comp(icoords,ns_field->newwave),
                          Rect_coords(icoords,ns_field->newwave)[0], 
                          Rect_coords(icoords,ns_field->newwave)[1]);
                    clean_up(ERROR);
                    **/
                    // TMP, should be solved by looking at the geometry
                    den = Params(s2)->dens;
                    mu = Params(s2)->shear_visc;
                    printf("supply with dens = %g, mu = %g\n", den, mu);
                }
            }

            /*
            if(fabs(den) < 0.00000001)
            {
                printf("ERROR: press_grad_and_visc, ic[%d %d] den = %g\n",
                    icoords[0], icoords[1], den);
                clean_up(ERROR);
            }
            */

            /* OLD
            if(ns_field->pt_sten == NULL)
            {
                mu = Rect_visc(icoords,ns_field);
                den = Rect_den(icoords,ns_field);
            }
            **/

            if(deb_npt)
            // if(icoords[0]== 40 && icoords[1] == 19)
            {
                printf("In press_grad_and_visc, ic[%d %d], j = %d, pbuf %d, mu = %g, den = %g,"
                     "\n",
                     icoords[0], icoords[1], j , pbuf, mu, den);
                printf("start = %d, end = %d\n", start, end);
                printf("Press j-1 = %13.12g, j = %13.12g, j+1 = %13.12g\n",
                          Press(vst->state[j-1]),
                          Press(vst->state[j]), 
                          Press(vst->state[j+1]));
                // printf("P_s1 = %13.12g\n", Press(s1));
                // printf("P_s3 = %13.12g\n", Press(s3));
            }

            if(idirs[0] == 0)
            {
                dp[0] = 0.5*(Press(vst->state[j+1]) - Press(vst->state[j-1]))/dh[0];
                dp[1] = 0.5*(Press(s3) - Press(s1))/dh[1];
                // diff[0] = mu*(vst->v[0][j+1] - 2.0*vst->v[0][j] + vst->v[0][j-1])/sqr(dh[0]);
                // diff[1] = mu*(v2[1] - 2.0*v1[1] + v0[1])/sqr(dh[1]);
                diff[0] = mu*(vst->v[0][j+1] - 2.0*vst->v[0][j] + vst->v[0][j-1])/sqr(dh[0]);
                diff[0] += mu*(v2[0] - 2.0*v1[0] + v0[0])/sqr(dh[1]);
                diff[1] = mu*(vst->v[1][j+1] - 2.0*vst->v[1][j] + vst->v[1][j-1])/sqr(dh[0]);
                diff[1] += mu*(v2[1] - 2.0*v1[1] + v0[1])/sqr(dh[1]);
            }
            else
            {
                dp[1] = 0.5*(Press(vst->state[j+1]) - Press(vst->state[j-1]))/dh[1];
                dp[0] = 0.5*(Press(s3) - Press(s1))/dh[0];
                // diff[1] = mu*(vst->v[1][j+1] - 2.0*vst->v[1][j] + vst->v[1][j-1])/sqr(dh[1]);
                // diff[0] = mu*(v2[0] - 2.0*v1[0] + v0[0])/sqr(dh[0]);
                diff[1] = mu*(vst->v[1][j+1] - 2.0*vst->v[1][j] + vst->v[1][j-1])/sqr(dh[1]);
                diff[1] += mu*(v2[1] - 2.0*v1[1] + v0[1])/sqr(dh[0]);
                diff[0] = mu*(vst->v[0][j+1] - 2.0*vst->v[0][j] + vst->v[0][j-1])/sqr(dh[1]);
                diff[0] += mu*(v2[0] - 2.0*v1[0] + v0[0])/sqr(dh[0]);
            }
            // if(deb_npt)
            if(isnan(diff[0]) || isnan(diff[1]) || isnan(dp[0]) || isnan(dp[1]))
            {
                printf("ERROR press_grad_and_visc, ic[%d %d], dp [%9.8g %9.8g], diff [%g %g]\n",
                     icoords[0], icoords[1], dp[0], dp[1], diff[0], diff[1]);
                clean_up(ERROR);
            }

            for(k = 0; k < dim; k++)
            {
                if(ns_field->pt_sten == NULL)
                {
                    Rect_gradP(icoords,ns_field)[k] = dp[k];
                    Rect_diff(icoords,k,ns_field) = diff[k];
                }
                else if(j == sten_rad)
                {
                    Rect_gradP(icoords,ns_field)[k] = dp[k];
                    Rect_diff(icoords,k,ns_field) = diff[k];
                }
                // Rect_gradP(icoords,ns_field)[k] = dp[k];
                // Rect_diff(icoords,k,ns_field) = diff[k];
            }

            if(deb_npt)
            {
                printf("ic[%d %d], dp of[%d] (%13.12g, %13.12g) diff (%13.12g, %13.12g)\n",
                     icoords[0], icoords[1], j, dp[0], dp[1], diff[0], diff[1]);
            }

            for(k = 0; k < dim; k++)
            {
                uLk  = uL[k]; 
                uRk  = uR[k]; 
                if((j + 1) != end)
                {
                    if(deb_npt && k == 0)
                    {
                        printf("uL %13.12g,dp %13.12g, den %13.12g, diff %13.12g\n",
                             uLk[j+1], dp[k], den, diff[k]); 
                    }

                    uLk[j+1] -= 0.5*dt*dp[k]/den;
                    uLk[j+1] += 0.5*dt*diff[k]/den;
              
                    if(isnan(uLk[j+1]))
                    {
                        printf("ERROR press_grad_and_visc, ic[%d %d], dp [%9.8g %9.8g], diff [%g %g], den = %g\n",
                            icoords[0], icoords[1], dp[0], dp[1], diff[0], diff[1], den);
                        clean_up(ERROR);
                    }
                }
                if(j != start)
                {
                    if(deb_npt && k == 0)
                    {
                        printf("uR %13.12g,dp %13.12g, den %13.12g, diff %13.12g\n",
                             uRk[j], dp[k], den, diff[k]); 
                    }

                    uRk[j] -= 0.5*dt*dp[k]/den;
                    uRk[j] += 0.5*dt*diff[k]/den;
                    if(isnan(uRk[j]))
                    {
                        printf("ERROR press_grad_and_visc,1 ic[%d %d], dp [%9.8g %9.8g], diff [%g %g] den = %g\n",
                            icoords[0], icoords[1], dp[0], dp[1], diff[0], diff[1], den);
                        clean_up(ERROR);
                    }
                }
            }
        }

        if(deb_npt)
        {
             printf("edge velocity, after include press grad, visco\n");
             for (j = start; j < end; ++j)
             {
                 (void) printf("%-4d L:",j);
                 for (k = 0; k < dim; ++k)
                     printf(" %14.13g",uL[k][j]);
                 printf(" ; R: ");
                 for (k = 0; k < dim; ++k)
                     printf(" %14.13g",uR[k][j]);
                 printf("\n");
             }
        }

}

// Compute uLk[j] = uk[j-1] + 0.5*(dn - dt*u[idirs[0]][j-1])*duk[j-1]
// Compute uRk[j] = uk[j] - 0.5*(dn + dt*u[idirs[0]][j])*duk[j] 
// on cell edges. The velocity and density are reconstructed.
LOCAL void ns2d_normal_vel_half_step(
        int         start,
        int         end,
        float       dt,
        float       *dh,
        int         swp_num,
        int         *iperm,
        int         dim,
        Upwind_NS   *upwind)
{
        int             i, j, k, idirs[MAXD];
        int             kmax;
        float           **u, **du, **uL, **uR; 
        float           **sgnl, **sgnr;
        float           *duk;
        float           *sgnrk, *sgnr1, *sgnlk, *sgnl1;
        float           *uk, *uLk, *uRk;
        float           dtdni, dn;
        float           **source_sum, **source, *ssk;

        if(upwind->vst->n_chem != 0)
            kmax = dim + 2 + upwind->vst->n_chem; // v, rho, p, chem
        else
            kmax = dim+1;
        // kmax = dim+1;

        // Source
        source = upwind->source;
        source_sum = upwind->source_sum;

        u = upwind->u;
        du = upwind->du;
        uL = upwind->uL;
        uR = upwind->uR;

        for (j = 0; j < dim; ++j)
            idirs[j] = iperm[(j+swp_num)%dim];

        dtdni = dt/dh[idirs[0]];
        dn = dh[idirs[0]];

        /* add up source term contributions for half time-step */
        /* For the current implementation, source correspond to
         * the momemtum eqn with primitive variable u[0,1,2].
         */ 

        if(deb_npt)
        {
            printf("IN ns2d_normal_vel_half_step, source %d, upwind->src %d\n",
                     source, upwind->src);
        }
 
        if ((source != NULL) && (upwind->src != NULL))
        {
            for (k = 0; k < kmax; ++k)
            {
                ssk = source_sum[k];
                for (j = start - 1; j < end; ++j)
                {
                    ssk[j] = source[k][j]*0.5 * dt;
                }
            }
        }
        else
        {
            for (k = 0; k < kmax; ++k)
            {
                ssk = source_sum[k];
                for (j = start - 1; j < end; ++j)
                    ssk[j] = 0.0;
            }
        }

        /* compute uL */
        for (k = 0; k < kmax; ++k)
        {
            uk = u[k];
            uLk = uL[k];
            duk = du[k];
            // Source
            ssk = source_sum[k];
            for (j = start; j < end; ++j)
            {
                // uLk[j] = uk[j-1] + 0.5*(dn - dt*u[idirs[0]][j-1])*duk[j-1];
                uLk[j] = uk[j-1] + 0.5*(1.0 - dtdni*u[idirs[0]][j-1])*duk[j-1] + ssk[j-1];
            }
        }

        Vec_NS_Gas_field_set(&upwind->VLst,v) = YES;
        Vec_NS_Gas_field_set(&upwind->VLst,rho) = YES;
        // Chemical
        if(upwind->vst->n_chem != 0)
        {
            Vec_NS_Gas_field_set(&upwind->VLst,p) = YES;
            Vec_NS_Gas_field_set(&upwind->VLst,chem) = YES;
        }

        /* compute uR */
        for (k = 0; k < kmax; ++k)
        {
            uk = u[k];
            uRk = uR[k];
            duk = du[k];
            ssk = source_sum[k];
            for (j = start; j < end; ++j)
            {
                // uRk[j] = uk[j] - 0.5*(dn + dt*u[idirs[0]][j])*duk[j];
                uRk[j] = uk[j] - 0.5*(1.0 + dtdni*u[idirs[0]][j])*duk[j] + ssk[j];
            }
        }

        Vec_NS_Gas_field_set(&upwind->VRst,v) = YES;
        Vec_NS_Gas_field_set(&upwind->VRst,rho) = YES;
        // chemical
        if(upwind->vst->n_chem != 0)
        {
            Vec_NS_Gas_field_set(&upwind->VRst,p) = YES;
            Vec_NS_Gas_field_set(&upwind->VRst,chem) = YES;
        }

        if(debugging("half_step") || deb_npt)
        {
             printf("half step velocity\n"); 
             for (j = start; j < end; ++j)
             {
                 (void) printf("%-4d L:",j);
                 for (k = 0; k < kmax; ++k)
                     printf(" %10.9g",uL[k][j]);
                 printf(" ; R: ");
                 for (k = 0; k < kmax; ++k)
                     printf(" %10.9g",uR[k][j]);
                 printf("\n");
             }
        }
}

// Compute transerve derivative using 3-point
//(v\hat{U}_y)_{i,j} = 1/(\delta y) \hat{v}_{i,j}\delta U_y
LOCAL void ns2d_transerve_vel_step(
        int         start,
        int         end,
        float       dt,
        float       *dh,
        int         swp_num,
        int         *iperm,
        int         dim,
        Upwind_NS   *upwind)
{
        int             i, j, k, idirs[MAXD];
        int             kmax;
        float           **u, **uL, **uR;
        float           *uk, *uLk, *uRk;
        float           dtdni, dn;
        Locstate        s1, s2, s3;
        Vec_NS_Gas      *vst = upwind->vst;
        Vec_NS_Gas      **Tvst = upwind->Tvst;
        float           v0[40], v1[40], v2[40], slope[40];

        for (j = 0; j < dim; ++j)
            idirs[j] = iperm[(j+swp_num)%dim];

        // kmax = dim+1;
        if(upwind->vst->n_chem != 0)
            kmax = dim + 2 + upwind->vst->n_chem; // v, rho, p, chem
        else
            kmax = dim+1;

        u = upwind->u;
        uL = upwind->uL;
        uR = upwind->uR;

        dn = dh[idirs[1]];

        for(j = start; j < end; j++)
        {
            s1 = Tvst[0]->state[j];
            s2 = vst->state[j];
            s3 = Tvst[1]->state[j];
            for(i = 0; i < dim; i++)
            {
                v0[i] = Tvst[0]->v[i][j];
                v1[i] = vst->v[i][j];
                v2[i] = Tvst[1]->v[i][j];
            }
            v0[i] = Tvst[0]->rho[j]; 
            v1[i] = vst->rho[j]; 
            v2[i] = Tvst[1]->rho[j]; 
            // chemical
            if(upwind->vst->n_chem != 0)
            {
                i++;
                v0[i] = Tvst[0]->p[j];
                v1[i] = vst->p[j];
                v2[i] = Tvst[1]->p[j];
                i++;
                for(k = 0; k < upwind->vst->n_chem; k++)
                {
                    v0[i] = Tvst[0]->chem[k][j];
                    v1[i] = vst->chem[k][j];
                    v2[i] = Tvst[1]->chem[k][j];
                    i++;
                }
            }
            limited_vel_slope(kmax,v0,v1,v2,slope);

            // TMP
            if(deb_npt)
            {
                printf("ns2d_transerve_vel_step, dir[%d] j[%d] slope [%10.9g %10.9g]\n",
                     idirs[0], j, slope[0], slope[1]);
                printf("v0[%10.9g, %10.9g], v1[%10.9g, %10.9g], v2[%10.9g, %10.9g]\n",
                 v0[0], v0[1], v1[0], v1[1], v2[0], v2[1]);
            }

            for(k = 0; k < kmax; k++)
            {
                if((j + 1) != end)
                    uL[k][j+1] += -0.5*dt*u[idirs[1]][j]*slope[k]/dn; 
                if(j != start)
                    uR[k][j] += -0.5*dt*u[idirs[1]][j]*slope[k]/dn;
                // OLD
                // uL[k][j] += -0.5*dt*u[idirs[1]][j-1]*slope[k]/dn; 
                // uR[k][j] += -0.5*dt*u[idirs[1]][j]*slope[k]/dn;
            }
            // icoords[idirs[0]] = j + pbuf;
        }

        if(deb_npt)
        {
            printf("edge velocity, after include trans deriv\n");
            for (j = start; j < end; ++j)
            {
                (void) printf("%-4d L:",j);
                for (k = 0; k < kmax; ++k)
                    printf(" %10.9g",uL[k][j]);
                printf(" ; R: ");
                for (k = 0; k < kmax; ++k)
                    printf(" %10.9g",uR[k][j]);
                printf("\n");
            }
        }
}

//  A Cell-Centered Cartesian Grid Projection
// Method for the Incompressible Euler
// equations in complex geometry, J. Bell, P. Colella
LOCAL void limited_vel_slope(
        int   dim,
        float *vl,
        float *v0,
        float *vr,
        float *slope)
{
        int   i;
        float b[40], f[40], c[40], d;
        float sign;

        for(i = 0; i < dim; i++)
        {
            b[i] = v0[i] - vl[i];    
            f[i] = vr[i] - v0[i];    
            c[i] = 0.5*(vr[i] - vl[i]);    
            if(b[i]*f[i] > 0.0)
            {
                sign = (c[i] == 0.0) ? 0.0 : (c[i] > 0.0) ? 1.0 : -1.0;    
                d = 2.0*min(fabs(b[i]), fabs(f[i]));
                slope[i] = sign*min(fabs(c[i]), d);
            }
            else
            {
                slope[i] = 0.0;
            }
        }
}
        
/* Now only velocity and density are reconstructed 
*/
LOCAL void ns2d_unsplit_bct_linear_reconstructor(
        int         start,
        int         end,
        Upwind_NS   *upwind)
{
        int       dim, j, k;
        int       kmax;
        float     *bk, *ck, *fk;
        float     **u, **backward, **central, **forward;
        float     *uk;

        dim = upwind->dim;

        if(upwind->vst->n_chem != 0)
            kmax = dim + 2 + upwind->vst->n_chem; // v, rho, p, chem
        else
            kmax = dim+1; 

        u = upwind->u;
        backward = upwind->backward;
        central = upwind->central;
        forward = upwind->forward;

        // TMP
        /*
        {
            printf("ns2d_unsplit_bct_linear_reconstructor, velocity\n");
            for (j = start; j < end; ++j)
            {
                printf("pos[%d]: v[", j);
                for(k = 0; k < dim; k++)
                {
                    uk = u[k];
                    printf(" %g, ", uk[j]);
                }
                printf("]\n");
            }
        }
        */

        for (k = 0; k < kmax; ++k)
        {
            bk = backward[k]; ck = central[k]; fk = forward[k];
            uk = u[k];
            for (j = start; j < end; ++j)
            {
                bk[j] = (uk[j] - uk[j-1]);
                ck[j] = 0.5 * (uk[j+1] - uk[j-1]);
                fk[j] = (uk[j+1] - uk[j]);
                /*
                if(debug_flag == YES && k == 1)
                {
                    printf("At J[%d] bk, ck, fk, %g %g %g\n", j, bk[j], ck[j], fk[j]);
                }
                */
            }
        }

        limiter_limiting(upwind->du,start,end,upwind);
}

LOCAL void limiter_limiting(
      float       **slope,
      int         start,
      int         end,
      Upwind_NS   *upwind)
{
      int         dim = upwind->dim;
      int         j, k, kmax;
      float       **backward, **central, **forward;
      float       *bk, *ck, *fk;
      float       *slopek; 
      float       sign, dl, abk, afk, ack;
      float       fact = 2.0; // 2.0 is van leer, 1.0 is min_mod

      // kmax = dim;
      if(upwind->vst->n_chem != 0)
          kmax = dim + 2 + upwind->vst->n_chem; // v, rho, p, chem
      else
          kmax = dim+1;

      backward = upwind->backward;
      central = upwind->central;
      forward = upwind->forward;

      for (k = 0; k < kmax; ++k)
      {
          bk = backward[k]; ck = central[k]; fk = forward[k];
          slopek = slope[k];
          for (j = start; j < end; ++j)
          {
              if (bk[j]*fk[j] > 0.0)
              {
                  abk = fabs(bk[j]);
                  afk = fabs(fk[j]);
                  dl = fact*min(abk,afk);
                  sign = (ck[j] == 0.0) ? 0.0 : (ck[j] > 0.0) ? 1.0 : -1.0;
                  ack = fabs(ck[j]);
                  slopek[j] = sign*min(ack,dl);
                  // slopek[j] = sign*min(abk,afk); /* mini_mod() */
              }
              else
              {
                  slopek[j] = 0.0;
              }
          }
      }
}
      
LOCAL void set_NS_gas_storage(
	float       **u,
        Vec_NS_Gas  *Vst,
        Upwind_NS   *upwind)
{
        int   dim, i;
        dim = upwind->dim;

        zero_scalar(Vst,sizeof(Vec_NS_Gas));

        ASSIGN_ARRAY_POINTER(Vst,rho,u[upwind->index.density]);
        ASSIGN_ARRAY_POINTER(Vst,p,u[upwind->index.pressure]);
        // Vst->rho = u[upwind->index.density];
        // Vst->p = u[upwind->index.pressure];

        VECTOR(Vst,v,dim,sizeof(float*));
        for(i = 0; i < dim; i++)
            Vst->v[i] = u[upwind->index.v[i]];

        if(Vst->n_chem != 0)
        {
            VECTOR(Vst,chem,Vst->n_chem,sizeof(float*));
            for(i = 0; i < Vst->n_chem; i++)
                Vst->chem[i] = u[upwind->index.chem[i]];
        }
}

LOCAL int  g_load_NS_state_vectors(
        int             swp_num,
        int             *iperm,
        Vec_NS_Gas      *vst,
        int             imin,
        int             imax,
        Wave            *wv,
        Wave            *newwv,
        int             *icrds,
        int             pbuf)
{
        Locstate        rst;
        Locstate        *state = vst->state;
        float           **coords = vst->coords;
        float           *rho = vst->rho;
        float           *p = vst->p;
        float           *m[MAXD], *v[MAXD];
        int             i, j, k;
        int             dim = wv->rect_grid->dim;
        int             idirs[MAXD];
        int             icoords[MAXD];
        int             start, end, non_obst;
        float           **chem = NULL;

        // clear_Vec_Gas_set_flags(vst);
        clear_vec_ns_gas_set_flags(vst);

        if(vst->n_chem != 0)
            chem = vst->chem;

        for (j = 0; j < dim; ++j)
        {
            idirs[j] = iperm[(j+swp_num)%dim];
            m[j] = vst->m[j];
            v[j] = vst->v[j];
            icoords[j] = icrds[j];
        }

        for (i = imin; i < imax; ++i)
        {
            icoords[idirs[0]] = i + pbuf;
            state[i] = rst = Rect_state(icoords,wv);
            coords[i] = Rect_coords(icoords,wv);
            rho[i] = Dens(rst);
            p[i] = Press(rst);
            for (j = 0; j < dim; ++j)
                v[j][i] = Vel(rst)[j];   // not permute with sweep direction.
            // Chemical components
            for (j = 0; j < vst->n_chem; ++j)
                chem[j][i] = Chem(rst)[j];
        }

        Vec_NS_Gas_field_set(vst,rho) = YES;
        Vec_NS_Gas_field_set(vst,p) = YES;
        // Vec_NS_Gas_field_set(vst,m) = YES;
        Vec_NS_Gas_field_set(vst,v) = YES;
        Vec_NS_Gas_field_set(vst,state) = YES;
        Vec_NS_Gas_field_set(vst,coords) = YES;
        // Chemical components
        Vec_NS_Gas_field_set(vst,chem) = YES;

        if(vst->n_chem != 0)
            vst->mu_chem = mu_chem;

        // NEW
        NS_set_params_jumps(vst,imin,imax);
        if ((Params(state[0]) == NULL) && (vst->nprms <= 1))
        {
            int nrad = vsten_radius(wv);
            copy_states_to_new_time_level(idirs,wv,newwv,imin+nrad,
                                          imax-nrad,icoords,pbuf);
            return CONSTANT_IN_TIME;
        }
        // g_init_obstacle_states(vst,imin,dim);

        for (i = 0; i < vst->nprms; ++i)
        {
            if (is_obstacle_state(vst->state[vst->prms_jmp[i]]))
            {
                start = vst->prms_jmp[i];
                end = vst->prms_jmp[i + 1];
      
                non_obst = (start == imin) ? end : start - 1;
                for (j = start; j < end; ++j)
                {
                    vst->rho[j] = vst->rho[non_obst];
                    vst->p[j] = vst->p[non_obst];
                    for (k = 0; k < dim; ++k)
                        vst->v[k][j] = vst->v[k][non_obst]; 
                    vst->state[j] = vst->state[non_obst];
                    // Chemical components
                    for (k = 0; k < vst->n_chem; ++k)
                        vst->chem[k][j] = vst->chem[k][non_obst];

                }
            }
        }
        return DYNAMIC_IN_TIME;
        // END NEW
}

LOCAL  void    NS_set_params_jumps(
        Vec_NS_Gas      *vst,
        int             imin,
        int             imax)
{
        Locstate        *state = vst->state;
        int             *prms_jmp = vst->prms_jmp;
        int             i, j, nprms;

        if (Vec_Gas_field_set(vst,prms_jmp))
            return;

        nprms = 0;
        j = imin;
        while (j < imax)
        {
            prms_jmp[nprms++] = j;
            for (i = j+1; i < imax; ++i)
                if (Different_params(state[i],state[j]))
                    break;
            j = i;
        }
        prms_jmp[nprms] = imax;
        Vec_Gas_field_set(vst,prms_jmp) = YES;

        vst->nprms = nprms;
}               /*end set_params_jumps*/

LOCAL   Vec_NS_Gas* alloc_ns_vgas(
        Vec_NS_Gas* vst,
        int      vctr_size,
        int      dim,
        int      n_chem)
{
        if (vst == NULL)
        {
            scalar(&vst,sizeof(Vec_Gas));
            vst->alloc.vst = YES;
        }

        MATRIX(vst,m,dim,vctr_size,FLOAT);
        MATRIX(vst,v,dim,vctr_size,FLOAT);
        VECTOR(vst,rho,vctr_size,FLOAT);
        VECTOR(vst,p,vctr_size,FLOAT);
        VECTOR(vst,state,vctr_size,sizeof(Locstate));
        VECTOR(vst,coords,vctr_size,sizeof(float*));
        VECTOR(vst,prms_jmp,vctr_size+1,INT);
        vst->n_chem = n_chem; 
        if(n_chem != 0)
        {
            int i;
            VECTOR(vst,chem,n_chem,sizeof(float*)); //for chemical component
            for(i = 0; i < n_chem; i++)
            {
                vector(&(vst->chem[i]), vctr_size,FLOAT);
            }
        }
        return vst;
}               /*end g_alloc_vgas*/

LOCAL void free_ns_vgas(
        Vec_NS_Gas* vst)
{
        Set_free(vst,m);
        Set_free(vst,v);
        Set_free(vst,rho);
        Set_free(vst,p);
        Set_free(vst,state);
        Set_free(vst,coords);
        Set_free(vst,prms_jmp);
        if(vst->n_chem != 0)
        {
            int i;
            for(i = 0; i < vst->n_chem; i++)
            {
                free(vst->chem[i]); 
            }
            Set_free(vst,chem); 
        }

        if(vst->alloc.vst == YES)
            free(vst);
}

LOCAL void clear_vec_ns_gas_set_flags(
        Vec_NS_Gas* vst)
{
        Vec_NS_Gas_field_set(vst,rho) = NO;
        Vec_NS_Gas_field_set(vst,p) = NO;
        Vec_NS_Gas_field_set(vst,m) = NO;
        Vec_NS_Gas_field_set(vst,v) = NO;
        Vec_NS_Gas_field_set(vst,state) = NO;
        Vec_NS_Gas_field_set(vst,coords) = NO;
        Vec_NS_Gas_field_set(vst,chem) = NO;
}

/* Set vector field u^* on cell edge for projection
 * Some interpolation has not been thought through.

 * The edge velocities of boundary cells is set to be zero
 Cell Case: crxing dir[0,1,2,3] = [1, 0, 0, 1]
 Below this cell: obstacle.
 The edge velocity of [1][0] (lower side) should be zero 
*/

/*
Used augmented grid size. The ghost states are set by
scatter_diff_conv_vel()
*/

LOCAL void cell_edge_vector_u_ver4(
        float             dt,
        NS_Field_Stencil  *ns_field)
{
        int             smin[MAXD], smax[MAXD];
        Wave            *newwv = ns_field->wave;
        Wave            *wv = ns_field->newwave;
        TRI_GRID        *ntg = wave_tri_soln(ns_field->newwave)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        Front           *nfr = ns_field->newfr, *fr = ns_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD], icrds[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vnew[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh = wv->rect_grid->h;
        int             vset;
        RECT_GRID       *gr = wv->rect_grid;

        if(debugging("NS_vel") && NO)
        {
            printf("--------------Entered cell_edge_vector_u_ver4\n");
            printf("Crank-Nicolson velocity:\n");
            for (iy = 0; iy < gmax[1]; iy++)
            {
                printf("Row-cell[%d]: ", iy);
                for (ix = 0; ix < min(4,gmax[0]); ix++)
                {
                    ic[0] = ix; ic[1] = iy;
                    printf("(%d %g %g) ", ix, Rect_u_star(ic,ns_field).v[0],
                             Rect_u_star(ic,ns_field).v[1]);
                }
                printf("\n");
            }
        }

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }
        smin[0] = smin[1] = 0;
        smax[0] = gmax[0]; smax[1] = gmax[1];
        // TMP
        // printf("cell_edge_vector_u_ver4, x-size[%d %d], y-size[%d %d]\n",
        //        smin[0], smax[0], smin[1], smax[1]);

        // Vertical edge 
        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            // for (ix = smin[0]+1;  ix < smax[0];  ++ix)
            for (ix = smin[0];  ix <= smax[0];  ++ix)
            {
                ic[0] = ix-1; ic[1] = iy;
                ic1[0] = ix; ic1[1] = iy;


                // comp = Rect_comp(ic,newwv);
                // comp1 = Rect_comp(ic1,newwv);

                // printf("on ix, iy [%d %d], comp %d\n", ix, iy, comp);

                // if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                //    comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                //    comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    for(i = 0; i < dim; i++)
                    {
                        v_line_vel(ic1,ns_field).v[i] = vnew[i] = 
                          0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]);
                    }
                }
                /* else 
                {
                    // Cell on the right side of the edge
                    // Dirichlet boundary type, might need to re-consider
                    if(blk_edge[ix][iy].edge[0][0].etype == D_SURF)
                    {
                        for(i = 0; i < dim; i++)
                            v_line_vel(ic1,ns_field).v[i] = 0.0;
                    }

                    // If one of the cell is obsctacle,
                    // Set the velocity on this common edge to be zero
                    // for(i = 0; i < dim; i++)
                    //     v_line_vel(ic1,ns_field).v[i] = 0.0;
                }
                */
            }
        }

        // Left vertical edge
        /****
        if(! buffered_boundary_type(rect_boundary_type(nfr->interf,0,0)))
        {
            ic[0] = ix = smin[0];
            prev_side = WEST;
            next_side = EAST;
            for (iy = smin[1];  iy < smax[1];  ++iy)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[1] = iy;
                comp = Rect_comp(ic,newwv);

                if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                     continue;

            // Neumann boundary type, set to 0.0
            // TMP
            // printf("cell_edge_vector_u_ver2, cell[%d %d], left edge_type = %d, N_SURF = %d\n",
            //                 ix, iy, blk_edge[ix][iy].edge[0][0].etype, N_SURF);

                if(Rect_blk_edge(icrds,wv).edge[0][0].etype == N_SURF)
                {
                    for(i = 0; i < dim; i++)
                        v_line_vel(ic,ns_field).v[i] = 0.0;
                    continue;
                }

                sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);
            
                sten->fr  = nfr;
                sten->wave = newwv;
                sten->newwave = newwv;
                ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
                if(sten->nc[-1] != 0)
                {
                    if(sten->nc[-1] > 1)
                    {
                        for(i = 0; i < sten->nc[-1]; i++)
                        {
                            if(wave_type(sten->crx[-1][i]->hs) == NEUMANN_BOUNDARY ||
                               wave_type(sten->crx[-1][i]->hs) == DIRICHLET_BOUNDARY)
                            {
                                newcrx = sten->crx[-1][i];
                                if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                    comp = negative_component(newcrx->hs);
                                else
                                    comp = positive_component(newcrx->hs);
                                break;
                            }
                        }
                    }
                    else
                        newcrx = sten->crx[-1][0];
                    vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
                }
                else
                {
                    printf("ERROR: cell_edge_vector_u_ver4(), left boundary, no crossing\n");
                    clean_up(ERROR);
                }

                // Dirichlet 
                for(i = 0; i < dim; i++)
                    v_line_vel(ic,ns_field).v[i] = vnew[i];

                // printf("cell[%d %d], left side: vnew[%g %g]\n",
                //        ic[0], ic[1], vnew[0], vnew[1]);
            }
        }

        // printf("cell_edge_vector_u_ver4(), right boundary before\n");
        
        // Right vertical edge
        if(! buffered_boundary_type(rect_boundary_type(nfr->interf,0,1)))
        {
            ic[0] = ix = smax[0]-1;
            ic1[0] = smax[0];
            prev_side = WEST;
            next_side = EAST;
            for (iy = smin[1];  iy < smax[1];  ++iy)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic1[1] = ic[1] = iy;
                comp = Rect_comp(ic,newwv);

                if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                     continue;

                // Neumann boundary type, set to 0.0
                if(Rect_blk_edge(icrds,wv).edge[0][1].etype == N_SURF)
                {
                    for(i = 0; i < dim; i++)
                        v_line_vel(ic1,ns_field).v[i] = 0.0;
                    continue;
                }

                sten_pt_loc_in_dir(ic,0,end,wv,YES,sten);
                
                sten->fr  = nfr;
                sten->wave = newwv;
                sten->newwave = newwv;
                ncomp = set_NS_stencil_comps_and_crosses(0,0,iperm,sten,end);
                if(sten->nc[1] != 0)
                {
                    if(sten->nc[1] > 1)
                    {
                        // printf("WARNING: cell_edge_vector_u_ver4, sten->nc[1] = %d\n",
                        //       sten->nc[1]);
                        // printf("wave type %d\n", wave_type(sten->crx[1][0]->hs));
                        for(i = 0; i < sten->nc[1]; i++)
                        {
                            if(wave_type(sten->crx[1][i]->hs) == NEUMANN_BOUNDARY ||
                               wave_type(sten->crx[1][i]->hs) == DIRICHLET_BOUNDARY)
                            {
                                newcrx = sten->crx[1][i];
                                if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                    comp = negative_component(newcrx->hs);
                                else
                                    comp = positive_component(newcrx->hs);

                                // printf("comp = %d\n", comp);
                                // print_general_vector("point crds",Coords(sten->p[1]),dim,"\n"); 
                                // print_curve(Curve_of_hs(newcrx->hs));
                                // print_crxings(newcrx, NO);
                                // clean_up(ERROR);
                                break;
                            }
                        } 
                    }
                    else
                        newcrx = sten->crx[1][0];
                    vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
                }
                else
                {
                    printf("ERROR: cell_edge_vector_u_ver4(), left boundary, no crossing\n");
                    clean_up(ERROR);
                }

                // Dirichlet 
                for(i = 0; i < dim; i++)
                    v_line_vel(ic1,ns_field).v[i] = vnew[i];
            }
        }
        ***/

        // Horizontal edge 
        // for (iy = smin[1]+1;  iy < smax[1];  ++iy)
        for (iy = smin[1];  iy <= smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                ic[0] = ix; ic[1] = iy-1;
                ic1[0] = ix; ic1[1] = iy;

                // comp = Rect_comp(ic,newwv);
                // comp1 = Rect_comp(ic1,newwv);

                /**
                if(ix == 3 && (iy == 1 || iy == 8 || iy ==9))
                {
                    printf("**************\n");
                    printf("cell_edge_vector_u_ver2 on ix, iy [%d %d], ctype = %d\n", ix, iy,
                                    blk_edge[ix][iy].ctype);
                    printf("comp = %d, comp1 = %d [ix][iy].edge[1][0].etype =%d\n", 
                          comp, comp1,blk_edge[ix][iy].edge[1][0].etype);
                    printf("******END********\n");
                }
                **/ 

                // if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                //    comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                //    comp1 != POTS_COMP && comp != POTS_COMP)
                {
                    for(i = 0; i < dim; i++)
                    {
                        h_line_vel(ic1,ns_field).v[i] = vnew[i] = 
                             0.5*(Rect_u_star(ic,ns_field).v[i] + Rect_u_star(ic1,ns_field).v[i]);
                    }

                    /**
                    if(ix == 0 && (iy == 0 || iy == 1))
                    {
                        printf("cell_edge_vector_u_ver4(): set ic1[%d %d] h_vel (%g %g) case 0\n",
                             ic1[0], ic1[1], h_line_vel(ic1,ns_field).v[0], 
                             h_line_vel(ic1,ns_field).v[1]);
                        printf("ic[%d %d] ic1[%d %d] cell center vel(%g %g) (%g %g)\n",
                                  ic[0], ic[1], ic1[0], ic1[1], 
                               Rect_u_star(ic,ns_field).v[0],
                               Rect_u_star(ic,ns_field).v[1], Rect_u_star(ic1,ns_field).v[0],
                              Rect_u_star(ic1,ns_field).v[1]); 
                    }
                    **/
                }
                /**
                else 
                {
                    // Cell on the upper side of the edge
                    // If one of the cell is obsctacle, 
                    // Set the velocity on this common edge to be zero

                    // Dirichlet boundary type, might need to re-consider
                    if(blk_edge[ix][iy].edge[1][0].etype == D_SURF)
                    {
                        // for(i = 0; i < dim; i++)
                        //     h_line_vel(ic1,ns_field).v[i] = 0.0;
                    }
                    for(i = 0; i < dim; i++)
                        h_line_vel(ic1,ns_field).v[i] = 0.0;
                }
                **/
            }
        }

        // bottom horizontal edge
        /****
        ic[1] = iy = smin[1];
        prev_side = SOUTH;
        next_side = NORTH;
        if(! buffered_boundary_type(rect_boundary_type(nfr->interf,1,0)))
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic[0] = ix;
                comp = Rect_comp(ic,newwv);

                if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                     continue;

                // Neumann boundary type, set to 0.0
                if(Rect_blk_edge(icrds,wv).edge[1][0].etype == N_SURF)
                {
                    for(i = 0; i < dim; i++)
                        h_line_vel(ic,ns_field).v[i] = 0.0;
                    continue;
                }

                sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);

                sten->fr  = nfr;
                sten->wave = newwv;
                sten->newwave = newwv;
                ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
                if(sten->nc[-1] != 0)
                {
                    if(sten->nc[-1] > 1)
                    {
                        for(i = 0; i < sten->nc[-1]; i++)
                        {
                            if(wave_type(sten->crx[-1][i]->hs) == NEUMANN_BOUNDARY ||
                               wave_type(sten->crx[-1][i]->hs) == DIRICHLET_BOUNDARY)
                            {
                                newcrx = sten->crx[-1][i];
                                if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                    comp = negative_component(newcrx->hs);
                                else
                                    comp = positive_component(newcrx->hs);
                                break;
                            }
                        }
                    }
                    else
                        newcrx = sten->crx[-1][0];
                    // newcrx = sten->crx[-1][0];
                    vset = velocity_from_boundary(sten,newcrx,sten->p[-1],
                           comp,sten->worksp[-1],vnew);
                }
                else
                {
                    printf("ERROR: cell_edge_vector_u_ver4(), left boundary, no crossing\n");
                    clean_up(ERROR);
                }
                // Dirichlet 
                for(i = 0; i < dim; i++)
                    h_line_vel(ic,ns_field).v[i] = vnew[i];

                // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
                //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
            }
        }

        // top horizontal edge
        if(! buffered_boundary_type(rect_boundary_type(nfr->interf,1,1)))
        {
            ic[1] = iy = smax[1]-1;
            ic1[1] =  smax[1];

            prev_side = SOUTH;
            next_side = NORTH;
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                icrds[0] = ix; icrds[1] = iy;
                ic1[0] = ic[0] = ix;
                comp = Rect_comp(ic,newwv);

                if(Rect_blk_edge(icrds,wv).ctype == EXTERNAL)
                     continue;

                // Neumann boundary type, set to 0.0
                if(Rect_blk_edge(icrds,wv).edge[1][1].etype == N_SURF)
                {
                    for(i = 0; i < dim; i++)
                        h_line_vel(ic1,ns_field).v[i] = 0.0;
                    continue;
                }

                sten_pt_loc_in_dir(ic,1,end,wv,YES,sten);
            
                sten->fr  = nfr;
                sten->wave = newwv;
                sten->newwave = newwv;
                ncomp = set_NS_stencil_comps_and_crosses(1,1,iperm,sten,end);
                if(sten->nc[1] != 0)
                {
                    if(sten->nc[1] > 1)
                    {
                        for(i = 0; i < sten->nc[1]; i++)
                        {
                            if(wave_type(sten->crx[1][i]->hs) == NEUMANN_BOUNDARY ||
                               wave_type(sten->crx[1][i]->hs) == DIRICHLET_BOUNDARY)
                            {
                                newcrx = sten->crx[1][i];
                                if(is_exterior_comp(positive_component(newcrx->hs),nfr->interf))
                                    comp = negative_component(newcrx->hs);
                                else
                                    comp = positive_component(newcrx->hs);
                                break;
                            }
                        }
                    }
                    else
                        newcrx = sten->crx[1][0];
                    // newcrx = sten->crx[1][0];
                    vset = velocity_from_boundary(sten,newcrx,sten->p[1],
                           comp,sten->worksp[1],vnew);
                }
                else
                {
                    printf("ERROR: cell_edge_vector_u_ver4(), top boundary, no crossing\n");
                    clean_up(ERROR);
                }

                // Dirichlet 
                for(i = 0; i < dim; i++)
                    h_line_vel(ic1,ns_field).v[i] = vnew[i];

                // printf("ic1[%d %d] set h_line_vel as Dirichlet vnew[%g %g]\n",
                //       ic1[0], ic1[1], vnew[0], vnew[1]);

                // printf("cell[%d %d], vold[%g %g], vnew[%g %g]\n",
                //         ic[0], ic[1], vold[0], vold[1], vnew[0], vnew[1]);
            }
        }
        ****/

        // TMP
        if(debugging("NS_vel"))
        {
            printf("----------------------------------------\n");
            printf("\n\ncell_edge_vector_u_ver4******vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            {
                if(j ==0 || j == gmax[1]) 
                { 

                printf("row%d: ", j);
                for(i = 0; i < gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf("%d %g %g; ", i,
                        h_line_vel(ic, ns_field).v[0], h_line_vel(ic, ns_field).v[1]);
                }
                // ic[0] = gmax[0]-1; ic[1] = j;
                // printf(" (%d, %9.8g %9.8g) ", ic[0],
                //         h_line_vel(ic, ns_field).v[0], h_line_vel(ic, ns_field).v[1]);
                printf("\n");
                }
            }
            /**
            printf("\n\ncell_edge_vector_u_ver4*******vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf("%d %g %g; ", i,
                       v_line_vel(ic, ns_field).v[0],v_line_vel(ic, ns_field).v[1]);
                }
                // ic[0] = gmax[0]; ic[1] = j;
                // printf(" (%d, %g %g) ", ic[0], 
                //       v_line_vel(ic, ns_field).v[0],v_line_vel(ic, ns_field).v[1]);
                printf("\n");
            }
            **/
            printf("----------------------------------------\n");
        }
}

/* 
 * grid_den_viscosity: Does count partial 
 * (including physical boundary) cells as
 * full cells. This is like the ghost cell method.
 * The boundary geometry is simplified.
 * The density is set into Rect_state() for 
 * half-time step reconstruction
*/
LOCAL void grid_den_viscosity_ver1(
        NS_Field_Stencil  *ns_field,
        Wave              *wv,
        Front             *fr)
{
        int             smin[MAXD], smax[MAXD];
        // Wave            *newwv = ns_field->newwave;
        // TRI_GRID        *ntg = wave_tri_soln(ns_field->newwave)->tri_grid;
        TRI_GRID        *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        // Front           *nfr = ns_field->newfr;
        int             ix, iy, ic[MAXD], i, j, k, i1;
        Gas_param       **prms_list;
        int             nprms;
        COMPONENT       *comps, comp, comp1;
        float           **den_vis, *a_frac;
        int             deb_flag = NO;
        RECT_GRID       *gr = wv->rect_grid;

        if(ns_field->n_eos == 0)
        {
            nprms = return_params_list(&prms_list);
            vector(&comps, nprms, sizeof(COMPONENT));
            matrix(&den_vis,nprms,2,sizeof(float));
 
            k = 0;
            for (i=0; i<nprms; i++)
            {
                if(prms_list[i]->eos->_NS_eos == YES)
                {
                    comps[k] = prms_list[i]->comp;
                    den_vis[k][0] = prms_list[i]->dens;
                    den_vis[k][1] = prms_list[i]->shear_visc;
                    k++;
                }
            }
            ns_field->n_eos = k;
            ns_field->comps = comps;
            ns_field->den_vis = den_vis;
        }
        else
        {
            k = ns_field->n_eos;
            comps = ns_field->comps;
            den_vis = ns_field->den_vis;
        }

        // TMP
        /***
        printf("In grid_den_viscosity_ver1, ns_field->n_eos = %d\n",
                ns_field->n_eos);
        for(i = 0; i < k; i++)
        {
            printf("param[%d] dens %g, viscosity %g, comp %d\n", i, 
                prms_list[i]->dens, prms_list[i]->shear_visc, prms_list[i]->comp);
        }
        ***/

        for(i = 0; i < dim; i++) 
        {
            smin[i] = -gr->lbuf[i];
            smax[i] = gr->gmax[i] + gr->ubuf[i];
        }

        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                ic[0] = ix; ic[1] = iy;

                // TMP
                /**
                if((ix == 2 && iy == 8) || 
                   (ix == 3 && iy == 8) ||
                    (ix == 4 && iy == 8) )
                {
                    deb_flag = YES;
                    printf("\ngrid_den_viscosity_ver1 cell[%d %d] ctype = %d\n", 
                             ix, iy, Rect_blk_edge(ic,wv).ctype);
                    printf("Rect_comp = %d, %d (newwv), config = %d\n\n", 
                          Rect_comp(ic,wv), Rect_comp(ic,ns_field->newwave),
                          Rect_blk_edge(ic,wv).config);
                }
                else
                    deb_flag = NO;
                **/
                
                Rect_den(ic,ns_field) = 0.0; 
                Rect_visc(ic,ns_field) = 0.0; 

                if(Rect_blk_edge(ic,wv).ctype == EXTERNAL)
                    continue;
                comp = Rect_comp(ic,wv);
                a_frac = Rect_blk_edge(ic,wv).a_frac;

                if(Rect_blk_edge(ic,wv).ctype == PARTIAL)
                {
                    for(i = 0; i < k; i++)
                    {
                        if(comp == comps[i])
                            break;
                    }
                    if(i >= k)
                    {
                        printf("ERROR grid_den_viscosity_ver1 cell[%d %d] ctype = %d, comp = %d\n", 
                             ix, iy, Rect_blk_edge(ic,wv).ctype, comp);
                        printf("case ---1\n");
                        clean_up(ERROR);
                    }
                    if(Rect_blk_edge(ic,wv).config == 0 ||
                       Rect_blk_edge(ic,wv).config == 1 ||
                       Rect_blk_edge(ic,wv).config == 3 ||
                       Rect_blk_edge(ic,wv).config == 4 ||
                       Rect_blk_edge(ic,wv).config == 5)
                    {
                        Rect_den(ic,ns_field) = den_vis[i][0]; 
                        Rect_visc(ic,ns_field) = den_vis[i][1]; 

                        if(deb_flag == YES)
                        {
                            printf("case 0, ic[%d %d] a_frac = %9.8g\n", ix, iy, 
                                den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[0]);
                        }
                        Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                    }
                    /***
                    else if(Rect_blk_edge(ic,wv).config == 3)
                    {
                        Rect_den(ic,ns_field) = den_vis[i][0]; 
                        Rect_visc(ic,ns_field) = den_vis[i][1]; 

                        if(deb_flag == YES)
                        {
                            printf("case 1, ic[%d %d] a_frac = %9.8g\n", ix, iy,
                                    den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[1]);
                        }
                        Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                    }
                    ***/
                    else if(Rect_blk_edge(ic,wv).config == 2 ||
                            Rect_blk_edge(ic,wv).config == 6)
                    {
                        if(Rect_blk_edge(ic,wv).corn_comp[0][0] == comp)
                        {
                            Rect_den(ic,ns_field) = den_vis[i][0]; 
                            Rect_visc(ic,ns_field) = den_vis[i][1]; 

                            if(deb_flag == YES)
                            {
                                printf("case 2, ic[%d %d] a_frac = %9.8g\n", ix, iy,
                                    den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[0]);
                            }
                            Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                        }  
                        else
                        {
                            Rect_den(ic,ns_field) = den_vis[i][0]; 
                            Rect_visc(ic,ns_field) = den_vis[i][1]; 

                            if(deb_flag == YES)
                            {
                                printf("case 3, ic[%d %d] a_frac = %9.8g\n", ix, iy,
                                    den_vis[i][0]*Rect_blk_edge(ic,wv).a_frac[1]);
                            }
                            Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                        }
                    }
                    else
                    {
                        printf("ERROR: update_den_viscosity_ver1 handle case: PARTIAL config %d\n",
                             Rect_blk_edge(ic,wv).config);
                        printf("IN update_den_viscosity, [%d][%d].ctype = %d\n",
                           ix, iy, Rect_blk_edge(ic,wv).ctype);
                        for(i = 0; i < 2; i++)
                        {
                            for(j = 0; j < 2; j++)
                            {
                                printf("blk->edge[%d][%d]->type = %d ,",
                                 i, j, Rect_blk_edge(ic,wv).edge[i][j].etype);
                            }
                            printf("\n");
                        }
                        clean_up(ERROR);
                    }
                }
                else if(Rect_blk_edge(ic,wv).ctype == INTERNAL)
                {
                    for(i = 0; i < k; i++)
                    {
                        if(comp == comps[i])
                        {
                            Rect_den(ic,ns_field) = den_vis[i][0];
                            Rect_visc(ic,ns_field) = den_vis[i][1];
                            break;
                        } 
                    }

                    /**
                    if(deb_flag == YES)
                    {
                        printf("Internal, frac[0] %9.8g, frac[1] = %9.8g += %9.8g\n",
                            a_frac[0], a_frac[1], a_frac[1]+a_frac[0]);
                    }
                    **/
                    Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                }
                else if(Rect_blk_edge(ic,wv).ctype == INTER_PART)
                {
                    // TMP
                    // if(ic[0] == 0 || ic[0] == 1)
                    //     printf("Internal, frac cell[%d %d]\n", ic[0], ic[1]);

                    if(Rect_blk_edge(ic,wv).config == 1 ||
                       Rect_blk_edge(ic,wv).config == 3)    
                    {
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][0] == comps[i])
                                break;
                        }
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][0] == comps[i1])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                        /**
                        if(deb_flag == YES)
                        {
                            printf("case 1-3, frac[0] %9.8g, frac[1] = %9.8g += %9.8g\n",
                                a_frac[0], a_frac[1], a_frac[1]+a_frac[0]);
                        }
                        **/ 
                        Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                    }
                    else if(Rect_blk_edge(ic,wv).config == 4)
                    {
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][1] == comps[i])
                                break;
                        }
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][1] == comps[i1])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                        /**
                        if(deb_flag == YES)
                        {
                            printf("case-4 frac(%9.8g %9.8g) += %9.8g, den_vis[%d][0] = %9.8g"
                                              " den_vis[%d][0] = %9.8g\n", 
                                a_frac[0], a_frac[1], a_frac[1]+a_frac[0], 
                                i, den_vis[i][0], i1, den_vis[i1][0]);
                        }
                        **/
                        Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                    }
                    else if(Rect_blk_edge(ic,wv).config == 5)
                    {
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][1] == comps[i1])
                                break;
                        }
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][1] == comps[i])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];
                        /**
                        if(deb_flag == YES)
                        {
                            printf("case 5, frac[0] %9.8g, frac[1] = %9.8g += %9.8g,"
                                  " den_vis[%d][0] = %9.8g den_vis[%d][0] = %9.8g\n",
                                a_frac[0], a_frac[1], a_frac[1]+a_frac[0],
                                   i, den_vis[i][0], i1, den_vis[i1][0]);
                        }
                        **/
                        Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);
                    }
                    else if(Rect_blk_edge(ic,wv).config == 2 ||
                            Rect_blk_edge(ic,wv).config == 6)
                    {
                        for(i = 0; i < k; i++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[0][0] == comps[i])
                                break;
                        }
                        for(i1 = 0; i1 < k; i1++)
                        {
                            if(Rect_blk_edge(ic,wv).corn_comp[1][1] == comps[i1])
                                break;
                        }
                        Rect_den(ic,ns_field) = den_vis[i][0]*a_frac[0] + 
                                                den_vis[i1][0]*a_frac[1];
                        Rect_visc(ic,ns_field) = den_vis[i][1]*a_frac[0] + 
                                                den_vis[i1][1]*a_frac[1];

                        Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);

                        /**
                        if(deb_flag == YES)
                        {
                            printf("case 2-6, frac[0] %9.8g, frac[1] = %9.8g += %9.8g,"
                                      " den[%d %d] = %11.10g\n", 
                                a_frac[0], a_frac[1], a_frac[1]+a_frac[0], ic[0], ic[1], Dens(Rect_state(ic,wv)));
                        }
                        **/
                    }
                    else
                    {
                        printf("ERROR: update_den_viscosity handle case INTER_PART: config %d\n",
                                 Rect_blk_edge(ic,wv).config);
                        printf("IN grid_den_viscosity_ver1, [%d][%d].ctype = %d\n",
                           ix, iy, Rect_blk_edge(ic,wv).ctype);
                        for(i = 0; i < 2; i++)
                        {
                            for(j = 0; j < 2; j++)
                            {
                                printf("blk->edge[%d][%d]->type = %d ,",
                                 i, j, Rect_blk_edge(ic,wv).edge[i][j].etype);
                            }
                            printf("\n");
                        }
                        clean_up(ERROR);
                    }
                    // TMP
                    /**
                    if(ic[0] == 0 || ic[0] == 1 || ic[0] == 49 || ic[0] == 48)
                    {
                        printf("IN grid_den_viscosity_ver1, [%d][%d].ctype = %d, dens %12.11g, visc %12.11g\n",
                           ix, iy, Rect_blk_edge(ic,wv).ctype, Rect_den(ic,ns_field), Rect_visc(ic,ns_field));
                    }
                    **/
                }
                else
                {
                    printf("ERROR: grid_den_viscosity handle case:\n");
                    printf("IN grid_den_viscosity, [%d][%d].ctype = %d\n",
                       ix, iy, Rect_blk_edge(ic,wv).ctype);
                    for(i = 0; i < 2; i++)
                    {
                        for(j = 0; j < 2; j++)
                        {
                            printf("blk->edge[%d][%d]->type = %d ,",
                             i, j, Rect_blk_edge(ic,wv).edge[i][j].etype);
                        }
                        printf("\n");
                    }
                    clean_up(ERROR);
                }
                // DEBUG TMP
                if(deb_flag == YES) 
                {
                    printf("IN grid_den_viscosity, [%d][%d] comp %d, den = %g, mu = %g\n",
                       ix, iy, comp, Rect_den(ic,ns_field), Rect_visc(ic,ns_field));
                }
            }
        }

        // free(comps);
        // free(den_vis); 
        /**
        {
           int ic2[MAXD];  
           ic[0] = 0; ic[1] = 0;
           ic2[0] = 0; ic2[1] = gmax[1]-1;
           printf("Cell density (%d %d, %g), (%d %d, %g)\n", 
                 ic[0], ic[1], Rect_den(ic,ns_field),
                 ic2[0], ic2[1], Rect_den(ic2,ns_field));
        }
        **/

        if(debugging("NS_vel"))
        {
            printf("-----Cell Density-------------\n");
            // for (iy = smin[1];  iy < smax[1];  ++iy)
            /**
            for (iy = 6;  iy <= 13;  ++iy)
            {
                printf("Row[%d] dens: ",iy);
                for (ix = 0;  ix < gmax[0];  ++ix)
                {
                    ic[0] = ix; ic[1] = iy;
                    printf("(%d %g)", ix, Rect_den(ic,ns_field));
                }
                printf("\n");
            }
            **/
        
            printf("-----Cell Density difference-------------\n");
            for (iy = 6;  iy <= 13;  ++iy)
            {
                printf("Row[%d] dens-diff: ",iy);
                for (ix = 0;  ix < (gmax[0])/2;  ++ix)
                {
                    int ic2[MAXD];  
                    ic[0] = ix; ic[1] = iy;
                    ic2[0] = gmax[0]-ix-1; ic2[1] = iy;
                    printf("(%d %g)", ix, Rect_den(ic,ns_field)-Rect_den(ic2,ns_field));
                }
                printf("\n");
            }
        }
}

/* set projection stencil to solve
 * DG\phi = DU before construction of convective derivatives.
 * \phi on the left boundary is set to be zero.
 * This is to use Dirichlet boundary condition on the left side.
 * On the partial cell, if a partial edge is the common
 * boundary of a obstacle cell and an interior(partial) cell,
 * this cell edge is treated as boundary. If a partial edge
 * is the common boundary of two interior(partial) cells, this
 * edge is treated as an interior edge. (allow fluild to flow through).
 * phis of the left side cells are set to be zero. 
 */
LOCAL void init_stencil_element2D_ver1(
         BLK_EDGE      **blk_edge,
         P_CELL2D      **p_cell,
         Front         *fr,
         Wave          *wv,
         NS_Field_Stencil *ns_field,
         int           add_time,
         float         dt)
{
       RECT_GRID     *gr = wv->rect_grid;
       int        xmax,ymax;
       int        ix,iy;
       int        dir,j,k,nor;
       int        icoords[MAXD], ic[MAXD];
       float      *coords,coord[MAXD];      // coords is the cell center
                                         // and coord is the cell edge center
       float      *h = gr->h;
       float      rweight = 0.0, dt_coeff = 1.0;

       xmax = gr->gmax[0];
       ymax = gr->gmax[1];

       if(add_time == YES)
           dt_coeff = dt;

       for (iy = 0;  iy < ymax;  ++iy)
       {
           for (ix = 1;  ix < xmax;  ++ix)
           {
               icoords[0] = ix; 
               icoords[1] = iy;
               coords = Rect_coords(icoords,wv);

               for(k = 0; k < 3; ++k)
                  for(j = 0; j < 3; ++j)
               {
                  {
                     p_cell[ix][iy].set[k][j] = NO;
                     p_cell[ix][iy].stencil[k][j] = 0.0;
                  }
               }

               p_cell[ix][iy].r_side = 0.0;

               // TMP
               /**
               if(ix == 9 && (iy == 9 || iy == 0))
                   deb_sten = YES;
               else  
                   deb_sten = NO;
               **/

               if(blk_edge[ix][iy].ctype == EXTERNAL)
                   continue; 

               if(blk_edge[ix][iy].ctype == PARTIAL)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);

                           if(blk_edge[ix][iy].edge[dir][j].etype == D_SURF ||
                              blk_edge[ix][iy].edge[dir][j].etype == N_SURF ||
                              blk_edge[ix][iy].edge[dir][j].etype == PARTIAL_SURF)
                           {
                               coord[0] = blk_edge[ix][iy].edge[dir][j].cen[0];
                               coord[1] = blk_edge[ix][iy].edge[dir][j].cen[1];
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == FULL_SURF)
                           {
                               coord[dir] = coords[dir]+ nor*h[dir]/2; 
                               coord[(dir+1)%2] = coords[(dir+1)%2];
                           }
                           else
                               continue;
 
                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].edge[dir][j].etype == FULL_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == NO) 
                                   regular_fd_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight);
                               else
                               {
                                   bdry_cell_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                        icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                                   // printf("WARNING: init_stencil_element2D, full_edge,"
                                   //     " but the outter side is body cell[%d %d]\n",
                                   //             ix, iy);
                                   // clean_up(ERROR);
                               }
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == D_SURF ||
                                   blk_edge[ix][iy].edge[dir][j].etype == N_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == NO) 
                                   regular_fd_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight);
                               else
                                   bdry_cell_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == PARTIAL_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == NO)
                                   regular_fd_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight);
                               else
                                   bdry_cell_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, partial edge 0,"
                               //     " implementing\n");
                               // clean_up(ERROR);
                           }
                           else
                           {
                               printf("ERROR: init_stencil_element2D, partial edge,"
                                   " implementing\n");
                               clean_up(ERROR);
                           }
                       }
                   }
               }
               else if(blk_edge[ix][iy].ctype == INTERNAL)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           coord[dir] = coords[dir]+ nor*h[dir]/2; coord[(dir+1)%2] = coords[(dir+1)%2];

                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].comp_label[dir][j] == NO)
                           {
                               regular_fd_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                          icoords,ns_field,rweight);
                           }
                           else
                           {
                               bdry_cell_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //     " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                       }
                   }
               }
               else if(blk_edge[ix][iy].ctype == INTER_PART)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           coord[dir] = coords[dir]+ nor*h[dir]/2; coord[(dir+1)%2] = coords[(dir+1)%2];

                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].comp_label[dir][j] == NO)
                           {
                               regular_fd_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight);
                           }
                           else
                           {
                               bdry_cell_sten(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //     " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                       }
                   }                   
                   /***
                   printf("ERROR: init_stencil_element2D handle case INTER_PART:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, blk_edge[ix][iy].ctype);
                   for(i = 0; i < 2; i++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             i, j, blk_edge[ix][iy].edge[i][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
                   ***/
               }
               else
               {
                   printf("ERROR: init_stencil_element2D handle case:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, blk_edge[ix][iy].ctype);
                   for(dir = 0; dir < 2; dir++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             dir, j, blk_edge[ix][iy].edge[dir][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
               }
           }
       }

}

/***
LOCAL void enforce_neumann_BC_on_nor_advec_vel(
       NS_Field_Stencil *ns_field)
{
       Wave       *newwv = ns_field->newwave;
       RECT_GRID  *gr = newwv->rect_grid;
       TRI_GRID   *ntg = wave_tri_soln(newwv)->tri_grid;
       BLK_EDGE   *blk_edge = ntg->blk_edge;
       int        xmax,ymax;
       int        ix,iy;
       int        dir,j,k,nor;
       int        icoords[2], ic[2];
       float      *coords,coord[2];      // coords is the cell center
                                         // and coord is the cell edge center
       float      *h = gr->h;

       xmax = gr->gmax[0];
       ymax = gr->gmax[1];

       for (iy = 0;  iy < ymax;  ++iy)
       {
           for (ix = 1;  ix < xmax;  ++ix)
           {
               icoords[0] = ix; icoords[1] = iy;
               coords = Rect_coords(icoords,newwv);

               if(Rect_blk_edge(icoords,newwv).ctype == EXTERNAL)
                   continue;

               if(blk_edge[ix][iy].ctype == PARTIAL)
               {
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           if(blk_edge[ix][iy].edge[dir][j].etype == N_SURF)
                           {
                               if(dir == 0)
                               {
                                   printf("ERROR: enforce_neumann_BC_on_nor_advec_vel, implement\n");
                                   clean_up(ERROR);
                               }
                               else
                               {
                                   ic[0] = ix; ic[1] = iy;
                                   if(j == 1)
                                       ic[1]++;
                                   n_vel_h_edge(ic, ns_field) = 0.0;
                               }
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == PARTIAL_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == YES)
                               {
                               }
                           }
                       }
                   }
               }               
           }
       }

        // TMP
        // if(debugging("NS_vel") && NO)
        {
            printf("enforce_neumann_BC_on_nor_advec_vel---normal vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i < gmax[0]; i++)
                for(k = 6; k < gmax[0]; k++)
                {
                    ic[0] = k;
                    ic[1] = j;
                    printf(" (%d, %10.9g) ", k, n_vel_h_edge(ic, ns_field));
                }
                printf("\n");
            }
        }

}
***/

/* set projection stencil to solve
 * DG\phi = DU before construction of convective derivatives.
 * \phi on the left boundary is set to be zero.
 * This is to use Dirichlet boundary condition on the left side.
 * On the partial cell, if a partial edge is the common
 * boundary of a obstacle cell and an interior(partial) cell,
 * this cell edge is treated as boundary. If a partial edge
 * is the common boundary of two interior(partial) cells, this
 * edge is treated as an interior edge. (allow fluild to flow through).
 * Phis of the right side cells are set to be zero. 
 */

LOCAL void init_stencil_element2D_ver2(
         BLK_EDGE      **blk_edge,
         P_CELL2D      **p_cell,
         Front         *fr,
         Wave          *wv,
         NS_Field_Stencil *ns_field,
         int           add_time,
         float         dt)
{
       RECT_GRID     *gr = wv->rect_grid;
       int        xmax,ymax;
       int        ix,iy;
       int        dir,j,k,nor;
       int        icoords[MAXD], ic[MAXD];
       float      *coords,coord[MAXD];      // coords is the cell center
                                         // and coord is the cell edge center
       float      *h = gr->h;
       float      rweight = 0.0, dt_coeff = 1.0;
       int        zero_nv = NO; 

       xmax = gr->gmax[0];
       ymax = gr->gmax[1];

       if(add_time == YES)
           dt_coeff = dt;

       for (iy = 0;  iy < ymax;  ++iy)
       {
           for (ix = 0;  ix < xmax-1;  ++ix)
           {
               icoords[0] = ix; 
               icoords[1] = iy;
               coords = Rect_coords(icoords,wv);

               for(k = 0; k < 3; ++k)
                  for(j = 0; j < 3; ++j)
               {
                  {
                     p_cell[ix][iy].set[k][j] = NO;
                     p_cell[ix][iy].stencil[k][j] = 0.0;
                  }
               }

               p_cell[ix][iy].r_side = 0.0;

               // TMP
               /**
               if(ix == 10 && (iy == 2 || iy == 7))
               {
                   deb_sten = YES;
                   printf("\n--------------cell[%d %d] init_stencil_element2D_ver2 type = %d\n",
                           ix, iy, blk_edge[ix][iy].ctype);
               }
               else  
                   deb_sten = NO;
               **/ 

               if(blk_edge[ix][iy].ctype == EXTERNAL)
                   continue; 

               if(blk_edge[ix][iy].ctype == PARTIAL)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);
                           // Domain boundary, homo. Neumann BC.
                           if(ix == 0 && nor == -1 && dir == 0)
                               zero_nv = YES;
                           else if(ix == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else if(iy == 0 && nor == -1 && dir == 1)
                               zero_nv = YES;
                           else if(iy == ymax-1 && nor == 1 && dir == 1)
                               zero_nv = YES;
                           else
                               zero_nv = NO;


                           if(deb_sten == YES)
                           {
                               printf("partial dir[%d] side[%d] edge_type = %d\n",
                                    dir, j, blk_edge[ix][iy].edge[dir][j].etype);
                           }

                           if(blk_edge[ix][iy].edge[dir][j].etype == D_SURF ||
                              blk_edge[ix][iy].edge[dir][j].etype == N_SURF ||
                              blk_edge[ix][iy].edge[dir][j].etype == PARTIAL_SURF)
                           {
                               coord[0] = blk_edge[ix][iy].edge[dir][j].cen[0];
                               coord[1] = blk_edge[ix][iy].edge[dir][j].cen[1];
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == FULL_SURF)
                           {
                               coord[dir] = coords[dir]+ nor*h[dir]/2; 
                               coord[(dir+1)%2] = coords[(dir+1)%2];
                           }
                           else
                               continue;
 
                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].edge[dir][j].etype == FULL_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == NO) 
                                   regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                               else
                               {
                                   bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                        icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                                   // printf("WARNING: init_stencil_element2D_ver2, full_edge,"
                                   //     " but the outter side is body cell[%d %d]\n",
                                   //             ix, iy);
                                   // clean_up(ERROR);
                               }
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == D_SURF ||
                                   blk_edge[ix][iy].edge[dir][j].etype == N_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == NO) 
                                   regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                               else
                                   bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                           }
                           else if(blk_edge[ix][iy].edge[dir][j].etype == PARTIAL_SURF)
                           {
                               if(blk_edge[ix][iy].comp_label[dir][j] == NO)
                                   regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                               else
                                   bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, partial edge 0,"
                               //     " implementing\n");
                               // clean_up(ERROR);
                           }
                           else
                           {
                               printf("ERROR: init_stencil_element2D, partial edge,"
                                   " implementing\n");
                               clean_up(ERROR);
                           }
                       }
                   }
               }
               else if(blk_edge[ix][iy].ctype == INTERNAL)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           if(deb_sten == YES)
                           {
                               printf("Set::::: dir[%d] side[%d] edge_type = %d\n",
                                    dir, j, blk_edge[ix][iy].edge[dir][j].etype);
                           }

                           nor = (j == 0? -1 : 1);
                           coord[dir] = coords[dir]+ nor*h[dir]/2; coord[(dir+1)%2] = coords[(dir+1)%2];

                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                               // TMP
                               if(deb_sten == YES)
                               {
                                   printf("v_line_rhor = %9.8g\n", v_line_rhor(ic,ns_field));
                               }
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].comp_label[dir][j] == NO)
                           {
                               regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                          icoords,ns_field,rweight,zero_nv);
                           }
                           else
                           {
                               bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //     " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                       }
                   }
               }
               else if(blk_edge[ix][iy].ctype == INTER_PART)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           if(deb_sten == YES)
                           {
                               printf("Inter_part dir[%d] side[%d] edge_type = %d\n",
                                    dir, j, blk_edge[ix][iy].edge[dir][j].etype);
                           }

                           nor = (j == 0? -1 : 1);
                           coord[dir] = coords[dir]+ nor*h[dir]/2; coord[(dir+1)%2] = coords[(dir+1)%2];

                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           if(blk_edge[ix][iy].comp_label[dir][j] == NO)
                           {
                               regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                           }
                           else
                           {
                               bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //     " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                       }
                   }                   
                   /***
                   printf("ERROR: init_stencil_element2D handle case INTER_PART:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, blk_edge[ix][iy].ctype);
                   for(i = 0; i < 2; i++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             i, j, blk_edge[ix][iy].edge[i][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
                   ***/
               }
               else
               {
                   printf("ERROR: init_stencil_element2D handle case:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, blk_edge[ix][iy].ctype);
                   for(dir = 0; dir < 2; dir++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             dir, j, blk_edge[ix][iy].edge[dir][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
               }
           }
       }

}

/* set projection stencil to solve
 * DG\phi = DU.
 * \phi on the right boundary is set to be zero.
 * This is to use Dirichlet boundary condition on the right side.
 * On the partial cell, if a partial edge is the common
 * boundary of a obstacle cell and an interior(partial) cell,
 * this cell edge is treated as boundary. If a partial edge
 * is the common boundary of two interior(partial) cells, this
 * edge is treated as an interior edge. (allow fluild to flow through).
 * Ghost cells are used to set phis on the cells outside the right boundary
 * with linear extrapolation. Other sides uses Homo Neumann boundary condition,
 * which is \grad \phi \cdot n = w \cdot n, where w = 0. 
 * Argument add_time = YES indicates that the projection is approximate, 
 * Therefore density comes from cell center values.
 */

/**
  IMPORTANT: For RT problem, the bottom and top boundary is Homo Dirichlet.
  init_stencil_element2D_ver3() should be modified correspondingly.
**/
// 042107, try set top line pressure = 1
LOCAL void init_stencil_element2D_ver3(
         Wave          *wv_blk,
         P_CELL2D      **p_cell,
         Front         *fr,
         Wave          *wv,
         NS_Field_Stencil *ns_field,
         int           add_time,
         float         dt)
{
       RECT_GRID     *gr = wv->rect_grid;
       int        xmax,ymax;
       int        ix,iy;
       int        dir,j,k,nor;
       int        icoords[MAXD], ic[MAXD], icrds[MAXD];
       float      *coords,coord[MAXD];      // coords is the cell center
                                         // and coord is the cell edge center
       float      *h = gr->h;
       float      rweight = 0.0, dt_coeff = 1.0;
       int        zero_nv = NO;

       xmax = gr->gmax[0];
       ymax = gr->gmax[1];

       for (iy = 0;  iy < ymax;  ++iy)
       // for (iy = 0;  iy < ymax-1;  ++iy)
       {
           // for (ix = 0;  ix < xmax-1;  ++ix)
           for (ix = 0;  ix < xmax;  ++ix)
           {
               icrds[0] = ix; icrds[1] = iy;
               icoords[0] = ix; 
               icoords[1] = iy;
               coords = Rect_coords(icoords,wv);

               for(k = 0; k < 3; ++k)
                  for(j = 0; j < 3; ++j)
               {
                  {
                     p_cell[ix][iy].set[k][j] = NO;
                     p_cell[ix][iy].stencil[k][j] = 0.0;
                  }
               }

               p_cell[ix][iy].r_side = 0.0;

               // TMP
               /**
               if((ix == 0 && iy == 0)
                  )
               {
                   deb_sten = YES;
                   printf("\n--------------cell[%d %d] init_stencil_element2D_ver3 type = %d\n",
                           ix, iy, Rect_blk_edge(icrds,wv_blk).ctype);
               }
               else  
                   deb_sten = NO;
               **/

               if(Rect_blk_edge(icrds,wv_blk).ctype == EXTERNAL)
                   continue; 

               if(Rect_blk_edge(icrds,wv_blk).ctype == PARTIAL)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           nor = (j == 0? -1 : 1);

                           // Domain boundary, homo. Neumann BC.
                           /**
                           if(icrds[0] == 0 && nor == -1 && dir == 0) 
                               zero_nv = YES;
                           else if(icrds[0] == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else if(icrds[1] == 0 && nor == -1 && dir == 1)
                               zero_nv = YES;
                           else if(icrds[1] == ymax-1 && nor == 1 && dir == 1)
                               zero_nv = YES;
                           else
                               zero_nv = NO;
                           if(icrds[0] == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else
                               zero_nv = NO;
                           **/

                           if(deb_sten == YES)
                           {
                               printf("partial dir[%d] side[%d] edge_type = %d\n",
                                    dir, j, Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype);
                           }

                           if(Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == D_SURF ||
                              Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == N_SURF ||
                              Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == PARTIAL_SURF)
                           {
                               coord[0] = Rect_blk_edge(icrds,wv_blk).edge[dir][j].cen[0];
                               coord[1] = Rect_blk_edge(icrds,wv_blk).edge[dir][j].cen[1];
                           }
                           else if(Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == FULL_SURF)
                           {
                               coord[dir] = coords[dir]+ nor*h[dir]/2; 
                               coord[(dir+1)%2] = coords[(dir+1)%2];
                           }
                           else
                               continue;
 
                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }
                       
                           // For approximate projection
                           // if(add_time == YES)
                           // rweight = dt_coeff/Dens(Rect_state(icoords,wv));

                           if(Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == FULL_SURF)
                           {
                               // if(Rect_blk_edge(icrds,wv_blk).comp_label[dir][j] == NO) 
                               regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                               /***
                               else
                               {
                                   // bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                   //      icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                                   // 040407, original = YES
                                   bdry_cell_sten_ver3(&(p_cell[ix][iy]),coord,h,dir,nor,
                                      icoords,ns_field,Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype,rweight,YES);
                                   printf("WARNING: init_stencil_element2D_ver2, full_edge,"
                                       " but the outter side is body cell[%d %d]\n",
                                               ix, iy);
                                    clean_up(ERROR);
                               }
                               **/
                           }
                           else if(Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == D_SURF)
                           {
                               if(Rect_blk_edge(icrds,wv_blk).comp_label[dir][j] == NO) 
                                   regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                               else
                               { 
                                   // bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                   //  icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                                   // 040407, original = NO
                                   bdry_cell_sten_ver3(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype,rweight,NO);
                               }
                           }
                           else if(Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == N_SURF)
                           {
                                   // bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                   //  icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                                   // 040407, original = YES
                                   // printf("cell[%d %d] call N_SURF used bdry_cell_sten_ver3\n", ix, iy);
                               bdry_cell_sten_ver3(&(p_cell[ix][iy]),coord,h,dir,nor,
                                icoords,ns_field,Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype,
                                    rweight,YES); // 042307, change YES to NO
                           }
                           else if(Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype == PARTIAL_SURF)
                           {
                               /**
                               if(Rect_blk_edge(icrds,wv_blk).comp_label[dir][j] == NO)
                                   regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                               else
                               {
                                   // bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                   //  icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                                   // original YES
                                   bdry_cell_sten_ver3(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype,
                                         rweight,YES);
                               }
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               **/
                               printf("ERROR: init_stencil_element2D_ver3, partial edge 0,"
                                   " implementing\n");
                               clean_up(ERROR);
                           }
                           else
                           {
                               printf("ERROR: init_stencil_element2D, partial edge,"
                                   " implementing\n");
                               clean_up(ERROR);
                           }
                       }
                   }
               }
               else if(Rect_blk_edge(icrds,wv_blk).ctype == INTERNAL)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           if(deb_sten == YES)
                           {
                               printf("Set::::: dir[%d] side[%d] edge_type = %d\n",
                                    dir, j, Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype);
                           }

                           nor = (j == 0? -1 : 1);
                           coord[dir] = coords[dir]+ nor*h[dir]/2; coord[(dir+1)%2] = coords[(dir+1)%2];

                           // Domain boundary, homo. Neumann BC.
                           /***
                           if(icrds[0] == 0 && nor == -1 && dir == 0)
                               zero_nv = YES;
                           else if(icrds[0] == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else if(icrds[1] == 0 && nor == -1 && dir == 1)
                               zero_nv = YES;
                           else if(icrds[1] == ymax-1 && nor == 1 && dir == 1)
                               zero_nv = YES;
                           else
                               zero_nv = NO;
                           if(icrds[0] == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else
                               zero_nv = NO;
                           ***/ 

                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                               // TMP
                               if(deb_sten == YES)
                               {
                                   printf("internal v_line_rhor = %9.8g\n", v_line_rhor(ic,ns_field));
                               }
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           // For approximate projection
                           // if(add_time == YES)
                           // rweight = dt_coeff/Dens(Rect_state(icoords,wv));

                           // if(Rect_blk_edge(icrds,wv_blk).comp_label[dir][j] == NO)
                           regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                          icoords,ns_field,rweight,zero_nv);
                           /**
                           else
                           {
                               // bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                               //      icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // Original = YES
                               bdry_cell_sten_ver3(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype,
                                          rweight,NO);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                                   " but the outter side is body cell\n");
                               clean_up(ERROR);
                           }
                           **/
                       }
                   }
               }
               else if(Rect_blk_edge(icrds,wv_blk).ctype == INTER_PART)
               {
                   // cells only have four edges
                   for(dir = 0; dir < 2; ++dir)
                   {
                       for(j = 0; j < 2; ++j)
                       {
                           if(deb_sten == YES)
                           {
                               printf("Inter_part dir[%d] side[%d] edge_type = %d\n",
                                    dir, j, Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype);
                           }

                           nor = (j == 0? -1 : 1);
                           coord[dir] = coords[dir]+ nor*h[dir]/2; coord[(dir+1)%2] = coords[(dir+1)%2];

                           // Domain boundary, homo. Neumann BC.
                           /***
                           if(icrds[0] == 0 && nor == -1 && dir == 0)
                               zero_nv = YES;
                           else if(icrds[0] == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else if(icrds[1] == 0 && nor == -1 && dir == 1)
                               zero_nv = YES;
                           else if(icrds[1] == ymax-1 && nor == 1 && dir == 1)
                               zero_nv = YES;
                           else
                               zero_nv = NO;
                           if(icrds[0] == xmax-1 && nor == 1 && dir == 0)
                               zero_nv = YES;
                           else
                               zero_nv = NO;
                           ***/

                           ic[0] = ix; ic[1] = iy;
                           if(dir == 0)
                           {
                               if(j == 1) ic[0]++;
                               rweight = dt_coeff*v_line_rhor(ic,ns_field);
                           }
                           else
                           {
                               if(j == 1) ic[1]++;
                               rweight = dt_coeff*h_line_rhor(ic,ns_field);
                           }

                           // For approximate projection
                           // if(add_time == YES)
                           // rweight = dt_coeff/Dens(Rect_state(icoords,wv));

                           // if(Rect_blk_edge(icrds,wv_blk).comp_label[dir][j] == NO)
                           // {
                               regular_fd_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                                       icoords,ns_field,rweight,zero_nv);
                           /**
                           }
                           else
                           {
                               // bdry_cell_sten_ver2(&(p_cell[ix][iy]),coord,h,dir,nor,
                               //      icoords,ns_field,blk_edge[ix][iy].edge[dir][j].etype,rweight);
                               // Original YES
                               bdry_cell_sten_ver3(&(p_cell[ix][iy]),coord,h,dir,nor,
                                    icoords,ns_field,Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype,
                                         rweight,NO);
                               // 120106, looks like partial surf is not right
                               // partial_surf_sten(&(p_cell[ix][iy]),coord,h,nor,
                               //      icoords,ns_field,i,j,blk_edge,rweight);
                               // printf("ERROR: init_stencil_element2D, INTERNAL CELL full_edge,"
                               //      " but the outter side is body cell\n");
                               // clean_up(ERROR);
                           }
                           **/
                       }
                   }                   
               }
               else
               {
                   printf("ERROR: init_stencil_element2D handle case:\n");
                   printf("IN init_stencil_element2D, [%d][%d].ctype = %d\n",
                       ix, iy, Rect_blk_edge(icrds,wv_blk).ctype);
                   for(dir = 0; dir < 2; dir++)
                   {
                       for(j = 0; j < 2; j++)
                       {
                           printf("blk->edge[%d][%d]->type = %d ,",
                             dir, j, Rect_blk_edge(icrds,wv_blk).edge[dir][j].etype);
                       }
                       printf("\n");
                   }
                   clean_up(ERROR);
               }

               if(deb_sten == YES)
               {
                   int i, j;
                   for(i = 0; i < 3; i++)
                   {
                       for(j = 0; j < 3; j++)
                       {
                           printf("stencil[%d][%d] = %g (%d),", i, j, p_cell[ix][iy].stencil[i][j],
                                   p_cell[ix][iy].set[i][j]);
                       }
                       printf("\n");
                   }
                   printf("right side = %10.9g\n", p_cell[ix][iy].r_side);
               }

               if(add_time == YES)
               {
                   p_cell[ix][iy].r_side /= dt;
               }
           }
       }

}

/* Phis of right side boundary cells are set to zero 
 */
LOCAL void regular_fd_sten_ver2(
           P_CELL2D  *p_cell,
           float     *coord,
           float     *h,
           int       dir,
           int       nor,
           int       *ic,
           NS_Field_Stencil *ns_field,
           float     rweight,
           int       zero_nv)
{
           int       eic[MAXD];
           float     nv = 0.0;

           assign(eic,ic,sizeof(int)*2);
           if(zero_nv == NO)
           {
               if(dir == 0)
               {
                   if(nor == 1)
                       eic[0]++;
                   nv = v_line_vel(eic,ns_field).v[0];
               }
               else
               {
                   if(nor == 1)
                       eic[1]++;
                   nv = h_line_vel(eic,ns_field).v[1];
               }
           }

           // 030707 p_cell->r_side += h[(i+1)%2]*nv*nor;
           p_cell->r_side += h[dir]*nv*nor;

           // TMP
           if(deb_sten == YES)
           {
               printf("--regular_fd_sten_ver2 ic[%d %d] dir[%d] nor %d, +r_side %g of(%d %d), nor_v %g\n",
                       ic[0], ic[1], dir, nor, nv*h[dir]*nor, eic[0], eic[1], nv);
           }

           p_cell->stencil[1][1] -= 1.0*rweight;
           p_cell->set[1][1] = YES;
 
           if(dir == 0)
           {
               if((ic[0] == gmax[0]-1 && nor == 1))
               {
                   // Right side BDRY
                   if(PERIOD_BC == YES)
                   {
                       p_cell->set[1+nor][1] = YES;
                       p_cell->stencil[1+nor][1] += 1.0*rweight;
                   }
                   else
                   {
                       // Homo Neumann
                       p_cell->set[1+nor][1] = NO;
                       p_cell->stencil[1][1] += 1.0*rweight;
                       // In right side ghost cell, \phi = 0
                   }
               }
               else if((ic[0] == 0 && nor == -1)) 
               {
                   // Left side BDRY
                   if(PERIOD_BC == YES)
                   {
                       p_cell->set[1+nor][1] = YES;
                       p_cell->stencil[1+nor][1] += 1.0*rweight;
                   }
                   else
                   {
                       // Homo Neumann
                       p_cell->set[1+nor][1] = NO;
                       p_cell->stencil[1][1] += 1.0*rweight;
                   }
               }
               else
               {  
                   p_cell->set[1+nor][1] = YES;
                   p_cell->stencil[1+nor][1] += 1.0*rweight;
               }

               if(deb_sten == YES)
               {
                   printf("X-regular_fd_stenx-dir set[%d][%d]=%d, coeff add= %g, zero_nv = %d\n",
                             1+nor,1, p_cell->set[1+nor][1], rweight, zero_nv);
               }
           }
           else
           {

               if((ic[1] == 0 && nor == -1))
               {
                   // Homo Neumann
                   p_cell->set[1][1+nor] = NO;
                   p_cell->stencil[1][1] += 1.0*rweight;
               }
               // NEW, the top line pressure = 0, 042107
               /**
               else if(ic[1] == gmax[1]-2 && nor == 1)
               {
                   p_cell->set[1][1+nor] = NO;
               }
               **/
               // END NEW
               else if((ic[1] == gmax[1]-1 && nor == 1))
               {
                   // Homo Dirichlet, for RT problem
                   p_cell->set[1][1+nor] = NO;
                   p_cell->stencil[1][1] -= 1.0*rweight;
               }
               else
               {
                   p_cell->set[1][1+nor] = YES;
                   p_cell->stencil[1][1+nor] += 1.0*rweight;
               }

               if(deb_sten == YES)
               {
                   printf("Y-regular_fd_stenx-dir set[%d][%d]=%d, coeff add = %9.8g, zero_nv = %d\n",
                             1, 1+nor, p_cell->set[1][1+nor], rweight, zero_nv);
                   // clean_up(ERROR);
               }
           }
}


/* finite volume discretization of 
 * DG\phi = DU, here \phi is a scalar,
 * U is vector, D is divergence operator,
 * and G is gradient.
 * regular_fd_sten_ver2(), bdry_cell_sten_ver2();
 * DG = L_{compact}; 
 * Note: the standard 9-point laplacian is 
 * replaced by a compact 5-point Laplacian
 * Phis of right side boundary cells are set to zero,
 * linear extrapolation is used to set the ghost cell value.
 */

LOCAL void bdry_cell_sten_ver2(
           P_CELL2D  *p_cell,
           float     *coord,
           float     *h,
           int       i,
           int       nor,
           int       *ic,
           NS_Field_Stencil *ns_field,
           ELEMENT_TYPE  etype,
           float       rweight)
{
           int       eic[MAXD];
           float     nv;

           assign(eic,ic,sizeof(int)*2);
           if(i == 0)
           {
               if(nor == 1)
                   eic[0]++;
               nv = v_line_vel(eic,ns_field).v[0];
           }
           else
           {
               if(nor == 1)
                   eic[1]++;
               nv = h_line_vel(eic,ns_field).v[1];
           }

           // Normal vel. should be dependent on the
           // boundary type, this is handled already.
           // We assume correct boundary vel is
           // stored in ns_field.
           // NEUMANN u_normal = 0; DIRICHLET u_nor = u_bdry.
           // p_cell->r_side += h[(i+1)%2]*nv*dir; 
           // The boundary condition for phi is : (n \cdot \grad phi = 0)

           // TMP
           /**
           if(ic[0] == 0 && ic[1] == 0)
           {
               printf("ic[%d %d] dir[%d] nor = %d, add to r_side = %g, bdry_cell_sten\n",
                       ic[0], ic[1], i, dir, h[(i+1)%2]*nv*dir);
           }
           **/

           // Original 
           // p_cell->r_side += h[(i+1)%2]*nv*dir; 
           // NEW, it's equivalent to reflect normal vel
           // 030707 p_cell->r_side += h[(i+1)%2]*nv*dir; 
           p_cell->r_side += h[(i)]*nv*nor; 

           // TMP
           if(deb_sten == YES)
           {
               printf("--bdry_cell_sten_ver2 ic[%d %d] dir[%d], nor %d, add to r_side %9.8g\n",
                       ic[0], ic[1], i, nor, h[(i+1)%2]*nv*nor);
           }

           /** 030707
           if(i == 0)
               p_cell->set[1+dir][1] = NO;
           else
               p_cell->set[1][1+dir] = NO;
           **/

           p_cell->stencil[1][1] -= 1.0*rweight;
           p_cell->set[1][1] = YES;

           if(i == 0)
           {
               /**
               if(ic[0] != gmax[0]-1 || (ic[0] == gmax[0]-1 && nor != 1))
               {
                   p_cell->set[1+nor][1] = YES;
                   p_cell->stencil[1+nor][1] += 1.0*rweight;
               }
               **/ 
               if((ic[0] == gmax[0]-1 && nor == 1))
               {
                   // Dirichlet
                   p_cell->set[1+nor][1] = NO;
                   p_cell->stencil[1][1] -= 1.0*rweight;
               }
               else
               {
                   // Neumann
                   p_cell->set[1+nor][1] = NO;
                   p_cell->stencil[1][1] += 1.0*rweight;
                   p_cell->r_side -= h[(i)]*nv*nor*nor*rweight; 
               }
           }
           else
           {
               // Neumann
               p_cell->set[1][1+nor] = NO;
               p_cell->stencil[1][1] += 1.0*rweight;
               p_cell->r_side -= h[(i)]*nv*nor*nor*rweight; 
           }

}

/* finite volume discretization of 
 * DG\phi = DU, here \phi is a scalar,
 * U is vector, D is divergence operator,
 * and G is gradient.
 * regular_fd_sten_ver2(), bdry_cell_sten_ver3();
 * DG = L_{compact}; 
 * Note: the standard 9-point laplacian is 
 * replaced by a compact 5-point Laplacian
 * \phis of right side boundary cells are set to zero,
 * linear extrapolation is used to set the ghost cell value.
 * The homogeneous Neumann bdry conditions are applied to
   \phis on the other boundaries.
 */

LOCAL void bdry_cell_sten_ver3(
           P_CELL2D  *p_cell,
           float     *coord,
           float     *h,
           int       dir,
           int       nor,
           int       *ic,
           NS_Field_Stencil *ns_field,
           ELEMENT_TYPE  etype,
           float       rweight,
           int         zero_nv)
{
           int       eic[MAXD];
           float     nv = 0.0;

           assign(eic,ic,sizeof(int)*2);
           if(zero_nv == NO)
           {
               if(dir == 0)
               {
                   if(nor == 1)
                       eic[0]++;
                   nv = v_line_vel(eic,ns_field).v[0];
               }
               else
               {
                   if(nor == 1)
                       eic[1]++;
                   nv = h_line_vel(eic,ns_field).v[1];
               }
           }

           // Normal vel. should be dependent on the
           // boundary type, this is handled already.
           // We assume correct boundary vel is
           // stored in ns_field.
           // NEUMANN u_normal = 0; DIRICHLET u_nor = u_bdry.

           // TMP
           /**
           if(ic[0] == 0 && ic[1] == 0)
           {
               printf("ic[%d %d] dir[%d] nor = %d, add to r_side = %g, bdry_cell_sten\n",
                       ic[0], ic[1], i, dir, h[(i+1)%2]*nv*dir);
           }
           **/

           p_cell->r_side += h[(dir)]*nv*nor; 

           // TMP
           if(deb_sten == YES)
           {
               printf("--bdry_cell_sten_ver3 ic[%d %d] dir[%d], nor %d, add to r_side %9.8g, zero_nv = %d\n",
                       ic[0], ic[1], dir, nor, h[(dir+1)%2]*nv*nor, zero_nv);
           }

           p_cell->stencil[1][1] -= 1.0*rweight;
           p_cell->set[1][1] = YES;

           if(dir == 0)
           {
               if((ic[0] == gmax[0]-1 && nor == 1))
               {
                   // Homo Dirichlet, for blood problem
                   p_cell->set[1+nor][1] = NO;
                   p_cell->stencil[1][1] -= 1.0*rweight;
               }
               else
               {
                   // Homo Neumann
                   p_cell->set[1+nor][1] = NO;
                   p_cell->stencil[1][1] += 1.0*rweight;
                   // p_cell->r_side -= h[(i)]*nv*nor*rweight; 
               }

               printf("--bdry_cell_sten_ver3 ic[%d %d] dir[%d], nor %d, add to r_side %9.8g, zero_nv = %d\n",
                       ic[0], ic[1], dir, nor, h[(dir+1)%2]*nv*nor, zero_nv);
               printf("RT problem should not go into here\n");
               clean_up(ERROR);
           }
           else
           {
               if((ic[1] == gmax[1]-1 && nor == 1))
               {
                   // Homo Dirichlet, for RT problem
                   p_cell->set[1][1+nor] = NO;
                   p_cell->stencil[1][1] -= 1.0*rweight;
               }
               else
               {
                   // Homo Neumann
                   p_cell->set[1][1+nor] = NO;
                   p_cell->stencil[1][1] += 1.0*rweight;
                   // p_cell->r_side -= h[(i)]*nv*nor*rweight; 
               }
           }

}

// Adjust normal velocity on cell edges,
// Physical boundary cell edges are also included.
LOCAL void adjust_nor_vel_on_edge_ver2(
        NS_Field_Stencil  *ns_field)
{
        int        i, j, k;
        int        ic[MAXD], ic1[MAXD], ic2[MAXD], ic3[MAXD];
        float      dv;
        Wave       *wv = ns_field->newwave;
        Front      *nfr = ns_field->newfr;
        TRI_GRID   *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE   *blk_edge = ntg->blk_edge;
        float      *dh = wv->rect_grid->h;
        int        *gmax = wv->rect_grid->gmax;
        float      rweight, grad_phi1, grad_phi2;
        COMPONENT  comp, comp1, comp2, comp3;
        int        smin[MAXD], smax[MAXD];
        int        debug_flag = NO;
       

        for(i = 0; i < dim; i++)
        {
            smin[i] = -wv->rect_grid->lbuf[i];
            smax[i] = gmax[i] + wv->rect_grid->ubuf[i];
        }
        smin[0] = smin[1] = 0; smax[1] = gmax[1]; smax[0] = gmax[0];

        // ic[] is edge index. compute on horizontal edge
        for(i = smin[0]; i < smax[0]; i++)
        {
            for(j = smin[1]+1; j < smax[1]; j++)
            // for(j = smin[1]; j <= smax[1]; j++)
            {
                ic2[0] = ic[0] = i;
                ic2[1] = ic[1] = j;
                ic2[1]--;

                // comp1 = Rect_comp(ic2,wv);
                // comp = Rect_comp(ic,wv);

                // if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                //    comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                //    comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    rweight = h_line_rhor(ic, ns_field);
                    dv = (Rect_phi(ic,ns_field)-Rect_phi(ic2,ns_field))/dh[1]*rweight;
                    // TMP
                    /**
                    if((ic[0] == 0 && ic[1] == 5) ||
                       (ic[0] == 1 && ic[1] == 5))
                    {
                        printf("In adjust_nor_vel_on_edge_ver2, horizontal edge[%d %d] before %g\n", 
                            ic[0], ic[1], n_vel_h_edge(ic, ns_field));
                        printf("dv = %9.8g, rweight = %9.8g, n_vel_h_edge %9.8g, phi1 %9.8g, phi2 %9.8g\n",
                             dv, rweight, n_vel_h_edge(ic, ns_field), 
                            Rect_phi(ic,ns_field), Rect_phi(ic2,ns_field));
                        printf("ic[%d %d], ic2[%d %d] ans %g\n", ic[0], ic[1], ic2[0] , ic2[1],
                                 n_vel_h_edge(ic, ns_field) - dv);
                    }
                    **/
                    n_vel_h_edge(ic, ns_field) -= dv; 
                }
            }
        }

        // ic[] is edge index. compute on vertical edge
        for(i = smin[0]+1; i < smax[0]; i++)
        // for(i = smin[0]; i <= smax[0]; i++)
        {
            for(j = smin[1]; j < smax[1]; j++)
            {
                ic2[0] = ic[0] = i;
                ic2[1] = ic[1] = j;
                ic2[0]--;

                // comp1 = Rect_comp(ic2,wv);
                // comp = Rect_comp(ic,wv);
                // if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                //    comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                //    comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    rweight = v_line_rhor(ic, ns_field);
                    dv = (Rect_phi(ic,ns_field)-Rect_phi(ic2,ns_field))/dh[0]*rweight;
                    n_vel_v_edge(ic, ns_field) -= dv;
                    // printf("normal vel edge[%d][%d] = %g  (phi %g %g,) ic2[%d %d], dv = %g\n", 
                    //                  ic[0],ic[1],n_vel_v_edge(ic, ns_field),
                    //                   Rect_phi(ic,ns_field),
                    //                   Rect_phi(ic2,ns_field), ic2[0], ic2[1], dv);
                }
            }
        }

        // ic[] is edge index. compute on vertical edge
 
        if(debugging("NS_vel"))
        {
            printf("+++++++++++++++++++++++++++++++++++++\n");
            printf("\nadjust_nor_vel_on_edge*******vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            // for(j = 10; j <= 14; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i < min(12,gmax[0]); i++)
                for(i = 0; i < 4; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d %9.8g)", i, n_vel_h_edge(ic, ns_field));            
                }
                printf("\n");
            }
            printf("\nadjust_nor_vel_on_edge*******vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            // for(j = 10; j <= 14; j++)
            {
                printf("row%d: ", j);
                // for(i = 0; i <= min(12,gmax[0]); i++)
                for(i = 0; i <= 3; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d %9.8g)", i, n_vel_v_edge(ic, ns_field));
                }
                printf("\n");
            }
            printf("+++++++++++++++++++++++++++++++++++++\n");
        }

        // printf("ERROR: stop in adjust_nor_vel_on_edge_ver2\n");
        // clean_up(ERROR);
}

// Adjust normal velocity on cell edges,
// Physical boundary cell edges are also included.
// According to the cell component,
// Grad\phi is linearly extrapolated if necessary.

LOCAL void adjust_nor_vel_on_edge_ver3(
        NS_Field_Stencil  *ns_field)
{
        int        i, j, k;
        int        ic[MAXD], ic1[MAXD], ic2[MAXD], ic3[MAXD];
        float      dv;
        Wave       *wv = ns_field->newwave;
        Front      *nfr = ns_field->newfr;
        TRI_GRID   *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE   *blk_edge = ntg->blk_edge;
        float      *dh = wv->rect_grid->h;
        int        *gmax = wv->rect_grid->gmax;
        float      rweight, grad_phi1, grad_phi2;
        COMPONENT  comp, comp1, comp2, comp3;
        int        smin[MAXD], smax[MAXD];
        int        debug_flag = NO;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        // ic[] is edge index. compute on horizontal edge
        for(i = smin[0]; i < smax[0]; i++)
        {
            for(j = smin[1]; j < smax[1]+1; j++)
            {
                ic1[0] = ic[0] = i;
                ic1[1] = ic[1] = j;
                ic1[1]--;

                comp1 = Find_rect_comp_ver2(ic1,wv);
                comp = Find_rect_comp_ver2(ic,wv);

                rweight = h_line_rhor(ic, ns_field);
                dv = 0.0;

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                   comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    dv = (Rect_phi(ic,ns_field)-Rect_phi(ic1,ns_field))/dh[1];
                }
                else if((comp == ext_comp || is_excluded_comp(comp,nfr->interf)) &&
                        (comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                       )
                {
                    // extrapolate from below to above to get grad \phi
                    ic2[0] = ic1[0];
                    ic2[1] = ic1[1];
                    ic2[1]--;
                    ic3[0] = ic2[0];
                    ic3[1] = ic2[1];
                    ic3[1]--;
                    comp2 = Find_rect_comp_ver2(ic2,wv);
                    comp3 = Find_rect_comp_ver2(ic3,wv);
                    if((comp2 != ext_comp && (! is_excluded_comp(comp2,nfr->interf))) &&
                       (comp3 != ext_comp && (! is_excluded_comp(comp3,nfr->interf)))
                      )
                    {
                        grad_phi2 = (Rect_phi(ic2,ns_field)-Rect_phi(ic3,ns_field))/dh[1];
                        grad_phi1 = (Rect_phi(ic1,ns_field)-Rect_phi(ic2,ns_field))/dh[1];
                        dv = 2.0*grad_phi1 - grad_phi2;
                    }
                }
                else if((comp != ext_comp && (! is_excluded_comp(comp,nfr->interf))) &&
                        (comp1 == ext_comp && (is_excluded_comp(comp1,nfr->interf)))
                       )
                {
                    // extrapolate from above to below to get grad \phi
                    ic2[0] = ic[0];
                    ic2[1] = ic[1];
                    ic2[1]++;
                    ic3[0] = ic2[0];
                    ic3[1] = ic2[1];
                    ic3[1]++;
                    comp2 = Find_rect_comp_ver2(ic2,wv);
                    comp3 = Find_rect_comp_ver2(ic3,wv);

                    // TMP
                    if(debug_flag == YES)
                    {
                        printf("ic2[%d %d], ic3[%d %d] ic[%d %d]\n", 
                                  ic2[0], ic2[1], ic3[0], ic3[1], ic[0], ic[1]);
                    }
                    if((comp2 != ext_comp && (! is_excluded_comp(comp2,nfr->interf))) &&
                       (comp3 != ext_comp && (! is_excluded_comp(comp3,nfr->interf)))
                      )
                    {
                        grad_phi2 = (Rect_phi(ic3,ns_field)-Rect_phi(ic2,ns_field))/dh[1];
                        grad_phi1 = (Rect_phi(ic2,ns_field)-Rect_phi(ic,ns_field))/dh[1];
                        dv = 2.0*grad_phi1 - grad_phi2;
                        // TMP
                        if(debug_flag == YES)
                        {
                            printf("grad  3, 2, 0 = %g %g %g\n", Rect_phi(ic3,ns_field),
                                      Rect_phi(ic2,ns_field),
                                      Rect_phi(ic,ns_field));
                        }
                    }
                }

                // TMP
                if(debug_flag == YES)
                    printf("dv = %g, n_vel_h_edge = %g\n", dv, n_vel_h_edge(ic, ns_field)); 

                n_vel_h_edge(ic, ns_field) -= dv*rweight; 
            } 
        }

        // ic[] is edge index. compute on vertical edge
        for(i = smin[0]; i < smax[0]+1; i++)
        {
            for(j = smin[1]; j < smax[1]; j++)
            {
                ic1[0] = ic[0] = i;
                ic1[1] = ic[1] = j;
                ic1[0]--;

                comp1 = Find_rect_comp_ver2(ic1,wv);
                comp = Find_rect_comp_ver2(ic,wv);

                rweight = v_line_rhor(ic, ns_field);
                dv = 0.0;

                if(comp != ext_comp && (! is_excluded_comp(comp,nfr->interf)) &&
                   comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)) &&
                   comp != POTS_COMP && comp1 != POTS_COMP)
                {
                    dv = (Rect_phi(ic,ns_field)-Rect_phi(ic1,ns_field))/dh[0];
                }
                else if((comp == ext_comp || is_excluded_comp(comp,nfr->interf)) &&
                        (comp1 != ext_comp && (! is_excluded_comp(comp1,nfr->interf)))
                       )
                {
                    // extrapolate from left to right to get grad \phi
                    ic2[0] = ic1[0];
                    ic2[1] = ic1[1];
                    ic2[0]--;
                    ic3[0] = ic2[0];
                    ic3[1] = ic2[1];
                    ic3[0]--;
                    comp2 = Find_rect_comp_ver2(ic2,wv);
                    comp3 = Find_rect_comp_ver2(ic3,wv);
                    if((comp2 != ext_comp && (! is_excluded_comp(comp2,nfr->interf))) &&
                       (comp3 != ext_comp && (! is_excluded_comp(comp3,nfr->interf)))
                      )
                    {
                        grad_phi2 = (Rect_phi(ic2,ns_field)-Rect_phi(ic3,ns_field))/dh[0];
                        grad_phi1 = (Rect_phi(ic1,ns_field)-Rect_phi(ic2,ns_field))/dh[0];
                        dv = 2.0*grad_phi1 - grad_phi2;
                    }
                }
                else if((comp != ext_comp && (! is_excluded_comp(comp,nfr->interf))) &&
                        (comp1 == ext_comp && (is_excluded_comp(comp1,nfr->interf)))
                       )
                {
                    // extrapolate from right to left to get grad \phi
                    ic2[0] = ic[0];
                    ic2[1] = ic[1];
                    ic2[0]++;
                    ic3[0] = ic2[0];
                    ic3[1] = ic2[1];
                    ic3[0]++;
                    comp2 = Find_rect_comp_ver2(ic2,wv);
                    comp3 = Find_rect_comp_ver2(ic3,wv);
                    if((comp2 != ext_comp && (! is_excluded_comp(comp2,nfr->interf))) &&
                       (comp3 != ext_comp && (! is_excluded_comp(comp3,nfr->interf)))
                      )
                    {
                        grad_phi2 = (Rect_phi(ic3,ns_field)-Rect_phi(ic2,ns_field))/dh[0];
                        grad_phi1 = (Rect_phi(ic2,ns_field)-Rect_phi(ic,ns_field))/dh[0];
                        dv = 2.0*grad_phi1 - grad_phi2;
                    }
                }
                n_vel_v_edge(ic, ns_field) -= dv*rweight;
            }
        }
 
        if(debugging("NS_vel"))
        {
            printf("\nadjust_nor_vel_on_edge_ver3*******vel on horizontal edge\n");
            for(j = 0; j <= gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i < min(6,gmax[0]); i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d %7.5g) ", i, n_vel_h_edge(ic, ns_field));            
                }
                printf("\n");
            }
            printf("\nadjust_nor_vel_on_edge_ver3*******vel on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= min(6,gmax[0]); i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" (%d %7.5g) ", i, n_vel_v_edge(ic, ns_field));
                }
                printf("\n");
            }
        }

        // printf("ERROR: stop in adjust_nor_vel_on_edge_ver3\n");
        // clean_up(ERROR);
}


LOCAL void half_step_den(
        NS_Field_Stencil   *ns_field,
        Wave               *wv)
{
        int             smin[MAXD], smax[MAXD];
        TRI_GRID        *ntg;
        BLK_EDGE        *blk_edge;
        Front           *nfr = ns_field->newfr, *fr = ns_field->fr;
        int             ix, iy, ic[MAXD], i, j, k, i1, dir, ic1[MAXD];
        COMPONENT       comp, comp1, ncomp;
        float           vold[MAXD], vnew[MAXD];
        static          Stencil *sten = NULL;
        int             end = 1, **icoords;
        int             *gmax, iperm[2] = {0, 1}; // No need to permute
        CRXING          *oldcrx, *newcrx;
        float           *dh;
        int             vset;
        Wave            *newwv = ns_field->newwave; 
        Wave            *oldwv = ns_field->wave; 
        int             deb_flag = NO;

        if (sten == NULL)
            sten = alloc_stencil(3,nfr);
        icoords = sten->icoords;

        gmax = wv->rect_grid->gmax;

        for(i = 0; i < dim; i++)
        {
            smin[i] = -lbuf[i];
            smax[i] = gmax[i] + ubuf[i];
        }

        // ic[] is edge index. First compute on horizontal edge
        for(i = 0; i < gmax[0]; i++)
        {
            ic[0] = i;
            for(j = 0; j <= gmax[1]; j++)
            {
                ic[1] = j;
                if(n_vel_h_edge(ic, ns_field) > 0.0)
                    h_line_rhor(ic,ns_field) = 1.0/h_line_vel_L(ic,ns_field).v[dim];
                else if(n_vel_h_edge(ic, ns_field) < 0.0)
                    h_line_rhor(ic,ns_field) = 1.0/h_line_vel_R(ic,ns_field).v[dim];
                else
                    h_line_rhor(ic,ns_field) = 
                            2.0/(h_line_vel_R(ic,ns_field).v[dim] + h_line_vel_L(ic,ns_field).v[dim]);
            }
        }

        // ic[] is edge index. Now compute on vertical edge
        for(i = 0; i <= gmax[0]; i++)
        {
            ic[0] = i;
            for(j = 0; j < gmax[1]; j++)
            {
                ic[1] = j;
                if(n_vel_v_edge(ic, ns_field) > 0.0)
                    v_line_rhor(ic,ns_field) = 1.0/v_line_vel_L(ic,ns_field).v[dim];
                else if(n_vel_v_edge(ic, ns_field) < 0.0)
                    v_line_rhor(ic,ns_field) = 1.0/v_line_vel_R(ic,ns_field).v[dim];
                else
                    v_line_rhor(ic,ns_field) = 
                           2.0/(v_line_vel_L(ic,ns_field).v[dim] + v_line_vel_R(ic,ns_field).v[dim]);
            }
        }      

        // Average to get half step cell center density
        for(i = 0; i < gmax[0]; i++)
        {
            ic[0] = i;
            for(j = 0; j < gmax[1]; j++)
            {
                ic[1] = j;
                Rect_den(ic,ns_field) = 0.5*(Dens(Rect_state(ic,oldwv)) + Dens(Rect_state(ic,newwv)));
            }
        }

        // TMP
        if(debugging("NS_vel") && NO)
        {
            printf("In edge_1_over_half_step_den, show 1/rho on horizontal edge\n");            
            for(j = 0; j <= gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d %g ", i, h_line_rhor(ic, ns_field)); 
                }
                printf("\n");
            }
            printf("In edge_1_over_half_step_den, show 1/rho on vertical edge\n");
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d %g ", i, v_line_rhor(ic, ns_field));
                }
                printf("\n");
            }
        }
            // TMP
            /**
            printf("In edge_1_over_half_step_den, show rho on CELL center\n");            
            for(j = 0; j < gmax[1]; j++)
            {
                printf("row%d: ", j);
                for(i = 0; i <= gmax[0]/2; i++)
                {
                    ic[0] = i;
                    ic[1] = j;
                    printf(" %d %g ", i, Rect_den(ic,ns_field));
                }
                printf("\n");
            }
            **/
}


LOCAL void include_chem_src(
        int         start,
        int         end,
        float       dt,
        float       *dh,
        int         swp_num,
        int         *iperm,
        int         dim,
        Upwind_NS   *upwind,
        NS_Field_Stencil   *ns_field,
        int         *icoords,
        int         pbuf)
{
        int             i, j, k, idirs[MAXD];
        int             kmin, kmax, sten_rad = 0;
        float           **u, **uL, **uR;
        float           *uk, *uLk, *uRk;
        float           dtdni, dn, diff;
        Locstate        s1, s2, s3;
        Vec_NS_Gas      *vst = upwind->vst;
        Vec_NS_Gas      **Tvst = upwind->Tvst;
        int             ic[MAXD];
        float           v0[40], v1[40], v2[40];
        float           *mu = mu_chem;
        COMPONENT       comp;

        for (j = 0; j < dim; ++j)
        {
            idirs[j] = iperm[(j+swp_num)%dim];
            ic[j] = icoords[j];
        }
        
        kmin = dim + 2;
        kmax = dim + 2 + upwind->vst->n_chem;
        u = upwind->u;
        uL = upwind->uL;
        uR = upwind->uR;

        // Find the chem associated with the comp
        // Find the stencil radius for irregular points
        if(ns_field->pt_sten != NULL)
            sten_rad = upwind->sten_rad;

        for(j = start; j < end; j++)
        {
            s1 = Tvst[0]->state[j];
            s2 = vst->state[j];
            s3 = Tvst[1]->state[j];

            for(k = kmin, i =0; k < kmax; k++, i++)
            {
                v0[i] = Tvst[0]->chem[i][j];
                v1[i] = vst->chem[i][j];
                v2[i] = Tvst[1]->chem[i][j];
            }

            if(ns_field->pt_sten != NULL)
            {
                icoords[idirs[0]] = ic[idirs[0]] + j - sten_rad;
            }
            else
                icoords[idirs[0]] = j + pbuf;

            // Find the chemical  in the (ghost, irregular) cell
	    /***
            if((icoords[0] < -lbuf[0] || icoords[0] >=gmax[0]+ubuf[0]  ||
                icoords[1] < -lbuf[1] || icoords[1] >=gmax[1]+ubuf[1]) ||
               (Rect_comp(icoords,ns_field->newwave) == ext_comp ||
                is_excluded_comp(Rect_comp(icoords,ns_field->newwave),
                ns_field->newfr->interf)) ||
                Rect_comp(icoords,ns_field->newwave) == POTS_COMP)
            {
                if(ns_field->pt_sten != NULL)
                {
                    comp = ns_field->pt_sten->newcomp;
                    {
                        printf("ERROR: include_chem_src, point outside domain, sten i = %d\n", i);
                        printf("ic[%d %d], start = %d, end = %d, pbuf %d, sten_rad = %d\n",
                         icoords[0], icoords[1], start, end, pbuf, sten_rad);
                        printf("comp = %d, newcomp %d\n",
                             Rect_comp(icoords,ns_field->newwave), ns_field->pt_sten->newcomp);
                        (void) printf("STENCIL: "); print_Stencil(ns_field->pt_sten);
                        for (i = -2; i <= 2; ++i)
                        {
                            (void) printf("state[%d]: comp[%d]",i, ns_field->pt_sten->comp[i]);
                            (*ns_field->fr->print_state)(ns_field->pt_sten->st[i]);
                        }
                    }
                    clean_up(ERROR);
                }
                else
                {
                    printf("ERROR: include_chem_src, point outside domain,\n");
                    printf("ic[%d %d], start = %d, end = %d, pbuf %d, sten_rad = %d\n",
                         icoords[0], icoords[1], start, end, pbuf, sten_rad);
                    printf("comp = %d\n", Rect_comp(icoords,ns_field->newwave));
                    clean_up(ERROR);
                }
            }
	    ***/ 

            for(k = kmin, i =0; k < kmax; k++, i++)
            {
                if(idirs[0] == 0)
                {
                    diff = mu[i]*(vst->chem[i][j+1] - 2.0*vst->chem[i][j] + vst->chem[i][j-1])/sqr(dh[0]);                
                    diff += mu[i]*(v2[0] - 2.0*v1[0] + v0[0])/sqr(dh[1]);
                }
                else
                {
                    diff = mu[i]*(vst->chem[i][j+1] - 2.0*vst->chem[i][j] + vst->chem[i][j-1])/sqr(dh[1]);                
                    diff += mu[i]*(v2[0] - 2.0*v1[0] + v0[0])/sqr(dh[0]);
                }
                // if(deb_npt)
                if(isnan(diff))
                {
                    printf("ERROR include_chem_src, ic[%d %d], dp [%9.8g %9.8g], diff [%g %g]\n",
                     icoords[0], icoords[1], diff);
                    clean_up(ERROR);
                }

                uLk  = uL[k];
                uRk  = uR[k];
                if((j + 1) != end)
                {
                    uLk[j+1] += 0.5*dt*diff;

                    if(isnan(uLk[j+1]))
                    {
                        printf("ERROR include_chem_src, ic[%d %d], diff [%g ]\n",
                            icoords[0], icoords[1], diff);
                        clean_up(ERROR);
                    }
                }
                if(j != start)
                {
                    uRk[j] += 0.5*dt*diff;
                    if(isnan(uRk[j]))
                    {
                        printf("ERROR include_chem_src,1 ic[%d %d], diff [%g ]\n",
                            icoords[0], icoords[1], diff);
                        clean_up(ERROR);
                    }
                }
            }
        }

        /**
        if(deb_npt)
        {
             printf("edge velocity, after include press grad, visco\n");
             for (j = start; j < end; ++j)
             {
                 (void) printf("%-4d L:",j);
                 for (k = 0; k < dim; ++k)
                     printf(" %9.8g",uL[k][j]);
                 printf(" ; R: ");
                 for (k = 0; k < dim; ++k)
                     printf(" %9.8g",uR[k][j]);
                 printf("\n");
             }
        }
        **/
}


EXPORT void g_init_chem_coeff()
{
        int i;
        screen("Enter num of chemicals: ");
        (void) Scanf("%d\n",&n_chem);
        if(n_chem != 0)
        {
            screen("Chemical diffusion: ");		 
            vector(&mu_chem, n_chem, sizeof(float));
            for(i = 0; i < n_chem-1; i++)
                Scanf("%f ",&mu_chem[i]);
            Scanf("%f \n",&mu_chem[i]);
        }
}

LOCAL int excluded_comp(
        COMPONENT   comp,
        INTERFACE   *intfc) 
{
        if( comp != ext_comp && (! is_excluded_comp(comp,intfc)) &&
            comp != POTS_COMP)
            return NO;

        return YES;
}


/*
 * grid_den_viscosity: Does count partial
 * (including physical boundary) cells as
 * full cells. This is like the ghost cell method.
 * The boundary geometry is simplified.
 * The density is set into Rect_state() for
 * half-time step reconstruction
*/
LOCAL void grid_den_viscosity_full_cell_ver(
        NS_Field_Stencil  *ns_field,
        Wave              *wv,
        Front             *fr)
{
        int             smin[MAXD], smax[MAXD];
        TRI_GRID        *ntg = wave_tri_soln(wv)->tri_grid;
        BLK_EDGE        *blk_edge = ntg->blk_edge;
        // Front           *nfr = ns_field->newfr;
        int             ix, iy, ic[MAXD], i, j, k, i1;
        Gas_param       **prms_list;
        int             nprms;
        COMPONENT       comp, comp1, cpmcomp;
        COMPONENT       *comps;
        float           **den_vis, *a_frac;
        RECT_GRID       *gr = wv->rect_grid; 
        int             deb_flag = NO;

        if(ns_field->n_eos == 0)
        {
            nprms = return_params_list(&prms_list);
            vector(&comps, nprms, sizeof(COMPONENT));
            matrix(&den_vis,nprms,2,sizeof(float));

            k = 0;
            for (i=0; i<nprms; i++)
            {
                if(prms_list[i]->eos->_NS_eos == YES)
                {
                    comps[k] = prms_list[i]->comp;
                    den_vis[k][0] = prms_list[i]->dens;
                    den_vis[k][1] = prms_list[i]->shear_visc;
                    k++;
                }
            }
            ns_field->n_eos = k;
            ns_field->comps = comps;
            ns_field->den_vis = den_vis;
        }
        else
        {
            k = ns_field->n_eos;
            comps = ns_field->comps;
            den_vis = ns_field->den_vis;
        }

        /**
        for(i = 0; i < k; i++)
        {
            printf("param[%d] dens %g, viscosity %g, comp %d\n", i,
                prms_list[i]->dens, prms_list[i]->shear_visc, prms_list[i]->comp);
        }
        **/

        for(i = 0; i < dim; i++)
        {
            smin[i] = -gr->lbuf[i];
            smax[i] = gr->gmax[i] + gr->ubuf[i];
        }

        for (iy = smin[1];  iy < smax[1];  ++iy)
        {
            for (ix = smin[0];  ix < smax[0];  ++ix)
            {
                ic[0] = ix; ic[1] = iy;

                /***
                if(ix == 53 && (iy == 4))
                {
                    deb_flag = YES;
                    printf(" grid_den_viscosity_full_cell_ver cell[%d %d] ctype = %d\n",
                             ix, iy, blk_edge[ix][iy].ctype);
                    printf("Rect_comp = %d, %d (newwv)\n",
                          Rect_comp(ic,wv), Rect_comp(ic,ns_field->newwave));
                    printf("blk_edge[ix][iy].ctype = %d\n",
                              blk_edge[ix][iy].ctype);
                }
                else
                    deb_flag = NO;
                **/

                Rect_den(ic,ns_field) = 0.0;
                Rect_visc(ic,ns_field) = 0.0;

                comp = Rect_comp(ic,wv);

                // if(blk_edge[ix][iy].ctype == EXTERNAL &&
                //    cpmcomp != POTS_COMP)
                //     continue;
                if(Rect_blk_edge(ic,wv).ctype == EXTERNAL &&
                   excluded_comp(comp, fr->interf))
                    continue;

                for(i = 0; i < k; i++)
                {
                    if(comp == comps[i])
                        break;
                }
                if(i >= k)
                {
                    printf("ERROR grid_den_viscosity_full_cell_ver "
                            "cell[%d %d] ctype = %d, comp = %d\n",
                         ix, iy, Rect_blk_edge(ic,wv).ctype, comp);
                    clean_up(ERROR);
                }

                Rect_den(ic,ns_field) = den_vis[i][0];
                Rect_visc(ic,ns_field) = den_vis[i][1];
                Dens(Rect_state(ic,wv)) = Rect_den(ic,ns_field);

                // DEBUG TMP
                /***
                if(deb_flag == YES)
                {
                    printf("IN grid_den_viscosity_full_cell_ver,"
                            " [%d][%d] comp %d, den = %g, mu = %g\n",
                       ix, iy, comp, Rect_den(ic,ns_field), Rect_visc(ic,ns_field));
                }
                ***/
            }
        }

        // free(comps);
        // free(den_vis);

        if(debugging("NS_vel") && NO)
        {
            printf("-----Cell Density-------------\n");
            for (iy = smin[1];  iy < smax[1];  ++iy)
            {
                printf("Row[%d] dens: ",iy);
                for (ix = smin[0];  ix < min(4,smax[0]);  ++ix)
                {
                    ic[0] = ix; ic[1] = iy;
                    printf("(%d, %9.8g) ", ix, Rect_den(ic,ns_field));
                }
                printf("\n");
            }
        }
}

